/*
 * Name: cache.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-05
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cache.c,v 1.1.1.1 1999/04/12 13:56:33 cs Exp $
 */

#include "stdheaders.h"
#include "cache.h"
#include "file.h"
#include "logging.h"
#include "process.h"
#include "hashtable.h"
#include "cfgparse.h"
#include "datacache.h"
#include "dircache.h"
#include "main.h"
#include "cifsmod.h"

/* ------------------------------------------------------------------------- */

void	cacheInit(void *obj, cacheInterface_t *interface)
{
cacheSuper_t	*cache = obj;

	cache->interface = interface;
	cache->opReady = procCondNew();
	cache->refCount = 1;
}

/* ------------------------------------------------------------------------- */

void	cacheDestroy(void *obj)
{
cacheSuper_t	*cache = obj;

	if(cache->opRunning != 0)
		logPrintf(LOGM_ERROR, "+++ cache: destroying while running\n");
	if(cache->owner != NULL && cache->opReady != NULL)
		cacheSetOwner(cache, NULL);
	if(cache->owner != NULL)
		fileSetCache(cache->owner, NULL);
	if(cache->opReady != NULL)
		procCondFree(cache->opReady);
}

/* ------------------------------------------------------------------------- */

void	cacheSetOwner(void *obj, struct cifsFile *owner)
{
cacheSuper_t	*cache = obj, *current;

	if(owner != NULL){
		current = fileCache(owner);
		if(current == obj)		/* nothing to be done */
			return;
		if(current != NULL){	/* owner already has a cache */
			cacheSetOwner(current, NULL);
		}
	}
	if(cache != NULL){
		if(cache->opRunning != 0)
			logPrintf(LOGM_ERROR, "+++ cache: changing owner while running\n");
		if(cache->owner != NULL){		/* cache already has an owner */
			fileSetCache(cache->owner, NULL);
		}
		cache->owner = owner;
		cacheVirginate(cache);
	}
	if(owner != NULL){
		fileSetCache(owner, cache);
	}
}

/* ------------------------------------------------------------------------- */
/* -------------------- interface for cache management --------------------- */
/* ------------------------------------------------------------------------- */

typedef struct cachePool{
	btsize		pageSize;	/* must be first entry for search */
	btint		refCount;
	btint		numPages;
    lru_t		*cacheLru;	/* lru-list of caches */
}cachePool_t;

typedef struct shareRep{
	void		*share;
	cachePool_t	*pool;
}shareRep_t;

static hashtable_t	*dcachesByShare = NULL;	/* hashtable of (shareRep_t *) */
static hashtable_t	*dcachesBySize = NULL;	/* hashtable of (cachePool_t *) */
static lru_t		*dirCacheLru = NULL;

/* ------------------------------------------------------------------------- */

static void	cachePoolFree(cachePool_t *pool)
{
void	*cache;

	if(pool != NULL){
		htabRemoveObjectEqualTo(dcachesBySize, pool);
		if(pool->cacheLru != NULL){
			while((cache = lruOldestRelativeTo(pool->cacheLru, NULL)) != NULL){
				lruRemoveObject(pool->cacheLru, cache);
				cacheRelease(cache);
			}
			lruFree(pool->cacheLru);
		}
		memset(pool, 0, sizeof(*pool));
		memFree(pool);
	}
}

static cachePool_t	*cachePoolNew(btsize pageSize, btint numPages, btint numCaches)
{
cachePool_t	*volatile pool;
btint		i;
void		*cache;

EXC_TRY
	pool = memCalloc(sizeof(cachePool_t));
	pool->pageSize = pageSize;
	pool->numPages = numPages;
	pool->cacheLru = lruNew(0);
	for(i=0;i<numCaches;i++){
		cache = dcacheNew(pageSize, numPages);
		lruAddObject(pool->cacheLru, cache);
	}
EXC_CATCH
	cachePoolFree(pool);
	EXC_RETHROW();
EXC_ENDTRY
	htabAddObject(dcachesBySize, pool);
	return pool;
}

static inline void	cachePoolRetain(cachePool_t *pool)
{
	pool->refCount++;
}

static inline void	cachePoolRelease(cachePool_t *pool)
{
	if(--pool->refCount == 0){
		cachePoolFree(pool);
	}else if(pool->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

static btint	cachePoolCompare(void *o1, void *o2)
{
cachePool_t	*this = o1, *other = o2;

	return this->pageSize - other->pageSize;
}

static btu32	cachePoolHashVal(void *obj)
{
cachePool_t	*pool = obj;

	return (pool->pageSize >> 10) ^ pool->pageSize >> 13;
}

static btint	srepCompare(void *o1, void *o2)
{
shareRep_t	*this = o1, *other = o2;

	return (btu8 *)this->share - (btu8 *)other->share;
}

static btu32	srepHashVal(void *obj)
{
shareRep_t	*srep = obj;

	return ((btu32)srep->share >> 6) ^ ((btu32)srep->share >> 9);
}

static collectionObjectI_t	cachePoolInterface = {
	cachePoolCompare,
	cachePoolHashVal,
	NULL,
	NULL,
};

static collectionObjectI_t	shareRepInterface = {
	srepCompare,
	srepHashVal,
	NULL,
	NULL,
};

/* ------------------------------------------------------------------------- */

void	cacheNeedPagesize(btsize size, void *share)
{
shareRep_t		*srep;
cachePool_t		*pool;
btint			i, numCaches, cacheSize, numEntries = -1;
void			*cache;
cfgProperty_t	*path[8];

	if(dcachesByShare == NULL)
		dcachesByShare = htabNew(&cachePoolInterface, 8, 50);
	if(dirCacheLru == NULL){
		dirCacheLru = lruNew(0);
		if(numEntries < 0)
			numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		numCaches = cfgIntValue(cfgQueryFromArray("initialDirCaches", numEntries, path), 8);
		for(i=0;i<numCaches;i++){
			cache = dirCacheNew();
			lruAddObject(dirCacheLru, cache);
		}
	}
	if((srep = htabFindObjectEqualTo(dcachesByShare, &share)) != NULL){
		if(srep->pool->pageSize == size)
			return;	/* same size as already allocated */
		htabRemoveObjectEqualTo(dcachesByShare, srep);
		cachePoolRelease(srep->pool);
	}else{
		srep = memCalloc(sizeof(*srep));
		srep->share = share;
	}
	if(dcachesBySize == NULL)
		dcachesBySize = htabNew(&shareRepInterface, 16, 50);
	if((pool = htabFindObjectEqualTo(dcachesBySize, &size)) == NULL){
		if(numEntries < 0)
			numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		numCaches = cfgIntValue(cfgQueryFromArray("initialDataCaches", numEntries, path), 8);
		cacheSize = cfgIntValue(cfgQueryFromArray("dataCacheSize", numEntries, path), 32768);
		pool = cachePoolNew(size, cacheSize / size, numCaches);
	}
	srep->pool = pool;
	cachePoolRetain(pool);
	htabAddObject(dcachesByShare, srep);
}

void	cacheShareFreed(void *share)
{
shareRep_t	*srep;

	if(dcachesByShare == NULL)
		return;	/* nothing to be searched */
	if((srep = htabFindObjectEqualTo(dcachesByShare, &share)) != NULL){
		htabRemoveObjectEqualTo(dcachesByShare, srep);
		cachePoolRelease(srep->pool);
		memset(srep, 0, sizeof(srep));
		memFree(srep);
	}
}

void	cacheAcquire(btint type, struct cifsFile *file)
{
void		*share;
shareRep_t	*srep = NULL;
void		*cache;
lru_t		*lru;

	if(type == CACHE_TYPE_DATA){
		share = fileShare(file);
		if((srep = htabFindObjectEqualTo(dcachesByShare, &share)) == NULL){
			DEBUG_TRAP();
		}
		lru = srep->pool->cacheLru;
	}else{	/* type must be CACHE_TYPE_DIR */
		lru = dirCacheLru;
	}
	if((cache = fileCache(file)) == NULL){
		while((cache = lruOldestRelativeTo(lru, cache)) != NULL){
			if(cacheOperationsRunning(cache) == 0)	/* cache is idle */
				break;
		}
		if(cache == NULL){							/* no idle cache found */
			if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
				logPrintf(LOGM_RARE, "file: short of %s cache objects, creating new\n", type == CACHE_TYPE_DATA ? "data" : "dir");
			if(type == CACHE_TYPE_DATA){
				cache = dcacheNew(srep->pool->pageSize, srep->pool->numPages);
			}else{	/* type must be CACHE_TYPE_DIR */
				cache = dirCacheNew();
			}
			lruAddObject(lru, cache);
		}
		cacheSetOwner(cache, file);
	}
    lruTouch(lru, cache);
}

/* ------------------------------------------------------------------------- */
