/*
 * Name: cache.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-03
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cache.h,v 1.1.1.1 1999/04/12 13:56:33 cs Exp $
 */

#ifndef __cache_H_INCLUDED__
#define __cache_H_INCLUDED__

#include "basictypes.h"
#include "lru.h"

/*
General Description:
Superclass for data- and directory-caches, implements the common functionality
and method dispatchers for polymorphism.
*/

/* ------------------------------------------------------------------------- */

typedef struct cacheInterface{
	void	(*invalidate)(void *cache);
	void	(*virginate)(void *cache);
	void	(*writeBack)(void *cache, void *caller);
	void	(*free)(void *cache);
}cacheInterface_t;

typedef struct cacheSuper{
	lruObject_t			lruPointers;
	cacheInterface_t	*interface;
	struct cifsFile		*owner;			/* file this cache belongs to */
	btint				forUser;		/* uid for which cache is valid */
	bti32				opRunning;		/* number of operations running */
    struct procCond		*opReady;		/* fires when an operation is ready */
	btint				refCount;
}cacheSuper_t;

#define	CACHE_TYPE_DATA		0	/* datacache */
#define	CACHE_TYPE_DIR		1	/* dircache */
#define	CACHE_TYPE_COUNT	2	/* number of types */

/* ------------------------------------------------------------------------- */

void	cacheInit(void *obj, cacheInterface_t *interface);
/* This function initializes the generic part (superclass) of all caches.
 */
void	cacheDestroy(void *obj);
/* This function frees all memory allocated in cacheInit().
 */
/* inline bti32	cacheOperationsRunning(void *this); */
/* Returns the number of operations that are currently running for this
 * cache.
 */
/* void	cacheInvalidate(void *cache);*/
/* Marks all information in the cache as invalid. Read data must be re-fetched
 * from the data source and data that is not written back is lost.
 */
/*void	cacheWriteBack(dcache_t *this, void *caller);*/
/* This function writes back all dirty entries in the cache.
 * Callback: no arguments (NULL)
 */
void	cacheSetOwner(void *cache, struct cifsFile *owner);
/* This function is used to remove the cache from a file or add it to a file.
 * If the file (owner) changes, the cache is implicitly removed from the
 * previously owning file
 */
 
/* -------------------- interface for cache management --------------------- */

void	cacheNeedPagesize(btsize size, void *share);
void	cacheShareFreed(void *share);
/* These two functions are used to keep track of which page sizes are
 * currently needed. This allows freeing unused caches when shares are
 * freed.
 */
void	cacheAcquire(btint type, struct cifsFile *file);
/* This function returns an unused cache of the "correct" size determined by
 * the file's share and sets the cache's owner to 'file'. If no cache is
 * currently free, a new one is allocated. The cache is automatically added
 * to the file.
 */

/* ------------------------------------------------------------------------- */
/* ---------------------------- inline methods ----------------------------- */
/* ------------------------------------------------------------------------- */

static inline bti32	cacheOperationsRunning(void *this)
{
cacheSuper_t	*cache = this;

	return cache->opRunning;
}

#define	cacheInvalidate(c)	(((cacheSuper_t *)(c))->interface->invalidate((void *)(c)))
#define	cacheVirginate(c)	(((cacheSuper_t *)(c))->interface->virginate((void *)(c)))
#define	cacheWriteBack(c, caller)	(((cacheSuper_t *)(c))->interface->writeBack((void *)(c), caller))
#define	cacheFree(c)	(((cacheSuper_t *)(c))->interface->free((void *)(c)))

static inline void	cacheRetain(void *this)
{
cacheSuper_t	*cache = this;

	cache->refCount++;
}

static inline void	cacheRelease(void *this)
{
cacheSuper_t	*cache = this;

	--cache->refCount;
	if(cache->refCount == 0){
		cacheFree(cache);
	}else if(cache->refCount < 0){
		DEBUG_TRAP();
	}
}

/* The following two are macros instead of inline functions because we
 * must not include file.h here.
 */
#define	cacheStartOperation(cache)									\
	do{																\
		cacheRetain(cache);											\
		((cacheSuper_t *)(cache))->opRunning++;						\
		if(((cacheSuper_t *)(cache))->owner != NULL)				\
			fileRetain(((cacheSuper_t *)(cache))->owner);			\
	}while(0)

#define	cacheEndOperation(cache, doDecrement)						\
	do{																\
		if(doDecrement) ((cacheSuper_t *)(cache))->opRunning--;		\
		procCondBroadcast(((cacheSuper_t *)(cache))->opReady, NULL);\
		if(((cacheSuper_t *)(cache))->owner != NULL)				\
			fileRelease(((cacheSuper_t *)(cache))->owner);			\
		cacheRelease(cache);										\
	}while(0)

/* ------------------------------------------------------------------------- */

#endif /* __cache_H_INCLUDED__ */
