/*
 * Name: cifs.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-20
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cifs.c,v 1.3 1999/08/01 12:57:09 cs Exp $
 */


#include "stdheaders.h"
#include "logging.h"
#include "cifs.h"
#include "smb.h"
#include "server.h"
#include "smbnumbers.h"
#include "errcodes.h"
#include "unicode.h"
#include "process.h"
#include "system.h"
#include "pslnetbios.h"
#include "main.h"
#include "cfgparse.h"
#include "pwhash.h"


/* ------------------------------------------------------------------------- */

#define	DECODER_START(excOnError)				\
	memset(&res, 0, sizeof(res));				\
    if(smb->rxHeader == NULL){					\
        res.err.class = ERRCLS_CIFS_LOCAL;	\
        res.err.code = ERR_CIFS_LOCtimeout;	\
        goto decode_exception;					\
    }else{										\
        res.err.class = smb->rxHeader->err.class;\
        res.err.code = smb->rxHeader->err.code;	\
		res.flags = smb->rxHeader->flags;		\
		res.flags2 = smb->rxHeader->flags2;		\
        res.uid = smb->rxHeader->uid;			\
        res.tid = smb->rxHeader->tid;			\
		if((excOnError) && res.err.class != 0)	\
			goto decode_exception;				\
    }

#define	DECODER_ERROR(msg)													\
	do{	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))									\
			logPrintf(LOGM_ERROR, "cifs: unmarshalling error: %s\n", msg);	\
        res.err.class = ERRCLS_CIFS_LOCAL;								\
        res.err.code = ERR_CIFS_LOCbogusdata;								\
        goto decode_exception;												\
    }while(0)

#define	SUBDECODER_ERROR(msg)												\
	do{	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))									\
			logPrintf(LOGM_ERROR, "cifs: unmarshalling error: %s\n", msg);	\
        res->err.class = ERRCLS_CIFS_LOCAL;								\
        res->err.code = ERR_CIFS_LOCbogusdata;								\
        return 1;															\
    }while(0)

#define	DECODER_FINISH														\
	decode_exception:														\
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))									\
		logPrintf(LOGM_CIFSTRACE, "cifs trace: [%d] rval = %d/%d\n", smb->mid, res.err.class, res.err.code);\
	if(ref != NULL) ((procCtx_t *)ref)->process(ref, &res);

#define	CIFS_TRACE(fname)													\
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))									\
		logPrintf(LOGM_CIFSTRACE, "cifs trace: %s()\n", fname)

#define	CIFS_TRACE_I(fname, x)												\
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))									\
		logPrintf(LOGM_CIFSTRACE, "cifs trace: %s(%d)\n", fname, x)


/* ------------------------------------------------------------------------- */

static void	cifsDecodeNegotiate(void *ref, struct smbRequest *smb)
{
cifsResNegotiate_t	res;
btu8				*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    res.dialect = smbGetWord(&p);
	if(res.dialect == CIFS_DIALECT_LM2){
		res.securityMode = smbGetWord(&p);
		res.maxXmit = smbGetWord(&p);
		res.maxMpx = smbGetWord(&p);
		res.maxVc = smbGetWord(&p);
		res.rawMode = smbGetWord(&p);
		res.maxRaw = res.rawMode != 0 ? 65536 : 0;
		res.sessionKey = smbGetLong(&p);
		p += 4;	/* skip current time and date */
		res.gmtOffset = (bti16)smbGetWord(&p);
		res.encryptKeyLen = smbGetWord(&p);
		p += 2;	/* reserved */
	}else if(res.dialect == CIFS_DIALECT_NTLM){
		res.securityMode = *p++;
		res.maxMpx = smbGetWord(&p);
		res.maxVc = smbGetWord(&p);
		res.maxXmit = smbGetLong(&p);
		res.maxRaw = smbGetLong(&p);
		res.rawMode = res.maxRaw > 0 ? 3 : 0;
		res.sessionKey = smbGetLong(&p);
		res.capabilities = smbGetLong(&p);
		p += 8;		/* skip system time and date */
		res.gmtOffset = (bti16)smbGetWord(&p);
		res.encryptKeyLen = *p++;
	}else if(res.dialect == CIFS_DIALECT_CORE){
		res.maxXmit = 1024;
		res.maxMpx = 1;
		res.maxVc = 1;
		res.gmtOffset = sysGmtOffset();	/* in minutes west of GMT */
	}else{
		DECODER_ERROR("Could not negotiate dialect");
	}
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in NEGOTIATE");
	if(res.encryptKeyLen > 0){
		/* check against rxBufferLimit, not rxDataBytes because rxDataBytes
		   may be sent wrong by samba1.9.17 */
		if(res.encryptKeyLen > smb->rxBufferLimit - (smb->rxBuffer + smb->rxDataBase))
			DECODER_ERROR("Not enough data bytes in NEGOTIATE");
		if(res.encryptKeyLen > 64)
			DECODER_ERROR("More than 64 bytes encryption key in NEGOTIATE");
		memcpy(res.encryptKey, smb->rxBuffer + smb->rxDataBase, res.encryptKeyLen);
	}
	if((res.securityMode & CIFS_SECMODE_PWENCRYPT) == 0)
		res.encryptKeyLen = 0;
	DECODER_FINISH
}

void	cifsNegotiate(struct smbRequest *smb, void *caller)
{
int			i;
/* the dialects array must match the CIFS_DIALECT_* defines in cifs.h! */
static char	*dialects[] = {
        "PC NETWORK PROGRAM 1.0",
        "LM1.2X002",
        "NT LM 0.12",
        NULL
	};

	CIFS_TRACE("cifsNegotiate");
	smbStartCommand(smb, 0, 0, 0, cifsDecodeNegotiate, caller);
    smbStartNormalSmb(smb, SMB_COM_NEGOTIATE);
    smbStartParameters(smb, -1);
    smbStartData(smb);
    for(i=0;dialects[i]!=NULL;i++){
		smbAddByte(smb, 2);
        smbAddAsciiFromAscii(smb, dialects[i]);
    }
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeSessionSetup(void *ref, struct smbRequest *smb)
{
cifsResSessionSetup_t	res;
btu8					*p, *base;
btuni					serverOS[256], serverLanMan[256];

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	p += 4;		/* skip andX stuff */
    res.action = smbGetWord(&p);
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in NEGOTIATE");
    p = base = smb->rxBuffer + smb->rxDataBase;

	serverOS[0] = serverLanMan[0] = 0;
	if((res.flags2 & CIFS_FLAG2_UNICODE) != 0){
		if(((int)p & 1) != 0)
			p++;	/* align to word boundary */

		if(p - base < smb->rxDataBytes){
			res.serverOS = smbGetUniString(&p);
		}else
			res.serverOS = serverOS;
		if(p - base < smb->rxDataBytes)
			res.serverLanMan = smbGetUniString(&p);
		else
			res.serverLanMan = serverLanMan;
	}else{
		if(p - base < smb->rxDataBytes){
			convServerAsciiToUni(p, serverOS, 256);
			p += strlen(p) + 1;
		}
		res.serverOS = serverOS;
		if(p - base < smb->rxDataBytes)
			convServerAsciiToUni(p, serverLanMan, 256);
		res.serverLanMan = serverLanMan;
	}
	DECODER_FINISH
}

void	cifsSessionSetup(struct smbRequest *smb, void *caller, btu16 uid, btuni *username, btuni *plainPw, btu8 *lmHash, btu8 *ntHash, btbool pwIsEncrypted, btu8 *domain)
{
btsize		maxXmit;
smbVar_t	pass1len, pass2len;
btsize		passBase;
btu8		passHash[21], passReply[24];
btbool		useUnicodeStore;

	CIFS_TRACE("cifsSessionSetup");
	useUnicodeStore = smb->server->useUnicode;
	smb->server->useUnicode = 0;	/* don't use Unicode in this early stage */
	smbStartCommand(smb, 0, 0, 0, cifsDecodeSessionSetup, caller);
    smbStartNormalSmb(smb, SMB_COM_SESSION_SETUP_ANDX);
    smbStartParameters(smb, 0xff);
	if((maxXmit = smb->server->maxXmit) > 32768){	/* don't request more than that many bytes */
		maxXmit = 32768;
    }else if(maxXmit < 1024){
        maxXmit = 1024;
    }
	smbAddWord(smb, maxXmit);
	smbAddWord(smb, smb->server->windowSize);
	smbAddWord(smb, 0);	/* virtual connection number */
	smbAddLong(smb, smb->server->sessionKey);
	smbAddVariable(smb, &pass1len, 2, 0);	/* case sensitive password length */
	if(smb->server->dialect >= CIFS_DIALECT_NTLM){
		smbAddVariable(smb, &pass2len, 2, 0);	/* case sensitive password length */
		smbAddLong(smb, 0);	/* reserved */
		smbAddLong(smb, CIFS_CLIENT_CAP_UNICODE | CIFS_CLIENT_CAP_LEVEL_II_OPLOCKS);
    }else{
		smbAddLong(smb, 0);	/* reserved */
    }
	smbStartData(smb);
	passBase = smbCurrentOffset(smb);
	if(pwIsEncrypted){
		memset(passHash, 0, sizeof(passHash));
		memcpy(passHash, lmHash, 16);
		pwhResponse(passHash, srvChallenge(smb->server), passReply);
		smbAddBlock(smb, passReply, 24);
	}else{
		smbAddStringFromUni(smb, plainPw);
	}
	smbWriteVariable(smb, &pass1len, smbCurrentOffset(smb) - passBase);
	if(smb->server->dialect >= CIFS_DIALECT_NTLM){
		passBase = smbCurrentOffset(smb);
		if(pwIsEncrypted){
			memset(passHash, 0, sizeof(passHash));
			memcpy(passHash, ntHash, 16);
			pwhResponse(passHash, srvChallenge(smb->server), passReply);
			smbAddBlock(smb, passReply, 24);
		}
		smbWriteVariable(smb, &pass2len, smbCurrentOffset(smb) - passBase);
	}
	smbAddStringFromUni(smb, username);
	smbAddStringFromAscii(smb, domain == NULL ? (btu8 *)"" : domain);
	smbAddStringFromAscii(smb, "UNIX");		/* native OS */
	smbAddStringFromAscii(smb, "SHARITY");	/* client software name */
	smbEndCommand(smb);
	smb->server->useUnicode = useUnicodeStore;	/* restore */
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeLogoff(void *ref, struct smbRequest *smb)
{
cifsRes_t	res;
btu8		*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	p += 4;		/* skip andX stuff */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in Logoff");
	DECODER_FINISH
}

void	cifsLogoff(struct smbRequest *smb, void *caller, btu16 uid)
{
	CIFS_TRACE("cifsLogoff");
	smbStartCommand(smb, uid, 0, 0, cifsDecodeLogoff, caller);
    smbStartNormalSmb(smb, SMB_COM_LOGOFF_ANDX);
    smbStartParameters(smb, 0xff);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeTreeConnect(void *ref, struct smbRequest *smb)
{
cifsRes_t	res;
btu8		*p, *base;
btu16		maxXmit;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	maxXmit = smbGetWord(&p);
	res.tid = smbGetWord(&p);
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in TreeConnect");
	if(smb->server->dialect < CIFS_DIALECT_LM2){
		smb->server->maxXmit = maxXmit;
	}
	DECODER_FINISH
}

static void	cifsDecodeTreeConnectX(void *ref, struct smbRequest *smb)
{
cifsRes_t	res;
btu8		*p, *base;
btu16		support;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	p += 4;		/* skip andX stuff */
	if(smb->rxParameterBytes >= 6){
		support = smbGetWord(&p);
	}
	/* what shall we do with this support stuff? */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in TreeConnectAndX");
	/* don't read optional service and filesystem info because we don't need it */
	DECODER_FINISH
}

void	cifsTreeConnect(struct smbRequest *smb, void *caller, btu16 uid, btu8 *passwd, int pwlen, btuni *path, btbool useAndX, btu8 *service)
{
btbool	useUnicodeBackup;

	CIFS_TRACE("cifsTreeConnect");
	smbStartCommand(smb, uid, 0, 0, useAndX ? cifsDecodeTreeConnectX : cifsDecodeTreeConnect, caller);
	useUnicodeBackup = smb->server->useUnicode;
	smb->server->useUnicode = 0;	/* hack to send this SMB in ASCII in any case */
    smbStartNormalSmb(smb, useAndX ? SMB_COM_TREE_CONNECT_ANDX : SMB_COM_TREE_CONNECT);
    smbStartParameters(smb, useAndX ? 0xff : -1);
	if(useAndX){
		smbAddWord(smb, 0);	/* disconnect flag */
		smbAddWord(smb, pwlen);
		smbStartData(smb);
		smbAddBlock(smb, passwd, pwlen);
		smbAddAsciiFromUni(smb, path);
	}else{
		smbStartData(smb);
		smbAddByte(smb, 4);
		smbAddAsciiFromUni(smb, path);
		smbAddByte(smb, 4);
		smbAddBlock(smb, passwd, pwlen);
		smbAddByte(smb, 4);
	}
	smbAddAsciiFromAscii(smb, service);
	smbEndCommand(smb);
	smb->server->useUnicode = useUnicodeBackup;
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeTreeDisconnect(void *ref, struct smbRequest *smb)
{
cifsRes_t	res;

    DECODER_START(1)
	DECODER_FINISH
} 

void	cifsTreeDisconnect(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid)
{
	CIFS_TRACE("cifsTreeDisconnect");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeTreeDisconnect, caller);
    smbStartNormalSmb(smb, SMB_COM_TREE_DISCONNECT);
    smbStartParameters(smb, -1);
	smbStartData(smb);
	smbEndCommand(smb);
}


/* ------------------------------------------------------------------------- */

static void	cifsDecodeQueryInformation(void *ref, struct smbRequest *smb)
{
cifsResQueryFile_t	res;
btu8				*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	res.attributes = smbGetWord(&p);
	res.lastWrite = smbGetLong(&p);
	res.size = smbGetLong(&p);
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeQueryInformation");
	DECODER_FINISH
}

void	cifsQueryInformation(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path)
{
	CIFS_TRACE("cifsQueryInformation");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeQueryInformation, caller);
    smbStartNormalSmb(smb, SMB_COM_QUERY_INFORMATION);
    smbStartParameters(smb, -1);
	smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeQueryPathInfoStd(void *ref, struct smbRequest *smb)
{
cifsResQueryFile_t	res;
btu8				*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxDataBase;
    res.ctimeDate = smbGetWord(&p);
    res.ctimeTime = smbGetWord(&p);
    res.atimeDate = smbGetWord(&p);
    res.atimeTime = smbGetWord(&p);
    res.mtimeDate = smbGetWord(&p);
    res.mtimeTime = smbGetWord(&p);
	res.size = smbGetLong(&p);
	res.allocSize = smbGetLong(&p);
	res.attributes = smbGetWord(&p);
    /* not interested in EA size */
	if(p - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough data in cifsDecodeQueryPathInfoStd");
	DECODER_FINISH
}

static void	cifsDecodeQueryPathInfoUnix(void *ref, struct smbRequest *smb)
{
cifsResQueryFile_t	res;
btu8				*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxDataBase;
    res.size = smbGetLarge(&p);
    res.ctime = smbGetLarge(&p);
    res.atime = smbGetLarge(&p);
    res.mtime = smbGetLarge(&p);
	res.fileUid = smbGetLarge(&p);
	res.fileGid = smbGetLarge(&p);
    res.type = smbGetLong(&p);
    res.devMajor = smbGetLarge(&p);
    res.devMinor = smbGetLarge(&p);
    res.inum = smbGetLarge(&p);
    res.mode = smbGetLarge(&p);
    res.nlinks = smbGetLarge(&p);
	if(p - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough data in cifsDecodeQueryPathInfoUnix");
	DECODER_FINISH
}

static void	cifsDecodeQueryPathInfoBasic(void *ref, struct smbRequest *smb)
{
cifsResQueryFile_t	res;
btu8				*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxDataBase;
    res.ctime = smbGetLarge(&p);
    res.atime = smbGetLarge(&p);
    res.mtime = smbGetLarge(&p);
	p += 8;			/* skip 'changeTime' */
	res.attributes = smbGetWord(&p);
	p += 6;			/* padding */
	res.allocSize = smbGetLarge(&p);
	res.size = smbGetLarge(&p);
	if(p - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough data in cifsDecodeQueryPathInfoBasic");
	DECODER_FINISH
}


void	cifsQueryPathInfo(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu32 infoLevel)
{
void	*decoder;

	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))
		logPrintf(LOGM_CIFSTRACE, "cifs trace: cifsQueryPathInfo(level=0x%x)\n", infoLevel);
	switch(infoLevel){
		case SMB_INFO_STANDARD:			decoder = cifsDecodeQueryPathInfoStd; 	break;
		case SMB_QUERY_FILE_ALL_INFO:
		case SMB_QUERY_FILE_BASIC_INFO:	decoder = cifsDecodeQueryPathInfoBasic;	break;
		case SMB_QUERY_FILE_UNIX_BASIC:	decoder = cifsDecodeQueryPathInfoUnix;	break;
		default:
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "cifs cifsQueryPathInfo(): level=0x%x not known\n", infoLevel);
			decoder = cifsDecodeQueryPathInfoStd;
			break;
	}
	smbStartCommand(smb, uid, 0, tid, decoder, caller);
    smbStartTrans2Smb(smb, TRANS2_QUERY_PATH_INFORMATION, 64, 2048, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, infoLevel);
    smbAddLong(smb, 0);
    smbAddStringFromUni(smb, path);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeFindAttributes(void *ref, struct smbRequest *smb)
{
cifsResQueryFile_t	res;
btu8				*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	p += 2;	/* skip searchID */
	if(smbGetWord(&p) != 1){	/* not one entry found */
        res.err.class = ERRCLS_CIFS_DOS;
        res.err.code = ERR_CIFS_DOSbadpath;
        goto decode_exception;
	}
	p += 6;	/* skip end of search, eaErrorOffset, lastNameOffset */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeFindAttributes");
    p = base = smb->rxBuffer + smb->rxDataBase;
    res.ctimeDate = smbGetWord(&p);
    res.ctimeTime = smbGetWord(&p);
    res.atimeDate = smbGetWord(&p);
    res.atimeTime = smbGetWord(&p);
    res.mtimeDate = smbGetWord(&p);
    res.mtimeTime = smbGetWord(&p);
	res.size = smbGetLong(&p);
	res.allocSize = smbGetLong(&p);
	res.attributes = smbGetWord(&p);
	if(p - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough data in cifsDecodeFindAttributes");
	DECODER_FINISH
}

void	cifsFindAttributes(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path)
{
	CIFS_TRACE("cifsFindAttributes");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeFindAttributes, caller);
    smbStartTrans2Smb(smb, TRANS2_FIND_FIRST2, 64, 32768, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM | CIFS_ATTR_DIR);
    smbAddWord(smb, 1);
    smbAddWord(smb, 1);	/* automatic close after this request */
    smbAddWord(smb, SMB_INFO_STANDARD);	/* information level */
    smbAddLong(smb, 0);	/* what's SearchStorageType ??? */
    smbAddStringFromUni(smb, path);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeQueryFsInfo(void *ref, struct smbRequest *smb)
{
cifsResQueryFsInfo_t	res;
btu8					*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxDataBase;
	res.fileSystemId = smbGetLong(&p);
	res.sectorsPerAU = smbGetLong(&p);
	res.totalAU = smbGetLong(&p);
	res.availableAU = smbGetLong(&p);
	res.bytesPerSector = smbGetWord(&p);
	if(p - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough parameters in QueryFsInfo");
	DECODER_FINISH
}

void	cifsQueryFsInfo(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid)
{
	CIFS_TRACE("cifsQueryFsInfo");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeQueryFsInfo, caller);
    smbStartTrans2Smb(smb, TRANS2_QUERY_FS_INFORMATION, 64, 2048, 0);
    smbStartParameters(smb, -1);
	smbAddWord(smb, SMB_INFO_ALLOCATION);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodecifsQueryInfoDisk(void *ref, struct smbRequest *smb)
{
cifsResQueryFsInfo_t	res;
btu8					*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    res.totalAU = smbGetWord(&p);
    res.sectorsPerAU = smbGetWord(&p);
    res.bytesPerSector = smbGetWord(&p);
    res.availableAU = smbGetWord(&p);
    res.fileSystemId = 0;
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodecifsQueryInfoDisk");
	DECODER_FINISH
}

void	cifsQueryInfoDisk(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid)
{
	CIFS_TRACE("cifsQueryInfoDisk");
	smbStartCommand(smb, uid, 0, tid, cifsDecodecifsQueryInfoDisk, caller);
    smbStartNormalSmb(smb, SMB_COM_QUERY_INFORMATION_DISK);
    smbStartParameters(smb, -1);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeSearch(void *ref, struct smbRequest *smb)
{
cifsResFile_t		entry;
cifsResReaddir_t	res;
btu8				*p, *q, *base, *resumeBase;
int					i, blockLen;
btuni				name[13], *n;
btu8				*s;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    res.entryCount = smbGetWord(&p);
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeSearch");
    p = base = smb->rxBuffer + smb->rxDataBase;
    p++;	/* skip variable block identifier */
    blockLen = smbGetWord(&p);
	if(p + blockLen - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough data in cifsDecodeSearch");
    base = p;
    if(res.entryCount > blockLen / 43)
		DECODER_ERROR("Directory Content not consistent in cifsDecodeSearch");
    resumeBase = p;
    for(i=0;i<res.entryCount;i++){
        q = p;
        p += 21;	/* skip resume key */
		entry.infoLevel = SMB_INFO_STANDARD;
		entry.dosMode = *p++;
		entry.ctime = entry.atime = entry.wtime = smbGetWord(&p);
		entry.cdate = entry.adate = entry.wdate = smbGetWord(&p);
		entry.size = smbGetLong(&p);
		entry.allocSize = entry.size;
        s = p;
        p += 13;
        n = name;
        while(*s != 0 && *s != ' '){
            *n++ = convCharServerAsciiToUni(&s);
        }
        *n++ = 0;
        entry.name = name;
        if(!((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry))
            break;
    	resumeBase = q;
    }
    res.resumeSize = 21;
    memcpy(res.resumeKey, resumeBase, res.resumeSize);
	DECODER_FINISH
}


void	cifsSearch(struct smbRequest *smb, void *caller, btbool (*resultCallback)(void *ref, cifsResFile_t *rec), btu16 uid, btu16 tid, btuni *path, int maxResults, btsize resumeSize, btu8 *resumeKey)
{
	CIFS_TRACE("cifsSearch");
    smb->usedByCifs2 = resultCallback;
	smbStartCommand(smb, uid, 0, tid, cifsDecodeSearch, caller);
    smbStartNormalSmb(smb, SMB_COM_SEARCH);
    smbStartParameters(smb, -1);
    smbAddWord(smb, maxResults);
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM | CIFS_ATTR_DIR);
    smbStartData(smb);
    smbAddByte(smb, 4);
	if(resumeSize > 0){
		smbAddStringFromAscii(smb, "");
	}else{
		smbAddStringFromUni(smb, path);	/* doc says ASCII, but that's wrong */
	}
    smbAddByte(smb, 5);
    smbAddWord(smb, resumeSize);
    smbAddBlock(smb, resumeKey, resumeSize);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeFindFirstUnix(void *ref, struct smbRequest *smb)
{
cifsResFile_t		entry;
cifsResReaddir_t	res;
btu8				*p, *q, *base;
int					i, nameLen;
btuni				name[256];

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    res.searchId = smbGetWord(&p);
    res.entryCount = smbGetWord(&p);
    res.eof = smbGetWord(&p) != 0;
    p += 4;	/* skip EaErrorOffset and LastNameOffset */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeFindFirstUnix");
    p = base = smb->rxBuffer + smb->rxDataBase;
    for(i=0;i<res.entryCount;i++){
        q = p + smbGetLong(&p);	/* offset to next */
        p += 52;	/* skip resume key, dates and sizes */
		entry.infoLevel = SMB_FIND_FILE_UNIX;
        entry.unixType = smbGetLong(&p);
        p += 16;	/* skip dev major and minor */
        entry.unixInum = smbGetLarge(&p);
        p += 16;	/* skip permissions and nlinks */
        nameLen = *p++;
        if((res.flags2 & CIFS_FLAG2_UNICODE) != 0){
            if(((int)p & 1) != 0)
                p++;	/* align to word boundary */
            entry.name = smbGetUniString(&p);
        }else{
            convServerAsciiToUni(p, name, 256);
        	entry.name = name;
        }
        p = q;
        if(p - base > smb->rxDataBytes)
            DECODER_ERROR("Not enough data in cifsDecodeFindFirstUnix");
        if(!((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry))
            break;
    }
	DECODER_FINISH
}

static void	cifsDecodeFindFirstNT(void *ref, struct smbRequest *smb)
{
cifsResFile_t		entry;
cifsResReaddir_t	res;
btu8				*p, *base, *q, *s, store[2];
int					i, nameLen, offset;
btuni				name[1024];
btbool				doContinue;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    if((smb->usedByCifs3 & 1) == 0)
    	res.searchId = smbGetWord(&p);
    res.entryCount = smbGetWord(&p);
    res.eof = smbGetWord(&p) != 0;
    p += 4;	/* skip EaErrorOffset and LastNameOffset */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeFindFirstNT");
    p = base = smb->rxBuffer + smb->rxDataBase;
    for(i=0;i<res.entryCount;i++){
		q = p;
		offset = smbGetLong(&p);
		entry.resumeKey = smbGetLong(&p);
		entry.infoLevel = SMB_FIND_FILE_DIRECTORY_INFO;
		entry.crtime64 = smbGetLarge(&p);
		entry.atime64 = smbGetLarge(&p);
		entry.wtime64 = smbGetLarge(&p);
		entry.chtime64 = smbGetLarge(&p);
		entry.size = smbGetLarge(&p);
		entry.allocSize = smbGetLarge(&p);
		entry.dosMode = smbGetLong(&p);
		nameLen = smbGetLong(&p);
		if(nameLen > 1023)
			nameLen = 1023;	/* emergency brake */
        if((res.flags2 & CIFS_FLAG2_UNICODE) != 0){
			if(((int)&p[nameLen] & 1) != 0)
                nameLen++;	/* align end of string to word boundary */
			if(((int)p & 1) != 0)
                p++;	/* align to word boundary */
			s = p;
			if(s - base + nameLen > smb->rxDataBytes)
				DECODER_ERROR("Not enough data in cifsDecodeFindFirstNT");
			store[0] = p[nameLen];
			store[1] = p[nameLen+1];
			p[nameLen] = 0;	/* 0-terminate string */
			p[nameLen+1] = 0;
            entry.name = smbGetUniString(&p);
			doContinue = ((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry);
			s[nameLen] = store[0];	/* restore original data */
			s[nameLen+1] = store[1];
       }else{
			s = p;
			if(s - base + nameLen > smb->rxDataBytes)
				DECODER_ERROR("Not enough data in cifsDecodeFindFirstNT");
			store[0] = p[nameLen];
			p[nameLen] = 0;	/* 0-terminate string */
            convServerAsciiToUni(p, name, 1024);
            entry.name = name;
			doContinue = ((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry);
			s[nameLen] = store[0];	/* restore original data */
        }
        if(!doContinue)
            break;
		p = q + offset;
    }
	DECODER_FINISH
}

static void	cifsDecodeFindFirstNTBoth(void *ref, struct smbRequest *smb)
{
cifsResFile_t		entry;
cifsResReaddir_t	res;
btu8				*p, *base, *q, *s, store[2];
int					i, nameLen, offset;
btuni				name[1024];
btbool				doContinue;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    if((smb->usedByCifs3 & 1) == 0)
    	res.searchId = smbGetWord(&p);
    res.entryCount = smbGetWord(&p);
    res.eof = smbGetWord(&p) != 0;
    p += 4;	/* skip EaErrorOffset and LastNameOffset */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeFindFirstNTBoth");
    p = base = smb->rxBuffer + smb->rxDataBase;
    for(i=0;i<res.entryCount;i++){
		q = p;
		offset = smbGetLong(&p);
		entry.resumeKey = smbGetLong(&p);
		entry.infoLevel = SMB_FIND_FILE_BOTH_DIRECTORY_INFO;
		entry.crtime64 = smbGetLarge(&p);
		entry.atime64 = smbGetLarge(&p);
		entry.wtime64 = smbGetLarge(&p);
		entry.chtime64 = smbGetLarge(&p);
		entry.size = smbGetLarge(&p);
		entry.allocSize = smbGetLarge(&p);
		entry.dosMode = smbGetLong(&p);
		nameLen = smbGetLong(&p);
		if(nameLen > 1023)
			nameLen = 1023;	/* emergency brake */
		p += 30;	/* skip EaSize, ShortNameLength, Reserved and ShortNmae */
        if((res.flags2 & CIFS_FLAG2_UNICODE) != 0){
			if(((int)&p[nameLen] & 1) != 0)
                nameLen++;	/* align end of string to word boundary */
			if(((int)p & 1) != 0)
                p++;	/* align to word boundary */
			s = p;
			if(s - base + nameLen > smb->rxDataBytes)
				DECODER_ERROR("Not enough data in cifsDecodeFindFirstNTBoth");
			store[0] = p[nameLen];
			store[1] = p[nameLen+1];
			p[nameLen] = 0;	/* 0-terminate string */
			p[nameLen+1] = 0;
            entry.name = smbGetUniString(&p);
			doContinue = ((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry);
			s[nameLen] = store[0];	/* restore original data */
			s[nameLen+1] = store[1];
       }else{
			s = p;
			if(s - base + nameLen > smb->rxDataBytes)
				DECODER_ERROR("Not enough data in cifsDecodeFindFirstNTBoth");
			store[0] = p[nameLen];
			p[nameLen] = 0;	/* 0-terminate string */
            convServerAsciiToUni(p, name, 1024);
            entry.name = name;
			doContinue = ((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry);
			s[nameLen] = store[0];	/* restore original data */
        }
        if(!doContinue)
            break;
		p = q + offset;
    }
	DECODER_FINISH
}


static void	cifsDecodeFindFirst(void *ref, struct smbRequest *smb)
{
cifsResFile_t		entry;
cifsResReaddir_t	res;
btu8				*p, *base, *q;
int					i, nameLen;
btuni				name[256];

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    if((smb->usedByCifs3 & 1) == 0)
    	res.searchId = smbGetWord(&p);
    res.entryCount = smbGetWord(&p);
    res.eof = smbGetWord(&p) != 0;
    p += 4;	/* skip EaErrorOffset and LastNameOffset */
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeFindFirst");
    p = base = smb->rxBuffer + smb->rxDataBase;
    for(i=0;i<res.entryCount;i++){
		entry.infoLevel = SMB_INFO_STANDARD;
   	 	if((smb->usedByCifs3 & 2) != 0)
			entry.resumeKey = smbGetLong(&p);
		entry.cdate = smbGetWord(&p);
		entry.ctime = smbGetWord(&p);
		entry.adate = smbGetWord(&p);
		entry.atime = smbGetWord(&p);
		entry.wdate = smbGetWord(&p);
		entry.wtime = smbGetWord(&p);
		entry.size = smbGetLong(&p);
		entry.allocSize = smbGetLong(&p);
		entry.dosMode = smbGetWord(&p);
        nameLen = *p++;	/* nameLen does not include terminating 0 */
        if((res.flags2 & CIFS_FLAG2_UNICODE) != 0){
            if(((int)p & 1) != 0)
                p++;	/* align to word boundary */
			if(nameLen > 1 && p[nameLen-1] == 0 && p[nameLen-2] == 0)
				nameLen -= 2;	/* netApp gives length including terminating 0 */
			if(p - base + nameLen + 2 > smb->rxDataBytes)
				DECODER_ERROR("Not enough data in cifsDecodeFindFirst");
			p[nameLen] = p[nameLen + 1] = 0;	/* NT does not always 0-terminate strings */
			q = p;
            entry.name = smbGetUniString(&p);
        	p = q + nameLen + 2;				/* there's always room for the terminating 0 */
        }else{
			if(nameLen > 0 && p[nameLen-1] == 0)
				nameLen--;	/* netApp gives length including terminating 0 */
			if(p - base + nameLen + 1 > smb->rxDataBytes)
				DECODER_ERROR("Not enough data in cifsDecodeFindFirst");
			p[nameLen] = 0;	/* NT does not always 0-terminate strings */
            convServerAsciiToUni(p, name, nameLen + 1);
            entry.name = name;
        	p += nameLen + 1;					/* there's always room for the terminating 0 */
        }
        if(p - base > smb->rxDataBytes)
            DECODER_ERROR("Not enough data in cifsDecodeFindFirst");
        if(!((btbool (*)(void *ref, cifsResFile_t *rec))smb->usedByCifs2)(ref, &entry))
            break;
    }
	DECODER_FINISH
}

void	cifsFindFirst(struct smbRequest *smb, void *caller, btbool (*resultCallback)(void *ref, cifsResFile_t *rec), btu16 uid, btu16 tid, btuni *path, int maxResults, btint infoLevel, btbool autoclose, btbool useResumeKeys)
{
btint	flag;

	CIFS_TRACE("cifsFindFirst");
    smb->usedByCifs2 = resultCallback;
    smb->usedByCifs3 = useResumeKeys ? 2 : 0;
	switch(infoLevel){
	case SMB_INFO_STANDARD:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirst, caller);
		break;
	case SMB_FIND_FILE_UNIX:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirstUnix, caller);
		break;
	case SMB_FIND_FILE_DIRECTORY_INFO:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirstNT, caller);
		break;
	case SMB_FIND_FILE_BOTH_DIRECTORY_INFO:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirstNTBoth, caller);
		break;
	default:
		logPrintf(LOGM_ERROR, "*** cifsFind: illegal info level %d\n", infoLevel);
		break;
	}
    smbStartTrans2Smb(smb, TRANS2_FIND_FIRST2, 64, 32768, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM | CIFS_ATTR_DIR);
    smbAddWord(smb, maxResults);
	flag = autoclose ? 2 : 0;
	if(useResumeKeys)
		flag |= 4;
    smbAddWord(smb, flag);	/* automatic close on EOF */
    smbAddWord(smb, infoLevel);	/* information level */
    smbAddLong(smb, 0);	/* what's SearchStorageType ??? */
    smbAddStringFromUni(smb, path);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

void	cifsFindNext(struct smbRequest *smb, void *caller, btbool (*resultCallback)(void *ref, cifsResFile_t *rec), btu16 uid, btu16 tid, btuni *path, int maxResults, btu16 searchId, btint infoLevel, btbool autoclose, btbool useResumeKeys, btu32 resumeKey)
{
btint	flag;

	CIFS_TRACE("cifsFindNext");
    smb->usedByCifs2 = resultCallback;
    smb->usedByCifs3 = useResumeKeys ? 3 : 1;
	switch(infoLevel){
	case SMB_INFO_STANDARD:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirst, caller);
		break;
	case SMB_FIND_FILE_UNIX:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirstUnix, caller);
		break;
	case SMB_FIND_FILE_DIRECTORY_INFO:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirstNT, caller);
		break;
	case SMB_FIND_FILE_BOTH_DIRECTORY_INFO:
		smbStartCommand(smb, uid, 0, tid, cifsDecodeFindFirstNTBoth, caller);
		break;
	default:
		logPrintf(LOGM_ERROR, "*** cifsFind: illegal info level %d\n", infoLevel);
		break;
	}
    /* Share the decoder with the findfirst command. The difference is in usedByCifs3! */
    smbStartTrans2Smb(smb, TRANS2_FIND_NEXT2, 64, 32768, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, searchId);
    smbAddWord(smb, maxResults);
    smbAddWord(smb, infoLevel);
    smbAddLong(smb, resumeKey);		/* resume key */
	flag = autoclose ? 2 : 0;
	flag |= useResumeKeys ? 4 : 8;
    smbAddWord(smb, flag);	/* [automatic close on EOF] / resume from previous ending place / useResumeKeys */
    smbAddStringFromAscii(smb, "");	/* no resume file name */
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeFindClose(void *ref, struct smbRequest *smb)
{
cifsRes_t	res;
    
    DECODER_START(1)
	DECODER_FINISH
}

void	cifsFindClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 searchId)
{
	CIFS_TRACE("cifsFindClose");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeFindClose, caller);
    smbStartNormalSmb(smb, SMB_COM_FIND_CLOSE2);
    smbStartParameters(smb, -1);
    smbAddWord(smb, searchId);
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static int	decodeOplock(int flags)
{
int	oplock = CIFS_OPLOCK_NONE;

	if(flags & CIFS_FLAG_EXCLUSIVE_LOCK){
		oplock = CIFS_OPLOCK_EXCLUSIVE;
		if(flags & CIFS_FLAG_BATCH_LOCK){
			oplock = CIFS_OPLOCK_BATCH;
		}
	}
	return oplock;
}

static void	cifsDecodeOpenCore(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;
btu8			*p, *base;

	DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    
	res.oplock = 0; /*decodeOplock(res.flags); oplocks don't work with core open*/
    res.fid = smbGetWord(&p);
    res.attributes = smbGetWord(&p);
    res.wtime = smbGetLong(&p);
    res.size = smbGetLong(&p);
    res.grantedWrite = (smbGetWord(&p) & SMB_ACC_MASK_ACCESS) == SMB_ACC_RW;
    res.didOpen = 1;

    if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeOpenCore");
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))
		logPrintf(LOGM_CIFSTRACE, "cifs: openCore fid=%d\n", res.fid);
	DECODER_FINISH
}

void	cifsOpenCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool readonly)
{
	CIFS_TRACE("cifsOpenCore");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenCore, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN);
    smbStartParameters(smb, -1);
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | (readonly ? SMB_ACC_READ : SMB_ACC_RW));
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static btbool	cifsSubdecodeOpen(smbRequest_t *smb, cifsResFileOp_t *res)
{
btu8			*p, *base;

    p = base = smb->rxBuffer + smb->rxParameterBase;
    p += 4;	/* skip andx stuff */
    res->fid = smbGetWord(&p);
    res->attributes = smbGetWord(&p);
    res->wtime = smbGetLong(&p);
    res->size = smbGetLong(&p);
    res->grantedWrite = (smbGetWord(&p) & SMB_ACC_MASK_ACCESS) == SMB_ACC_RW;
    p += 4;	/* skip file type and state of named pipe */
    p += 2;	/* skip action taken */
	res->oplock = decodeOplock(res->flags);
	res->didOpen = 1;
    if(p - base > smb->rxParameterBytes)
		SUBDECODER_ERROR("Not enough parameters in cifsDecodeOpen");
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))
		logPrintf(LOGM_CIFSTRACE, "cifs: open fid=%d\n", res->fid);
    return 0;
}

static void	cifsDecodeOpen(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
    if(cifsSubdecodeOpen(smb, &res))
        goto decode_exception;
	DECODER_FINISH
}

void	cifsOpen(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool readonly, btbool truncate, btbool create, btu16 attributes, btu32 ctime, btoffset truncateLen)
{
btu16	openFunction = 0;
    
	CIFS_TRACE("cifsOpen");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeOpen, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN_ANDX);
    smbStartParameters(smb, 0xff);
    smbAddWord(smb, srvOpLock(smb->server) | 1);	/* request additional infos */
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | (readonly ? SMB_ACC_READ : SMB_ACC_RW));
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbAddWord(smb, attributes);	/* create attributes */
    smbAddLong(smb, ctime);			/* creation time */
	openFunction = truncate ? SMB_OPEN_TRUNCATE : SMB_OPEN_OPEN;
	if(create)
		openFunction |= SMB_OPEN_CREATE;
    smbAddWord(smb, openFunction);
    smbAddLong(smb, (btu32)truncateLen);	/* truncate length */
    smbAddLong(smb, 0);	/* reserved */
    smbAddLong(smb, 0);	/* reserved */
    smbStartData(smb);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static btbool	cifsSubdecodeClose(smbRequest_t *smb, cifsResFileOp_t *res)
{
    res->didClose = 1;
    return 0;
}

static void	cifsDecodeClose(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
    cifsSubdecodeClose(smb, &res);
	DECODER_FINISH
}

void	cifsClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu32 mtime)
{
	CIFS_TRACE_I("cifsClose", fid);
	smbStartCommand(smb, uid, 0, tid, cifsDecodeClose, caller);
    smbStartNormalSmb(smb, SMB_COM_CLOSE);
    smbStartParameters(smb, -1);
    smbAddWord(smb, fid);
    smbAddLong(smb, mtime);
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeReadCore(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;
btu8			*p, *base;
btu16			len;

	DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    res.dataLen = smbGetWord(&p);
    if(res.dataLen > smb->usedByCifs3)	/* more bytes read than requested */
        res.dataLen = smb->usedByCifs3;
    res.didRW = 1;
    if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeReadCore");
    p = base = smb->rxBuffer + smb->rxDataBase;
    p++;	/* skip buffer format = 0x01 */
    len = smbGetWord(&p);	/* read data len */
    res.data = p;
    p += res.dataLen;
    if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) && len != res.dataLen){	/* why encoded twice? */
        logPrintf(LOGM_DEBUG, "cifs decodeReadCore: inconsistent len: %d vs %d\n", res.dataLen, len);
    }
    if(p - base > smb->rxDataBytes)
		DECODER_ERROR("Not enough data in cifsDecodeReadCore");
	DECODER_FINISH
}

void	cifsReadCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btoffset fileOffset, btu32 len)
{
	CIFS_TRACE_I("cifsReadCore", fid);
	smbStartCommand(smb, uid, 0, tid, cifsDecodeReadCore, caller);
    smbStartNormalSmb(smb, SMB_COM_READ);
    smbStartParameters(smb, -1);
    smbAddWord(smb, fid);
    smbAddWord(smb, len);
    smbAddLong(smb, fileOffset);
    smbAddWord(smb, 0);	/* don't know how much will be read next time */
    smb->usedByCifs3 = len;
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static btbool	cifsSubdecodeRead(smbRequest_t *smb, cifsResFileOp_t *res)
{
btu8			*p, *base;
int				dataOffset;

	p = base = smb->rxBuffer + smb->rxParameterBase;
    p += 4;	/* skip andX stuff */
    p += 6;	/* skip 'remaining', compaction mode and reserved parameter */
    res->dataLen = smbGetWord(&p);
    if(res->dataLen > smb->usedByCifs3)	/* more bytes read than requested */
        res->dataLen = smb->usedByCifs3;
    dataOffset = smbGetWord(&p);
    res->didRW = 1;
    if(p - base > smb->rxParameterBytes)
		SUBDECODER_ERROR("Not enough parameters in cifsDecodeReadCore");
    base = smb->rxBuffer + smb->rxDataBase;
    p = smb->rxBuffer + dataOffset;
    res->data = p;
    p += res->dataLen;
    if(p - base > smb->rxDataBytes)
		SUBDECODER_ERROR("Not enough data in cifsSubdecodeRead");
    return 0;
}

static void	cifsDecodeRead(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
    if(cifsSubdecodeRead(smb, &res))
        goto decode_exception;
	DECODER_FINISH
}

void	cifsRead(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btoffset fileOffset, btu32 len)
{
btoffset	offsetHigh;
    
	CIFS_TRACE_I("cifsRead", fid);
	smbStartCommand(smb, uid, 0, tid, cifsDecodeRead, caller);
    smbStartNormalSmb(smb, SMB_COM_READ_ANDX);
    smbStartParameters(smb, 0xff);
    smbAddWord(smb, fid);
    smbAddLong(smb, fileOffset);
    smbAddWord(smb, len);
    smbAddWord(smb, 0);	/* minCount */
    smbAddLong(smb, 0);	/* reserved */
    smbAddWord(smb, 0);	/* don't know how much will be read next time */
    offsetHigh = fileOffset >> 16;
    offsetHigh >>= 16;				/* do right shift in two steps to avoid compiler warning if btoffset == btu32 */
    if(offsetHigh != 0)
        smbAddLong(smb, offsetHigh);
    smb->usedByCifs3 = len;
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

void	cifsDecodeOpenRead(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(0)
    if(smbAndxCommand(smb) == SMB_COM_READ_ANDX){	/* open was processed, even on error */
    	if(cifsSubdecodeOpen(smb, &res))
        	goto decode_exception;
        if(res.err.code == 0){					/* no error */
			smbAdvanceToAndx(smb);
            if(cifsSubdecodeRead(smb, &res))
                goto decode_exception;
        }
    }
	DECODER_FINISH
}

void	cifsOpenRead(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btoffset fileOffset, btu32 len, btbool readonly)
{
btoffset	offsetHigh;

	CIFS_TRACE("cifsOpenRead");
    smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenRead, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN_ANDX);
/* open command */
    smbStartParameters(smb, SMB_COM_READ_ANDX);
    smbAddWord(smb, srvOpLock(smb->server) | 1);	/* request additional infos */
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | (readonly ? SMB_ACC_READ : SMB_ACC_RW));
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbAddWord(smb, 0);	/* create attributes */
    smbAddLong(smb, 0);	/* creation time */
    smbAddWord(smb, SMB_OPEN_OPEN);	/* open function: normal open, no create or truncate */
    smbAddLong(smb, 0);	/* truncate length */
    smbAddLong(smb, 0);	/* reserved */
    smbAddLong(smb, 0);	/* reserved */
    smbStartData(smb);
    smbAddStringFromUni(smb, path);
/* read command */
    smbStartParameters(smb, 0xff);
    smbAddWord(smb, -1);	/* implicitly taken from open */
    smbAddLong(smb, fileOffset);
    smbAddWord(smb, len);
    smbAddWord(smb, 0);	/* minCount */
    smbAddLong(smb, 0);	/* reserved */
    smbAddWord(smb, 0);	/* don't know how much will be read next time */
    offsetHigh = fileOffset >> 16;
    offsetHigh >>= 16;				/* do right shift in two steps to avoid compiler warning if btoffset == btu32 */
    if(offsetHigh != 0)
        smbAddLong(smb, offsetHigh);
    smb->usedByCifs3 = len;
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeOpenReadClose(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(0)
    if(smbAndxCommand(smb) == SMB_COM_READ_ANDX){	/* open was processed, even on error */
    	if(cifsSubdecodeOpen(smb, &res))
        	goto decode_exception;
        smbAdvanceToAndx(smb);
        if(smbAndxCommand(smb) == SMB_COM_CLOSE){	/* read was processed, even on error */
            if(cifsSubdecodeRead(smb, &res))
                goto decode_exception;
            if(res.err.code == 0){					/* no error */
				smbAdvanceToAndx(smb);
                if(cifsSubdecodeClose(smb, &res))
                    goto decode_exception;
            }
        }
    }
	DECODER_FINISH
}

void	cifsOpenReadClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btoffset fileOffset, btu32 len, btbool readonly)
{
btoffset	offsetHigh;

	CIFS_TRACE("cifsOpenReadClose");
    smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenReadClose, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN_ANDX);
/* open command */
    smbStartParameters(smb, SMB_COM_READ_ANDX);
    smbAddWord(smb, srvOpLock(smb->server) | 1);	/* request additional infos */
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | (readonly ? SMB_ACC_READ : SMB_ACC_RW));
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbAddWord(smb, 0);	/* create attributes */
    smbAddLong(smb, 0);	/* creation time */
    smbAddWord(smb, SMB_OPEN_OPEN);	/* open function: normal open, no create or truncate */
    smbAddLong(smb, 0);	/* truncate length */
    smbAddLong(smb, 0);	/* reserved */
    smbAddLong(smb, 0);	/* reserved */
    smbStartData(smb);
    smbAddStringFromUni(smb, path);
/* read command */
    smbStartParameters(smb, SMB_COM_CLOSE);
    smbAddWord(smb, -1);	/* implicitly taken from open */
    smbAddLong(smb, fileOffset);
    smbAddWord(smb, len);
    smbAddWord(smb, 0);	/* minCount */
    smbAddLong(smb, 0);	/* reserved */
    smbAddWord(smb, 0);	/* don't know how much will be read next time */
    offsetHigh = fileOffset >> 16;
    offsetHigh >>= 16;				/* do right shift in two steps to avoid compiler warning if btoffset == btu32 */
    if(offsetHigh != 0)
        smbAddLong(smb, offsetHigh);
    smb->usedByCifs3 = len;
    smbStartData(smb);
/* close command */
    smbStartParameters(smb, -1);
    smbAddWord(smb, -1);	/* implicitly taken from read */
    smbAddLong(smb, 0);		/* let the server set the modification date */
    smbStartData(smb);
	smbEndCommand(smb);
}


/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void	cifsDecodeWriteCore(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;
btu8			*p, *base;

	DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
    res.dataLen = smbGetWord(&p);
    res.didRW = 1;
    if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeWriteCore");
	DECODER_FINISH
}

void	cifsWriteCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu8 *buffer, btoffset fileOffset, btu32 len)
{
	CIFS_TRACE_I("cifsWriteCore", fid);
	smbStartCommand(smb, uid, 0, tid, cifsDecodeWriteCore, caller);
    smbStartNormalSmb(smb, SMB_COM_WRITE);
    smbStartParameters(smb, -1);
    smbAddWord(smb, fid);
    smbAddWord(smb, len);
    smbAddLong(smb, fileOffset);
    smbAddWord(smb, 0);	/* don't know how much will be written next time */
    smbStartData(smb);
    smbAddByte(smb, 1);
    smbAddWord(smb, len);
    smbAddBlock(smb, buffer, len);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static btbool	cifsSubdecodeWrite(smbRequest_t *smb, cifsResFileOp_t *res)
{
btu8			*p, *base;

	p = base = smb->rxBuffer + smb->rxParameterBase;
    p += 4;	/* skip andX stuff */
    res->dataLen = smbGetWord(&p);
    res->didRW = 1;
    if(p - base > smb->rxParameterBytes)
		SUBDECODER_ERROR("Not enough parameters in cifsSubdecodeWrite");
    return 0;
}

static void	cifsDecodeWrite(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
    if(cifsSubdecodeWrite(smb, &res))
        goto decode_exception;
	DECODER_FINISH
}

void	cifsWrite(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu8 *buffer, btoffset fileOffset, btu32 len, btbool writeThrough)
{
btoffset	offsetHigh;
smbVar_t	dataStart;
    
	CIFS_TRACE_I("cifsWrite", fid);
	smbStartCommand(smb, uid, 0, tid, cifsDecodeWrite, caller);
    smbStartNormalSmb(smb, SMB_COM_WRITE_ANDX);
    smbStartParameters(smb, 0xff);
    smbAddWord(smb, fid);
    smbAddLong(smb, fileOffset);
    smbAddLong(smb, 0);	/* reserved */
    smbAddWord(smb, writeThrough ? 1 : 0);	/* write mode */
    smbAddWord(smb, 0);	/* remaining */
    smbAddWord(smb, 0);	/* reserved */
    smbAddWord(smb, len);
    smbAddVariable(smb, &dataStart, 2, 0);
    offsetHigh = fileOffset >> 16;
    offsetHigh >>= 16;				/* do right shift in two steps to avoid compiler warning if btoffset == btu32 */
    if(offsetHigh != 0)
        smbAddLong(smb, offsetHigh);
    smbStartData(smb);
    smbAddAlignment(smb, 4);
    smbWriteVariable(smb, &dataStart, smbCurrentOffset(smb));
    smbAddBlock(smb, buffer, len);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

void	cifsDecodeOpenWrite(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(0)
    if(smbAndxCommand(smb) == SMB_COM_WRITE_ANDX){	/* open was processed, even on error */
    	if(cifsSubdecodeOpen(smb, &res))
        	goto decode_exception;
        if(res.err.code == 0){					/* no error */
			smbAdvanceToAndx(smb);
            if(cifsSubdecodeWrite(smb, &res))
                goto decode_exception;
        }
    }
	DECODER_FINISH
}

void	cifsOpenWrite(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu8 *buffer, btoffset fileOffset, btu32 len, btbool readonly, btbool writeThrough)
{
btoffset	offsetHigh;
smbVar_t	dataStart;

	CIFS_TRACE("cifsOpenWrite");
    smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenWrite, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN_ANDX);
/* open command */
    smbStartParameters(smb, SMB_COM_WRITE_ANDX);
    smbAddWord(smb, srvOpLock(smb->server) | 1);	/* request additional infos */
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | (readonly ? SMB_ACC_READ : SMB_ACC_RW));
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbAddWord(smb, 0);	/* create attributes */
    smbAddLong(smb, 0);	/* creation time */
    smbAddWord(smb, SMB_OPEN_OPEN);	/* open function: normal open, no create or truncate */
    smbAddLong(smb, 0);	/* truncate length */
    smbAddLong(smb, 0);	/* reserved */
    smbAddLong(smb, 0);	/* reserved */
    smbStartData(smb);
    smbAddStringFromUni(smb, path);
/* write command */
    smbStartParameters(smb, 0xff);
    smbAddWord(smb, -1);	/* implicitly taken from open */
    smbAddLong(smb, fileOffset);
    smbAddLong(smb, 0);	/* reserved */
    smbAddWord(smb, writeThrough ? 1 : 0);	/* write mode */
    smbAddWord(smb, 0);	/* remaining */
    smbAddWord(smb, 0);	/* reserved */
    smbAddWord(smb, len);
    smbAddVariable(smb, &dataStart, 2, 0);
    offsetHigh = fileOffset >> 16;
    offsetHigh >>= 16;				/* do right shift in two steps to avoid compiler warning if btoffset == btu32 */
    if(offsetHigh != 0)
        smbAddLong(smb, offsetHigh);
    smbStartData(smb);
    smbAddAlignment(smb, 4);
    smbWriteVariable(smb, &dataStart, smbCurrentOffset(smb));
    smbAddBlock(smb, buffer, len);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeOpenWriteClose(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(0)
    if(smbAndxCommand(smb) == SMB_COM_WRITE_ANDX){	/* open was processed, even on error */
    	if(cifsSubdecodeOpen(smb, &res))
        	goto decode_exception;
        smbAdvanceToAndx(smb);
        if(smbAndxCommand(smb) == SMB_COM_CLOSE){	/* read was processed, even on error */
            if(cifsSubdecodeWrite(smb, &res))
                goto decode_exception;
            if(res.err.code == 0){					/* no error */
				smbAdvanceToAndx(smb);
                if(cifsSubdecodeClose(smb, &res))
                    goto decode_exception;
            }
        }
    }
	DECODER_FINISH
}

void	cifsOpenWriteClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu8 *buffer, btoffset fileOffset, btu32 len, btbool readonly, btbool writeThrough)
{
btoffset	offsetHigh;
smbVar_t	dataStart;

	CIFS_TRACE("cifsOpenWriteClose");
    smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenWriteClose, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN_ANDX);
/* open command */
    smbStartParameters(smb, SMB_COM_WRITE_ANDX);
    smbAddWord(smb, srvOpLock(smb->server) | 1);	/* request additional infos */
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | (readonly ? SMB_ACC_READ : SMB_ACC_RW));
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbAddWord(smb, 0);	/* create attributes */
    smbAddLong(smb, 0);	/* creation time */
    smbAddWord(smb, SMB_OPEN_OPEN);	/* open function: normal open, no create or truncate */
    smbAddLong(smb, 0);	/* truncate length */
    smbAddLong(smb, 0);	/* reserved */
    smbAddLong(smb, 0);	/* reserved */
    smbStartData(smb);
    smbAddStringFromUni(smb, path);
/* write command */
    smbStartParameters(smb, SMB_COM_CLOSE);
    smbAddWord(smb, -1);	/* implicitly taken from open */
    smbAddLong(smb, fileOffset);
    smbAddLong(smb, 0);	/* reserved */
    smbAddWord(smb, writeThrough ? 1 : 0);	/* write mode */
    smbAddWord(smb, 0);	/* remaining */
    smbAddWord(smb, 0);	/* reserved */
    smbAddWord(smb, len);
    smbAddVariable(smb, &dataStart, 2, 0);
    offsetHigh = fileOffset >> 16;
    offsetHigh >>= 16;				/* do right shift in two steps to avoid compiler warning if btoffset == btu32 */
    if(offsetHigh != 0)
        smbAddLong(smb, offsetHigh);
    smbStartData(smb);
    smbAddAlignment(smb, 4);
    smbWriteVariable(smb, &dataStart, smbCurrentOffset(smb));
    smbAddBlock(smb, buffer, len);
/* close command */
    smbStartParameters(smb, -1);
    smbAddWord(smb, -1);	/* implicitly taken from read */
    smbAddLong(smb, 0);		/* let the server set the modification date */
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void cifsDecodeSetInformation(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
	DECODER_FINISH
}

void	cifsSetInformation(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu16 attributes, btu32 wtime)
{
int		i;
    
	CIFS_TRACE("cifsSetInformation");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeSetInformation, caller);
    smbStartNormalSmb(smb, SMB_COM_SET_INFORMATION);
    smbStartParameters(smb, -1);
    smbAddWord(smb, attributes);
    smbAddLong(smb, wtime);
    for(i=0;i<5;i++){	/* 5 reserved words */
        smbAddWord(smb, 0);
    }
    smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void cifsDecodeSetPathInfo(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
    /* ignore the reply since it is not documented */
	DECODER_FINISH
}

void	cifsSetPathInfoStd(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu16 attributes, btu16 atime, btu16 adate, btu16 mtime, btu16 mdate, btu16 ctime, btu16 cdate)
{
	CIFS_TRACE("cifsSetPathInfoStd");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeSetPathInfo, caller);
    smbStartTrans2Smb(smb, TRANS2_SET_PATH_INFORMATION, 64, 2048, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, SMB_INFO_STANDARD);
    smbAddLong(smb, 0);
    smbAddStringFromUni(smb, path);
	smbStartData(smb);
    smbAddWord(smb, cdate);
    smbAddWord(smb, ctime);
    smbAddWord(smb, adate);
    smbAddWord(smb, atime);
    smbAddWord(smb, mdate);
    smbAddWord(smb, mtime);
    smbAddLong(smb, 0);	/* file size */
    smbAddLong(smb, 0);	/* allocation size */
    smbAddWord(smb, attributes);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void cifsDecodeSetFileInfo(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
    /* ignore the reply since it is not documented */
	DECODER_FINISH
}

void	cifsSetFileInfoStd(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu16 attributes, btu16 atime, btu16 adate, btu16 mtime, btu16 mdate, btu16 ctime, btu16 cdate)
{
	CIFS_TRACE("cifsSetFileInfoStd");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeSetFileInfo, caller);
    smbStartTrans2Smb(smb, TRANS2_SET_FILE_INFORMATION, 64, 2048, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, fid);
    smbAddWord(smb, SMB_INFO_STANDARD);
    smbAddWord(smb, 0);
	smbStartData(smb);
    smbAddWord(smb, cdate);
    smbAddWord(smb, ctime);
    smbAddWord(smb, adate);
    smbAddWord(smb, atime);
    smbAddWord(smb, mdate);
    smbAddWord(smb, mtime);
    smbAddLong(smb, 0);	/* file size */
    smbAddLong(smb, 0);	/* allocation size */
    smbAddWord(smb, attributes);
	smbEndCommand(smb);
}


/* ------------------------------------------------------------------------- */

void	cifsSetInformation2(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu16 atime, btu16 adate, btu16 mtime, btu16 mdate)
{
	CIFS_TRACE("cifsSetInformation2");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeSetPathInfo, caller);
    smbStartNormalSmb(smb, SMB_COM_SET_INFORMATION2);
    smbStartParameters(smb, -1);
    smbAddWord(smb, fid);
    smbAddWord(smb, 0);	/* cdate */
    smbAddWord(smb, 0);	/* ctime */
    smbAddWord(smb, adate);
    smbAddWord(smb, atime);
    smbAddWord(smb, mdate);
    smbAddWord(smb, mtime);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

void	cifsSetPathInfoUnix(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu32 mode, bti64 atime, bti64 mtime, bti64 fileUid, bti64 fileGid)
{
	CIFS_TRACE("cifsSetPathInfoUnix");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeSetPathInfo, caller);
    smbStartTrans2Smb(smb, TRANS2_SET_PATH_INFORMATION, 64, 2048, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, SMB_QUERY_FILE_UNIX_BASIC);
    smbAddLong(smb, 0);
    smbAddStringFromUni(smb, path);
	smbStartData(smb);
	smbAddLarge(smb, 0);	/* file size */
	smbAddLarge(smb, 0);	/* change time */
	smbAddLarge(smb, atime);
	smbAddLarge(smb, mtime);
	smbAddLarge(smb, fileUid);
	smbAddLarge(smb, fileGid);
	smbAddLong(smb, 0);		/* type of file: can't change */
	smbAddLarge(smb, 0);	/* dev major */
	smbAddLarge(smb, 0);	/* dev minor */
	smbAddLarge(smb, 0);	/* inode number */
	smbAddLarge(smb, mode);
	smbAddLarge(smb, 0);	/* number of links */
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void cifsDecodeCreateFileCore(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;
btu8			*p, *base;

	DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	res.oplock = decodeOplock(res.flags);
    res.fid = smbGetWord(&p);
    res.didOpen = 1;
    if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in cifsDecodeCreateFileCore");
	DECODER_FINISH
}

void	cifsCreateFileCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu16 attributes, btu32 ctime)
{
	CIFS_TRACE("cifsCreateFileCore");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeCreateFileCore, caller);
    smbStartNormalSmb(smb, SMB_COM_CREATE_NEW);
    smbStartParameters(smb, -1);
    smbAddWord(smb, attributes);
    smbAddLong(smb, ctime);
    smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void cifsDecodeMkdirCore(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
	DECODER_FINISH
}

void	cifsMkdirCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path)
{
	CIFS_TRACE("cifsMkdirCore");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeMkdirCore, caller);
    smbStartNormalSmb(smb, SMB_COM_CREATE_DIRECTORY);
    smbStartParameters(smb, -1);
    smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	cifsDecodeOpenClose(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(0)
    if(smbAndxCommand(smb) == SMB_COM_CLOSE){	/* open was processed, even on error */
    	if(cifsSubdecodeOpen(smb, &res))
        	goto decode_exception;
        if(res.err.code == 0){					/* no error */
            smbAdvanceToAndx(smb);
            if(cifsSubdecodeClose(smb, &res))
                goto decode_exception;
        }
    }
	DECODER_FINISH
}

void	cifsOpenClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool truncate, btbool create, btu16 attributes, btu32 ctime, btoffset truncateLen)
{
btu16	openFunction;
    
	CIFS_TRACE("cifsOpenClose");
    smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenClose, caller);
    smbStartNormalSmb(smb, SMB_COM_OPEN_ANDX);
/* open command */
    smbStartParameters(smb, SMB_COM_CLOSE);

    smbAddWord(smb, 1);	/* request additional infos */
    smbAddWord(smb, SMB_ACC_SHARE_DENY_NONE | SMB_ACC_RW);
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    smbAddWord(smb, attributes);	/* create attributes */
    smbAddLong(smb, ctime);			/* creation time */
	openFunction = truncate ? SMB_OPEN_TRUNCATE : SMB_OPEN_OPEN;
	if(create)
		openFunction |= SMB_OPEN_CREATE;
    smbAddWord(smb, openFunction);
    smbAddLong(smb, (btu32)truncateLen);	/* truncate length */
    smbAddLong(smb, 0);	/* reserved */
    smbAddLong(smb, 0);	/* reserved */
    smbStartData(smb);
    smbAddStringFromUni(smb, path);
/* close command */
    smbStartParameters(smb, -1);
    smbAddWord(smb, -1);	/* fid implicitly taken from read */
    smbAddLong(smb, ctime);	/* set the modification date */
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void cifsDecodeRemoveCore(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
	DECODER_FINISH
}

void	cifsRemoveCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool isDir)
{
	CIFS_TRACE("cifsRemoveCore");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeRemoveCore, caller);
    smbStartNormalSmb(smb, isDir ? SMB_COM_DELETE_DIRECTORY : SMB_COM_DELETE);
    smbStartParameters(smb, -1);
    if(!isDir){
        smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM);
    }
    smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, path);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void cifsDecodeRename(void *ref, struct smbRequest *smb)
{
cifsResFileOp_t	res;

	DECODER_START(1)
	DECODER_FINISH
}

void	cifsRename(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *frompath, btuni *topath)
{
	CIFS_TRACE("cifsRename");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeRename, caller);
    smbStartNormalSmb(smb, SMB_COM_RENAME);
    smbStartParameters(smb, -1);
    smbAddWord(smb, CIFS_ATTR_HIDDEN | CIFS_ATTR_SYSTEM | CIFS_ATTR_DIR);
    smbStartData(smb);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, frompath);
    smbAddByte(smb, 4);
    smbAddStringFromUni(smb, topath);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

void	cifsDecodeOplockBreak(struct smbHeader *hdr, btu8 *data, btsize len, btu16 *fid, int *lockMode)
{
btu8			*p, *base, lockType;

    p = base = data + hdr->parameterOffset;
	p += 4;	/* skip andX stuff */
	*fid = smbGetWord(&p);
	lockType = *p++;
	switch(*p++){
		case 0:		*lockMode = CIFS_OPLOCK_NONE;	break;
		default:	*lockMode = CIFS_OPLOCK_L2;		break;
	}
	if((lockType & LOCKING_ANDX_OPLOCK_RELEASE) == 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		logPrintf(LOGM_ERROR, "cifs: cifsDecodeOplockBreak: not an OPLOCK break (type=0x%02x)\n", lockType);
	}
	if(p - base > 2 * hdr->parameterWords && LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		logPrintf(LOGM_ERROR, "cifs: cifsDecodeOplockBreak: not enough parameters\n");
	}
}

void	cifsAckOplockBreak(struct smbRequest *smb, btu16 tid, btu16 fid, int lockLevel)
{
	CIFS_TRACE("cifsAckOplockBreak");
	smbStartCommand(smb, 0, 0, tid, NULL, NULL);
    smbStartNormalSmb(smb, SMB_COM_LOCKING_ANDX);
    smbStartParameters(smb, 0xff);
	smbAddWord(smb, fid);
	smbAddByte(smb, LOCKING_ANDX_OPLOCK_RELEASE);
	switch(lockLevel){
		default:
		case CIFS_OPLOCK_NONE:		lockLevel = 0;	break;
		case CIFS_OPLOCK_L2:		lockLevel = 1;	break;
		case CIFS_OPLOCK_EXCLUSIVE:	lockLevel = 1;	break;
		case CIFS_OPLOCK_BATCH:		lockLevel = 1;	break;
	}
	smbAddByte(smb, lockLevel);
	smbAddLong(smb, 0);		/* timeout */
	smbAddWord(smb, 0);		/* number of unlocks */
	smbAddWord(smb, 0);		/* number of locks */
    smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */


static void	cifsDecodeOpenNT(void *ref, struct smbRequest *smb)
{
cifsResOpenNT_t	res;
btu8			*p, *base;

    DECODER_START(1)
    p = base = smb->rxBuffer + smb->rxParameterBase;
	p += 4;		/* skip andX stuff */
	switch(*p++){
		default:
		case 0:	res.oplock = CIFS_OPLOCK_NONE;		break;
		case 1:	res.oplock = CIFS_OPLOCK_EXCLUSIVE;	break;
		case 2:	res.oplock = CIFS_OPLOCK_BATCH;		break;
		case 3:	res.oplock = CIFS_OPLOCK_L2;		break;
	}
	res.fid = smbGetWord(&p);
	res.createAction = smbGetLong(&p);
	res.ctime = smbGetLarge(&p);
	res.atime = smbGetLarge(&p);
	res.mtime = smbGetLarge(&p);
	res.lastChanged = smbGetLarge(&p);
	res.extAttr = smbGetLong(&p);
	res.allocSize = smbGetLarge(&p);
	res.size = smbGetLarge(&p);
	res.ftype = smbGetWord(&p);
	res.devState = smbGetWord(&p);
	res.isDir = *p++;
	if(p - base > smb->rxParameterBytes)
		DECODER_ERROR("Not enough parameters in NT_CREATE_ANDX");
	DECODER_FINISH
}

void	cifsOpenNT(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu32 parentFid, btuni *name, btbool doCreate)
{
smbVar_t	nameLen;
btsize		namebase;

	CIFS_TRACE("cifsOpenNT");
	smbStartCommand(smb, uid, 0, tid, cifsDecodeOpenNT, caller);
    smbStartNormalSmb(smb, SMB_COM_NT_CREATE_ANDX);
    smbStartParameters(smb, 0xff);
	smbAddByte(smb, 0);		/* reserved */
	smbAddVariable(smb, &nameLen, 2, 0);
	smbAddLong(smb, 2);		/* flags */
	smbAddLong(smb, parentFid);
	smbAddLong(smb, 0);		/* accessMask */
	smbAddLarge(smb, 0);	/* initial allocation size */
	smbAddLong(smb, 0);		/* extended attributes */
	smbAddLong(smb, SMB_NTCR_FILE_SHARE_READ | SMB_NTCR_FILE_SHARE_WRITE | SMB_NTCR_FILE_SHARE_DELETE);		/* share access */
	smbAddLong(smb, SMB_NTCR_OPEN_EXISTING);
	smbAddLong(smb, 0);		/* create options */
	smbAddLong(smb, 0);		/* security impersonation level */
	smbAddByte(smb, 0);		/* security flags */
	smbStartData(smb);
	namebase = smbCurrentOffset(smb);
	smbAddStringFromUni(smb, name);
	smbWriteVariable(smb, &nameLen, smbCurrentOffset(smb) - namebase);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
