/*
 * Name: cifs.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-20
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: cifs.h,v 1.2 1999/05/30 21:18:56 cs Exp $
 */

#ifndef __cifs_H_INCLUDED__
#define __cifs_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This module implements the marshalling and unmarshalling routines for all
SMB commands that are used. All functions are "processes" in the sense
defined by "proc.h". The types of the return values can most reliably be
found as the type of the variable 'res' in the implementation.
*/

/* ------------------------------------------------------------------------- */

#define	CIFS_DIALECT_CORE	0
#define	CIFS_DIALECT_LM2	1
#define	CIFS_DIALECT_NTLM	2
/* the CIFS_DIALECT_* macros must match the 'dialects[]' array in cifs.c,
 * function 'cifsNegotiate()'!
 */

#define	CIFS_GENERIC_RESULT		\
	btError	err;				\
	btu8	flags;				\
	btu16	flags2;				\
    btu16	uid;				\
    btu16	tid

/* ------------------------------------------------------------------------- */



typedef struct cifsRes{
    CIFS_GENERIC_RESULT;
}cifsRes_t;

typedef struct cifsResNegotiate{
    CIFS_GENERIC_RESULT;
    int		dialect;
	btu8	securityMode;
	btu16	maxMpx;
	btu16	maxVc;
	btu32	maxXmit;
	btu32	maxRaw;
	btu16	rawMode;
	btu32	sessionKey;
	btu32	capabilities;
	bti32	gmtOffset;
	btu16	encryptKeyLen;
	btu8	encryptKey[64];
}cifsResNegotiate_t;

typedef struct cifsResSessionSetup{
    CIFS_GENERIC_RESULT;
	btu16	action;
	btuni	*serverOS;
	btuni	*serverLanMan;
}cifsResSessionSetup_t;

typedef struct cifsResQueryFile{
    CIFS_GENERIC_RESULT;
    btu32	lastWrite;	/* for core command only */
    bti64	atime;		/* for unix extension / basic info only */
    bti64	mtime;		/* for unix extension / basic info only */
    bti64	ctime;		/* for unix extension / basic info only */
    btu16	atimeTime;	/* not for unix extension */
    btu16	atimeDate;	/* not for unix extension */
    btu16	mtimeTime;	/* not for unix extension */
    btu16	mtimeDate;	/* not for unix extension */
    btu16	ctimeTime;	/* not for unix extension */
    btu16	ctimeDate;	/* not for unix extension */
    btu64	size;
    btu64	allocSize;	/* only for trans2 except unix */
    btu16	attributes;	/* not for unix extension */
/* for unix extension only: */
    btu64	fileUid;
    btu64	fileGid;
    btu32	mode;
    int		type;
    btu32	inum;
    btu32	nlinks;
    btu32	devMajor;
    btu32	devMinor;
}cifsResQueryFile_t;

typedef struct cifsResQueryFsInfo{
    CIFS_GENERIC_RESULT;
	btu32	fileSystemId;
	btu32	sectorsPerAU;
	btu32	totalAU;
	btu32	availableAU;
	btu16	bytesPerSector;
}cifsResQueryFsInfo_t;

typedef struct cifsResFile{
    btuni	*name;
	btint	infoLevel;
	/* file attributes read from response: */
	btu32		dosMode;
	btoffset	size;
	btoffset	allocSize;
	btu16		wtime;
	btu16		wdate;
	btu16		ctime;
	btu16		cdate;
	btu16		atime;
	btu16		adate;
	bti64		wtime64;
	bti64		atime64;
	bti64		chtime64;
	bti64		crtime64;
	bti64		unixInum;
	btu32		unixType;
	btu32		resumeKey;
}cifsResFile_t;

typedef struct cifsResReaddir{
    CIFS_GENERIC_RESULT;
    int		entryCount;
    int		resumeSize;
    btu8	resumeKey[24];
    btbool	eof;
    btu16	searchId;
}cifsResReaddir_t;

typedef struct cifsResFileOp{
    CIFS_GENERIC_RESULT;
    btu8	oplock;		/* open: type of oplock granted */
    btu16	fid;		/* open: file ID */
    btu16	attributes;	/* open: file attributes */
    btu32	wtime;		/* open: last write time */
    btu32	size;		/* open: file size in bytes */
    btbool	grantedWrite;/* open: write access granted */
    btu8	*data;		/* read: pointer to data */
    btu32	dataLen;	/* number of bytes read/written */
    btbool	didOpen;	/* open executed (important in chained requests) */
    btbool	didRW;		/* read/write executed (important in chained requests) */
    btbool	didClose;	/* close executed (important in chained requests) */
}cifsResFileOp_t;


typedef struct cifsResOpenNT{
    CIFS_GENERIC_RESULT;
	btu8	oplock;
	btu16	fid;
	btu32	createAction;
	btu64	ctime;
	btu64	atime;
	btu64	mtime;
	btu64	lastChanged;
	btu32	extAttr;
	btu64	allocSize;
	btu64	size;
	btu16	ftype;
	btu16	devState;
	btbool	isDir;
}cifsResOpenNT_t;

/* ------------------------------------------------------------------------- */

struct smbRequest;
struct smbHeader;

void	cifsNegotiate(struct smbRequest *smb, void *caller);
void	cifsSessionSetup(struct smbRequest *smb, void *caller, btu16 uid, btuni *username, btuni *plainPw, btu8 *lmHash, btu8 *ntHash, btbool pwIsEncrypted, btu8 *domain);
void	cifsLogoff(struct smbRequest *smb, void *caller, btu16 uid);
void	cifsTreeConnect(struct smbRequest *smb, void *caller, btu16 uid, btu8 *passwd, int pwlen, btuni *path, btbool useAndX, btu8 *service);
void	cifsTreeDisconnect(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid);

void	cifsQueryInformation(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path);
void	cifsQueryPathInfo(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu32 infoLevel);
void	cifsFindAttributes(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path);

void	cifsSetInformation(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu16 attributes, btu32 wtime);
void	cifsSetPathInfoStd(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu16 attributes, btu16 atime, btu16 adate, btu16 mtime, btu16 mdate, btu16 ctime, btu16 cdate);
void	cifsSetFileInfoStd(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu16 attributes, btu16 atime, btu16 adate, btu16 mtime, btu16 mdate, btu16 ctime, btu16 cdate);
void	cifsSetInformation2(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu16 atime, btu16 adate, btu16 mtime, btu16 mdate);
void	cifsSetPathInfoUnix(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu32 mode, bti64 atime, bti64 mtime, bti64 fileUid, bti64 fileGid);

void	cifsQueryFsInfo(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid);
void	cifsQueryInfoDisk(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid);

void	cifsSearch(struct smbRequest *smb, void *caller, btbool (*resultCallback)(void *ref, cifsResFile_t *rec), btu16 uid, btu16 tid, btuni *path, int maxResults, btsize resumeSize, btu8 *resumeKey);
void	cifsFindFirst(struct smbRequest *smb, void *caller, btbool (*resultCallback)(void *ref, cifsResFile_t *rec), btu16 uid, btu16 tid, btuni *path, int maxResults, btint infoLevel, btbool autoclose, btbool useResumeKeys);
void	cifsFindNext(struct smbRequest *smb, void *caller, btbool (*resultCallback)(void *ref, cifsResFile_t *rec), btu16 uid, btu16 tid, btuni *path, int maxResults, btu16 searchId, btint infoLevel, btbool autoclose, btbool useResumeKeys, btu32 resumeKey);
void	cifsFindClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 searchId);

void	cifsOpenCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool readonly);
void	cifsOpen(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool readonly, btbool truncate, btbool create, btu16 attributes, btu32 ctime, btoffset truncateLen);
void	cifsClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu32 mtime);

void	cifsOpenReadClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btoffset fileOffset, btu32 len, btbool readonly);
void	cifsOpenRead(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btoffset fileOffset, btu32 len, btbool readonly);
void	cifsRead(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btoffset fileOffset, btu32 len);
void	cifsReadCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btoffset fileOffset, btu32 len);

void	cifsOpenWriteClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu8 *buffer, btoffset fileOffset, btu32 len, btbool readonly, btbool writeThrough);
void	cifsOpenWrite(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu8 *buffer, btoffset fileOffset, btu32 len, btbool readonly, btbool writeThrough);
void	cifsWrite(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu8 *buffer, btoffset fileOffset, btu32 len, btbool writeThrough);
void	cifsWriteCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu16 fid, btu8 *buffer, btoffset fileOffset, btu32 len);

void	cifsCreateFileCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btu16 attributes, btu32 ctime);
void	cifsMkdirCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path);
void	cifsOpenClose(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool truncate, btbool create, btu16 attributes, btu32 ctime, btoffset truncateLen);
void	cifsRemoveCore(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *path, btbool isDir);
void	cifsRename(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btuni *frompath, btuni *topath);

void	cifsDecodeOplockBreak(struct smbHeader *hdr, btu8 *data, btsize len, btu16 *fid, int *lockMode);
void	cifsAckOplockBreak(struct smbRequest *smb, btu16 tid, btu16 fid, int lockLevel);

/* functions not currently used: */
void	cifsOpenNT(struct smbRequest *smb, void *caller, btu16 uid, btu16 tid, btu32 parentFid, btuni *name, btbool doCreate);


/* ------------------------------------------------------------------------- */

#endif	/* __cifs_H_INCLUDED__ */
