/*
 * Name: datacache.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-02
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: datacache.c,v 1.2 1999/07/22 17:53:53 cs Exp $
 */


#include "stdheaders.h"
#include "datacache.h"
#include "process.h"
#include "blockpool.h"
#include "file.h"
#include "smbnumbers.h"
#include "logging.h"
#include "evntloop.h"
#include "cifsconst.h"
#include "user.h"
#include "share.h"
#include "server.h"
#include "backend.h"

/* ------------------------------------------------------------------------- */

/*
Terms used in this module:
Valid and Dirty bits for each byte:
  - A byte is dirty, if it needs writeback.
  - A byte is locked, if the writeback has started and has not now
    completed. Locked bytes are always valid, they may be used as reply for
	reads.
  - A page becomes valid by reading from the file or writing through it.
  - A byte becomes dirty by writing through it.
  - A byte becomes locked if the writeback begins and the dirty bit is
    cleared.
  - A byte becomes un-dirty when the writeback command is sent.
  - A byte becomes un-locked when the writeback command completes
  - A page becomes un-valid if the validity timeout occures or if it is
    invalidated.
The locked- and dirty-bits are held in bitarrays in the cache pages. Each
cache page also holds the general bits allDirty, anyDirty, allLocked and
anyLocked to avoid tedious searches through the entire bitarrays for frequent
cases.
 */

/* ------------------------------------------------------------------------- */

static btint	pageBaseCompare(void *a, void *b)
{
dcachePage_t	*this = a, *other = b;

    return this->base - other->base;
}

static btu32	pageHash(void *obj)
{
dcachePage_t	*this = obj;
btu32			hval;

	hval = this->base / 1024;
    hval ^= (hval >> 2) ^ (hval >> 4);
    return hval;
}

static inline void	pageFree(dcachePage_t *this)
{
    if(this->data != NULL)
        memFree(this->data);
    if(this->dirty != NULL)
        memFree(this->dirty);
    if(this->locked != NULL)
        memFree(this->locked);
    memFree(this);
}

static void	pageRetain(void *obj)
{
dcachePage_t	*this = obj;

    this->refCount++;
}

static void	pageRelease(void *obj)
{
dcachePage_t	*this = obj;

	this->refCount--;
    if(this->refCount == 0){
        pageFree(this);
	}else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

static collectionObjectI_t	cachePageInterface = {
    pageBaseCompare,
    pageHash,
    pageRetain,
    pageRelease
};

/* ------------------------------------------------------------------------- */

extern cacheInterface_t	dataCacheInterface;	/* forward declaration */

dcache_t	*dcacheNew(btu32 volatile pageSize, btu32 volatile pages)
{
dcache_t		*volatile cache = NULL;
dcachePage_t	*page;
int				i;

    if((pageSize & (pageSize - 1)) != 0){	/* not a power of 2 */
        EXC_THROW_MSG(excInternalError, "dcache: cache page size not a power of 2");
    }
EXC_TRY
	cache = memCalloc(sizeof(dcache_t));
	cacheInit(&cache->general, &dataCacheInterface);
    cache->pages = htabNew(&cachePageInterface, pages + pages/2, 80);
    cache->pagesLru = lruNew(0);
    cache->pageSize = pageSize;
    for(i=0;i<pages;i++){
        page = memCalloc(sizeof(dcachePage_t));
        page->base = i * pageSize;
        htabAddObject(cache->pages, page);
        lruAddObject(cache->pagesLru, page);
        page->data = memAlloc(pageSize);
        page->dirty = memCalloc(pageSize/8);
        page->locked = memCalloc(pageSize/8);
    }
	/* some reasonable defaults: */
	cache->windowSize = CONST_MAX_WINDOW_SIZE;
    cache->readAheadPages = 1;
EXC_CATCH
	dcacheFree(cache);
	EXC_RETHROW();
EXC_ENDTRY
	return cache;
}

/* ------------------------------------------------------------------------- */

void	dcacheFree(void *this)
{
dcache_t *cache = this;
    if(cache != NULL){
		cacheDestroy(&cache->general);
        if(cache->pagesLru != NULL)
        	lruFree(cache->pagesLru);
        if(cache->pages != NULL)
            htabFree(cache->pages);	/* this also frees the cache pages */
        memFree(cache);
    }
}

/* ------------------------------------------------------------------------- */
/* ------------------------ bit array manipulation ------------------------- */
/* ------------------------------------------------------------------------- */

static btbool	bitarrIsAllZero(btu8 *arr, btsize numBits)
{
btu32	*p = (btu32 *)arr;
btsize	i, words = numBits / 32;	/* numBits must be multiple of 32 */

    for(i=0;i<words;i++){
        if(*p++ != 0)
            return 0;
    }
    return 1;
}

/* ------------------------------------------------------------------------- */

static void	bitarrSetRange(btu8 *arr, btsize startBitIndex, btsize numBits, btbool value)
{
btsize	fillStart, startRemainder, fillEnd, endRemainder;
btu8	startMask, endMask;
   
    fillStart = startBitIndex / 8 + 1;
    startRemainder = startBitIndex % 8;
    fillEnd = (startBitIndex + numBits) / 8;
    endRemainder = (startBitIndex + numBits) % 8;
    if(fillEnd > fillStart)
    	memset(arr + fillStart, value ? 0xff : 0, fillEnd - fillStart);
	startMask = ~((1<<startRemainder) - 1);
	endMask = (1<<endRemainder) - 1;
	if(value){
		if(fillStart - 1 == fillEnd){
			arr[fillEnd] |= startMask & endMask;
		}else{
			arr[fillStart - 1] |= startMask;
			if(endMask != 0)
				arr[fillEnd] |= endMask;
		}
	}else{
		if(fillStart - 1 == fillEnd){
			arr[fillEnd] &= ~(startMask & endMask);
		}else{
			arr[fillStart - 1] &= ~startMask;
			if(endMask != 0)
				arr[fillEnd] &= ~endMask;
		}
	}
}

/* ------------------------------------------------------------------------- */

static inline void	bitarrSetAll(btu8 *arr, btsize numBits, btbool value)
{
btsize	numBytes = numBits / 8;		/* numBits must be multiple of 8 */

	memset(arr, value ? 0xff : 0, numBytes);
}

static inline btbool	bitarrBitIsSet(btu8 *arr, btsize bitnum)
{
	return (arr[bitnum/8] & (1 << (bitnum % 8))) != 0;
}

/* ------------------------------------------------------------------------- */

#define	BITS_ALL_ZERO		0
#define	BITS_ALL_ONE		1
#define	BITS_BLOCK_ONE		2
#define	BITS_TOO_COMPLEX	3

static btint	bitarrClassify(btu8 *arr, btsize numBits, btsize *start, btsize *end)
{
btint	i, j, nextI = 0, bytes = numBits / 8;
btu8	b, state = 0;
btsize	index[2];

    for(i=0;i<bytes;i++){
        b = arr[i];
        if(b == state){	/* all bits are part of the block */
            continue;
        }
        for(j=0;j<8;j++){
            if(((b ^ state) & (1<<j)) != 0){
                if(nextI >= ELEMENTS(index)){
                    nextI++;
                    goto outerBreak;
                }
                index[nextI++] = i * 8 + j;
                state = ~state;
            }
        }
    }
	if(nextI == 0){
		*start = 0;
		*end = numBits;
		return BITS_ALL_ZERO;
	}
    if(nextI == 1){
        index[nextI++] = numBits;	/* block reaches to the end */
        if(index[0] == 0)
            nextI = BITS_ALL_ONE;
    }
outerBreak:
	*start = index[0];
	*end = index[1];
	return nextI;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

void	dcacheSetWindowSize(dcache_t *this, btu8 winSize)
{
    this->windowSize = winSize > CONST_MAX_WINDOW_SIZE ? CONST_MAX_WINDOW_SIZE : winSize;
}

/* ------------------------------------------------------------------------- */

void	dcacheInvalidate(void *obj)
{
dcache_t		*this = obj;
dcachePage_t	*p = NULL;

    while((p = lruOldestRelativeTo(this->pagesLru, p)) != NULL){
		p->isValid = 0;
    }
    this->readFileSize = 0;
}

/* ------------------------------------------------------------------------- */

void	dcacheInit(void *obj)
{
dcache_t		*this = obj;
dcachePage_t	*p = NULL;

    while((p = lruOldestRelativeTo(this->pagesLru, p)) != NULL){
		p->anyLocked = p->allLocked = 0;
		p->anyDirty = p->allDirty = 0;
		p->isValid = 0;
		bitarrSetAll(p->dirty, this->pageSize, 0);
		bitarrSetAll(p->locked, this->pageSize, 0);
    }
    this->readFileSize = this->writeFileSize = 0;
}

/* ------------------------------------------------------------------------- */

void	dcacheTruncate(dcache_t *this, btoffset newFileSize)
{
dcachePage_t	*p = NULL;
btu32			offset;

    while((p = lruOldestRelativeTo(this->pagesLru, p)) != NULL){
        if(p->base >= newFileSize){	/* page entirely out of file */
			p->isValid = 0;
			p->anyDirty = p->allDirty = 0;
			bitarrSetAll(p->dirty, this->pageSize, 0);
        }else if(p->base + this->pageSize > newFileSize){	/* EOF within this page */
            offset = newFileSize - p->base;
			bitarrSetRange(p->dirty, offset, this->pageSize - offset, 0);
			p->allDirty = 0;
        }
    }
	this->readFileSize = this->writeFileSize = newFileSize;
}

/* ------------------------------------------------------------------------- */

typedef struct cacheCtx{
    PROC_VARSET;
/* parameters: */
    btu8			*buffer;
    btoffset		fileOffset;
    btu32			len;
    cifsUser_t		*user;
	btint			op;
	btbool			writeThrough;
/* context: */
	btsize			start;
	btsize			end;
    dcache_t		*cache;
    dcachePage_t	*page;

    bti32			readAheadPages;	/* number of pages to read ahead */
	btoffset		pageBase;	/* file offset of beginning of current page */
    btu32			pageOffset;	/* offset where to start copying in current page */
    btu8			*memBase;	/* current copy destination */
	btu8			*memEof;	/* expected (or detected) end of this read */
    btError			err;
    struct{
        btu8			*memBase;
        btu32			pageOffset;
        dcachePage_t	*page;
    }				wait[CONST_MAX_WINDOW_SIZE];
}cacheCtx_t;

static blockpool_t	*cachePool = NULL;

static void	cachePoolFree(void *block)
{
    poolReturnBlock(cachePool, block);
}

/* ------------------------------------------------------------------------- */

static dcachePage_t	*acquirePage(dcache_t *cache, btoffset base)
{
dcachePage_t	*p, refPage;

    refPage.base = base;
    if((p = htabFindObjectEqualTo(cache->pages, &refPage)) == NULL){
        while((p = lruOldestRelativeTo(cache->pagesLru, p)) != NULL){
            if(p->readRunning || p->writeRunning || p->anyDirty)
                continue;
            pageRetain(p);
            htabRemoveObjectEqualTo(cache->pages, p);
            p->base = base;
            htabAddObject(cache->pages, p);
            pageRelease(p);
			p->isValid = 0;
            p->anyLocked = p->allLocked = 0;
            p->anyDirty = p->allDirty = 0;
			bitarrSetAll(p->dirty, cache->pageSize, 0);	/* both should not be necessary, but stay on the safe side */
			bitarrSetAll(p->locked, cache->pageSize, 0);
            break;
        }
	}else if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB)){
        logPrintf(LOGM_CACHEDEB, "cache: acquire page 0x%08x: found\n", (int)base);
    }
    if(p != NULL)
		lruTouch(cache->pagesLru, p);
    return p;
}

/* ------------------------------------------------------------------------- */

static inline void	checkOutdated(dcache_t *cache, dcachePage_t *page)
{
bttime	age, maxAge;

	if(!page->isValid)
		return;		/* nothing to check */
	maxAge = fileMaxDataCacheAge(cache->general.owner);
    if(maxAge == 0){	/* expiry immediately */
		page->isValid = 0;
	}else if(maxAge > 0){	/* negative maxAge = expire never */
		age = elpNow() - page->lastSync;
		if(age > maxAge || age < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
				logPrintf(LOGM_CACHEDEB, "cache: page 0x%08x outdated, age=%d\n", (int)page->base, age);
			page->isValid = 0;
		}else if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
			logPrintf(LOGM_CACHEDEB, "cache: page 0x%08x still valid, age=%d\n", (int)page->base, age);
	}
}

/* ------------------------------------------------------------------------- */

static void	makePageValidProc(void *obj, void *arg)
{
cacheCtx_t		*this = obj;
dcache_t		*cache = this->cache;
dcachePage_t	*page = this->page;
filePrivRes_t	*rval = arg;
int				i;

    PROC_BEGIN
    page->readRunning = 1;
	cacheStartOperation(cache);
    if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
        logPrintf(LOGM_CACHEDEB, "cache: make page valid 0x%08x\n", (int)page->base);
	while(page->anyLocked && !page->allLocked){		/* don't overlap reads with running writes! */
		PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
	}
	checkOutdated(cache, page);
    if(page->isValid || page->allDirty || page->allLocked){	/* page _is_ valid */
        page->readError.code = page->readError.class = 0;
        if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
            logPrintf(LOGM_CACHEDEB, "cache: page 0x%08x has been all valid\n", (int)page->base);
	}else{	/* not entirely valid, get the page and merge it in */
        PROC_WAITFOR_X(2, fileRWUncached(cache->general.owner, FILE_OP_READ, this->writeThrough, NULL, page->base, cache->pageSize, this->user, this));
		if(page->isValid){
/* Page may have become valid due to write. In this case we must discard the
 * received data because it is outdated.
 */
        	page->readError.code = page->readError.class = 0;
		}else{
			if(rval->err.class == 0){
				if(rval->dataLen > cache->pageSize)		/* don't trust server with dataLen */
					rval->dataLen = cache->pageSize;
				if(page->anyDirty || page->anyLocked){	/* we have to merge the page */
					if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
						logPrintf(LOGM_CACHEDEB, "cache: merging into partially valid page 0x%08x\n", (int)page->base);
					for(i=0;i<rval->dataLen;i++){
						if(!bitarrBitIsSet(page->dirty, i) && !bitarrBitIsSet(page->locked, i))
							page->data[i] = rval->data[i];		/* copy only if not dirty and not locked */
					}
					for(;i<cache->pageSize;i++){
						if(!bitarrBitIsSet(page->dirty, i) && !bitarrBitIsSet(page->locked, i))
							page->data[i] = 0;			/* fill the rest with zeros */
					}
				}else{	/* no merge necessary, we can simply copy it */
					memcpy(page->data, rval->data, rval->dataLen);
					memset(page->data + rval->dataLen, 0, cache->pageSize - rval->dataLen);
				}
				if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
					logPrintf(LOGM_CACHEDEB, "cache: page now valid 0x%08x (errclass=%d)\n", (int)page->base, rval->err.class);
				page->isValid = 1;
				page->lastSync = elpNow();
				if(page->base + rval->dataLen > cache->readFileSize)	/* move EOF */
					cache->readFileSize = page->base + rval->dataLen;
			}
			page->readError = rval->err;
		}
    }
    page->readRunning = 0;
	cache->general.opRunning--;
    PROC_RETURN_AND_CLEANUP(NULL,
		cacheEndOperation(cache, 0);
	);
    PROC_END
}

static void makePageValid(dcache_t *this, dcachePage_t *page, cifsUser_t *user, btbool writeFollows)
{
cacheCtx_t	*ctx;
    
    if(page->readRunning){	/* update is already running */
		if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
			logPrintf(LOGM_CACHEDEB, "cache: makeValid: update page 0x%08x already running\n", (int)page->base);
        return;
	}
    PROC_INIT(ctx, makePageValidProc, NULL, cachePool, cachePoolFree);
    ctx->cache = this;
    ctx->page = page;
    ctx->user = user;
	ctx->writeThrough = writeFollows;
    PROC_START(ctx, NULL);
}


static inline btoffset	cacheFileSize(dcache_t *this)
{
    if(this->readFileSize > this->writeFileSize)
        return this->readFileSize;
    else
        return this->writeFileSize;
}

/* ------------------------------------------------------------------------- */

static void	pageWriteBackProc(void *obj, void *arg)
{
cacheCtx_t		*this = obj;
dcache_t		*cache = this->cache;
dcachePage_t	*page = this->page;
filePrivRes_t	*rval = arg;
btsize			start, end;
int				dirtyStatus;

	PROC_BEGIN
    while(page->writeRunning){					/* don't interfer with other writes */
        if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
            logPrintf(LOGM_CACHEDEB, "cache: write back 0x%08x: waiting for previous write\n", (int)page->base);
        PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
    }
    page->writeRunning = 1;
	cacheStartOperation(cache);
	if(!page->anyDirty){			/* not really dirty */
		dirtyStatus = BITS_ALL_ZERO;
	}else if(page->allDirty){		/* completely dirty */
		dirtyStatus = BITS_ALL_ONE;
	}else{
		dirtyStatus = bitarrClassify(page->dirty, cache->pageSize, &start, &end);
	}
	if(dirtyStatus == BITS_ALL_ZERO){
		if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
			logPrintf(LOGM_CACHEDEB, "cache: write back: not dirty!!! 0x%08x\n", (int)page->base);
		page->anyDirty = page->allDirty = 0;
		goto writeAbort;						/* should never happen */
	}else if(dirtyStatus == BITS_ALL_ONE || dirtyStatus == BITS_TOO_COMPLEX){	/* make valid before complete writeback */
		if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
			logPrintf(LOGM_CACHEDEB, "cache: write back: make valid 0x%08x\n", (int)page->base);
		makePageValid(cache, page, this->user, 1);	/* if dirtyStatus == BITS_ALL_ONE, isValid must also be true */
		while(page->readRunning){	/* wait for makePageValid to succeed */
			PROC_WAITFOR_X(2, procCondWait(cache->general.opReady, this));
		}
		if(page->readError.class != 0){			/* error occured while updating page */
			page->writeError = page->readError;	/* declare it as a write error */
			goto writeAbort;
		}
		start = 0;				/* write back the entire page */
		end = cache->pageSize;
	} /* if(dirtyStatus == BITS_BLOCK_ONE): start and end are already set appropriately */
    page->anyDirty = page->allDirty = 0;	/* page _will_ be clean after next operation: */
	bitarrSetAll(page->dirty, cache->pageSize, 0);
    if(page->base + end >= cacheFileSize(cache)){	/* if we hit EOF */
        end = cacheFileSize(cache) - page->base;
    }
	page->anyLocked = 1;
	bitarrSetRange(page->locked, start, end - start, 1);	/* lock the range we write */
	if(start == 0 && end == cache->pageSize){
		page->isValid = 1;
    	page->lastSync = elpNow();
		page->allLocked = 1;
	}
    if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
        logPrintf(LOGM_CACHEDEB, "cache: writing back 0x%08x from %d to %d\n", (int)page->base, start, end);
	this->start = start;	/* store for use after write */
	this->end = end;
    if(end > start){	/* write needed */
        PROC_WAITFOR_X(3, fileRWUncached(cache->general.owner, FILE_OP_WRITE, this->writeThrough, page->data+start, page->base+start, end-start, this->user, this));
        page->writeError = rval->err;
    }else{
        page->writeError.code = page->writeError.class = 0;
    }
	bitarrSetRange(page->locked, this->start, this->end - this->start, 0);	/* unlock the range we wrote */
	page->allLocked = 0;
	page->anyLocked = !bitarrIsAllZero(page->locked, cache->pageSize);
	if(page->anyLocked){
		DEBUG_TRAP();	/* all locks must be zero! */
	}
/* The previous lines could be simplified because only one write can run at a time.
 * However, we want to simplify later extensions and thus leave the 'correct' version.
 */
writeAbort:
    page->writeRunning = 0;
	cache->general.opRunning--;
    PROC_RETURN_AND_CLEANUP(NULL,
		cacheEndOperation(cache, 0);
	);
    PROC_END
}

static void	pageWriteBack(dcache_t *this, dcachePage_t *page, cifsUser_t *user, bttime delay)
{
cacheCtx_t	*ctx;
    
    PROC_INIT(ctx, pageWriteBackProc, NULL, cachePool, cachePoolFree);
    ctx->cache = this;
    ctx->page = page;
    ctx->user = user;
	if(delay == 0){
		ctx->proc.process(ctx, NULL);
	}else{
		elpCallAtCancelPrev(elpNow() + delay, ctx->proc.process, ctx);
	}
}

/* ------------------------------------------------------------------------- */

static void	dcacheWriteBackProc(void *obj, void *arg)
{
cacheCtx_t		*this = obj;
dcache_t		*cache = this->cache;
dcachePage_t	*p;

    PROC_BEGIN
/* ### elpCallAtCancelPrev() with elpNow() if we ever delay writeback */
	for(;;){	/* pages will become clean if we wait long enough.... */
		p = NULL;
		while((p = lruOldestRelativeTo(cache->pagesLru, p)) != NULL){
			if(p->anyDirty || p->writeRunning)
				break;
		}
		if(p == NULL)	/* no page was dirty */
			break;
        PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
	}
    PROC_END
}

static void	dcacheWriteBack(void *obj, void *caller)
{
cacheCtx_t	*ctx;
dcache_t	*cache = obj;

    if(cache->general.opRunning <= 0){
        PROC_IMMEDIATE_RET(caller, NULL);
    }
    PROC_INIT(ctx, dcacheWriteBackProc, caller, cachePool, cachePoolFree);
    ctx->cache = cache;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	cancelFollowingWaits(cacheCtx_t *this, dcachePage_t *page)
{
int		i;

    for(i=0;i<this->cache->windowSize;i++){
        if(this->wait[i].page != NULL){
            if(this->wait[i].page->base > page->base)
                this->wait[i].page = NULL;
        }
    }
}

static void	deliverPage(cacheCtx_t *this, dcachePage_t *page, btu8 *memBase, btu32 pageOffset)
{
bti32		copyLen, pageLen, memLen;
btbool		hitEof = 0;
dcache_t	*cache = this->cache;

    if(page->readError.code != 0){			/* an error occured while reading this page */
        this->err = page->readError;		/* store error for return value */
        this->memEof = memBase;				/* stop reading further pages */
        cancelFollowingWaits(this, page);	/* don't wait for further pages */
    }else{
        if(page->base + cache->pageSize > cacheFileSize(cache)){
            hitEof = 1;
            pageLen = cacheFileSize(cache) - page->base - pageOffset;
        }else
            pageLen = cache->pageSize - pageOffset;
        memLen = this->memEof - memBase;
        copyLen = pageLen < memLen ? pageLen : memLen;
        copyLen = copyLen < 0 ? 0 : copyLen;
        memcpy(memBase, page->data + pageOffset, copyLen);
        if(hitEof){							/* this page hit the EOF */
            if(this->memEof > memBase + copyLen)
            	this->memEof = memBase + copyLen;
	        cancelFollowingWaits(this, page);	/* don't wait for more pages */
        }
    }
}

static void	readBlockProc(void *obj, void *arg)
{
cacheCtx_t		*this = obj;
dcache_t		*cache = this->cache;
dcachePage_t	*page;
int				i;
beFileRes_t		res;

    PROC_BEGIN
    this->pageBase = this->fileOffset & ~(cache->pageSize - 1);
    this->pageOffset = this->fileOffset & (cache->pageSize - 1);
    this->memBase = this->buffer;
    this->memEof = this->buffer + this->len;
    for(;;){
        while(this->memBase < this->memEof){
			for(i=0;i<cache->windowSize && this->wait[i].page != NULL;i++);	/* find a free wait entry */
			if(i >= cache->windowSize)			/* no free entry found */
				break;
            if((page = acquirePage(cache, this->pageBase)) == NULL)			/* get a cache page */
                break;	/* no page available */
			makePageValid(cache, page, this->user, 0);
			this->wait[i].page = page;
			this->wait[i].memBase = this->memBase;
			this->wait[i].pageOffset = this->pageOffset;
            this->pageBase += cache->pageSize;
            this->memBase += cache->pageSize - this->pageOffset;
            this->pageOffset = 0;
        }
        for(i=0;i<cache->windowSize;i++){
            if(this->wait[i].page != NULL && !this->wait[i].page->readRunning){	/* page became ready */
                if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
                    logPrintf(LOGM_CACHEDEB, "cache: page 0x%08x arrived\n", this->wait[i].page->base);
                deliverPage(this, this->wait[i].page, this->wait[i].memBase, this->wait[i].pageOffset);
                this->wait[i].page = NULL;
            }
        }
        if(this->memBase >= this->memEof){								/* if EOF reached */
            while(this->readAheadPages > 0 && this->err.class == 0){	/* do read ahead if no error occured */
                if((page = acquirePage(cache, this->pageBase)) == NULL)
                    break;												/* no room for read ahead */
				if(page->writeRunning)	/* don't interfer with write */
					break;
				makePageValid(cache, page, this->user, 0);
                this->readAheadPages--;
                this->pageBase += cache->pageSize;
            }
            for(i=0;i<cache->windowSize && this->wait[i].page == NULL;i++);
            if(i >= cache->windowSize)		/* if nothing to wait for */
                break;						/* break the main loop */
        }
        PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
    }
    res.err = this->err;
	res.dataLen = this->memEof - this->buffer;
	res.eof = res.dataLen < this->len;
    PROC_RETURN(&res);
    PROC_END
}

static void	readBlock(dcache_t *this, btu8 *dest, btoffset fileOffset, btu32 len, cifsUser_t *user, void *caller)
{
cacheCtx_t	*ctx;
    
    if(cachePool == NULL)
        cachePool = poolNew(sizeof(cacheCtx_t), sizeof(cacheCtx_t));
    PROC_INIT(ctx, readBlockProc, caller, cachePool, cachePoolFree);
    ctx->buffer = dest;
    ctx->fileOffset = fileOffset;
    ctx->len = len;
    ctx->cache = this;
    ctx->user = user;
    ctx->readAheadPages = this->readAheadPages;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	writeBlockProc(void *obj, void *arg)
{
cacheCtx_t		*this = obj;
dcache_t		*cache = this->cache;
beFileRes_t		res;
btoffset		wrSize;
btsize			copyLen;
dcachePage_t	*page;
int				i;

    PROC_BEGIN
    this->pageBase = this->fileOffset & ~(cache->pageSize - 1);
    this->pageOffset = this->fileOffset & (cache->pageSize - 1);
    this->memBase = this->buffer;
    this->memEof = this->buffer + this->len;
    wrSize = this->fileOffset + this->len;
    if(wrSize > cache->writeFileSize)
        cache->writeFileSize = wrSize;
    for(;;){
        while(this->memBase < this->memEof){
			for(i=0;i<cache->windowSize && this->wait[i].page != NULL;i++);	/* find a wait entry */
			if(i >= cache->windowSize)		/* no free entry found */
				break;
            if((page = acquirePage(cache, this->pageBase)) == NULL)
                break;						/* no page available */
            copyLen = cache->pageSize;
            if(this->fileOffset + this->len < this->pageBase + copyLen)
                copyLen = this->fileOffset + this->len - this->pageBase;
            copyLen -= this->pageOffset;
            memcpy(page->data + this->pageOffset, this->memBase, copyLen);
			bitarrSetRange(page->dirty, this->pageOffset, copyLen, 1);
			page->anyDirty = 1;
			if(this->pageOffset == 0 && copyLen == cache->pageSize){
				page->allDirty = page->isValid = 1;
				page->lastSync = elpNow();
			}
            pageWriteBack(cache, page, this->user, this->writeThrough ? 0 : cache->writebackDelay);
			if(this->writeThrough)
				this->wait[i].page = page;	/* we will wait for this page */
            this->pageBase += cache->pageSize;
            this->memBase += copyLen;
            this->pageOffset = 0;
        }
        for(i=0;i<cache->windowSize;i++){	/* check for ready pages */
            if(this->wait[i].page != NULL && !this->wait[i].page->writeRunning){	/* page became ready */
                if(LOG_MODE_IS_ACTIVE(LOGM_CACHEDEB))
                    logPrintf(LOGM_CACHEDEB, "cache: page 0x%08x written back\n", this->wait[i].page->base);
                if(this->wait[i].page->writeError.code != 0){
                    this->err = this->wait[i].page->writeError;
                }
                this->wait[i].page = NULL;
            }
        }
        if(this->memBase >= this->memEof){			/* if EOF reached */
            for(i=0;i<cache->windowSize && this->wait[i].page == NULL;i++);
            if(i >= cache->windowSize)				/* if nothing to wait for */
                break;								/* break the main loop */
        }
        PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
    }
	res.dataLen = this->len;
	res.err = this->err;
    PROC_RETURN(&res);
    PROC_END
}

static void	writeBlock(dcache_t *this, btu8 *src, btoffset fileOffset, btu32 len, btbool writeThrough, cifsUser_t *user, void *caller)
{
cacheCtx_t	*ctx;
    
    if(cachePool == NULL)
        cachePool = poolNew(sizeof(cacheCtx_t), sizeof(cacheCtx_t));
    PROC_INIT(ctx, writeBlockProc, caller, cachePool, cachePoolFree);
    ctx->buffer = src;
    ctx->fileOffset = fileOffset;
    ctx->len = len;
    ctx->cache = this;
    ctx->user = user;
	ctx->writeThrough = writeThrough;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	dcacheRWProc(void *obj, void *arg)
{
cacheCtx_t		*this = obj;
dcache_t		*cache = this->cache;
beFileRes_t		*rval = arg;

    PROC_BEGIN
	cacheStartOperation(cache);
	while(cache->general.forUser != userSmbUid(this->user)){
		if(cache->general.opRunning > 1){	/* don't flush while other operations running! */
			PROC_WAITFOR_X(1, procCondWait(cache->general.opReady, this));
		}else{
			dcacheInvalidate(cache);
			cache->general.forUser = userSmbUid(this->user);
		}
	}
	if(this->op == FILE_OP_WRITE){
		PROC_WAITFOR_X(2, writeBlock(cache, this->buffer, this->fileOffset, this->len, this->writeThrough, this->user, this));
	}else{
		PROC_WAITFOR_X(3, readBlock(cache, this->buffer, this->fileOffset, this->len, this->user, this));
	}
	cache->general.opRunning--;
    PROC_RETURN_AND_CLEANUP(rval,
		cacheEndOperation(cache, 0);
	);
	PROC_END
}

void	dcacheRW(dcache_t *cache, btint operation, btbool writeThrough, btu8 *buffer, btoffset offset, btsize len, cifsUser_t *user, void *caller)
{
cacheCtx_t	*ctx;
    
    if(cachePool == NULL)
        cachePool = poolNew(sizeof(cacheCtx_t), sizeof(cacheCtx_t));
    PROC_INIT(ctx, dcacheRWProc, caller, cachePool, cachePoolFree);
    ctx->buffer = buffer;
    ctx->fileOffset = offset;
    ctx->len = len;
    ctx->cache = cache;
    ctx->user = user;
	ctx->op = operation;
	ctx->writeThrough = writeThrough;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

cacheInterface_t	dataCacheInterface = {
	dcacheInvalidate,
	dcacheInit,
	dcacheWriteBack,
	dcacheFree,
};

/* ------------------------------------------------------------------------- */
