/*
 * Name: datacache.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-02
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: datacache.h,v 1.2 1999/07/22 17:53:54 cs Exp $
 */

#ifndef __datacache_H_INCLUDED__
#define __datacache_H_INCLUDED__

#include "basictypes.h"
#include "hashtable.h"
#include "lru.h"
#include "cache.h"

/*
General Description:
This module implements a read/write data cache. It's called by functions
from file.c (the read and write functions) and calls functions in file.c
(the uncached basic read/write functions). The current implementation is
statically bound to file.c by the fileRWUncached() call.
*/

/* ------------------------------------------------------------------------- */

struct generalFile;
struct cifsUser;

typedef struct dcachePage{
    lruObject_t		lruPointers;
    btoffset		base;			/* base offset in file of this page */
    bttime			lastSync;
    btint			refCount;
    btbool			isValid;		/* set after synchronizing the page from/to file */
    btbool			allDirty;		/* dirty array is all one */
    btbool			anyDirty;		/* dirty array is not all zero */
	btbool			allLocked;		/* locked array is all one */
	btbool			anyLocked;		/* locked array is not all zero */
    btbool			readRunning;	/* read operation running */
    btbool			writeRunning;	/* write back operations running */
    btError			readError;		/* last read error */
    btError			writeError;		/* last write error */
    btu8			*data;			/* the cached data */
    btu8			*dirty;			/* bit array with dirty bit for each byte */
    btu8			*locked;		/* bit array, marks bytes that must not be overwritten by read */
}dcachePage_t;

typedef struct dcache{
	cacheSuper_t		general;
    hashtable_t			*pages;
    lru_t				*pagesLru;
    btu32				pageSize;
    btu32				readAheadPages;	/* number of pages to read ahead */
	bttime				writebackDelay;	/* delay for write back */
    btu8				windowSize;
    btoffset			readFileSize;	/* file size determined by reading */
    btoffset			writeFileSize;	/* file size determined by writing */
}dcache_t;


/* ------------------------------------------------------------------------- */

dcache_t	*dcacheNew(btu32 volatile pageSize, btu32 volatile pages);
/* This function allocates a new data cache object. The cache sizes must
 * be known at allocation time.
 */
void	dcacheInit(void *obj);
/* Takes a used datacache object and initializes it as if it were newly
 * allocated.
 */
void	dcacheFree(void *this);
/* This function frees a data cache object.
 */
void	dcacheRW(dcache_t *cache, btint operation, btbool writeThrough, btu8 *buffer, btoffset offset, btsize len, struct cifsUser *user, void *caller);
/* This is the basic I/O function of the cache. Parameters should be self
 * explaining ('buffer' is also the destination for reads).
 * Callaback: beFileRes_t: err, dataLen
 */
void	dcacheTruncate(dcache_t *cache, btoffset newFileSize);
/* This function removes all cached data from the cache that's behind file
 * offset 'newFileSize'. It must be called when the file is truncated.
 */
void	dcacheInvalidate(void *cache);
/* This function marks all data in the cache as invalid. It does not change
 * the status of data that needs to be written back.
 */
void	dcacheSetWindowSize(dcache_t *this, btu8 winSize);
/* This function sets the maximum number of requests that may be outstanding
 * simultaneously in the transport layer. It's used as a hint for optimization.
 */
/* ------------------------------------------------------------------------- */
/* ---------------------------- inline methods ----------------------------- */
/* ------------------------------------------------------------------------- */

static inline void	dcacheSetReadaheadPages(dcache_t *this, int pages)
{
    this->readAheadPages = pages;
}

/* ------------------------------------------------------------------------- */

#endif	/* __datacache_H_INCLUDED__ */
