/*
 * Name: fileattr.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-08
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: fileattr.c,v 1.2 1999/07/22 17:53:58 cs Exp $
 */

#include "stdheaders.h"
#include "fileattr.h"
#include "share.h"
#include "file.h"
#include "user.h"
#include "server.h"
#include "foundationconst.h"
#include "evntloop.h"
#include "backend.h"


/* ------------------------------------------------------------------------- */
/* ----------------------- data conversion routines ------------------------ */
/* ------------------------------------------------------------------------- */

/* --------------- convert to internal representation --------------- */

inline bti64	attrTimeFromUnixTime(btu32 utime, cifsFile_t *file)
{
bti32	serverOffset = srvGmtOffset(shrServer(file->share));

    return ((bti64)utime + FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL + serverOffset * 60) * (10LL * 1000LL * 1000LL);
}

/* this looks horrible, but mktime() does not work properly on NEXTSTEP */
static btu32	secondsSince1970(btint year, btint month, btint day, btint hour, btint minute, btint second)
{
static btint	monthBase[] = {0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};
btint			days, seconds;

	seconds = (((hour * 60) + minute) * 60) + second;
	year += month / 12;
	month %= 12;
	days = (year - 70) * 365 + (year - 69)/4 + 1;
	days += monthBase[month];
	if(!(month > 1 && (year % 4) == 0 && ((year % 100) != 0 || ((year + 1900) % 1000) == 0))){	/* not >= March in leap year */
		days--;
	}
	days += day;
	return days * 86400 + seconds;
}

bti64	attrTimeFromDateTime(btu16 date, btu16 time, cifsFile_t *file)
{
btu32			utime;

	utime = secondsSince1970(
			(date >> 9) + 80,					/* year */
			((date >> 5) & 0xf) - 1,			/* month */
			(date & 0x1f) - 1,					/* day */
			time >> 11,							/* hour */
			(time >> 5) & 0x3f,					/* minute */
			(time & 0x1f) * 2					/* second */
		);
    return attrTimeFromUnixTime(utime, file);
}

bti64	attrTimeFromServerTime(bti64 serverTime, struct server *server)
{
	return serverTime + (bti64)(srvGmtOffset(server)) * (600LL * 1000LL * 1000LL);
}

bti64	attrTimeFromVfsTime(struct beFileTime *t)
{
    return ((bti64)t->seconds + FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL) * (10LL * 1000LL * 1000LL) + t->nseconds / 100;
}


/* --------------- convert to server representation --------------- */

btu32	attrUnixTimeFromTime(bti64 tm, cifsFile_t *file)
{
bti32	serverOffset = srvGmtOffset(shrServer(file->share));

    tm /= 10;							/* time is in microseconds now */
    tm -= FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL * 1000LL * 1000LL;	/* convert to 1970-based time */
    return tm / 1000000 - serverOffset * 60;
}

void	attrDateTimeFromTime(bti64 time, btu16 *timep, btu16 *datep, cifsFile_t *file)
{
struct tm	*tm;
time_t		uxTime;
    
	uxTime = attrUnixTimeFromTime(time, file);
    tm = gmtime(&uxTime);
    *timep = (tm->tm_sec >> 1) | (tm->tm_min << 5) | (tm->tm_hour << 11);
    *datep = tm->tm_mday | ((tm->tm_mon + 1) << 5) | ((tm->tm_year - 80) << 9);
}

bti64	attrServerTimeFromTime(bti64 serverTime, server_t *srvContext)
{
	return serverTime - (bti64)srvGmtOffset(srvContext) * (600LL * 1000LL * 1000LL);
}

/* ------------------- convert to client representation -------------------- */

void	attrFileVfsTime(struct beFileTime *vt, bti64 tm)
{
    tm -= FCONST_DAYS_FROM_NT_TO_UNIX_DATE * 86400LL * 1000LL * 10000LL;	/* convert to 1970-based time */
    vt->seconds = tm / 10000000;
    vt->nseconds = (tm % 10000000) * 100;
}

/* ------------------------- attribute conversion -------------------------- */

static inline btint	attrMask(btint xmap)
{
	switch(xmap & ~SRV_EXMAP_INVERT){
		default:
		case SRV_EXMAP_ARCHIVE:	return CIFS_ATTR_ARCHIVE;
		case SRV_EXMAP_SYSTEM:	return CIFS_ATTR_SYSTEM;
		case SRV_EXMAP_HIDDEN:	return CIFS_ATTR_HIDDEN;
		case SRV_EXMAP_ON:
		case SRV_EXMAP_OFF:		return 0;
	}
}

btu32	attrUnixModeFromDOSAttr(btint dosMode, btint xmap)
{
btu32	mode = 0600;

    if(dosMode & CIFS_ATTR_READONLY)
        mode &= ~0200;
	if((dosMode & attrMask(xmap)) != 0 || (xmap & ~SRV_EXMAP_INVERT) == SRV_EXMAP_ON)
		mode |= 0100;	/* map bit to execute */
	if((xmap & SRV_EXMAP_INVERT) != 0)	/* invert execute bit on demand */
		mode ^= 0100;
	if((dosMode & CIFS_ATTR_DIR) != 0)	/* directories must have execute bit set */
		mode |= 0100;
	return mode;
}

btint	attrDOSAttrFromUnixMode(btu32 mode, btint oldDosMode, btint xmap)
{
btint	mask, dosMode;

	if((oldDosMode & CIFS_ATTR_DIR) != 0)
		xmap = SRV_EXMAP_ON;	/* exec flag is always on for directories */
	if((xmap & SRV_EXMAP_INVERT) != 0)	/* invert execute bit on demand */
		mode ^= 0100;
	dosMode = 0;
	mask = attrMask(xmap);
	if((mode & 0100) != 0)	/* execute bit */
		dosMode |= mask;
	if((mode & 0200) == 0)	/* write bit */
		dosMode |= CIFS_ATTR_READONLY;
	mask |= CIFS_ATTR_READONLY;
	return (oldDosMode & ~mask) | dosMode;
}

/* ------------------------------------------------------------------------- */
/* -------------------------- superclass methods --------------------------- */
/* ------------------------------------------------------------------------- */

void	attrInit(void *obj, fileAttrInterface_t *interface)
{
fileAttr_t	*attr = obj;

	attr->interface = interface;
}

/* ------------------------------------------------------------------------- */

void	attrExit(void *obj)
{
fileAttr_t	*attr = obj;

	attr->interface = NULL;
}

/* ------------------------------------------------------------------------- */

btbool	attrGeneralValid(struct cifsFile *file, struct cifsUser *user)
{
bttime		age;
fileAttr_t	*attr = file->attributes;

	if(attr->updatedAt == 0)
		return 0;
	age = elpNow() - attr->updatedAt;
	return age < srvAttributesCacheTime(shrServer(fileShare(file))) && age >= 0;
}

/* ------------------------------------------------------------------------- */

void	attrGeneralInvalidate(struct cifsFile *file)
{
fileAttr_t	*attr = file->attributes;

	attr->updatedAt = 0;
}

/* ------------------------------------------------------------------------- */

void	attrGeneralTruncate(struct cifsFile *file, btoffset newSize)
{
fileAttr_t	*attr = file->attributes;

	if(attr->size != newSize){
		attr->size = newSize;
		/* the following is an approximation, but it should suffice */
		attr->allocSize = ((newSize / 32768) + 1) * 32768;
	}
	attr->wtime = elpNow64();
}

/* ------------------------------------------------------------------------- */

void	attrGeneralUpdateAfterWrite(struct cifsFile *file, btoffset offset, btsize bytesWritten)
{
btoffset	minSize = offset + bytesWritten;
fileAttr_t	*attr = file->attributes;

	if(attr->size < minSize){
		attr->size = minSize;
		/* the following is an approximation, but it should suffice */
		attr->allocSize = ((minSize / 32768) + 1) * 32768;
	}
	attr->wtime = elpNow64();
}

/* ------------------------------------------------------------------------- */

btint	attrDOSFromAttributes(struct beFileSetAttr *setAttr, struct cifsUser *userContext, struct server *serverContext)
{
	/* This function is only used to set the attributes of newly created
	 * files. It is thus sufficent to ignore the user context and use
	 * the user attributes.
	 */
	 return attrDOSAttrFromUnixMode(setAttr->setUnixMode ? setAttr->unixMode : 0666, 0, srvExecMapping(serverContext));
}

/* ------------------------------------------------------------------------- */
