/*
 * Name: fileattr.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-03
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: fileattr.h,v 1.3 1999/07/22 17:53:59 cs Exp $
 */

#ifndef __fileattr_H_INCLUDED__
#define __fileattr_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This is the (abstract) superclass of the file attribute classes. There is an
extra abstraction layer around attributes because this is the thing that
differs most among operating systems. Subclasses like fileAttrDOS,
fileAttrUnix and fileAttrNT are possible. The file attribute classes are
bound statically to the file class and the cache classes.
*/

/* ------------------------------------------------------------------------- */

struct server;
struct cifsUser;
struct cifsFile;
struct fileAttr;
struct beFileSetAttr;
struct beFileAttr;
struct beFileTime;

typedef struct fileAttrInterface{
	btbool	(*valid)(struct cifsFile *file, struct cifsUser *user);
	void	(*setWritable)(struct cifsFile *file, btbool writable, struct cifsUser *user, btbool noLocking, void *caller);
	void	(*makeValid)(struct cifsFile *file, struct cifsUser *user, void *caller);
	void	(*getAttributes)(struct cifsFile *file, struct cifsUser *user, struct beFileAttr *attr);
	void	(*setAttributes)(struct cifsFile *file, struct cifsUser *user, struct beFileSetAttr *setAttr, void *caller);
	btbool	(*getWritable)(struct cifsFile *file, struct cifsUser *user);
	void	(*readLink)(struct cifsFile *file, struct cifsUser *user, void *caller);
	void	(*touch)(struct cifsFile *file, struct cifsUser *user, void *caller);
	void	(*free)(void *attr);
	btu32   (*getWtime)(struct cifsFile *file);
	void    (*createLink)(struct cifsFile *parent, struct cifsUser *user, btuni *linkName, btuni *path, struct beFileSetAttr *setAttr, void *caller);
}fileAttrInterface_t;

typedef struct fileAttr{
	fileAttrInterface_t	*interface;
	bttime				updatedAt;	/* when these attributes were read (0 = never) */
	bti64				atime;		/* last access */
	bti64				wtime;		/* last write */
	btoffset			size;		/* file size */
	btoffset			allocSize;	/* allocated size */
	btint				attrLocked;	/* lock to avoid overlapping operations, signal is openSignal */
}fileAttr_t;


/* ------------------------------------------------------------------------- */
/* ---------------------- superclass implementations ----------------------- */
/* ------------------------------------------------------------------------- */
/* These functions are intended for use by subclasses! */

void	attrInit(void *obj, fileAttrInterface_t *interface);
void	attrExit(void *obj);
btbool	attrGeneralValid(struct cifsFile *file, struct cifsUser *user);
void	attrGeneralInvalidate(struct cifsFile *file);
void	attrGeneralTruncate(struct cifsFile *file, btoffset newSize);
void	attrGeneralUpdateAfterWrite(struct cifsFile *file, btoffset offset, btsize bytesWritten);


/* ------------------------------------------------------------------------- */
/* --------------------------- utility functions --------------------------- */
/* ------------------------------------------------------------------------- */

btint	attrDOSFromAttributes(struct beFileSetAttr *setAttr, struct cifsUser *userContext, struct server *serverContext);
bti64	attrTimeFromUnixTime(btu32 utime, struct cifsFile *file);
btu32	attrUnixTimeFromTime(bti64 tm, struct cifsFile *file);
void	attrDateTimeFromTime(bti64 time, btu16 *timep, btu16 *datep, struct cifsFile *file);
bti64	attrTimeFromServerTime(bti64 serverTime, struct server *server);
bti64	attrTimeFromDateTime(btu16 date, btu16 time, struct cifsFile *file);
bti64	attrTimeFromVfsTime(struct beFileTime *nfst);
bti64	attrServerTimeFromTime(bti64 serverTime, struct server *srvContext);
void	attrFileVfsTime(struct beFileTime *nfst, bti64 tm);
btu32	attrUnixModeFromDOSAttr(btint dosMode, btint xmap);
btint	attrDOSAttrFromUnixMode(btu32 mode, btint oldDosMode, btint xmap);


/* ------------------------------------------------------------------------- */
/* -------------------------- polymorphic methods -------------------------- */
/* ------------------------------------------------------------------------- */

#define	attrValid(file, user)								(((fileAttr_t *)((file)->attributes))->interface->valid(file, user))
/* Returns whether the attributes stored in the attribute object are valid
 * or if they require a reload with 'attrMakeValid()'.
 */
#define	attrSetWritable(file, writable, user, noLocking, caller)		(((fileAttr_t *)((file)->attributes))->interface->setWritable(file, writable, user, noLocking, caller))
/* Makes the file writable or writeprotected for the calling user, if
 * possible. 'noLocking' prevents the attribute writeback from grabbing
 * the open lock. That's necessary if it's called from an open-locked
 * environment.
 * Callback: err
 */
#define	attrInvalidate(file)								attrGeneralInvalidate(file)
/* Marks the attributes of 'file' as invalid. They will be reloaded the
 * next time they are needed.
 */
#define	attrTruncate(file, newSize)							attrGeneralTruncate(file, newSize)
/* This function informs the attribute object about a truncate operation
 * that occured on the file.
 */
#define	attrUpdateAfterWrite(file, offset, bytesWritten)	attrGeneralUpdateAfterWrite(file, offset, bytesWritten)
/* This function informs the attribute object about writes that occured
 * to the file. To avoid frequent reloads, the attributes may be changed
 * to what the client thinks the server would return.
 */
#define	attrMakeValid(file, user, caller)					(((fileAttr_t *)((file)->attributes))->interface->makeValid(file, user, caller))
/* This function reads the attributes from the file.
 * Callback: err
 */
#define	attrFileSize(file)									(((fileAttr_t *)(file)->attributes)->size)
/* Returns the size of the file.
 */
#define	attrGetAttributes(file, userCtx, readAttr)		(((fileAttr_t *)((file)->attributes))->interface->getAttributes(file, userCtx, readAttr))
/* This functions represents the attributes of the file as NFS attributes
 * (which are roughly Unix attributes). If semantic conversions must be
 * made, they are performed under the context of 'userCtx' and the file's
 * server. Returns the attributes in 'readAttr'.
 */
#define	attrSetAttributes(file, user, setAttr, caller)	(((fileAttr_t *)((file)->attributes))->interface->setAttributes(file, user, setAttr, caller))
/* This function sets the file's attributes based on the NFS attribute
 * representation 'setAttr'. Semantic conversion is done if needed under
 * the context of user and the file's server.
 * Callback: err
 */
#define	attrGetWritable(file, userCtx)						(((fileAttr_t *)((file)->attributes))->interface->getWritable(file, userCtx))
/* This function returns whether the currently loaded attributes allow
 * writing to the file for user 'userCtx'.
 */
#define	attrReadLink(file, user, caller)					(((fileAttr_t *)((file)->attributes))->interface->readLink(file, user, caller))
/* This function retrieves the path pointed to by the symbolic link
 * 'file'.
 * Callback: err, path
 */
#define	attrTouch(file, user, caller)						(((fileAttr_t *)((file)->attributes))->interface->touch(file, user, caller))
/* This function sets the modification date of 'file' to now.
 * Callback: err
 */
#define	attrFree(attr)										(((fileAttr_t *)(attr))->interface->free(attr))
/* This function frees the memory allocated for the attribute object.
 */
#define attrGetWtime(file)									(((fileAttr_t *)((file)->attributes))->interface->getWtime(file))
/* This function returns the last write time of the file, if it is cached,
 * -1 otherwise. The time is returned in 32 bit Unix-like CIFS representation.
 * It is used to set the modification time on file close.
 */
#define	attrCreateLink(parent, user, linkName, path, setAttr, caller)	(((fileAttr_t *)((parent)->attributes))->interface->createLink(parent, user, linkName, path, setAttr, caller))
/* This function creates a new symbolic link of name 'linkName' in directory
 * 'parent'. The newly created link will point to the path 'path'.
 * Callback: err
 */
  
/* ------------------------------------------------------------------------- */


#endif /* __fileattr_H_INCLUDED__ */
