/*
 * Name: fileattrdos.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-08
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: fileattrdos.c,v 1.6 1999/08/01 12:57:12 cs Exp $
 */

#include "stdheaders.h"
#include "hashtable.h"
#include "fileattrdos.h"
#include "file.h"
#include "cache.h"
#include "server.h"
#include "smb.h"
#include "cifs.h"
#include "unicode.h"
#include "process.h"
#include "user.h"
#include "blockpool.h"
#include "share.h"
#include "logging.h"
#include "evntloop.h"
#include "cifsconst.h"
#include "errcodes.h"
#include "backend.h"
#include "dircache.h"


#define       SYMLINK_HEADER_UNI      "!\000<\000s\000y\000m\000l\000i\000n\000k\000>\000"
#define       SYMLINK_HEADER          "!<symlink>"

/* ------------------------------------------------------------------------- */

extern fileAttrInterface_t	attrDOSInterface;	/* forward declaration */

typedef struct dirtyBits{
	unsigned	atime	:1;
	unsigned	wtime	:1;
	unsigned	ctime	:1;
	unsigned	mode	:1;
}dirtyBits_t;

/* ------------------------------------------------------------------------- */

typedef struct attrCtx{
    PROC_VARSET;
    cifsFile_t		*file;
    btu16			uid;
    btu16			tid;
	cifsUser_t		*user;
	dirtyBits_t		dirty;
	beFileSetAttr_t	setAttr;
	btbool			writable;
	btbool			noLocking;
	bti64			oldWtime;
	btuni			*name;
	btuni			*path;
	server_t		*server;
}attrCtx_t;

static blockpool_t *attrPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(attrPool, block);
}

/* ------------------------------------------------------------------------- */

void	*attrDOSNew(void)
{
fileAttrDOS_t	*attr = NULL;

	if(attrPool == NULL)
		attrPool = poolNew(sizeof(attrCtx_t), sizeof(attrCtx_t));
	attr = memCalloc(sizeof(fileAttrDOS_t));
	attrInit(&attr->general, &attrDOSInterface);
    return attr;
}

/* ------------------------------------------------------------------------- */

static void	attrDOSFree(void *obj)
{
	attrExit(obj);
	memFree(obj);
}

/* ------------------------------------------------------------------------- */

static btbool	attrDOSValid(struct cifsFile *file, cifsUser_t *user)
{
fileAttrDOS_t	*attr = file->attributes;

	if(!attrGeneralValid(file, user))
		return 0;
	return attr->updatedBy == userSmbUid(user);
}

/* ------------------------------------------------------------------------- */

static inline btint	censuredMode(btint dosMode)	/* mask out bits we can't set */
{
	return dosMode & ~(CIFS_ATTR_DIR | CIFS_ATTR_VOLUME);
}

/* ------------------------------------------------------------------------- */

static void	attrDOSWritebackProc(void *obj, void *arg)
{
attrCtx_t			*this = obj;
beFileRes_t			res;
cifsRes_t			*rval = arg;
smbRequest_t		*smb = arg;
cifsFile_t		*file = this->file;
fileAttrDOS_t		*attr = file->attributes;
server_t			*server = shrServer(file->share);
btuni				*path, buffer[CIFS_MAX_PATHLEN];
btu16				atime16, mtime16, adate16, mdate16, cdate16, ctime16;
btu32				wtime32;

	PROC_BEGIN
	fileRetain(file);
	if(!this->noLocking){
		while(file->openLock){	/* wait for open state to stabilize */
			PROC_WAITFOR_X(1, procCondWait(file->openSignal, this));
		}
		file->openLock++;		/* nail down open status */
	}
	if(file->openMode == FILE_CLOSED){	/* fid not valid */
		if((this->dirty.atime || this->dirty.wtime || this->dirty.ctime || this->dirty.mode) && srvUseRequest(server, SRV_USE_SETATTR_TRANS2SETPATH)){
			PROC_WAITFOR_X(2, srvWantRequest(server, this));
			ctime16 = atime16 = mtime16 = cdate16 = adate16 = mdate16 = 0;
			if(this->dirty.atime)
				attrDateTimeFromTime(attr->general.atime, &atime16, &adate16, file);
			if(this->dirty.wtime)
				attrDateTimeFromTime(attr->general.wtime, &mtime16, &mdate16, file);
			if(this->dirty.ctime)
				attrDateTimeFromTime(attr->ctime, &ctime16, &cdate16, file);
			path = fileBuildPath(file, NULL, 0, buffer, CIFS_MAX_PATHLEN);
			PROC_WAITFOR_X(3, cifsSetPathInfoStd(smb, this, this->uid, this->tid, path, censuredMode(attr->dosMode), atime16, adate16, mtime16, mdate16, ctime16, cdate16));
			if(rval->err.class == 0){
				memset(&this->dirty, 0, sizeof(this->dirty));
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR))
					logPrintf(LOGM_DEBUG_ATTR, "attrDos: cifsSetPathInfoStd() succeeded\n");
			}
		}
	}else{	/* fid is valid */
		if((this->dirty.atime || this->dirty.wtime || this->dirty.ctime || this->dirty.mode) && srvUseRequest(server, SRV_USE_SETATTR_TRANS2SETFILE)){
			PROC_WAITFOR_X(4, srvWantRequest(server, this));
			ctime16 = atime16 = mtime16 = cdate16 = adate16 = mdate16 = 0;
			if(this->dirty.atime)
				attrDateTimeFromTime(attr->general.atime, &atime16, &adate16, file);
			if(this->dirty.wtime)
				attrDateTimeFromTime(attr->general.wtime, &mtime16, &mdate16, file);
			if(this->dirty.ctime)
				attrDateTimeFromTime(attr->ctime, &ctime16, &cdate16, file);
			PROC_WAITFOR_X(5, cifsSetFileInfoStd(smb, this, this->uid, this->tid, file->fid, censuredMode(attr->dosMode), atime16, adate16, mtime16, mdate16, ctime16, cdate16));
			if(rval->err.class == 0){
				memset(&this->dirty, 0, sizeof(this->dirty));
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR))
					logPrintf(LOGM_DEBUG_ATTR, "attrDos: cifsSetFileInfoStd() succeeded\n");
			}
		}
		if((this->dirty.atime || this->dirty.wtime) && srvUseRequest(server, SRV_USE_SETATTR_SETFILE2)){
			PROC_WAITFOR_X(6, srvWantRequest(server, this));
			atime16 = mtime16 = adate16 = mdate16 = 0;
			if(this->dirty.atime)
				attrDateTimeFromTime(attr->general.atime, &atime16, &adate16, file);
			if(this->dirty.wtime)
				attrDateTimeFromTime(attr->general.wtime, &mtime16, &mdate16, file);
			PROC_WAITFOR_X(7, cifsSetInformation2(smb, this, this->uid, this->tid, file->fid, atime16, adate16, mtime16, mdate16));
			if(rval->err.class == 0){
				this->dirty.atime = this->dirty.wtime = 0;
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR))
					logPrintf(LOGM_DEBUG_ATTR, "attrDos: cifsSetInformation2() succeeded\n");
			}
		}
	}
	/* if none of the specific commands proceeded, try core method: */
	if((this->dirty.wtime || this->dirty.mode) && srvUseRequest(server, SRV_USE_SETATTR_CORETIME)){
		PROC_WAITFOR_X(8, srvWantRequest(server, this));
		wtime32 = this->dirty.wtime ? attrUnixTimeFromTime(attr->general.wtime, file) : 0;
		path = fileBuildPath(file, NULL, 0, buffer, CIFS_MAX_PATHLEN);
		/* last access = now because we access the file to set the attr. */
		attr->general.atime = elpNow64();
		PROC_WAITFOR_X(9, cifsSetInformation(smb, this, this->uid, this->tid, path, censuredMode(attr->dosMode), wtime32));
		if(rval->err.class == 0){
			this->dirty.wtime = this->dirty.mode = 0;
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR))
				logPrintf(LOGM_DEBUG_ATTR, "attrDos: cifsSetInformation(attr, time) succeeded\n");
		}
	}
	if(this->dirty.mode){	/* catch all */
		PROC_WAITFOR_X(10, srvWantRequest(server, this));
		path = fileBuildPath(file, NULL, 0, buffer, CIFS_MAX_PATHLEN);
		PROC_WAITFOR_X(11, cifsSetInformation(smb, this, this->uid, this->tid, path, censuredMode(attr->dosMode), 0));
		if(rval->err.class == 0){
			this->dirty.mode = 0;
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG_ATTR))
				logPrintf(LOGM_DEBUG_ATTR, "attrDos: cifsSetInformation(attr, 0) succeeded\n");
		}
	}
	if(rval != NULL){
		res.err = rval->err;
	}else
		res.err.code = res.err.class = 0;
    PROC_RETURN_AND_CLEANUP(&res,
		if(!this->noLocking){
			file->openLock--;
			procCondBroadcast(file->openSignal, NULL);
		}
		fileRelease(file);
	);
    PROC_END
}

static void	attrDOSWriteback(struct cifsFile *file, dirtyBits_t *dirty, cifsUser_t *user, btbool noLocking, void *caller)
{
attrCtx_t	*ctx;

	PROC_INIT(ctx, attrDOSWritebackProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->dirty = *dirty;
	ctx->uid = userSmbUid(user);
	ctx->noLocking = noLocking;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	attrDOSSetValues(fileAttrDOS_t *attr, struct cifsFile *file, btu16 updatedBy, btu16 dosMode, bti64 ctime, bti64 wtime, bti64 atime, btoffset size, btoffset allocSize)
{
bti64	oldWtime;

	oldWtime = attr->general.wtime;
	attr->general.atime = atime;
	attr->general.wtime = wtime;
	attr->ctime = ctime;
	attr->dosMode = dosMode;
	if(file->openMode != FILE_CLOSED && (file->oplockLevel != 0 || (srvOpLock(shrServer(file->share)) == 0 && file->lastWrite != 0 && elpNow() - file->lastWrite < CONST_ATTR_SIZE_HOLD_AFTER_WRITE))){
		/* we probably know the file size better... */
		/* NT or 95 may give bogus sizes if the file is open. Accept only size increases */
		if(size >= attr->general.size){
			attr->general.size = size;
			attr->general.allocSize = allocSize;
		}else if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "fileattrdos: server lied about file size of ->%s<- (%d instead of %d)\n", convUniToAscii(file->name), (int)size, (int)attr->general.size);
	}else{
		/* don't trust size == 0 if file is open: */
		if(file->openMode == FILE_CLOSED || size != 0){
			attr->general.size = size;
			attr->general.allocSize = allocSize;
		}
	}
	attr->updatedBy = updatedBy;
	attr->general.updatedAt = elpNow();

	if(file->children != NULL && file->cache.any != NULL && wtime != oldWtime){
		dirCacheDirWasModified(file->cache.any);	/* necessary to avoid conflict with dir cache of OS */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
			logPrintf(LOGM_DEBUG, "attrDOSMakeValid: flushing dir cache of ->%s<-\n", convUniToAscii(file->name));
			logPrintf(LOGM_DEBUG, "attrDOSMakeValid: wtime changed from 0x%08x%08x to 0x%08x%08x\n", (int)(oldWtime >> 32), (int)oldWtime, (int)(attr->general.wtime >> 32), (int)attr->general.wtime);
		}
	}
    if((attr->dosMode & CIFS_ATTR_DIR) != 0 && file->children == NULL){
		fileIsDirectory(file);
    }
}

/* ------------------------------------------------------------------------- */

static void	attrDOSMakeValidProc(void *obj, void *arg)
{
attrCtx_t			*this = obj;
beFileRes_t			res;
smbRequest_t		*smb = arg;
cifsFile_t			*file = this->file;
fileAttrDOS_t		*attr = file->attributes;
server_t			*server = shrServer(file->share);
btuni				*filePath, pathBuffer[CIFS_MAX_PATHLEN];
cifsResQueryFile_t	*rval = arg;
bti64				tval;

	PROC_BEGIN
	fileRetain(file);
    if(file->children == NULL && file->cache.any != NULL){	/* write back to obtain correct file size */
        PROC_WAITFOR_X(1, cacheWriteBack(file->cache.any, this));
    }
	while(attr->general.attrLocked){
        PROC_WAITFOR_X(2, procCondWait(file->openSignal, this));
	}
	if(attrDOSValid(file, this->user)){	/* if attributes became valid meanwhile */
		res.err.class = res.err.code = 0;
		fileRelease(file);
		PROC_RETURN(&res);
	}
	attr->general.attrLocked++;	
	this->oldWtime = attr->general.wtime;
	if(file->parent == NULL && srvFakeMountpointDate(server)){	/* fake root date */
		tval = elpNow64();
		attrDOSSetValues(attr, file, this->uid, CIFS_ATTR_DIR, tval, tval, tval, 0, 0);
		res.err.class = res.err.code = 0;
	}else if(srvUseRequest(server, SRV_USE_GETATTR_PATHINFO)){
		PROC_WAITFOR_X(5, srvWantRequest(server, this));
		filePath = fileBuildPath(file, NULL, 0, pathBuffer, CIFS_MAX_PATHLEN);
        PROC_WAITFOR_X(6, cifsQueryPathInfo(smb, this, this->uid, this->tid, filePath, SMB_INFO_STANDARD));
		if(rval->err.class == 0){
			attrDOSSetValues(attr, file, this->uid, rval->attributes,
					attrTimeFromDateTime(rval->ctimeDate, rval->ctimeTime, file),	/* ctime */
					attrTimeFromDateTime(rval->mtimeDate, rval->mtimeTime, file),	/* wtime */
					attrTimeFromDateTime(rval->atimeDate, rval->atimeTime, file),	/* atime */
					rval->size, rval->allocSize);
		}
		res.err = rval->err;
	}else if(srvUseRequest(server, SRV_USE_GETATTR_RDIR) && file->parent != NULL){	/* don't use it on root directory */
		PROC_WAITFOR_X(3, srvWantRequest(server, this));
		filePath = fileBuildPath(file, NULL, 0, pathBuffer, CIFS_MAX_PATHLEN);
        PROC_WAITFOR_X(4, cifsFindAttributes(smb, this, this->uid, this->tid, filePath));
		if(rval->err.class == 0){
			attrDOSSetValues(attr, file, this->uid, rval->attributes,
					attrTimeFromDateTime(rval->ctimeDate, rval->ctimeTime, file),	/* ctime */
					attrTimeFromDateTime(rval->mtimeDate, rval->mtimeTime, file),	/* wtime */
					attrTimeFromDateTime(rval->atimeDate, rval->atimeTime, file),	/* atime */
					rval->size, rval->allocSize);
		}
		res.err = rval->err;
	}else{
		PROC_WAITFOR_X(7, srvWantRequest(server, this));
		filePath = fileBuildPath(file, NULL, 0, pathBuffer, CIFS_MAX_PATHLEN);
        PROC_WAITFOR_X(8, cifsQueryInformation(smb, this, this->uid, this->tid, filePath));
		if(rval->err.class == 0){
			tval = attrTimeFromUnixTime(rval->lastWrite, file);
			attrDOSSetValues(attr, file, this->uid, rval->attributes, tval, tval, tval, rval->size, rval->allocSize);
		}
		res.err = rval->err;
	}
	if(res.err.class != 0)
		attr->general.updatedAt = 0;
	attr->general.attrLocked--;
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(file->openSignal, NULL);
		fileRelease(file);
	);
    PROC_END
}

static void	attrDOSMakeValid(struct cifsFile *file, cifsUser_t *user, void *caller)
{
attrCtx_t	*ctx;
beFileRes_t	res;

	if(attrDOSValid(file, user)){
		res.err.class = res.err.code = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, attrDOSMakeValidProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->user = user;
	ctx->uid = userSmbUid(user);
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	attrDOSSetWritableProc(void *obj, void *arg)
{
attrCtx_t			*this = obj;
cifsFile_t			*file = this->file;
fileAttrDOS_t		*attr = file->attributes;
beFileRes_t			res, *rval = arg;
smbRequest_t		*smb = arg;
cifsResQueryFile_t	*crval = arg;
dirtyBits_t			dirty;
btu16				oldMode;
btuni				*filePath, pathBuffer[CIFS_MAX_PATHLEN];

	PROC_BEGIN;
	if(!this->noLocking){
		while(attr->general.attrLocked){
			PROC_WAITFOR_X(1, procCondWait(file->openSignal, this));
		}
		attr->general.attrLocked++;
	}
	if(!attrDOSValid(file, this->user)){
/* Can't use attrDOSMakeValid() here because it tries to write back the cache
 * before it fetches the attributes. This will cause a deadlock because
 * setWritable is called from the cache writeback routine!
 */
		PROC_WAITFOR_X(2, srvWantRequest(shrServer(file->share), this));
		filePath = fileBuildPath(file, NULL, 0, pathBuffer, CIFS_MAX_PATHLEN);
        PROC_WAITFOR_X(3, cifsQueryInformation(smb, this, userSmbUid(this->user), shrTid(file->share), filePath));
		if(crval->err.class == 0){
            attr->dosMode = crval->attributes;
		}else{
			res.err = crval->err;
			rval = &res;
			goto doReturn;
		}
	}
	oldMode = attr->dosMode;
	if(this->writable)
		attr->dosMode &= ~CIFS_ATTR_READONLY;
	else
		attr->dosMode |= CIFS_ATTR_READONLY;
	if(attr->dosMode != oldMode){
		memset(&dirty, 0, sizeof(dirty));
		dirty.mode = 1;
		PROC_WAITFOR_X(4, attrDOSWriteback(file, &dirty, this->user, this->noLocking, this));
	}else{
		res.err.code = res.err.class = 0;
		rval = &res;
	}
doReturn:
	if(!this->noLocking)
		attr->general.attrLocked--;
    PROC_RETURN_AND_CLEANUP(rval,
		if(!this->noLocking)
			procCondBroadcast(file->openSignal, NULL);
	);
	PROC_END;
}


static void	attrDOSSetWritable(struct cifsFile *file, btbool writable, cifsUser_t *user, btbool noLocking, void *caller)
{
attrCtx_t	*ctx;

	PROC_INIT(ctx, attrDOSSetWritableProc, caller, attrPool, poolFreeBlock);
	ctx->writable = writable;
	ctx->file = file;
	ctx->user = user;
	ctx->noLocking = noLocking;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	attrDOSGetAttributes(struct cifsFile *file, cifsUser_t *user, struct beFileAttr *readAttr)
{
fileAttrDOS_t	*attr = file->attributes;
btint			xmap, mode;
server_t		*server;
btu16			mask;

	server = shrServer(file->share);
	mask = srvLinkMask(server);
	memset(readAttr, 0, sizeof(*readAttr));
	if((attr->dosMode & CIFS_ATTR_DIR) != 0){       /* it's a directory */
		readAttr->type = BE_FILETYPE_DIR;
	}else if(mask != 0 && (attr->dosMode & mask) == srvLinkMode(server)){  /* it's a link */
		readAttr->type = BE_FILETYPE_LNK;
		readAttr->size = attr->general.size;
		readAttr->allocatedSize = attr->general.allocSize;
		readAttr->unixMode = 0555;	/* at least read and execute for everyone */
	}else{  /* it's a regular file */
		readAttr->type = BE_FILETYPE_REG;
		readAttr->size = attr->general.size;
		readAttr->allocatedSize = attr->general.allocSize;
	}
	xmap = srvExecMapping(server);
	mode = attrUnixModeFromDOSAttr(attr->dosMode, xmap);
	readAttr->unixMode |= mode | (mode >> 3) | (mode >> 6);	/* map mode to ugo */
	readAttr->uid = userUnixUid(user);
	readAttr->gid = userUnixGid(user);
    attrFileVfsTime(&readAttr->atime, attr->general.atime);
    attrFileVfsTime(&readAttr->mtime, attr->general.wtime);
    attrFileVfsTime(&readAttr->ctime, attr->general.wtime);	/* ctime in Unix is attribute change time, ctime in DOS is creation time... */
}

/* ------------------------------------------------------------------------- */

static void	attrDOSSetAttributesProc(void *obj, void *arg)
{
attrCtx_t		*this = obj;
cifsFile_t		*file = this->file;
fileAttrDOS_t	*attr = file->attributes;
beFileSetAttr_t	*setAttr = &this->setAttr;
btint			xmap;
beFileRes_t		res, *rval = arg;
btu16			oldMode;
server_t		*server;

 	PROC_BEGIN;
	fileRetain(file);
	while(attr->general.attrLocked){
		PROC_WAITFOR_X(1, procCondWait(file->openSignal, this));
	}
	/* don't lock immediately: this would create a deadlock with attrDOSMakeValid() */
	if(setAttr->setUnixMode){
		if(!attrDOSValid(file, this->user)){
			PROC_WAITFOR_X(2, attrDOSMakeValid(file, this->user, this));
		}
		/* I think ignoring errors here is better than aborting on error because
		 * the attribute information could still be valid and it is not that
		 * important.
		 */
		server = shrServer(file->share);
		xmap = srvExecMapping(server);
		oldMode = attr->dosMode;
		attr->dosMode = attrDOSAttrFromUnixMode(setAttr->unixMode, attr->dosMode, xmap);
		if((oldMode & srvLinkMask(server)) == srvLinkMode(server))     /* if it was a link, preserve link attribute */
			attr->dosMode = (attr->dosMode & ~srvLinkMask(server)) | srvLinkMode(server);
		if(attr->dosMode != oldMode)
			this->dirty.mode = 1;
	}
    if(setAttr->setAtime){
        attr->general.atime = attrTimeFromVfsTime(&setAttr->atime);
		this->dirty.atime = 1;
    }
    if(setAttr->setMtime){
        attr->general.wtime = attrTimeFromVfsTime(&setAttr->mtime);
		this->dirty.wtime = 1;
    }
	/* The lock should not be set here because attrDOSMakeValid() calls the
	 * callback before it fires the signal
	 */
	attr->general.attrLocked++;
	if(this->dirty.mode || this->dirty.atime || this->dirty.wtime){
		PROC_WAITFOR_X(3, attrDOSWriteback(file, &this->dirty, this->user, 0, this));
		if(rval->err.class != 0)
			goto errorOccured;
	}
    if(setAttr->setSize){
		/* free attributes lock before fileTruncate() because it needs this lock */
		attr->general.attrLocked--;
		PROC_WAITFOR_X(4, fileTruncate(file, setAttr->size, this->user, this));
		attr->general.attrLocked++;
    }else{
		res.err.code = res.err.class = 0;
		rval = &res;
	}
errorOccured:
	attr->general.attrLocked--;
    PROC_RETURN_AND_CLEANUP(rval,
		procCondBroadcast(file->openSignal, NULL);
		fileRelease(file);
	);
	PROC_END;
}

static void	attrDOSSetAttributes(struct cifsFile *file, cifsUser_t *user, struct beFileSetAttr *setAttr, void *caller)
{
attrCtx_t	*ctx;

	PROC_INIT(ctx, attrDOSSetAttributesProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->user = user;
	ctx->setAttr = *setAttr;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static btbool	attrDOSGetWritable(struct cifsFile *file, struct cifsUser *user)
{
fileAttrDOS_t	*attr = file->attributes;

	return (attr->dosMode & CIFS_ATTR_READONLY) == 0;
}

/* ------------------------------------------------------------------------- */

static void	attrDOSReadLinkProc(void *obj, void *arg)
{
attrCtx_t		*this = obj;
beFileRes_t		res;
filePrivRes_t	*prval = arg;
btuni			buffer[CIFS_MAX_PATHLEN + 1];
btint			i, j;

	PROC_BEGIN;
	PROC_WAITFOR_X(1, fileRWUncached(this->file, FILE_OP_READ, 0, NULL, 0, CIFS_MAX_PATHLEN * sizeof(btuni), this->user, this));
	res.err = prval->err;
	if(res.err.class == 0){
		if(prval->dataLen >= 20 && memcmp(prval->data, SYMLINK_HEADER_UNI, 20) == 0){
			j = 20;
			for(i=0;i<CIFS_MAX_PATHLEN && j < prval->dataLen;i++){
				buffer[i] = prval->data[j++];
				buffer[i] |= ((btuni)prval->data[j++]) << 8;
			}
			buffer[i] = 0;
		}else if(prval->dataLen >= 10 && memcmp(prval->data, SYMLINK_HEADER, 10) == 0){
			prval->data[prval->dataLen - 1] = 0;    /* always 0-terminate! */
			convServerAsciiToUni(prval->data + 10, buffer, CIFS_MAX_PATHLEN + 1);
		}else{
			res.err.class = ERRCLS_CIFS_DOS;
			res.err.code = ERR_CIFS_DOSnoaccess;
		}
		res.path = buffer;
	}
	PROC_RETURN(&res);
	PROC_END;
}

static void	attrDOSReadLink(struct cifsFile *file, cifsUser_t *user, void *caller)
{
attrCtx_t	*ctx;
beFileRes_t	res;
  
	if(srvLinkMask(shrServer(file->share)) == 0){   /* no faked softlinks */
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSbadfunc;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, attrDOSReadLinkProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->user = user;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	attrDOSCreateLinkProc(void *obj, void *arg)
{
attrCtx_t		*this = obj;
beFileRes_t		res, *frval = arg;
filePrivRes_t	*prval = arg;
btu8			buffer[2 * CIFS_MAX_PATHLEN + 20];
btint			i, j;

	PROC_BEGIN;
	PROC_WAITFOR_X(1, fileCreateUnix(this->file, this->name, 0, 1, this->user, arg, this));
	if(frval->err.class != 0){
		res.err = frval->err;
		goto errorOccured;
	}
	this->file = frval->file;
	if(srvLinksAreUnicode(shrServer(this->file->share))){
		memcpy(buffer, SYMLINK_HEADER_UNI, (j=20));
		for(i=0;i<CIFS_MAX_PATHLEN && this->path[i] != 0;i++){
			buffer[j++] = this->path[i];
			buffer[j++] = this->path[i] >> 8;
		}
		buffer[j++] = 0;
		buffer[j++] = 0;  /* null terminate */
	}else{
		memcpy(buffer, SYMLINK_HEADER, 10);
		convUniToServerAscii(this->path, buffer + 10, 2 * CIFS_MAX_PATHLEN);
		j = strlen(buffer) + 1;
	}
	PROC_WAITFOR_X(6, fileRWUncached(this->file, FILE_OP_WRITE, 1, buffer, 0, j, this->user, this));
	if(prval->err.class != 0){
		fileRelease(this->file);
		this->file = NULL;
	}
	res.err = prval->err;
	res.file = this->file;
	errorOccured:;
	PROC_RETURN(&res);
	PROC_END;
}

static void	attrDOSCreateLink(struct cifsFile *parent, struct cifsUser *user, btuni *linkName, btuni *path, struct beFileSetAttr *setAttr, void *caller)
{
attrCtx_t	*ctx;
beFileRes_t	res;

	if(srvLinkMask(shrServer(parent->share)) == 0){ /* no faked softlinks */
		res.err.class = ERRCLS_CIFS_DOS;
		res.err.code = ERR_CIFS_DOSnoaccess;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, attrDOSCreateLinkProc, caller, attrPool, poolFreeBlock);
	ctx->file = parent;
	ctx->user = user;
	ctx->name = linkName;
	ctx->path = path;
	PROC_START(ctx, setAttr);
}

/* ------------------------------------------------------------------------- */

static void	attrDOSTouch(struct cifsFile *file, cifsUser_t *user, void *caller)
{
fileAttrDOS_t	*attr = file->attributes;
dirtyBits_t		dirty;

/* ### check srvCloseForSetattr() here if we open directories */
	attr->general.wtime = elpNow64();
	memset(&dirty, 0, sizeof(dirty));
	dirty.wtime = 1;
/* ### maybe acquire the attrLock lock? */
	attrDOSWriteback(file, &dirty, user, 0, caller);
}

/* ------------------------------------------------------------------------- */

static btu32	attrDOSWtime(struct cifsFile *file)
{
fileAttrDOS_t	*attr = file->attributes;

	if(attrGeneralValid(file, NULL)){
		return attrUnixTimeFromTime(attr->general.wtime, file);
	}
	return -1;
}


/* ------------------------------------------------------------------------- */

fileAttrInterface_t	attrDOSInterface = {
	attrDOSValid,
	attrDOSSetWritable,
	attrDOSMakeValid,
	attrDOSGetAttributes,
	attrDOSSetAttributes,
	attrDOSGetWritable,
	attrDOSReadLink,
	attrDOSTouch,
	attrDOSFree,
	attrDOSWtime,
	attrDOSCreateLink,
};

/* ------------------------------------------------------------------------- */
