/*
 * Name: fileattrunix.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-04-09
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: fileattrunix.c,v 1.5 1999/08/01 17:53:48 cs Exp $
 */

#include "stdheaders.h"
#include "hashtable.h"
#include "fileattrunix.h"
#include "file.h"
#include "smb.h"
#include "cifs.h"
#include "process.h"
#include "server.h"
#include "logging.h"
#include "blockpool.h"
#include "user.h"
#include "share.h"
#include "cache.h"
#include "cifsconst.h"
#include "unicode.h"
#include "evntloop.h"
#include "errcodes.h"
#include "backend.h"
#include "errcodes.h"
#include "dircache.h"

/* ------------------------------------------------------------------------- */

extern fileAttrInterface_t	attrUnixInterface;	/* forward declaration */

typedef struct dirtyBits{
	unsigned	atime	:1;
	unsigned	wtime	:1;
	unsigned	uid		:1;
	unsigned	gid		:1;
	unsigned	mode	:1;
}dirtyBits_t;

/* ------------------------------------------------------------------------- */

typedef struct attrCtx{
    PROC_VARSET;
    cifsFile_t		*file;
    btu16			uid;
    btu16			tid;
	cifsUser_t		*user;
	dirtyBits_t		dirty;
	btbool			writable;
	btbool			noLocking;
	beFileSetAttr_t	setAttr;
	bti64			oldWtime;
}attrCtx_t;

static blockpool_t *attrPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(attrPool, block);
}

/* ------------------------------------------------------------------------- */

void	*attrUnixNew(void)
{
fileAttrUnix_t	*attr = NULL;

	if(attrPool == NULL)
		attrPool = poolNew(sizeof(attrCtx_t), sizeof(attrCtx_t));
    attr = memCalloc(sizeof(fileAttrUnix_t));
	attrInit(&attr->general, &attrUnixInterface);
    return attr;
}

/* ------------------------------------------------------------------------- */

static void	attrUnixFree(void *obj)
{
	attrExit(obj);
	memFree(obj);
}

static inline btbool	attrUnixValid(cifsFile_t *file, cifsUser_t *user)
{
	return attrGeneralValid(file, user);
}

/* ------------------------------------------------------------------------- */

static void	attrUnixWritebackProc(void *obj, void *arg)
{
attrCtx_t		*this = obj;
cifsFile_t	*file = this->file;
fileAttrUnix_t	*attr = file->attributes;
server_t		*server = shrServer(file->share);
smbRequest_t	*smb = arg;
cifsRes_t		*rval = arg;
bti64			atime64, mtime64, uid, gid;
btu32			mode;
btuni			*path, buffer[CIFS_MAX_PATHLEN];
beFileRes_t		res;

	PROC_BEGIN;
	if(this->dirty.atime || this->dirty.wtime || this->dirty.uid || this->dirty.gid || this->dirty.mode){
		PROC_WAITFOR_X(1, srvWantRequest(server, this));
		atime64 = mtime64 = 0;
		uid = gid = -1;
		mode = -1;
		if(this->dirty.atime)
			atime64 = attrServerTimeFromTime(attr->general.atime, server);
		if(this->dirty.wtime)
			mtime64 = attrServerTimeFromTime(attr->general.wtime, server);
		if(this->dirty.uid)
			uid = attr->owner;
		if(this->dirty.gid)
			gid = attr->group;
		if(this->dirty.mode)
			mode = attr->mode;
		path = fileBuildPath(file, NULL, 0, buffer, CIFS_MAX_PATHLEN);
		PROC_WAITFOR_X(2, cifsSetPathInfoUnix(smb, this, this->uid, this->tid, path, attr->mode, atime64, mtime64, uid, gid));
	}
	if(rval != NULL)
		res.err = rval->err;
	else
		res.err.code = res.err.class = 0;
    PROC_RETURN(&res);
	PROC_END;
}

static void	attrUnixWriteback(struct cifsFile *file, dirtyBits_t *dirty, cifsUser_t *user, btbool noLocking, void *caller)
{
attrCtx_t	*ctx;

	PROC_INIT(ctx, attrUnixWritebackProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->dirty = *dirty;
	ctx->uid = userSmbUid(user);
/*	ctx->noLocking = noLocking;	currently not used */
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	attrUnixMakeValidProc(void *obj, void *arg)
{
attrCtx_t			*this = obj;
beFileRes_t			res;
smbRequest_t		*smb = arg;
cifsFile_t		*file = this->file;
fileAttrUnix_t		*attr = file->attributes;
server_t			*server = shrServer(file->share);
btuni				*path, buffer[CIFS_MAX_PATHLEN];
cifsResQueryFile_t	*rval = arg;
btbool				timeChanged;

	PROC_BEGIN
    if(file->children == NULL && file->cache.any != NULL){	/* write back to obtain correct file size */
        PROC_WAITFOR_X(1, cacheWriteBack(file->cache.any, this));
    }
	this->oldWtime = attr->general.wtime;
	PROC_WAITFOR_X(2, srvWantRequest(server, this));
	path = fileBuildPath(file, NULL, 0, buffer, CIFS_MAX_PATHLEN);
	PROC_WAITFOR_X(3, cifsQueryPathInfo(smb, this, this->uid, this->tid, path, SMB_QUERY_FILE_UNIX_BASIC));
	if(rval->err.class == 0){
		attr->general.size = rval->size;
		attr->general.allocSize = rval->allocSize;
		attr->general.atime = attrTimeFromServerTime(rval->atime, server);
		attr->general.wtime = attrTimeFromServerTime(rval->mtime, server);
		attr->chtime = attrTimeFromServerTime(rval->ctime, server);
		attr->type = rval->type;
		attr->mode = rval->mode;
		attr->nlinks = rval->nlinks;
		attr->devMajor = rval->devMajor;
		attr->devMinor = rval->devMinor;
		attr->owner = rval->fileUid;	/* mapping is done in conversion to NFS */
		attr->group = rval->fileGid;
		attr->general.updatedAt = elpNow();
	}
	timeChanged = attr->general.wtime != this->oldWtime;
	if(file->children != NULL && file->cache.any != NULL && (timeChanged || rval->err.class != 0)){
		dirCacheDirWasModified(file->cache.any);	/* necessary to avoid conflict with dir cache of OS */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
			logPrintf(LOGM_DEBUG, "attrUnixMakeValid: flushing dir cache of ->%s<-\n", convUniToAscii(file->name));
			if(timeChanged)
				logPrintf(LOGM_DEBUG, "attrUnixMakeValid: wtime changed from 0x%08x%08x to 0x%08x%08x\n", (int)(this->oldWtime >> 32), (int)this->oldWtime, (int)(attr->general.wtime >> 32), (int)attr->general.wtime);
		}
	}
    if(attr->type == SMB_UNIX_FILETYPE_DIR && file->children == NULL){
		fileIsDirectory(file);
    }
    res.err = rval->err;
    PROC_RETURN(&res);
    PROC_END
}

static void	attrUnixMakeValid(struct cifsFile *file, cifsUser_t *user, void *caller)
{
attrCtx_t	*ctx;
beFileRes_t	res;

	if(attrUnixValid(file, user)){
		res.err.class = res.err.code = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, attrUnixMakeValidProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->tid = shrTid(file->share);
	ctx->uid = userSmbUid(user);
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	attrUnixSetWritableProc(void *obj, void *arg)
{
attrCtx_t		*this = obj;
cifsFile_t	*file = this->file;
fileAttrUnix_t	*attr = file->attributes;
beFileRes_t		*rval = arg;
dirtyBits_t		dirty;

	PROC_BEGIN;
	if(!attrUnixValid(file, this->user)){
		PROC_WAITFOR_X(1, attrUnixMakeValid(file, this->user, this));
	}
	if(rval->err.class != 0){
		PROC_RETURN(rval);
	}
	if(this->writable){	/* Apply to owner, regardless of current user. If user != owner, attrUnixWriteback will fail */
		attr->mode |= 0x200;
	}else{
		attr->mode &= ~0x200;
	}
	memset(&dirty, 0, sizeof(dirty));
	dirty.mode = 1;
	PROC_WAITFOR_X(2, attrUnixWriteback(file, &dirty, this->user, this->noLocking, this));
	PROC_RETURN(rval);
	PROC_END;
}

static void	attrUnixSetWritable(struct cifsFile *file, btbool writable, cifsUser_t *user, btbool noLocking, void *caller)
{
attrCtx_t	*ctx;

	PROC_INIT(ctx, attrUnixSetWritableProc, caller, attrPool, poolFreeBlock);
	ctx->writable = writable;
	ctx->file = file;
	ctx->user = user;
	ctx->noLocking = noLocking;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	attrUnixGetAttributes(cifsFile_t *file, struct cifsUser *user, struct beFileAttr *readAttr)
{
fileAttrUnix_t	*attr = file->attributes;

	readAttr->type = attr->type;
	readAttr->unixMode = attr->mode;
	readAttr->numLinks = attr->nlinks;
	readAttr->uid = attr->owner;	/* ### translate */
	readAttr->gid = attr->group;	/* ### translate */
	readAttr->size = attr->general.size;
	readAttr->allocatedSize = attr->general.allocSize;
	readAttr->devMajor = attr->devMajor;
	readAttr->devMinor = attr->devMinor;
    attrFileVfsTime(&readAttr->atime, attr->general.atime);
    attrFileVfsTime(&readAttr->mtime, attr->general.wtime);
    attrFileVfsTime(&readAttr->ctime, attr->chtime);
}

/* ------------------------------------------------------------------------- */

static void	attrUnixSetAttributesProc(void *obj, void *arg)
{
attrCtx_t		*this = obj;
cifsFile_t	*file = this->file;
fileAttrUnix_t	*attr = file->attributes;
beFileRes_t		*rval = arg;
dirtyBits_t		dirty;
beFileSetAttr_t	*setAttr = &this->setAttr;

	PROC_BEGIN;
	memset(&dirty, 0, sizeof(dirty));
    if(setAttr->setUnixMode){
		attr->mode = setAttr->unixMode;
		dirty.mode = 1;
	}
    if(setAttr->setUid){
		attr->owner = setAttr->uid;		/* ### translate */
		dirty.uid = 1;
	}
    if(setAttr->setGid){
		attr->group = setAttr->gid;		/* ### translate */
		dirty.gid = 1;
	}
    if(setAttr->setAtime){
        attr->general.atime = attrTimeFromVfsTime(&setAttr->atime);
        dirty.atime = 1;
    }
    if(setAttr->setMtime){
        attr->general.wtime = attrTimeFromVfsTime(&setAttr->mtime);
        dirty.wtime = 1;
    }
	PROC_WAITFOR_X(1, attrUnixWriteback(file, &this->dirty, this->user, 0, this));
	if(rval->err.class != 0){
		PROC_RETURN(rval);
	}
    if(setAttr->setSize){
		PROC_WAITFOR_X(2, fileTruncate(file, setAttr->size, this->user, this));
    }
	PROC_RETURN(rval);
	PROC_END;
}

static void	attrUnixSetAttributes(struct cifsFile *file, cifsUser_t *user, struct beFileSetAttr *setAttr, void *caller)
{
attrCtx_t	*ctx;

	PROC_INIT(ctx, attrUnixSetAttributesProc, caller, attrPool, poolFreeBlock);
	ctx->file = file;
	ctx->user = user;
	ctx->setAttr = *setAttr;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static btbool	attrUnixGetWritable(struct cifsFile *file, cifsUser_t *userCtx)
{
fileAttrUnix_t	*attr = file->attributes;

	/* We should look at the user/group accessing the file and return the
	 * appropriate write flag. Since this function is only useful to find
	 * out the write protection status for the file's owner (because others
	 * can't change it), it is sufficient to return the owner's status.
	 * This is really contradicting the documentation, but we should get
	 * away with it.
	 */
	 return (attr->mode & 0200) != 0;
}

/* ------------------------------------------------------------------------- */

static void	attrUnixReadLink(struct cifsFile *file, cifsUser_t *user, void *caller)
{
beFileRes_t	res;			/* ### dummy implementation */
procCtx_t	*p = caller;

	res.err.class = ERRCLS_CIFS_DOS;
	res.err.code = ERR_CIFS_DOSnoaccess;
	p->process(p, &res);
}

/* ------------------------------------------------------------------------- */

static void   attrUnixCreateLink(struct cifsFile *parent, struct cifsUser *user, btuni *linkName, btuni *path, struct beFileSetAttr *setAttr, void *caller)
{
beFileRes_t	res;		/* ### dummy implementation */
procCtx_t	*p = caller;

	res.err.class = ERRCLS_CIFS_DOS;
	res.err.code = ERR_CIFS_DOSnoaccess;
	p->process(p, &res);
}

/* ------------------------------------------------------------------------- */

static void	attrUnixTouch(struct cifsFile *file, cifsUser_t *user, void *caller)
{
fileAttrUnix_t	*attr = file->attributes;
dirtyBits_t		dirty;

	attr->general.wtime = elpNow64();
	memset(&dirty, 0, sizeof(dirty));
	dirty.wtime = 1;
	attrUnixWriteback(file, &dirty, user, 0, caller);
}

/* ------------------------------------------------------------------------- */

static btu32	attrUnixGetWtime(struct cifsFile *file)
{
fileAttrUnix_t	*attr = file->attributes;

	if(attrGeneralValid(file, NULL)){
		return attrUnixTimeFromTime(attr->general.wtime, file);
	}
	return -1;
}

/* ------------------------------------------------------------------------- */

fileAttrInterface_t	attrUnixInterface = {
	attrGeneralValid,		/* user superclass method */
	attrUnixSetWritable,
	attrUnixMakeValid,
	attrUnixGetAttributes,
	attrUnixSetAttributes,
	attrUnixGetWritable,
	attrUnixReadLink,
	attrUnixTouch,
	attrUnixFree,
	attrUnixGetWtime,
	attrUnixCreateLink,
};

/* ------------------------------------------------------------------------- */
