/*
 * Name: lmb.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-01-03
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: lmb.c,v 1.3 1999/07/26 20:06:22 cs Exp $
 */

#include "stdheaders.h"
#include "hashtable.h"
#include "process.h"
#include "memalloc.h"
#include "evntloop.h"
#include "server.h"
#include "nbnsclient.h"
#include "blockpool.h"
#include "util.h"
#include "unicode.h"
#include "lmb.h"
#include "cfgparse.h"
#include "main.h"
#include "cifsmod.h"
#include "errcodes.h"
#include "logging.h"

/* ------------------------------------------------------------------------- */

typedef struct lmb{
	btu8		*domain;
	btu8		*srvName;
	bttime		createdAt;
	btint		refCount;
}lmb_t;

/* ------------------------------------------------------------------------- */

typedef struct{
	PROC_VARSET;
	btu8		domain[64];
	server_t	*srv;
}lmbCtx_t;

static blockpool_t *lmbPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(lmbPool, block);
}

/* ------------------------------------------------------------------------- */

static hashtable_t	*lmbsByDomain = NULL;
static hashtable_t	*lmbServers = NULL;

/* ------------------------------------------------------------------------- */

static void	lmbFree(lmb_t *lmb)
{
	if(lmb != NULL){
		if(lmb->srvName != NULL){
			lmbRemoveServer(lmbLookupServer(lmb->srvName));
			memFree(lmb->srvName);
		}
		if(lmb->domain != NULL)
			memFree(lmb->domain);
		memset(lmb, 0, sizeof(*lmb));
		memFree(lmb);
	}
}

static lmb_t	*lmbNew(btu8 *volatile domain, btu8 *volatile srvName)
{
lmb_t	*volatile lmb;

EXC_TRY
	lmb = memCalloc(sizeof(*lmb));
	lmb->refCount = 0;
	lmb->domain = utilStrDup(domain);
	lmb->srvName = utilStrDup(srvName);
	lmb->createdAt = elpNow();
EXC_CATCH
	lmbFree(lmb);
	EXC_RETHROW();
EXC_ENDTRY
	return lmb;
}

static void	lmbRetain(void *obj)
{
lmb_t	*lmb = obj;

	lmb->refCount++;
}

static void	lmbRelease(void *obj)
{
lmb_t	*lmb = obj;

	lmb->refCount--;
    if(lmb->refCount == 0){
        lmbFree(lmb);
    }else if(lmb->refCount < 0){
		DEBUG_TRAP();
	}
}

static btint	lmbCompare(void *obj1, void *obj2)
{
lmb_t	*this = obj1, *other = obj2;

	return utilStrCmp(this->domain, other->domain, 0);
}

static btu32	lmbHashVal(void *obj)
{
lmb_t	*lmb = obj;

	return utilStrHash(lmb->domain, 0);
}

/* ------------------------------------------------------------------------- */

static btint	srvCompare(void *obj1, void *obj2)
{
	return utilStrCmp(srvName(obj1), srvName(obj2), 0);
}
 
static btu32	srvHashVal(void *obj)
{
	return utilStrHash(srvName(obj), 0);
}

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	lmbsByDomainInterface = {
	lmbCompare,
	lmbHashVal,
	lmbRetain,
	lmbRelease
};

static collectionObjectI_t	lmbServersInterface = {
	srvCompare,
	srvHashVal,
	srvRetain,
	srvRelease
};

/* ------------------------------------------------------------------------- */

void	lmbForDomainProc(void *obj, void *arg)
{
lmbCtx_t		*this = obj;
lmb_t			*lmb;
bttime			now;
static bttime	maxAge;
static btu32	maxAgeChg = 0;
btu8			*srvName, *domainp;
nbnsRes_t		*rval = arg;
srvRes_t		*srval = arg;
lmbRes_t		res;
btint			i, numEntries;
cfgProperty_t	*path[8];

	PROC_BEGIN;
	srvName = NULL;
	domainp = this->domain;
	if(lmbsByDomain != NULL && (lmb = htabFindObjectEqualTo(lmbsByDomain, &domainp)) != NULL){
		if(CFG_DID_CHANGE(maxAgeChg)){
			numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
			maxAge = cfgIntValue(cfgQueryFromArray("lmbCacheTime", numEntries, path), 300000);
		}
		now = elpNow();
		if(now - lmb->createdAt >= 0 && now - lmb->createdAt < maxAge){
			srvName = lmb->srvName;
		}else{
			htabRemoveObjectEqualTo(lmbsByDomain, lmb);	/* is outdated */
		}
	}
	if(srvName == NULL){
		PROC_WAITFOR_X(1, nbnsResolve(this->domain, 0x1d, 0xffffffff, 1, this));
		if(rval->err.class != 0){
			res.err = rval->err;
			PROC_RETURN(&res);
		}
		PROC_WAITFOR_X(2, nbnsNodeStatus(rval->ipAddress, this->domain, 0x1d, this));
		if(rval->err.class == ERRCLS_CIFS_LOCAL && rval->err.code == ERR_CIFS_LOCtimeout){
			/* try broadcasting node status request if unicast failed */
			PROC_WAITFOR_X(3, nbnsNodeStatus(0xffffffff, this->domain, 0x1d, this));
		}
		if(rval->err.class != 0){
			res.err = rval->err;
			PROC_RETURN(&res);
		}
		srvName = rval->name;
		/* capitalize server name */
		for(i = 0; i < 15 && srvName[i] != 0; i++){
			if(i == 0)
				srvName[i] = asciiCharToupper(srvName[i]);
			else
				srvName[i] = asciiCharTolower(srvName[i]);
		}
		srvName[i] = 0;
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "lmb: found LMB = %s\n", srvName);
	}
	this->srv = srvWithName(srvName, 1);
	lmbAddServer(this->srv, this->domain);
	PROC_WAITFOR_X(4, srvConnect(this->srv, 0, this));
	if(srval->err.class != 0){
		lmbRemoveServer(this->srv);
		srvRelease(this->srv);
	}
	res.err = srval->err;
	res.server = this->srv;
	PROC_RETURN(&res);
	PROC_END;
}

void	lmbForDomain(btu8 *domain, void *caller)
{
lmbCtx_t	*ctx;

	if(lmbPool == NULL)
		lmbPool = poolNew(sizeof(lmbCtx_t), sizeof(lmbCtx_t));
	PROC_INIT(ctx, lmbForDomainProc, caller, lmbPool, poolFreeBlock);
	if(strlen(domain) >= sizeof(ctx->domain))
		domain[sizeof(ctx->domain) - 1] = 0;	/* emergency brake */
	strcpy(ctx->domain, domain);
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	lmbAddServer(struct server *srv, btu8 *domain)
{
	if(lmbsByDomain == NULL)
		lmbsByDomain = htabNew(&lmbsByDomainInterface, 4, 75);
	if(lmbServers == NULL)
		lmbServers = htabNew(&lmbServersInterface, 4, 75);
	htabAddObject(lmbsByDomain, lmbNew(domain, srvName(srv)));
	htabAddObject(lmbServers, srv);
}

/* ------------------------------------------------------------------------- */

void	lmbRemoveServer(struct server *srv)
{
	if(lmbServers != NULL && srv != NULL)
		htabRemoveObjectEqualTo(lmbServers, srv);
}

/* ------------------------------------------------------------------------- */

struct server	*lmbLookupServer(btu8 *srvName)
{
	if(lmbServers == NULL)
		return NULL;
	return htabFindObjectEqualTo(lmbServers, &srvName);
}

/* ------------------------------------------------------------------------- */

struct array	*lmbAllServers(void)
{
	if(lmbServers == NULL){
		return arrNew(&lmbServersInterface, 1);
	}else{
		return htabAllObjects(lmbServers);
	}
}

/* ------------------------------------------------------------------------- */

