/*
 * Name: message.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-21
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: message.h,v 1.1.1.1 1999/04/12 13:56:35 cs Exp $
 */

#ifndef __message_H_INCLUDED__
#define __message_H_INCLUDED__

#include "basictypes.h"
#include "fifo.h"

/*
General Description:
This module implements a container class for netbios messages. The main
purpose why this class is used and not dynamically allocated memory is the
possibility to reserve some bytes for the netbios header without any need
for the message creating procedures to know about that header. An other
advantage is the possibility to reuse message objects, saving the overhead
of repeated malloc/free operations. All marshalling functions in "smb.c" and
"smb.h" are based on this class.
*/

/* ------------------------------------------------------------------------- */

typedef struct msg{
	fifoObject_t	fifo;	/* we are subclass of fifo */
	btu8			*data;
	btsize			allocated;
	btsize			reserved;	/* reserved bytes at beginning of msg */
	btsize			size;
	btsize			readIndex;
	int				usageCount;	/* usages since last resize */
	int				refCount;
}nbmsg_t;

/* ------------------------------------------------------------------------- */

nbmsg_t	*msgNew(btsize reserved);

void	msgIncreaseSize(nbmsg_t *this, btsize size);
void	msgRetain(void *obj);
void	msgRelease(void *obj);

void	msgAddSpace(nbmsg_t *this, btsize bytes);
void	msgAddByte(nbmsg_t *this, btu8 data);
void	msgAddBlock(nbmsg_t *this, btu8 *data, btsize len);
void	msgSetByte(nbmsg_t *this, btsize index, btu8 data);
void	msgInsertBlock(nbmsg_t *this, btu8 *data, btsize len);


/* ------------------------------------------------------------------------- */

static inline void	msgEnsureSpace(nbmsg_t *this, btsize space)
{
	if(this->size + space > this->allocated)
		msgIncreaseSize(this, this->size + space);
}

static inline btsize	msgSize(nbmsg_t *this)
{
	return this->size - this->reserved;
}

static inline void	msgAddByteUnchecked(nbmsg_t *this, btu8 data)
{
	this->data[this->size++] = data;
}

static inline void	msgResetRead(nbmsg_t *this)
{
	this->readIndex = this->reserved;
}

static inline void	msgClear(nbmsg_t *this, btsize reserved)
{
	this->reserved = reserved;
	this->size = reserved;
	this->readIndex = this->reserved;
	if(this->size > this->allocated)
		msgIncreaseSize(this, this->size);
}

static inline void	msgPeekData(nbmsg_t *this, btu8 **data, btsize *len)
{
	*len = this->size - this->readIndex;
	*data = &this->data[this->readIndex];
}

static inline void	msgReadData(nbmsg_t *this, btu8 **data, btsize len)
{
	if(this->readIndex + len > this->size){
		len = this->size - this->readIndex;
	}
	*data = &this->data[this->readIndex];
	this->readIndex += len;
}

/* ------------------------------------------------------------------------- */

#endif	/* __message_H_INCLUDED__ */
