/*
 * Name: nbnsclient.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-11-17
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nbnsclient.c,v 1.3 1999/07/26 20:06:23 cs Exp $
 */

#include "stdheaders.h"
#include "nbnsclient.h"
#include "process.h"
#include "main.h"
#include "cfgparse.h"
#include "blockpool.h"
#include "smbnumbers.h"
#include "logging.h"
#include "evntloop.h"
#include "system.h"
#include "util.h"
#include "errcodes.h"
#include "cifsmod.h"
#include "unicode.h"


#define	NBNS_PORT	137

typedef struct nbnsCtx{
	PROC_VARSET;
	btu8	*block;
	btsize	len;
	btu32	destIp;
	btint	sockFd;
	bttime	absMaxTime;
	bttime	timeout;
	btu16	xid;
	btbool	abortIfTimeout;
	btu8	buffer[580];
}nbnsCtx_t;

static blockpool_t *nbnsPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(nbnsPool, block);
}

static btu16	xidCounter;

/* ------------------------------------------------------------------------- */

static void	nbnsRunQueryProc(void *obj, void *arg)
{
nbnsCtx_t			*this = obj;
struct sockaddr_in	txAddr, rxAddr;
btint				len, slen, numEntries;
btu16				xid;
nbnsRes_t			res;
cfgProperty_t		*path[8];
bttime				nextTimeout;
static bttime		maxTime, initialTimeout;
static btu32		timeoutsChg = 0;

	PROC_BEGIN;
	if((this->sockFd = socket(AF_INET, SOCK_DGRAM, 0)) < 0){
		res.err.class = ERRCLS_UNIX;
		res.err.code = errno;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nbns: socket(): [%d] %s\n", errno, sysStrerror(errno));
		goto errorOccured;
	}
#ifdef SO_BROADCAST
	len = 1;	/* switch on broadcast permission */
	if(setsockopt(this->sockFd, SOL_SOCKET, SO_BROADCAST, (void *)&len, sizeof(len)) < 0 && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "nbnsclient: Error setting broadcast option on socket: [%d] %s\n", errno, sysStrerror(errno));
	/* len used as dummy here */
#endif
	this->xid = ((btu16)this->block[0] << 8) | this->block[1];
	if(CFG_DID_CHANGE(timeoutsChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		initialTimeout = cfgIntValue(cfgQueryFromArray("nbnsInitialTimeout", numEntries, path), 100);
		maxTime = cfgIntValue(cfgQueryFromArray("nbnsTotalTimeout", numEntries, path), 1000);
	}
	this->timeout = initialTimeout;
	this->absMaxTime = elpNow() + maxTime;
	for(;;){
		PROC_WAITFOR_X(1, elpRegisterForFd(this->sockFd, ELP_FDMODE_WRITE, nbnsRunQueryProc, this));
		elpCancelForFd(this->sockFd, ELP_FDMODE_WRITE);
		memset(&txAddr, 0, sizeof(txAddr));
		txAddr.sin_family = AF_INET;
		txAddr.sin_port = htons(NBNS_PORT);
		txAddr.sin_addr.s_addr = htonl(this->destIp);
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nbns: sending request to 0x%08x: %02x %02x %02x %02x...\n", this->destIp, this->block[0], this->block[1], this->block[2], this->block[3]);
		if(sendto(this->sockFd, this->block, this->len, 0, (struct sockaddr *)&txAddr, sizeof(txAddr)) < 0){
			res.err.class = ERRCLS_UNIX;
			res.err.code = errno;
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "nbns: sendto(): [%d] %s\n", errno, sysStrerror(errno));
			goto errorOccured;
		}
		nextTimeout = elpNow() + this->timeout;
		if(nextTimeout - this->absMaxTime >= -50){
			this->abortIfTimeout = 1;
			nextTimeout = this->absMaxTime;
		}
		elpCallAtCancelPrev(nextTimeout, nbnsRunQueryProc, this);	/* setup timeout */
		PROC_WAITFOR_X(2, elpRegisterForFd(this->sockFd, ELP_FDMODE_READ, nbnsRunQueryProc, this));
		elpCancelCallAt(nbnsRunQueryProc, this);
		elpCancelForFd(this->sockFd, ELP_FDMODE_READ);
		if(arg != NULL){	/* no read timeout */
			slen = sizeof(rxAddr);
			if((len = recvfrom(this->sockFd, this->buffer, sizeof(this->buffer), 0, (struct sockaddr *)&rxAddr, &slen)) < 0){
				res.err.class = ERRCLS_UNIX;
				res.err.code = errno;
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "nbns: recvfrom(): [%d] %s\n", errno, sysStrerror(errno));
				goto errorOccured;
			}
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "nbns: received reply %02x %02x %02x %02x...\n", this->buffer[0], this->buffer[1], this->buffer[2], this->buffer[3]);
			xid = ((btu16)this->buffer[0] << 8) | this->buffer[1];
			if(xid != this->xid){
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
					logPrintf(LOGM_ERROR, "nbns: received xid=%d instead of %d, ignoring\n", xid, this->xid);
			}else{
				memset(&res.err, 0, sizeof(res.err));
				res.data = this->buffer;
				res.len = len;
				res.ipAddress = ntohl(rxAddr.sin_addr.s_addr);
				break;
			}
		}
		if(this->abortIfTimeout){
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCtimeout;
			goto errorOccured;
		}
		this->timeout *= 2;	/* ... and retry */
	}
errorOccured:
	if(this->sockFd >= 0){
		close(this->sockFd);
		this->sockFd = -1;
	}
	PROC_RETURN(&res);
	PROC_END;
}

static void	nbnsRunQuery(btu8 *block, btsize len, btu32 destIp, void *caller)
{
nbnsCtx_t	*ctx;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	PROC_INIT(ctx, nbnsRunQueryProc, caller, nbnsPool, poolFreeBlock);
	ctx->block = block;
	ctx->len = len;
	ctx->destIp = destIp;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

btsize	nbnsEncodeName(btu8 *dest, btsize bufSize, btu8 *name, btint tag)
{
btint			i, len, numEntries;
btu8			c, *s = dest;
cfgProperty_t	*path[8];
static btu32	scopeChg = 0;
static btu8		*scopeID;

	*dest++ = 32;
	for(i=0;i<15;i++){
		if(*name == 0)	/* name may be null-terminated string */
			break;
		c = asciiCharToupper(*name);
		name++;
		*dest++ = 'A' + ((c >> 4) & 0xf);
		*dest++ = 'A' + (c & 0xf);
	}
	for(;i<15;i++){		/* fill rest with spaces */
		*dest++ = 'A' + 2;
		*dest++ = 'A' + 0;
	}
	*dest++ = 'A' + ((tag >> 4) & 0xf);
	*dest++ = 'A' + (tag & 0xf);
	*dest++ = 0;
	bufSize -= 34;
	if(CFG_DID_CHANGE(scopeChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		scopeID = cfgStringValue(cfgQueryFromArray("scopeID", numEntries, path), NULL);
	}
	if(scopeID != NULL){
		for(i=len=0;bufSize > 0;i++,len++){
			switch(scopeID[i]){
			case 0:
				dest[-1] = len;
				if(len > 0)
					dest[len] = 0;
				goto outerBreak;
			case '.':
				dest[-1] = len;
				dest += len + 1;
				len = 0;
				bufSize--;
				break;
			default:
				dest[len] = scopeID[i];
				bufSize--;
				break;
			}
		}
	}
outerBreak:
    for(len = 1; *s != 0; s += *s + 1){
        len += *s + 1;
    }
    return len;
}

/* ------------------------------------------------------------------------- */

static btsize	skipName(btu8 *name, btsize remainingLen)
{
btu8	*p = name;
btint	len;

	for(;;){
		if(p - name > remainingLen)	/* limit to max. len */
			return remainingLen;
		len = *p++;
		if(len == 0 || len > 63)	/* end of name or pointer */
			return p - name;
		p += len;	/* skip this name */
	}
}

/* ------------------------------------------------------------------------- */

static void	nbnsResolveProc(void *obj, void *arg)
{
nbnsCtx_t	*this = obj;
nbnsRes_t	res, *rval = arg;
btint		rcode;
btbool		redirect;
btu8		*p;

	PROC_BEGIN;
	for(;;){
		PROC_WAITFOR_X(1, nbnsRunQuery(this->buffer, this->len, this->destIp, this));
		if(rval->err.class != 0){
			rval->ipAddress = 0;
			PROC_RETURN(rval);
		}
		if(rval->len < 23){	/* can't be a valid package */
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCbogusdata;
			PROC_RETURN(&res);
		}
		rcode = rval->data[3] & 0xf;
		redirect = (rval->data[2] & 0x4) == 0;
		if(rcode != 0){	/* this was an error, report as not found */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "nbns: nbnsResolve: error %d\n", rcode);
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOClookup;
			PROC_RETURN(&res);
		}
		if(redirect){	/* we are redirected to an other nameserver */
			if(this->abortIfTimeout){	/* misuse as counter: accept only one redirect */
				res.err.class = ERRCLS_CIFS_LOCAL;
				res.err.code = ERR_CIFS_LOClookup;
				PROC_RETURN(&res);
			}
			this->destIp = (btu32)(rval->data[rval->len - 4]) << 24;
			this->destIp |= (btu32)(rval->data[rval->len - 3]) << 16;
			this->destIp |= (btu32)(rval->data[rval->len - 2]) << 8;
			this->destIp |= rval->data[rval->len - 1];
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "nbns: nbnsResolve: redirect to ip=0x%08x\n", this->destIp);
			this->abortIfTimeout = 1;
		}else{	/* authoritative non-error answer */
			p = rval->data + 12;
			p += skipName(p, rval->len - 12);
			p += 10;	/* skip NB, IN, TTL and RDLENGTH */
			if(p - rval->data + 6 <= rval->len){
				p += 2;	/* skip flags field */
				res.ipAddress = (btu32)*p++ << 24;
				res.ipAddress |= (btu32)*p++ << 16;
				res.ipAddress |= (btu32)*p++ << 8;
				res.ipAddress |= *p++;
				if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
					logPrintf(LOGM_DEBUG, "nbns: nbnsResolve: found ip=0x%08x\n", res.ipAddress);
				memset(&res.err, 0, sizeof(res.err));
				PROC_RETURN(&res);
			}else{	/* not enough data */
				res.err.class = ERRCLS_CIFS_LOCAL;
				res.err.code = ERR_CIFS_LOCbogusdata;
				PROC_RETURN(&res);
			}
		}
	}
	PROC_END;
}

void	nbnsResolve(btu8 *name, btint tag, btu32 nbnsIp, btbool broadcast, void *caller)
{
nbnsCtx_t	*ctx;
btu8		*p;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "nbnsclient: resolving %s <%02x>\n", name, tag);
	PROC_INIT(ctx, nbnsResolveProc, caller, nbnsPool, poolFreeBlock);
	ctx->destIp = nbnsIp;
	p = ctx->buffer;
	*p++ = xidCounter >> 8;
	*p++ = xidCounter++;
	*p++ = 1;
	*p++ = broadcast ? 0x10 : 0;
	*p++ = 0;
	*p++ = 1;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	p += nbnsEncodeName(p, 256, name, tag);
	*p++ = 0;
	*p++ = 0x20;
	*p++ = 0;
	*p++ = 1;
	ctx->len = p - ctx->buffer;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	nbnsNodeStatusProc(void *obj, void *arg)
{
nbnsCtx_t	*this = obj;
nbnsRes_t	*rval = arg;
btint		rcode, i, entries;
btu8		*p;

	PROC_BEGIN;
	PROC_WAITFOR_X(1, nbnsRunQuery(this->buffer, this->len, this->destIp, this));
	if(rval->err.class != 0){
		PROC_RETURN(rval);
	}
	if(rval->len < 23){	/* can't be a valid package */
		rval->err.class = ERRCLS_CIFS_LOCAL;
		rval->err.code = ERR_CIFS_LOCbogusdata;
		PROC_RETURN(rval);
	}
	rcode = rval->data[3] & 0xf;
	if(rcode != 0){	/* this was an error */
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nbns: nbnsNodeStatus: error %d\n", rcode);
		rval->err.class = ERRCLS_CIFS_LOCAL;
		rval->err.code = ERR_CIFS_LOClookup;
		PROC_RETURN(rval);
	}
	p = rval->data + 12;
	p += skipName(p, rval->len - 12);
	p += 10;	/* skip NB, IN, TTL and RDLENGTH */
	entries = *p++;	/* number of NODE_NAME entries */
	for(i=0;i<entries;i++){
		if(p + 18 > rval->data + rval->len)	/* not an entire record */
			break;
		if(p[15] == 0x20 || p[15] == 0x00){	/* first 0x20 or 0x00 is name */
			for(i=0;i<15 && p[i] != 0x20;i++);
			p[i] = 0;
			rval->name = p;
			PROC_RETURN(rval);
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
			rcode = p[15];
			p[15] = 0;
			logPrintf(LOGM_DEBUG, "nbns: nbnsNodeStatus: skipping %s<%d>\n", p, rcode);
		}
		p += 18;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "nbns: nbnsNodeStatus: no valid 0x20 name found\n");
	rval->err.class = ERRCLS_CIFS_LOCAL;
	rval->err.code = ERR_CIFS_LOClookup;
	PROC_RETURN(rval);
	PROC_END;
}

void	nbnsNodeStatus(btu32 destIp, btu8 *name, btint tag, void *caller)
{
nbnsCtx_t	*ctx;
btu8		*p;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	PROC_INIT(ctx, nbnsNodeStatusProc, caller, nbnsPool, poolFreeBlock);
	ctx->destIp = destIp;
	p = ctx->buffer;
	*p++ = xidCounter >> 8;
	*p++ = xidCounter++;
	*p++ = 0;
	*p++ = 0;	/* should be 0x10 for broadcast, but we should not broadcast anyway */
	*p++ = 0;
	*p++ = 1;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	*p++ = 0;
	p += nbnsEncodeName(p, 256, name, tag);
	*p++ = 0;
	*p++ = 0x21;
	*p++ = 0;
	*p++ = 1;
	ctx->len = p - ctx->buffer;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

typedef struct nsCache{
	btu8	*hostname;
	btu32	ipAddress;
	bttime	updatedAt;
}nsCache_t;

static hashtable_t	*nsCache = NULL;

static btint	cacheCompare(void *o1, void *o2)
{
nsCache_t	*this = o1, *other = o2;

	return utilStrCmp(this->hostname, other->hostname, 0);
}

static btu32	cacheHashVal(void *obj)
{
nsCache_t	*this = obj;

	return utilStrHash(this->hostname, 0);
}

static collectionObjectI_t	cacheInterface = {
	cacheCompare,
	cacheHashVal,
	NULL,
	NULL,
};

/* ------------------------------------------------------------------------- */

static bttime	maxCacheTime(void)
{
static btu32	timeChg = 0;
static bttime	maxTime;
btint			numEntries;
cfgProperty_t	*path[8];

	if(CFG_DID_CHANGE(timeChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		maxTime = cfgIntValue(cfgQueryFromArray("nbnsCacheTime", numEntries, path), 60*1000);
	}
	return maxTime;
}

static void	cacheEntryFree(nsCache_t *c)
{
	if(c != NULL){
		if(c->hostname != NULL){
			htabRemoveObjectEqualTo(nsCache, c);
			memFree(c->hostname);
		}
		memset(c, 0, sizeof(*c));
		memFree(c);
	}
}

static nsCache_t	*cacheEntryAdd(btu8 *name, btu32 ip)
{
nsCache_t	*volatile c;

	c = htabFindObjectEqualTo(nsCache, &name);
	if(c != NULL){
		c->ipAddress = ip;
		c->updatedAt = elpNow();
		return c;
	}
EXC_TRY;
	c = memCalloc(sizeof(*c));
	c->hostname = utilStrDup(name);
	c->ipAddress = ip;
	c->updatedAt = elpNow();
	htabAddObject(nsCache, c);
EXC_CATCH;
	cacheEntryFree(c);
	EXC_RETHROW();
EXC_ENDTRY;
	return c;
}

static void	cacheCheckOutdated(void)
{
array_t		*allEntries;
btint		i, cnt;
nsCache_t	*c;
bttime		now, age, maxTime = maxCacheTime();

	allEntries = htabAllObjects(nsCache);
	cnt = arrObjectCount(allEntries);
	now = elpNow();
	for(i=0;i<cnt;i++){
		c = arrObjectAtIndex(allEntries, i);
		age = now - c->updatedAt;
		if(age < 0 || age > maxTime){
			cacheEntryFree(c);
		}
	}
	arrFree(allEntries);
}

/* ------------------------------------------------------------------------- */

static void	nbnsLookupProc(void *obj, void *arg)
{
nbnsCtx_t		*this = obj;
nbnsRes_t		*rval = arg;
btu32			ip, *res;
btint			numEntries;
cfgProperty_t	*path[8];
static btu32	cfgChg = 0;
static btbool	tryNetbios, tryDns;
static btu8		*wins;

	if(CFG_DID_CHANGE(cfgChg)){
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		tryNetbios = cfgEnumValue(cfgQueryFromArray("lookupTryNetbios", numEntries, path), cfgYesNoEnum, 1);
		tryDns = cfgEnumValue(cfgQueryFromArray("lookupTryDns", numEntries, path), cfgYesNoEnum, 1);
		wins = cfgStringValue(cfgQueryFromArray("nbnsWinsIp", numEntries, path), "255.255.255.255");
	}
	PROC_BEGIN;
	if(tryNetbios){
		if(wins != NULL && *wins != 0){
			if(!utilResolveName(wins, &ip))
				ip = 0xffffffff;
			this->destIp = ntohl(ip);
			PROC_WAITFOR_X(1, nbnsResolve(this->buffer, 0x20, this->destIp, 0, this));
			if(rval->err.class == 0){
				res = &rval->ipAddress;
				goto doReturn;
			}
		}
		/* try broadcast if wins failed */
		PROC_WAITFOR_X(2, nbnsResolve(this->buffer, 0x20, 0xffffffff, 1, this));
		if(rval->err.class == 0){
			res = &rval->ipAddress;
			goto doReturn;
		}
	}
	if(tryDns){
		if(utilResolveName(this->buffer, &ip)){
			ip = ntohl(ip);
			res = &ip;
			goto doReturn;
		}
	}
	res = NULL;
doReturn:
	if(res != NULL)
		cacheEntryAdd(this->buffer, *res);
	PROC_RETURN(res);	/* an error occured, but nobody cares what it was */
	PROC_END;
}

void	nbnsLookup(btu8 *name, void *caller)
{
nbnsCtx_t		*ctx;
nsCache_t		*c;
bttime			age;
static btint	lookupCount = 0;

    if(nbnsPool == NULL)
        nbnsPool = poolNew(sizeof(nbnsCtx_t), (btu8 *)&ctx->buffer - (btu8 *)ctx);
	if(nsCache == NULL)
		nsCache = htabNew(&cacheInterface, 32, 50);
	if(++lookupCount > 100){	/* clean cache every 100th call */
		lookupCount = 0;
		cacheCheckOutdated();
	}
	c = htabFindObjectEqualTo(nsCache, &name);
	if(c != NULL){
		age = elpNow() - c->updatedAt;
		if(age >= 0 && age < maxCacheTime()){	/* not outdated */
			PROC_IMMEDIATE_RET(caller, &c->ipAddress);
		}
	}
    PROC_INIT(ctx, nbnsLookupProc, caller, nbnsPool, poolFreeBlock);
	strncpy(ctx->buffer, name, sizeof(ctx->buffer));
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

