/*
 * Name: pslayer.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-05-21
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: pslayer.h,v 1.1.1.1 1999/04/12 13:56:34 cs Exp $
 */

#ifndef __pslayer_H_INCLUDED__
#define __pslayer_H_INCLUDED__

/*
General Description:
This header defines a unified interface between layers of the protocol stack.
The interface works as folows:

     -------------+                                      +-----------
                  | -------- setHigherLayer -----------> |
                  | -------- release ------------------> |
                  | -------- setConnectStatus ---------> | [psUpInterface]
                  | -------- connectStatus ------------> |
          upper   | -------- wantSend -----------------> |   lower
                  | -------- maxInserts ---------------> |
          layer   |                                      |   layer
                  | <------- connectStatusChanged ------ |
                  | <------- rxMessage ----------------- |
[psDownInterface] | <------- getTxMessage -------------- |
                  | <------- txReady ------------------- |
     -------------+                                      +-----------

It consists of two interfaces: the psUpInterface, which is a set of functions
called by the higher layer at the lower layer, and the psDownInterface, which
is a set of functions called by the lower layer at the higher layer.

Everything works as one would expect from the prototype definition, except
message sending. Message sending works as follows:
If the higher layer wants to send data, it calls wantSend() at the lower
layer. The lower layer stores this requests and arranges for sending the data.
When data can be sent (flowcontrol allows it), the getTxMessage() is called
at the higher level to get the data to be transmitted. If NULL is returned
by the higher level, there's no data to be sent. txReady() is called at the
higher level when the data has been sent.
*/

/* ------------------------------------------------------------------------- */

/* constants for connect status: */
#define	PSL_NOTCONNECTED	0
#define	PSL_CONNECTING		1
#define	PSL_CONNECTED		2
#define	PSL_DISCONNECTING	3

/* ------------------------------------------------------------------------- */

struct psUpInterface;
struct psDownInterface;

typedef struct psUpInterface{
	void	(*setHigherLayer)(void *obj, void *nextObj, struct psDownInterface *nextI);
	void	(*release)(void *obj);
	void	(*setConnectStatus)(void *obj, btint wantStatus, bttime timeout, void *caller);	/* callback: btError */
	btint	(*connectStatus)(void *obj);
	void	(*wantSend)(void *obj);
	void	(*maxInserts)(void *obj, btsize *preInsert, btsize *postInsert);
}psUpInterface_t;


typedef struct psDownInterface{
	void		(*connectStatusChanged)(void *obj, btint newStatus);
	void		(*rxMessage)(void *obj, btu8 *data, btsize len);
	struct msg	*(*getTxMessage)(void *obj);
	void		(*txReady)(void *obj, btint errcode);
}psDownInterface_t;

/* ------------------------------------------------------------------------- */

#endif /* __pslayer_H_INCLUDED__ */
