/*
 * Name: pslnetbios.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-05-21
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: pslnetbios.c,v 1.3 1999/07/22 17:54:03 cs Exp $
 */


#include "stdheaders.h"
#include "exception.h"
#include "pslnetbios.h"
#include "memalloc.h"
#include "evntloop.h"
#include "logging.h"
#include "system.h"
#include "process.h"
#include "message.h"
#include "blockpool.h"
#include "util.h"
#include "pslsocket.h"
#include "pslsslsocket.h"
#include "main.h"
#include "nbnsclient.h"
#include "errcodes.h"


/* ------------------------------------------------------------------------- */

#define	RXS_IDLE		0
#define	RXS_GETHDR		1
#define	RXS_GETDATA		2

#define	INITIAL_TRY_REALNAME	1

/* ------------------------------------------------------------------------- */

typedef struct netbCtx{
    PROC_VARSET;
    pslNetbios_t	*psl;
	btbool			tryRealName;
	bttime			timeout;
	btError			res;
}netbCtx_t;

static blockpool_t	*netbPool = NULL;

static void	freePoolBlock(void *block)
{
    poolReturnBlock(netbPool, block);
}

/* ------------------------------------------------------------------------- */

static void	netbSetConnectStatus(void *obj, btint wantStatus, bttime timeout, void *caller);

/* ------------------------------------------------------------------------- */

static void	netbFree(pslNetbios_t *psl)
{
	if(psl != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "netbios: freeing object at 0x%08x\n", (int)psl);
		if(psl->connectSignal != NULL){
			procCondFree(psl->connectSignal);
			psl->connectSignal = NULL;
		}
		if(psl->lowerObj != NULL){
			psl->lowerI->setHigherLayer(psl->lowerObj, NULL, NULL);
			psl->lowerI->setConnectStatus(psl->lowerObj, PSL_NOTCONNECTED, 0, NULL);
			psl->lowerI->release(psl->lowerObj);
			psl->lowerObj = NULL;
		}
		if(psl->rxData != NULL){
			memFree(psl->rxData);
			psl->rxData = NULL;
		}
		if(psl->passPhrase != NULL){
			memFree(psl->passPhrase);
			psl->passPhrase = NULL;
		}
		if(psl->fdForSSLSocket >= 0){
			close(psl->fdForSSLSocket);
			psl->fdForSSLSocket = -1;
		}
		memFree(psl);
	}
}

/* ------------------------------------------------------------------------- */

/* the following complicates searching for the pass phrase with grep in
 * /dev/kmem.
 */
static void	trivialCryptPhrase(pslNetbios_t *psl)
{
btsize	len, i;
btu8	ebuf[4];
btu32	e = (btu32)psl;

	if(psl->passPhrase != NULL){
		ebuf[0] = (btu8)(e >> 4);
		ebuf[1] = (btu8)(e >> 12);
		ebuf[2] = ebuf[0] - ebuf[1];
		ebuf[3] = ebuf[0] + ebuf[1];
		len = strlen(psl->passPhrase);
		for(i=0;i<len;i++)
			psl->passPhrase[i] ^= ebuf[i % 4];
	}
}

/* ------------------------------------------------------------------------- */

pslNetbios_t	*pslNetbiosNew(char *myName, char *remoteName, btu32 destIp, int port, struct server *server)
{
pslNetbios_t	*volatile psl = NULL;
int				nlen, rnlen;
char			*p;

    if(netbPool == NULL)
        netbPool = poolNew(sizeof(netbCtx_t), sizeof(netbCtx_t));
EXC_TRY
	if((p = strchr(myName, '.')) != NULL)
		nlen = p - myName;	/* strip off domain part */
	else
		nlen = strlen(myName);
	if(nlen > 16){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "netbios: local name ->%s<- longer than 16 chars\n", myName);
		nlen = 16;
	}
	if((p = strchr(remoteName, '.')) != NULL)
		rnlen = p - remoteName;	/* strip off domain part */
	else
		rnlen = strlen(remoteName);
	if(rnlen > 16){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "netbios: remote name ->%s<- longer than 16 chars\n", remoteName);
		rnlen = 16;
	}
	psl = memCalloc(sizeof(pslNetbios_t));
	psl->fdForSSLSocket = -1;
	memcpy(psl->myName, myName, nlen);
	memset(psl->myName + nlen, ' ', 16 - nlen);
	memcpy(psl->remoteName, remoteName, rnlen);
	memset(psl->remoteName + rnlen, ' ', 16 - rnlen);
	psl->destIp = destIp;
	psl->port = port;
	psl->rxState = RXS_IDLE;
	psl->rxDataAllocated = 2048;
	psl->connectSignal = procCondNew();
	psl->server = server;
	psl->refCount = 1;
	psl->lowerObj = pslSocketNew(destIp, port, psl->server);
	pslSocketUpInterface.setHigherLayer(psl->lowerObj, psl, &pslNetbiosDownInterface);
	psl->lowerI = &pslSocketUpInterface;
EXC_CATCH
	netbFree(psl);
	EXC_RETHROW();
EXC_ENDTRY
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "netbios: allocated new object at 0x%08x\n", (int)psl);
	return psl;
}

/* ------------------------------------------------------------------------- */

void	pslNetbiosSetPassPhrase(pslNetbios_t *psl, btu8 *passPhrase)
{
	if(psl->passPhrase != NULL)
		memFree(psl->passPhrase);
	psl->passPhrase = utilStrDup(passPhrase);
	trivialCryptPhrase(psl);
}

/* ------------------------------------------------------------------------- */

static inline void	pslNetbiosRetain(void *obj)
{
pslNetbios_t	*psl = obj;

	psl->refCount++;
}

static inline void	pslNetbiosRelease(void *obj)
{
pslNetbios_t	*psl = obj;

	psl->refCount--;
	if(psl->refCount == 0){
		netbFree(psl);
	}else if (psl->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

btbool	pslNetbiosConnectionEncrypted(pslNetbios_t *psl)
{
	return psl->lowerI != &pslSocketUpInterface;
}

/* ------------------------------------------------------------------------- */
/* ----------------------- message encoding/decoding ----------------------- */
/* ------------------------------------------------------------------------- */

static void	dumpMessage(btu8 *title, btu8 *data, btsize len)
{
int		i;
btu8	buffer[256];
int		lineLen = 4;

	while(len > 0){
		strcpy(buffer, title);
		for(i=0;i<lineLen && len>0;i++){
			sprintf(buffer + strlen(buffer), " %02x", *data);
			data++;
			len--;
		}
		lineLen = 16;	/* further lines will be 16 long */
		logPrintf(LOGM_NETBTRACE, "%s\n", buffer);
	}
}

/* ------------------------------------------------------------------------- */

#define	IS_SESSION_BLOCK(b)		((b) == 0 || ((b) >= 0x81 && (b) <= 0x85))
#define	IS_DATAGRAM_BLOCK(b)	(((b) >= 0x10 && (b) <= 0x16))

/* ------------------------------------------------------------------------- */

static char	*netbiosError(btu8 code)
{
	switch(code){
		case 0x80:	return "Not listening on called name";
		case 0x81:	return "Not listening for calling name";
		case 0x82:	return "Called name not present";
		case 0x83:	return "Called name present, but insufficient resources";
		case 0x8e:	return "Server requests SSL";
		default:	return "Unspecified error";
	}
}

/* ------------------------------------------------------------------------- */

static void	processData(pslNetbios_t *psl)
{
btError	res;

	switch(psl->rxHeader[0]){
	case 0:		/* session message */
		if(psl->rxData != NULL && psl->higherI != NULL)
			psl->higherI->rxMessage(psl->higherObj, psl->rxData, psl->rxDataLen);
		break;
	case 0x81:	/* session request */
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSERR))
			logPrintf(LOGM_NETBIOSERR, "netbios: received session request\n");
		break;
	case 0x82:	/* positive session response */
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSDEB))
			logPrintf(LOGM_NETBIOSDEB, "netbios: positive session response\n");
		res.class = res.code = 0;	/* connect succeeded */
		if(psl->rxSessionResponseCallback != NULL)
			psl->rxSessionResponseCallback(psl->rxContext, &res);
		break;
	case 0x83:	/* negative session response */
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSDEB))
			logPrintf(LOGM_NETBIOSDEB, "netbios: negative session response 0x%02x (%s)\n", psl->rxData != NULL ? psl->rxData[0] : 0, psl->rxData != NULL ? netbiosError(psl->rxData[0]) : "NULL");
		res.class = ERRCLS_CIFS_LOCAL;
		res.code = psl->rxData != NULL && psl->rxData[0] == 0x8e ? ERR_CIFS_LOCsslserver : ERR_CIFS_LOCnetbiosrefused;
		if(psl->rxSessionResponseCallback != NULL)
			psl->rxSessionResponseCallback(psl->rxContext, &res);
		break;
	case 0x84:	/* retarget session response */
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSERR))
			logPrintf(LOGM_NETBIOSERR, "netbios: can\'t handle retarget\n");
		res.class = ERRCLS_CIFS_LOCAL;
		res.code = ERR_CIFS_LOCnetbiosrefused;
		if(psl->rxSessionResponseCallback != NULL)
			psl->rxSessionResponseCallback(psl->rxContext, &res);
		break;
	case 0x85:	/* keep alive */
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSDEB))
			logPrintf(LOGM_NETBIOSDEB, "netbios: received keep alive\n");
		break;
	case 0x10:	/* direct unique datagram */
	case 0x11:	/* direct group datagram */
	case 0x12:	/* broadcast datagram */
	case 0x13:	/* datagram error */
	case 0x14:	/* datagram query request */
	case 0x15:	/* datagram positive query response */
	case 0x16:	/* datagram negative query response */
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSERR))
			logPrintf(LOGM_NETBIOSERR, "netbios: received datagram 0x%02x\n", psl->rxHeader[0]);
		break;
	}
}

/* ------------------------------------------------------------------------- */

static void	processHeader(pslNetbios_t *psl)
{
	psl->rxDataPtr = 0;
	if(IS_SESSION_BLOCK(psl->rxHeader[0])){
		psl->rxDataLen = psl->rxHeader[3] | ((btu32)psl->rxHeader[2] << 8);
		psl->rxDataLen |= (btu32)(psl->rxHeader[1] & 1) << 16;
		psl->rxState = RXS_GETDATA;
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSDEB))
			logPrintf(LOGM_NETBIOSDEB, "netbios: session 0x%02x len %d\n", psl->rxHeader[0], psl->rxDataLen);
	}else if(IS_DATAGRAM_BLOCK(psl->rxHeader[0])){
		psl->rxDataLen = psl->rxHeader[11] | ((btu32)psl->rxHeader[10]<<8);
		psl->rxState = RXS_GETDATA;
		if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSDEB))
			logPrintf(LOGM_NETBIOSDEB, "netbios: datagram 0x%02x len %d\n", psl->rxHeader[0], psl->rxDataLen);
	}
	if(psl->rxDataLen == 0){
		psl->rxState = RXS_IDLE;
		processData(psl);
	}
	EXC_TRY
		if(psl->rxDataLen > psl->rxDataAllocated || psl->rxData == NULL){
			while(psl->rxDataLen > psl->rxDataAllocated)
				psl->rxDataAllocated *= 2;
			psl->rxData = memRealloc(psl->rxData, psl->rxDataAllocated);
			if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
				logPrintf(LOGM_RESOURCE, "netbios: setting rxDataAlloc=%d\n", psl->rxDataAllocated);
		}
	EXC_CATCH
		memFree(psl->rxData);
		psl->rxData = NULL;
	EXC_ENDTRY
}

/* ------------------------------------------------------------------------- */

static inline btsize	processBytes(pslNetbios_t *psl, btu8 *data, btsize dataLen)
{
btsize	useBytes;

	switch(psl->rxState){
	case RXS_IDLE:
		psl->rxHeader[0] = *data;
		psl->rxHdrPtr = 1;
		psl->rxState = RXS_GETHDR;
		if(IS_SESSION_BLOCK(*data)){		/* session block */
			psl->rxHdrLen = 4;
		}else if(IS_DATAGRAM_BLOCK(*data)){	/* datagram block */
			psl->rxHdrLen = 14;
		}else{
			if(LOG_MODE_IS_ACTIVE(LOGM_NETBIOSERR))
				logPrintf(LOGM_NETBIOSERR, "netbios: illegal type 0x%02x\n", *data);
			psl->rxState = RXS_IDLE;
		}
		return 1;
	case RXS_GETHDR:
		useBytes = psl->rxHdrLen - psl->rxHdrPtr;
		if(useBytes > dataLen)
			useBytes = dataLen;
		memcpy(psl->rxHeader + psl->rxHdrPtr, data, useBytes);
		psl->rxHdrPtr += useBytes;
		if(psl->rxHdrPtr >= psl->rxHdrLen){
			psl->rxState = RXS_IDLE;	/* default for processHeader() */
			processHeader(psl);
		}
		return useBytes;
	case RXS_GETDATA:
		useBytes = psl->rxDataLen - psl->rxDataPtr;
		if(useBytes > dataLen)
			useBytes = dataLen;
		if(psl->rxData != NULL)
			memcpy(psl->rxData + psl->rxDataPtr, data, useBytes);
		psl->rxDataPtr += useBytes;
		if(psl->rxDataPtr >= psl->rxDataLen){
			psl->rxState = RXS_IDLE;	/* default for processData() */
			processData(psl);
		}
		return useBytes;
	}
	return 1;
}

/* ------------------------------------------------------------------------- */
/* ---------------------- interface to previous layer ---------------------- */
/* ------------------------------------------------------------------------- */

static void	netbRxMessage(void *obj, btu8 *data, btsize len)
{
pslNetbios_t	*psl = obj;
btsize			i;

	if(LOG_MODE_IS_ACTIVE(LOGM_NETBTRACE))
		dumpMessage("-> Rx:", data, len);
	for(i=0;i<len;)
		i += processBytes(psl, data + i, len - i);
}

/* ------------------------------------------------------------------------- */

static void	netbConnectStatusChanged(void *obj, btint newStatus)
{
pslNetbios_t	*psl = obj;
btError			err;

	if(newStatus == PSL_NOTCONNECTED && psl->cstatus != PSL_NOTCONNECTED){	/* connection lost */
		if(psl->rxSessionResponseCallback != NULL){
			err.class = ERRCLS_CIFS_LOCAL;
			err.code = ERR_CIFS_LOCconnlost;
			psl->rxSessionResponseCallback(psl->rxContext, &err);
		}
		netbSetConnectStatus(psl, PSL_NOTCONNECTED, 0, NULL);
	}
}

/* ------------------------------------------------------------------------- */

static struct msg	*netbGetTxMessage(void *obj)
{
pslNetbios_t	*psl = obj;
nbmsg_t			*msg = NULL;
btu8			header[4], *data;
btsize			len;
btu8			encName[256];

	if(psl->txSessionRequest){	/* transmit a session request */
		msg = msgNew(4);
		msgAddByte(msg, 0x81);	/* session request */
		msgAddByte(msg, 0);		/* flags */
		msgAddByte(msg, 0);		/* size high */
		msgAddByte(msg, 68);	/* size low */
		len = nbnsEncodeName(encName, sizeof(encName), psl->txSessionRqWithRealName ? psl->remoteName : (btu8 *)"*SMBSERVER", 0x20);
		msgAddBlock(msg, encName, len);
		len = nbnsEncodeName(encName, sizeof(encName), psl->myName, 0x20);
		msgAddBlock(msg, encName, len);
	}else if(psl->higherI != NULL){
		msg = psl->higherI->getTxMessage(psl->higherObj);
		if(msg != NULL){
			header[0] = 0;	/* message type */
			header[1] = (msgSize(msg) >> 16) & 1;
			header[2] = msgSize(msg) >> 8;
			header[3] = msgSize(msg);
			msgInsertBlock(msg, header, 4);		/* insert netbios header */
		}
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_NETBTRACE) && msg != NULL){
		msgPeekData(msg, &data, &len);
		dumpMessage("<- Tx:", data, len);
	}
	return msg;
}

/* ------------------------------------------------------------------------- */

static void	netbTxReady(void *obj, btint errcode)
{
pslNetbios_t	*psl = obj;
btError			err;

	if(psl->txSessionRequest){
		psl->txSessionRequest = 0;
		if(psl->rxSessionResponseCallback != NULL){
			err.class = errcode == 0 ? 0 :  ERRCLS_UNIX;
			err.code = errcode;
			psl->rxSessionResponseCallback(psl->rxContext, &err);
		}
	}else if(psl->higherI != NULL){
		psl->higherI->txReady(psl->higherObj, errcode);
	}
}

/* ------------------------------------------------------------------------- */
/* ------------------------ interface to next layer ------------------------ */
/* ------------------------------------------------------------------------- */

static void	netbSetHigherLayer(void *obj, void *nextObj, struct psDownInterface *nextI)
{
pslNetbios_t	*psl = obj;

	psl->higherI = nextI;
	psl->higherObj = nextObj;
}


/* ------------------------------------------------------------------------- */

static void	netbConnectProc(void *obj, void *arg)
{
netbCtx_t			*this = obj;
pslNetbios_t		*psl = this->psl;
btError				*rval = arg;
btError				res;

	PROC_BEGIN
	pslNetbiosRetain(psl);
	while(psl->cstatus == PSL_DISCONNECTING || psl->cstatus == PSL_CONNECTING){
		PROC_WAITFOR_X(1, procCondWait(psl->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(psl->cstatus == PSL_CONNECTED){
		res.class = res.code = 0;
		pslNetbiosRelease(psl);
		PROC_RETURN(&res);
	}
	psl->cstatus = PSL_CONNECTING;
	if(psl->fdForSSLSocket >= 0)	/* if an earlier connect can be completed */
		goto retrySSLConnect;
	this->tryRealName = INITIAL_TRY_REALNAME;
retry:
	if(psl->lowerI->connectStatus(psl->lowerObj) != PSL_CONNECTED){
		PROC_WAITFOR_X(2, psl->lowerI->setConnectStatus(psl->lowerObj, PSL_CONNECTED, this->timeout, this));
		if(rval->class != 0){
			res = *rval;
			goto errorOccured;
		}
	}
	psl->txSessionRequest = 1;
	psl->txSessionRqWithRealName = this->tryRealName;
	elpCallAtCancelPrev(elpNow() + this->timeout, netbConnectProc, this);	/* setup timeout */
	psl->rxSessionResponseCallback = netbConnectProc;	/* set callback to self */
	psl->rxContext = this;
	psl->lowerI->wantSend(psl->lowerObj);	/* schedule sending of session request */
	PROC_WAITFOR_X(3, ;);	/* first callback is success of write */
	if(rval == NULL){		/* if it was the timeout */
		res.class = ERRCLS_UNIX;
		res.code = ETIMEDOUT;
		goto errorOccured;
	}else if(rval->class != 0){
		res = *rval;
		goto errorOccured;
	}
	PROC_WAITFOR_X(4, ;);	/* wait for received message or timeout */
	elpCancelCallAt(netbConnectProc, this);
	psl->rxSessionResponseCallback = NULL;
	if(rval == NULL){	/* if it was the timeout */
		res.class = ERRCLS_UNIX;
		res.code = ETIMEDOUT;
		goto errorOccured;
	}else{				/* analyze the session response */
		if(rval->class == ERRCLS_CIFS_LOCAL && rval->code == ERR_CIFS_LOCnetbiosrefused){
			if(this->tryRealName){
				res.class = ERRCLS_CIFS_LOCAL;
				res.code = ERR_CIFS_LOCnetbiosrefused;
				goto errorOccured;
			}
			PROC_WAITFOR_X(5, psl->lowerI->setConnectStatus(psl->lowerObj, PSL_NOTCONNECTED, 0, this));
			this->tryRealName = 1;
			goto retry;
		}else if(rval->class == ERRCLS_CIFS_LOCAL && rval->code == ERR_CIFS_LOCsslserver){
#ifdef USE_SSL
			if(psl->lowerI == &pslSslSocketUpInterface){	/* if we already have SSL */
				res.class = ERRCLS_CIFS_LOCAL;
				res.code = ERR_CIFS_LOCnetbiosrefused;
				goto errorOccured;
			}
			/* we must still have a pslSocket object as lower layer */
			psl->fdForSSLSocket = pslSocketFd(psl->lowerObj);
retrySSLConnect:
			psl->lowerI->release(psl->lowerObj);	/* release socket object */
			PROC_WAITFOR_X(6, elpCallCancelPrev(netbConnectProc, this));	/* let old object be free'd */
			psl->lowerI = &pslSslSocketUpInterface;
			res.code = res.class = 0;
			trivialCryptPhrase(psl);	/* decrypt pass phrase */
EXC_TRY
			psl->lowerObj = pslSslSocketNew(psl->destIp, psl->port, psl->fdForSSLSocket, psl->server, psl->passPhrase);
			psl->lowerI->setHigherLayer(psl->lowerObj, psl, &pslNetbiosDownInterface);
EXC_CATCH	/* error recovery: */
			if(EXC_NAME == excSslNeedPassPhrase){
				res.class = ERRCLS_CIFS_LOCAL;
				res.code = ERR_CIFS_LOCneedpassphrase;
			}else if(EXC_NAME == memExcNomem){
				res.class = ERRCLS_FRAMEWORK;
				res.code = ERR_FRWnomem;
			}else{
				res.class = ERRCLS_CIFS_LOCAL;
				res.code = ERR_CIFS_LOCsslerror;
			}
			close(psl->fdForSSLSocket);
			psl->fdForSSLSocket = -1;
			psl->lowerObj = pslSocketNew(psl->destIp, psl->port, psl->server);	/* we need a socket object */
			psl->lowerI = &pslSocketUpInterface;
			pslSocketUpInterface.setHigherLayer(psl->lowerObj, psl, &pslNetbiosDownInterface);
			/* we can't use goto here because EXC_ENDTRY changes handler stack */
EXC_ENDTRY
			trivialCryptPhrase(psl);	/* encrypt pass phrase */
			if(res.class != 0)
				goto errorOccured;
			psl->fdForSSLSocket = -1;	/* SSL allocation did work */
			this->tryRealName = INITIAL_TRY_REALNAME;
			goto retry;	/* retry connect with SSL */
#else
retrySSLConnect:
			res.class = ERRCLS_CIFS_LOCAL;
			res.code = ERR_CIFS_LOChavenossl;
			goto errorOccured;
#endif
		}
	}
#ifdef USE_SSL
	if(psl->requireEncryption && psl->lowerI != &pslSslSocketUpInterface){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "netbios: Connection aborted because server does not support SSL\n");
		res.class = ERRCLS_CIFS_LOCAL;
		res.code = ERR_CIFS_LOCnosslserver;
		goto errorOccured;
	}
#else
	if(psl->requireEncryption){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "netbios: Data encryption not supported! No SSL support compiled in!\n");
		res.class = ERRCLS_CIFS_LOCAL;
		res.code = ERR_CIFS_LOChavenossl;
		goto errorOccured;
	}
#endif
	psl->cstatus = PSL_CONNECTED;
	res.class = res.code = 0;
    PROC_RETURN_AND_CLEANUP(&res,
		if(psl->higherI != NULL)
			psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
		procCondBroadcast(psl->connectSignal, NULL);
		pslNetbiosRelease(psl);
	);
errorOccured:
	this->res = res;
	elpCancelCallAt(netbConnectProc, this);
	psl->rxSessionResponseCallback = NULL;
	PROC_WAITFOR_X(7, psl->lowerI->setConnectStatus(psl->lowerObj, PSL_NOTCONNECTED, 0, this));
	if(psl->lowerI != &pslSocketUpInterface){	/* if SSL: switch back to normal socket */
		psl->lowerI->release(psl->lowerObj);	/* release socket object */
		PROC_WAITFOR_X(8, elpCallCancelPrev(netbConnectProc, this));	/* let old object be free'd */
		psl->lowerObj = pslSocketNew(psl->destIp, psl->port, psl->server);
		psl->lowerI = &pslSocketUpInterface;
		psl->lowerI->setHigherLayer(psl->lowerObj, psl, &pslNetbiosDownInterface);
	}
	psl->cstatus = PSL_NOTCONNECTED;
    PROC_RETURN_AND_CLEANUP(&this->res,
		procCondBroadcast(psl->connectSignal, NULL);
		pslNetbiosRelease(psl);
	);
    PROC_END
}


/* ------------------------------------------------------------------------- */

static void	netbDisconnectProc(void *obj, void *arg)
{
netbCtx_t			*this = obj;
pslNetbios_t		*psl = this->psl;
btError				res;

	PROC_BEGIN
	pslNetbiosRetain(psl);
	while(psl->cstatus == PSL_DISCONNECTING || psl->cstatus == PSL_CONNECTING){
		PROC_WAITFOR_X(1, procCondWait(psl->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(psl->cstatus == PSL_NOTCONNECTED){
		res.class = res.code = 0;
		pslNetbiosRelease(psl);
		PROC_RETURN(&res);
	}
	psl->cstatus = PSL_DISCONNECTING;
	PROC_WAITFOR_X(2, psl->lowerI->setConnectStatus(psl->lowerObj, PSL_NOTCONNECTED, 0, this));
#ifdef USE_SSL
	if(psl->lowerI == &pslSslSocketUpInterface){	/* if SSL: switch back to normal socket */
		psl->lowerI->release(psl->lowerObj);	/* release socket object */
		PROC_WAITFOR_X(3, elpCallCancelPrev(netbDisconnectProc, this));	/* let old object be free'd */
		psl->lowerObj = pslSocketNew(psl->destIp, psl->port, psl->server);
		psl->lowerI = &pslSocketUpInterface;
		psl->lowerI->setHigherLayer(psl->lowerObj, psl, &pslNetbiosDownInterface);
	}
#endif
	if(psl->fdForSSLSocket >= 0){
		close(psl->fdForSSLSocket);
		psl->fdForSSLSocket = -1;
	}
	psl->cstatus = PSL_NOTCONNECTED;
	res.class = res.code = 0;
    PROC_RETURN_AND_CLEANUP(&res,
		if(psl->higherI != NULL)
			psl->higherI->connectStatusChanged(psl->higherObj, psl->cstatus);
		procCondBroadcast(psl->connectSignal, NULL);
		pslNetbiosRelease(psl);
	);
	PROC_END
}

/* ------------------------------------------------------------------------- */

static void	netbSetConnectStatus(void *obj, btint wantStatus, bttime timeout, void *caller)
{
pslNetbios_t	*psl = obj;
netbCtx_t		*ctx;
btError			res;

	if(wantStatus != psl->cstatus){
		PROC_INIT(ctx, wantStatus == PSL_CONNECTED ? netbConnectProc : netbDisconnectProc, caller, netbPool, freePoolBlock);
		ctx->psl = psl;
		ctx->timeout = timeout;
		PROC_START(ctx, NULL);
	}else{
		res.class = res.code = 0;
		PROC_IMMEDIATE_RET(caller, &res);
	}
}

/* ------------------------------------------------------------------------- */

static btint	netbConnectStatus(void *obj)
{
pslNetbios_t	*psl = obj;

	return psl->cstatus;
}

/* ------------------------------------------------------------------------- */

static void	netbWantSend(void *obj)
{
pslNetbios_t	*psl = obj;

	psl->lowerI->wantSend(psl->lowerObj);
}

/* ------------------------------------------------------------------------- */

static void	netbMaxInserts(void *obj, btsize *preInsert, btsize *postInsert)
{
pslNetbios_t	*psl = obj;

	psl->lowerI->maxInserts(psl->lowerObj, preInsert, postInsert);
	*preInsert += 4;
}

/* ------------------------------------------------------------------------- */

psUpInterface_t	pslNetbiosUpInterface = {
	netbSetHigherLayer,
	pslNetbiosRelease,
	netbSetConnectStatus,
	netbConnectStatus,
	netbWantSend,
	netbMaxInserts
};

psDownInterface_t	pslNetbiosDownInterface = {
	netbConnectStatusChanged,
	netbRxMessage,
	netbGetTxMessage,
	netbTxReady
};

/* ------------------------------------------------------------------------- */
