/*
 * Name: raprpc.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-10-10
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: raprpc.c,v 1.1.1.1 1999/04/12 13:56:34 cs Exp $
 */


#include "stdheaders.h"
#include "logging.h"
#include "smb.h"
#include "server.h"
#include "smbnumbers.h"
#include "unicode.h"
#include "process.h"
#include "raprpc.h"
#include "errcodes.h"

/* ------------------------------------------------------------------------- */

#define	RAP_TRACE(fname)													\
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))									\
		logPrintf(LOGM_CIFSTRACE, "rap trace: %s()\n", fname)

#define	DECODER_ERROR(msg)													\
	do{	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))									\
			logPrintf(LOGM_ERROR, "rap: unmarshalling error: %s\n", msg);	\
        res.err.class = ERRCLS_CIFS_LOCAL;								\
        res.err.code = ERR_CIFS_LOCbogusdata;								\
        goto decode_exception;												\
    }while(0)

/* ------------------------------------------------------------------------- */

static void	rapDecodeEnumShares(void *ref, struct smbRequest *smb)
{
rapRes_t	res;
btu8		*p, *base;
btu16		cvtWord;
btint		remark, num, i;

	memset(&res, 0, sizeof(res));
    if(smb->rxHeader == NULL){
        res.err.class = ERRCLS_CIFS_LOCAL;
        res.err.code = ERR_CIFS_LOCtimeout;
        goto decode_exception;
    }else{
        res.err = smb->rxHeader->err;
		if(res.err.class != 0)
			goto decode_exception;
    }
	if(smb->rxParameterBytes < 8)
		DECODER_ERROR("Not enough parameters in rapDecodeEnumShares");
    p = base = smb->rxBuffer + smb->rxParameterBase;
	res.status = smbGetWord(&p);	/* function return status */
	cvtWord = smbGetWord(&p);		/* converter word */
	num = smbGetWord(&p);			/* number of entries returned */
	if(smb->rxDataBytes < num * 20)
		DECODER_ERROR("Not enough data in rapDecodeEnumShares");
    p = base = smb->rxBuffer + smb->rxDataBase;
	res.numEntries = num;
	if(num > 0)
		res.entries = memCalloc(sizeof(rapEnumEntry_t) * num);
	for(i=0;i<num;i++){
		res.entries[i].name = p;
		p += 14;
		res.entries[i].type = smbGetWord(&p);
		remark = smbGetWord(&p);
		if(remark != 0)
			res.entries[i].remark = base + (remark - cvtWord);
		p += 2;
		if(res.entries[i].remark != NULL && res.entries[i].remark - base > smb->rxDataBytes){
			res.entries[i].remark = NULL;
		}
	}
	
decode_exception:
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))
		logPrintf(LOGM_CIFSTRACE, "rap trace: [%d] rval = %d/%d\n", smb->mid, res.err.class, res.err.code);
	if(ref != NULL)
		((procCtx_t *)ref)->process(ref, &res);
	if(res.entries != NULL)
		memFree(res.entries);
}

void	rapEnumShares(struct smbRequest *smb, btu16 uid, btu16 tid, void *caller)
{
	RAP_TRACE("rapEnumShares");
	smbStartCommand(smb, uid, 0, tid, rapDecodeEnumShares, caller);
    smbStartTransSmb(smb, "\\PIPE\\LANMAN", 50, 65535, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, RAP_NET_SHARE_ENUM);
	smbAddAsciiFromAscii(smb, "WrLeh");		/* parameter descriptor */
	smbAddAsciiFromAscii(smb, "B13BWz");	/* data descriptor */
    smbAddWord(smb, 1);						/* info level */
    smbAddWord(smb, 65535);					/* size of receive buffer */
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */

static void	rapDecodeEnumServers(void *ref, struct smbRequest *smb)
{
rapRes_t	res;
btu8		*p, *base;
btu16		cvtWord;
btint		remark, num, i;

	memset(&res, 0, sizeof(res));
    if(smb->rxHeader == NULL){
        res.err.class = ERRCLS_CIFS_LOCAL;
        res.err.code = ERR_CIFS_LOCtimeout;
        goto decode_exception;
    }else{
        res.err = smb->rxHeader->err;
		if(res.err.class != 0)
			goto decode_exception;
    }
	if(smb->rxParameterBytes < 8)
		DECODER_ERROR("Not enough parameters in rapDecodeEnumServers");
    p = base = smb->rxBuffer + smb->rxParameterBase;
	res.status = smbGetWord(&p);	/* function return status */
	cvtWord = smbGetWord(&p);		/* converter word */
	num = smbGetWord(&p);			/* number of entries returned */
	if(smb->rxDataBytes < num * 26)
		DECODER_ERROR("Not enough data in rapDecodeEnumServers");
    p = base = smb->rxBuffer + smb->rxDataBase;
	res.numEntries = num;
	if(num > 0)
		res.entries = memCalloc(sizeof(rapEnumEntry_t) * num);
	for(i=0;i<num;i++){
		res.entries[i].name = p;
		p += 18;	/* skip name and version info */
		res.entries[i].type = smbGetLong(&p);
		remark = smbGetWord(&p);
		if(remark != 0)
			res.entries[i].remark = base + (remark - cvtWord);
		p += 2;		/* skip highbytes of char * */
		if(res.entries[i].remark != NULL && res.entries[i].remark - base > smb->rxDataBytes){
			res.entries[i].remark = NULL;
		}
	}
	
decode_exception:
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))
		logPrintf(LOGM_CIFSTRACE, "rap trace: [%d] rval = %d/%d\n", smb->mid, res.err.class, res.err.code);
	if(ref != NULL)
		((procCtx_t *)ref)->process(ref, &res);
	if(res.entries != NULL)
		memFree(res.entries);
}

void	rapEnumServers(struct smbRequest *smb, btu8 *workgroup, btbool listDomains, btu16 uid, btu16 tid, void *caller)
{
	RAP_TRACE("rapEnumServers");
	smbStartCommand(smb, uid, 0, tid, rapDecodeEnumServers, caller);
    smbStartTransSmb(smb, "\\PIPE\\LANMAN", 50, 65535, 0);
    smbStartParameters(smb, -1);
    smbAddWord(smb, RAP_NET_SERVER_ENUM);
	smbAddAsciiFromAscii(smb, "WrLehDz");	/* parameter descriptor */
	smbAddAsciiFromAscii(smb, "B16BBDz");	/* data descriptor */
    smbAddWord(smb, 1);						/* info level */
    smbAddWord(smb, 65535);					/* size of receive buffer */
	if(listDomains){
    	smbAddLong(smb, RAP_SENUM_TYPE_DOMAIN_ENUM);
	}else{
    	smbAddLong(smb, RAP_SENUM_TYPE_SERVER);
	}
	smbAddAsciiFromAscii(smb, workgroup);
	smbStartData(smb);
	smbEndCommand(smb);
}

/* ------------------------------------------------------------------------- */
