/*
 * Name: server.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-20
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: server.c,v 1.6 1999/08/01 12:57:14 cs Exp $
 */

#include "stdheaders.h"
#include "memalloc.h"
#include "evntloop.h"
#include "unicode.h"
#include "server.h"
#include "smb.h"
#include "cifs.h"
#include "user.h"
#include "share.h"
#include "array.h"
#include "blockpool.h"
#include "smbnumbers.h"
#include "process.h"
#include "logging.h"
#include "cifsconst.h"
#include "fifo.h"
#include "cfgparse.h"
#include "main.h"
#include "util.h"
#include "wildmat.h"
#include "pslnetbios.h"
#include "message.h"
#include "licman.h"
#include "file.h"
#include "lmb.h"
#include "nbnsclient.h"
#include "ipc.h"
#include "errcodes.h"
#include "uirep.h"
#include "mainmodule.h"
#include "cifsmod.h"
#include "ipcserver.h"
#include "guimanager.h"

/* ------------------------------------------------------------------------- */

typedef struct connCtx{
    PROC_VARSET;
    server_t	*server;
	btu32		srvIp;
	btu8		*srvName;
	btint		uid;
	ipcSrvClient_t	*client;
}connCtx_t;

static blockpool_t *connCtxPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(connCtxPool, block);
}

/* ------------------------------------------------------------------------- */

static btu8	*smbModeEnum[] = {	/* this must match SRV_USE_* in server.h! */
		"getattrFind",
		"getattrTrans2QueryPath",
		"attrUnix",
		"setattrTrans2SetFile",
		"setattrTrans2SetPath",
		"setattrSetFile2",
		"setattrCoreWithTime",
		"createOpenX",
		"openOpenX",
		"readReadX",
		"readOpenRead",
		"writeWriteX",
		"writeOpenWrite",
		"findUnix",
		"findTrans2",
		"fsinfoTrans2",
		"sessionSetup",
		"treeconAndX",
		"setDirDates",
		"findTrans2dirinfo",
		"findAutoclose",
		"findImplicitResume",
		"findTrans2bothdirinfo",
		NULL
};

static void	srvCheckDemands(void *ref, void *arg);

/* ------------------------------------------------------------------------- */

void	srvSetWindowsize(server_t *this, int newWinSize)
{
int		i, oldWinSize;

EXC_TRY
    if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
        logPrintf(LOGM_DEBUG, "server: setting window size to %d\n", newWinSize);
    }
    oldWinSize = this->windowSize;
    if(newWinSize != oldWinSize){
        if(newWinSize < oldWinSize){
            for(i=newWinSize;i<this->windowSize;i++){
                smbRequestFree(this->requests[i]);
                this->requests[i] = NULL;
            }
        }
        this->requests = memRealloc(this->requests, newWinSize * sizeof(smbRequest_t *));
        this->windowSize = newWinSize;
        if(newWinSize > oldWinSize){
            for(i=oldWinSize;i<newWinSize;i++){
                this->requests[i] = NULL;
                this->requests[i] = smbRequestNew(this);
            }
        }
    }
    fifoClear(this->sendRequests);

    for(i=0;i<this->windowSize;i++){
        fifoAddObject(this->sendRequests, this->requests[i]);	/* try tx with all smbs */
    }

EXC_CATCH
    for(i = 1; i < this->windowSize && this->requests[i] != NULL; i++){
        smbRequestFree(this->requests[i]);
    }
	this->windowSize = 1;	/* in emergency: reduce window size to 1 */
    EXC_RETHROW();
EXC_ENDTRY
}

/* ------------------------------------------------------------------------- */

static void	srvConnectStatusChanged(void *obj, btint newStatus);
static void	srvRxMessage(void *obj, btu8 *data, btsize len);
static nbmsg_t	*srvGetTxMessage(void *obj);
static void	srvTxReady(void *obj, btint errcode);

static psDownInterface_t	serverDownInterface ={
	srvConnectStatusChanged,
	srvRxMessage,
	srvGetTxMessage,
	srvTxReady
};

/* ------------------------------------------------------------------------- */

#define	MASK_ASSIGN(var, mask, value)	(var) = (value) ? (var) | (mask) : (var) & ~(mask)
#define	SET_ASCII(var, val)			\
	do{								\
		btu8	*s = (val);			\
		if(s != NULL){				\
			if((var) != NULL)		\
				memFree(var);		\
			(var) = utilStrDup(s);	\
		}							\
	}while(0)

static void	srvReadConfig(server_t *this, plist_t *srvClass)
{
plist_t			*current;
btbool			flag;
btsize			count, i, numEntries;
cfgProperty_t	*path[8];

	numEntries = cifsModServerCfgPath(this->name, srvClass, path, ELEMENTS(path));
	
	this->linkMode = cfgIntValue(cfgQueryFromArray("linkMode", numEntries, path), 0);
	this->linkModeMask = cfgIntValue(cfgQueryFromArray("linkModeMask", numEntries, path), 0);
	if(!cfgEnumValue(cfgQueryFromArray("enableFakeLinks", numEntries, path), cfgYesNoEnum, 0))
		this->linkModeMask = 0;
	this->linkMode &= this->linkModeMask;
	this->linksAreUnicode = cfgEnumValue(cfgQueryFromArray("linksAreUnicode", numEntries, path), cfgYesNoEnum, 1);
	if(!cfgEnumValue(cfgQueryFromArray("useUnicode", numEntries, path), cfgYesNoEnum, 0))
		this->useUnicode = 0;
	flag = cfgEnumValue(cfgQueryFromArray("requestOplock", numEntries, path), cfgYesNoEnum, 1);
	MASK_ASSIGN(this->txFlags, CIFS_FLAG_EXCLUSIVE_LOCK | CIFS_FLAG_BATCH_LOCK, flag);
	this->cacheValidUnlocked = cfgIntValue(cfgQueryFromArray("dataCacheTimeNoLock", numEntries, path), CONST_DCACHE_LIFETIME_UNLOCKED);
	this->closeDelayUnlocked = cfgIntValue(cfgQueryFromArray("closeDelayNoLock", numEntries, path), CONST_CLOSEDELAY_UNLOCKED);
	this->closeDelayLocked = cfgIntValue(cfgQueryFromArray("closeDelayExclusiveLock", numEntries, path), CONST_CLOSEDELAY_LOCKED);
	this->closeDelayBatchlock = cfgIntValue(cfgQueryFromArray("closeDelayBatchLock", numEntries, path), CONST_CLOSEDELAY_BATCH);
	this->readAheadUnlocked = cfgIntValue(cfgQueryFromArray("readAheadNoLock", numEntries, path), CONST_READAHEAD_UNLOCKED);
	this->readAheadLocked = cfgIntValue(cfgQueryFromArray("readAheadLocked", numEntries, path), CONST_READAHEAD_LOCKED);
	this->writeThroughUnlocked = !cfgEnumValue(cfgQueryFromArray("useWriteBackNoLock", numEntries, path), cfgYesNoEnum, 0);
	this->writeThroughLocked = !cfgEnumValue(cfgQueryFromArray("useWriteBackLocked", numEntries, path), cfgYesNoEnum, 0);
	this->caseSensitive = cfgEnumValue(cfgQueryFromArray("caseSensitive", numEntries, path), cfgYesNoEnum, 0);
	MASK_ASSIGN(this->txFlags, CIFS_FLAG_CASE_INSENSITIVE, !this->caseSensitive);
	this->fakeMountpointDate = cfgEnumValue(cfgQueryFromArray("fakeMountpointDate", numEntries, path), cfgYesNoEnum, 1);
	this->capitalizeShares = cfgEnumValue(cfgQueryFromArray("capitalizeShares", numEntries, path), cfgYesNoEnum, 1);
	this->closeForSetattr = cfgEnumValue(cfgQueryFromArray("closeForSetattr", numEntries, path), cfgYesNoEnum, 0);
	this->sslRequireEncryption = cfgEnumValue(cfgQueryFromArray("sslRequireEncryption", numEntries, path), cfgYesNoEnum, 0);
	this->sslCompatibility = cfgEnumValue(cfgQueryFromArray("sslCompatibility", numEntries, path), cfgYesNoEnum, 0);
	this->sslRequireServerCert = cfgEnumValue(cfgQueryFromArray("sslRequireServerCert", numEntries, path), cfgYesNoEnum, 0);
	this->sslVersion = cfgEnumValue(cfgQueryFromArray("sslVersion", numEntries, path), cifsModSslVersEnum, SRV_SSL_V23);
	SET_ASCII(this->sslCACertDir, cfgStringValue(cfgQueryFromArray("sslCACertDir", numEntries, path), ""));
	SET_ASCII(this->sslCACertFile, cfgStringValue(cfgQueryFromArray("sslCACertFile", numEntries, path), ""));
	SET_ASCII(this->sslKeyFile, cfgStringValue(cfgQueryFromArray("sslKeyFile", numEntries, path), ""));
	SET_ASCII(this->sslCertFile, cfgStringValue(cfgQueryFromArray("sslCertFile", numEntries, path), ""));
	SET_ASCII(this->sslCiphers, cfgStringValue(cfgQueryFromArray("sslCiphers", numEntries, path), NULL));
	this->attrCacheTime = cfgIntValue(cfgQueryFromArray("attributesCacheTime", numEntries, path), CONST_ATTRCACHE_LIFETIME);
	this->dirCacheTime = cfgIntValue(cfgQueryFromArray("dirCacheTime", numEntries, path), CONST_DIRCACHE_LIFETIME);
	this->maxCachedFiles = cfgIntValue(cfgQueryFromArray("maxCachedFiles", numEntries, path), 150);

/* srvReadConfig() is called twice, once during object allocation and once
 * after connecting. The following code therefore collects all disabled SMBs
 * of the two calls. During the first call, srvClass is NULL and the normal
 * lookup procedure is used to find "disableSmbs". During the second call,
 * if there's no server-specific configuration, "disableSmbs" will be found
 * in the server class configuration and the SMBs disabled there will be
 * added to the list of disabled SMBs.
 */
	current = cfgQueryFromArray("disableSmbs", numEntries, path);
	count = cfgQueryValueCount(current);
	for(i=0;i<count;i++)
		this->cfgProfile |= 1 << cfgEnumValue(cfgQueryAtIndex(i, current), smbModeEnum, 32);
}

/* ------------------------------------------------------------------------- */

static void	srvFree(server_t *this)
{
int		i;

    if(this != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "server: freeing object at 0x%08x\n", (int)this);
		if(this->name != NULL){
			licRemoveServer(this);
			lmbRemoveServer(this);
		}
       	if(this->lowerObj != NULL){
			this->lowerI->setHigherLayer(this->lowerObj, NULL, NULL);
			this->lowerI->release(this->lowerObj);
			this->lowerObj = NULL;
		}
		if(this->requests != NULL){
            for(i=0;i<this->windowSize;i++){
                if(this->requests[i] != NULL)
                	smbRequestFree(this->requests[i]);
            }
			memFree(this->requests);
        }
		if(this->nullSession != NULL){
			this->nullSession->server = NULL;	/* hack: avoid server release */
			userRelease(this->nullSession);
		}
		if(this->ipcShare != NULL){
			srvRemoveShare(this, this->ipcShare);
			this->ipcShare->server = NULL;		/* hack: avoid server release */
			shrRelease(this->ipcShare);
		}
        if(this->sendRequests != NULL)
            fifoFree(this->sendRequests);
        if(this->smbDemands != NULL)
            procCondFree(this->smbDemands);
        if(this->connectSignal != NULL)
            procCondFree(this->connectSignal);
        if(this->usersByUnixUid != NULL)
            htabFree(this->usersByUnixUid);
        if(this->allShares != NULL)
            arrFree(this->allShares);
		if(this->name != NULL)
			memFree(this->name);
		memset(this, 0, sizeof(server_t));
		memFree(this);
		elpCancelCall(srvCheckDemands, this);
    }
}

/* ------------------------------------------------------------------------- */

server_t	*srvNew(btu8 *srvName)
{
server_t	*volatile this = NULL;
btsize		len;
btuni		nullName[1] = {0}, nullPasswd[1] = {0}, ipcUncPath[71];
	
    if(connCtxPool == NULL)
        connCtxPool = poolNew(sizeof(connCtx_t), sizeof(connCtx_t));
EXC_TRY
    this = memCalloc(sizeof(server_t));
	this->refCount = 1;
	this->useUnicode = 1;	/* set to 0 by anything that objects Unicode */
    this->sendRequests = fifoNew();
    this->smbDemands = procCondNew();
    this->connectSignal = procCondNew();
    this->cstatus = PSL_NOTCONNECTED;
	this->name = utilStrDup(srvName);
    this->usersByUnixUid = htabNew(&userByUnixUidInterface, 16, 50);
	this->allShares = arrNew(NULL, 8);
	this->txFlags2 = CIFS_FLAG2_LONG_NAMES;
	this->txFlags = CIFS_FLAG_GUARANTEE_RXBUF | CIFS_FLAG_CASE_INSENSITIVE;
	this->cachePageSize = 512;	/* defaults until transfer size is known */
	srvReadConfig(this, NULL);
    this->windowSize = 1;
    this->requests = memCalloc(sizeof(smbRequest_t *));
    this->requests[0] = smbRequestNew(this);
	this->nullSession = userNew(-1, nullName, nullPasswd, NULL, NULL, 1);
	this->nullSession->server = this;	/* avoid userSetServer() method because it retains server */
	ipcUncPath[0] = ipcUncPath[1] = '\\';
	convLocalAsciiToUni(srvName, ipcUncPath + 2, ELEMENTS(ipcUncPath) - 2);
	len = uniStrLen(ipcUncPath);
	if(len < ELEMENTS(ipcUncPath) - 2)
		ipcUncPath[len++] = '\\';
	convLocalAsciiToUni("IPC$", ipcUncPath + len, ELEMENTS(ipcUncPath) - len);
	this->ipcShare = shrNew(ipcUncPath, this, 1);	
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "server: allocated new object at 0x%08x\n", (int)this);
EXC_CATCH
	srvFree(this);
	EXC_RETHROW();
EXC_ENDTRY
	return this;
}

server_t	*srvWithName(btu8 *srvName, btbool isLMB)
{
server_t	*srv;
btError		dummy;

	if((srv = lmbLookupServer(srvName)) == NULL)
		srv = licLookupServer(srvName);
	if(srv != NULL){
		srvRetain(srv);
		return srv;
	}
	srv = srvNew(srvName);
	if(!isLMB){
		licRegisterServer(srv, NULL, &dummy);
	}
	return srv;
}

/* ------------------------------------------------------------------------- */

void	srvRetain(void *obj)
{
server_t	*this = obj;

    this->refCount++;
}

void	srvRelease(void *obj)
{
server_t	*this = obj;

	this->refCount--;
    if(this->refCount == 0){
        srvFree(this);
    }else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

void	srvSetSessionSetupResult(server_t *this, btuni *serverOS, btuni *serverLM)
{
btu8		srvOS[256], srvLM[256], *cfgOS, *cfgLM;
plist_t		*srvClasses, *current, *section, *path[4];
btsize		i, count, numEntries;
btu32		capMask, capVal;

	if(!this->hadSessionSetup){
		this->hadSessionSetup = 1;
		convUniToLocalAscii(serverOS, srvOS, sizeof(srvOS));
		convUniToLocalAscii(serverLM, srvLM, sizeof(srvLM));
		if(LOG_MODE_IS_ACTIVE(LOGM_INFO)){
			logPrintf(LOGM_INFO, "Server OS: %s\n", srvOS);
			logPrintf(LOGM_INFO, "Server LM: %s\n", srvLM);
			logPrintf(LOGM_INFO, "Server Capabilities: 0x%08x\n", this->capabilities);
		}
		numEntries = cifsModGlobalCfgPath(path, ELEMENTS(path));
		srvClasses = cfgQueryFromArray("serverClasses", numEntries, path);
		count = cfgQueryValueCount(srvClasses);
		for(i=0;i<count;i++){
			current = cfgQueryAtIndex(i, srvClasses);
			cfgOS = cfgStringValue(cfgQueryWithName("OS", 1, current), NULL);
			cfgLM = cfgStringValue(cfgQueryWithName("LanManager", 1, current), NULL);
			capMask = cfgIntValue(cfgQueryWithName("capabilityMask", 1, current), 0);
			capVal = cfgIntValue(cfgQueryWithName("capabilityValue", 1, current), 0);
			section = cfgQueryWithName("config", 1, current);
			if(cfgOS == NULL || cfgLM == NULL || section == NULL){
				if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
					logPrintf(LOGM_ERROR, "server: config error in serverClasses, definition %d\n", i + 1);
				}
				continue;
			}
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "server: serverClasses: trying match with OS=\"%s\" LM=\"%s\" CM=0x%08x CV=0x%08x\n", cfgOS, cfgLM, capMask, capVal);
			if(!wildcardMatch(srvOS, cfgOS) || !wildcardMatch(srvLM, cfgLM))
				continue;
			if(((this->capabilities ^ capVal) & capMask) != 0)
				continue;
			if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
				logPrintf(LOGM_INFO, "server: serverClasses: match found with OS=\"%s\" LM=\"%s\" CM=0x%08x CV=0x%08x\n", cfgOS, cfgLM, capMask, capVal);
			srvReadConfig(this, section);
			this->cmdProfile &= ~this->cfgProfile;
			break;
		}
		
	}else if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "server: ignoring session setup result\n");
}

/* ------------------------------------------------------------------------- */

void	srvAddShare(server_t *this, struct share *share)
{
	if(arrFindObjectEqualTo(this->allShares, share) == NULL)
		arrAddObject(this->allShares, share);
}

/* ------------------------------------------------------------------------- */

static void	srvCheckUsed(server_t *srv)
{
	/* One share will always be registered: the IPC share. */
	if(htabObjectCount(srv->usersByUnixUid) == 0 && arrObjectCount(srv->allShares) <= 1)
		licRemoveServer(srv);
}

void	srvRemoveUser(server_t *this, struct cifsUser *user)
{
	htabRemoveObjectEqualTo(this->usersByUnixUid, user);
	srvCheckUsed(this);
	mainModGuiUpdate(cifsLoginListObject);
}

void	srvRemoveShare(server_t *this, struct share *share)
{
	arrRemoveObjectEqualTo(this->allShares, share);
	srvCheckUsed(this);
}

/* ------------------------------------------------------------------------- */

struct cifsUser *srvUserForUid(server_t *this, int unixUid)
{
cifsUser_t		dummyUser;

    dummyUser.unixUid = unixUid;
    return htabFindObjectEqualTo(this->usersByUnixUid, &dummyUser);
}

/* ------------------------------------------------------------------------- */

void	srvConnectionLost(server_t *this)
{
int			i, l, cstatus = this->cstatus;
array_t		*a;

    this->cstatus = PSL_NOTCONNECTED;
    procCondBroadcast(this->connectSignal, NULL);
    for(i=0;i<this->windowSize;i++){
		if(this->requests[i]->requestState != SMBS_IDLE){
            smbConnectionLost(this->requests[i]);
        }
    }
    if(cstatus == PSL_CONNECTED){
        a = htabAllObjects(this->usersByUnixUid);
        l = arrObjectCount(a);
        for(i=0;i<l;i++){
            userConnectionLost(arrObjectAtIndex(a, i));
        }
        arrFree(a);
		if(this->nullSession != NULL){
			userConnectionLost(this->nullSession);
		}
        l = arrObjectCount(this->allShares);
        for(i=0;i<l;i++){
            shrConnectionLost(arrObjectAtIndex(this->allShares, i));
        }
		if(this->ipcShare != NULL){
			shrConnectionLost(this->ipcShare);
		}
		if(this->lowerObj != NULL)
			this->lowerI->setConnectStatus(this->lowerObj, PSL_NOTCONNECTED, 0, NULL);
    }
}

/* ------------------------------------------------------------------------- */

static void	srvConnectProc(void *obj, void *arg)
{
connCtx_t			*this = obj;
server_t			*server = this->server;
smbRequest_t		*smb = arg;
cifsResNegotiate_t	*rval = arg;
userRes_t			*urval = arg;
btError				*netbRval = arg;
srvRes_t			res;
uirepRes_t			*prval = arg;
btint				port, *ipRval = arg, numEntries;
cfgProperty_t		*path[8];
btu8				*ipName, *myName, hostname[64];
btsize				dummy;

	PROC_BEGIN;
	srvRetain(server);
    while(server->cstatus == PSL_CONNECTING || server->cstatus == PSL_DISCONNECTING){
		PROC_WAITFOR_X(1, procCondWait(server->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(server->cstatus == PSL_CONNECTED){	/* if someone connected meanwhile */
		res.err.class = res.err.code = 0;
		srvRelease(server);
		PROC_RETURN(&res);
	}
    server->cstatus = PSL_CONNECTING;
	if(server->lowerObj == NULL){	/* no interface so far */
		numEntries = cifsModServerCfgPath(server->name, NULL, path, ELEMENTS(path));
		ipName = cfgStringValue(cfgQueryFromArray("ipAddress", numEntries, path), NULL);
		if(ipName != NULL && ipName[0] != 0){
			if(!utilResolveName(ipName, &this->srvIp))
				this->srvIp = 0;	/* keep IP at 0 if error occured */
		}
		if(this->srvIp == 0){
			PROC_WAITFOR_X(2, nbnsLookup(server->name, this));
			if(ipRval != NULL)
				this->srvIp = *ipRval;
		}
		if(this->srvIp == 0){
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOClookup;
			goto errorOccured;
		}
		gethostname(hostname, sizeof(hostname));
		hostname[63] = 0;
		numEntries = cifsModServerCfgPath(server->name, NULL, path, ELEMENTS(path));
		myName = cfgStringValue(cfgQueryFromArray("localNetbiosName", numEntries, path), hostname);
		port = cfgIntValue(cfgQueryFromArray("tcpPort", numEntries, path), CONST_NETBIOS_PORT);
		server->lowerObj = pslNetbiosNew(myName, server->name, this->srvIp, port, server);
		server->lowerI = &pslNetbiosUpInterface;
		server->lowerI->setHigherLayer(server->lowerObj, server, &serverDownInterface);
		server->lowerI->maxInserts(server->lowerObj, &server->msgPreInsert, &dummy);
	}
	pslNetbiosSetRequireEncryption(server->lowerObj, server->sslRequireEncryption);
retry:
    PROC_WAITFOR_X(3, server->lowerI->setConnectStatus(server->lowerObj, PSL_CONNECTED, srvConnectTimeout(server), this));
	if(netbRval->class == ERRCLS_CIFS_LOCAL && netbRval->code == ERR_CIFS_LOCneedpassphrase){
		if(this->client != NULL){
			btError	err;
			err.class = ERRCLS_CIFS_LOCAL;
			err.code = ERR_CIFS_LOCwrongphrase;
			PROC_WAITFOR_X(4, guimanDisplayError(this->client, &err, this));
		}
		PROC_WAITFOR_X(5, uirepGetPassPhrase(this->uid, server->name, this));
		if(prval->isOk){
			srvSetSslPassPhrase(server, prval->cifsPasswd);
			this->client = prval->client;
			goto retry;
		}
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCneedpassphrase;
		goto errorOccured;
	}else if(netbRval->class != 0){
		res.err = *netbRval;
		goto errorOccured;
	}
    PROC_WAITFOR_X(6, srvWantRequest(server, this));
    PROC_WAITFOR_X(7, cifsNegotiate(smb, this));
	if(rval->err.class != 0){
		res.err = rval->err;
		goto errorOccured;
	}
	srvSetNegotiationResult(server, rval);
	server->cstatus = PSL_CONNECTED;	/* server is connected now, required by userConnect */
	if(srvUseRequest(server, SRV_USE_SESSION_SETUP)){
		PROC_WAITFOR_X(8, userConnect(server->nullSession, this));
		if(urval->err.class == 0){
			PROC_WAITFOR_X(9, shrConnect(server->ipcShare, server->nullSession, this));
			/* ignore result */
		}
		/* success of server connect does not depend on ipcShare and nullSession */
	}
	memset(&res, 0, sizeof(res));
	goto okReturn;
errorOccured:
	server->cstatus = PSL_NOTCONNECTED;
	if(server->lowerObj != NULL)
		server->lowerI->setConnectStatus(server->lowerObj, PSL_NOTCONNECTED, 0, NULL);	/* disconnect in background */
okReturn:
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(server->connectSignal, NULL);
		srvRelease(server);
	);
    PROC_END;
}

/* ------------------------------------------------------------------------- */

void	srvConnect(server_t *this, btint passPhraseUid, void *caller)
{
srvRes_t	res;
connCtx_t	*ctx;

	res.err.class = res.err.code = 0;
    if(this->cstatus == PSL_CONNECTED){
        PROC_IMMEDIATE_RET(caller, &res);
    }
    PROC_INIT(ctx, srvConnectProc, caller, connCtxPool, poolFreeBlock);
    ctx->server = this;
	ctx->uid = passPhraseUid;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	srvSetNegotiationResult(server_t *this, cifsResNegotiate_t *res)
{
btu32	mask;

    if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
        logPrintf(LOGM_DEBUG, "Negotiate result: maxXmit=%d\n", res->maxXmit);
    }
	this->securityIsUserLevel = (res->securityMode & 0x01) != 0;
    this->canEncrypt = res->encryptKeyLen == sizeof(this->encryptKey) && (res->securityMode & 0x02) != 0;
    this->maxXmit = res->maxXmit;
    this->sessionKey = res->sessionKey;
    this->dialect = res->dialect;
	this->serverGmtOffset = res->gmtOffset;
	this->capabilities = res->capabilities;
	if(this->canEncrypt)
    	memcpy(this->encryptKey, res->encryptKey, sizeof(this->encryptKey));
    this->encryptKeyLen = res->encryptKeyLen;
	srvSetWindowsize(this, res->maxMpx > CONST_MAX_WINDOW_SIZE ? CONST_MAX_WINDOW_SIZE : res->maxMpx);
    this->cmdProfile = 0;
    switch(this->dialect){
        case CIFS_DIALECT_NTLM:
            if(res->capabilities & CIFS_SERVER_CAP_UNIX){
				this->cmdProfile |= (1 << SRV_USE_ATTR_UNIX);
				this->cmdProfile |= (1 << SRV_USE_RDIR_UNIX);
			}
            if(res->capabilities & CIFS_SERVER_CAP_NT_FIND){
				this->cmdProfile |= (1 << SRV_USE_RDIR_DIRINFO);
				this->cmdProfile |= (1 << SRV_USE_RDIR_BOTHDIRINFO);
			}
		/* fallthrough */
        case CIFS_DIALECT_LM2:
			this->cmdProfile |= (1 << SRV_USE_GETATTR_RDIR);
			this->cmdProfile |= (1 << SRV_USE_GETATTR_PATHINFO);

			this->cmdProfile |= (1 << SRV_USE_SETATTR_TRANS2SETFILE);
			this->cmdProfile |= (1 << SRV_USE_SETATTR_TRANS2SETPATH);
			this->cmdProfile |= (1 << SRV_USE_SETATTR_SETFILE2);
			this->cmdProfile |= (1 << SRV_USE_SETATTR_CORETIME);

			this->cmdProfile |= (1 << SRV_USE_CREATE_OPENX);
			this->cmdProfile |= (1 << SRV_USE_OPEN_OPENX);
			this->cmdProfile |= (1 << SRV_USE_READ_READX);
			this->cmdProfile |= (1 << SRV_USE_READ_OPENREAD);
			this->cmdProfile |= (1 << SRV_USE_WRITE_WRITEX);
			this->cmdProfile |= (1 << SRV_USE_WRITE_OPENWRITE);

			this->cmdProfile |= (1 << SRV_USE_RDIR_LM);
			this->cmdProfile |= (1 << SRV_USE_FSINFO_LM);
			this->cmdProfile |= (1 << SRV_USE_SESSION_SETUP);
			this->cmdProfile |= (1 << SRV_USE_TCON_ANDX);
		/* fallthrough */
        case CIFS_DIALECT_CORE:
        default:
			this->cmdProfile |= (1 << SRV_USE_DIRDATE_EMUL);
			this->cmdProfile |= (1 << SRV_USE_RDIR_AUTOCLOSE);
			this->cmdProfile |= (1 << SRV_USE_RDIR_IMPLICIT_RESUME);
            break;
    }
/* take care: Anything configured here may be overridden by srvReadConfig()
 * called after negotiation! */
	if((res->capabilities & CIFS_SERVER_CAP_UNICODE) == 0)
		this->useUnicode = 0;	/* use unicode only with unicode capable servers */
	this->cmdProfile &= ~this->cfgProfile;
	if(this->maxXmit < 1274){
		this->cachePageSize = 512;
	}else if(this->maxXmit < 2298){
		this->cachePageSize = 1024;
	}else if(this->maxXmit < 4346){
		this->cachePageSize = 2048;
	}else if(this->maxXmit < 8442){
		this->cachePageSize = 4096;
	}else{
		this->cachePageSize = 8192;
	}
	if(this->maxXmit - this->cachePageSize < (300 + CIFS_MAX_PATHLEN * 2)){
		mask = 1 << SRV_USE_WRITE_OPENWRITE;
		if((this->cmdProfile & mask) != 0 && LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "server: cannot use batched open and write because max_xmit too small\n");
		this->cmdProfile &= ~mask;
	}
}

/* ------------------------------------------------------------------------- */

static void	srvCheckDemands(void *ref, void *arg)
{
server_t		*this = ref;
int				r;

	while(procCondWaiting(this->smbDemands)){
        for(r=0;r<this->windowSize;r++){
            if(this->requests[r]->requestState == SMBS_IDLE)
                break;
        }
        if(r >= this->windowSize)	/* no SMB free */
            break;
        if(LOG_MODE_IS_ACTIVE(LOGM_SERVER_SMB)){
            logPrintf(LOGM_SERVER_SMB, "server: granted SMB block\n");
        }
        procCondSignal(this->smbDemands, this->requests[r]);
    }
}

/* ------------------------------------------------------------------------- */

void	srvWantRequest(server_t *this, void *caller)
{
    if(LOG_MODE_IS_ACTIVE(LOGM_SERVER_SMB))
        logPrintf(LOGM_SERVER_SMB, "server: request for SMB block\n");
	elpCallCancelPrev(srvCheckDemands, this);
    procCondWait(this->smbDemands, caller);
}

/* ------------------------------------------------------------------------- */

void	srvRequestStatusChanged(server_t *this, struct smbRequest *rq, int newstat)
{
	if(this->lowerI == NULL)
		return;
    if(this->lowerI->connectStatus(this->lowerObj) == PSL_CONNECTED)
		if(newstat == SMBS_IDLE && procCondWaiting(this->smbDemands))
			elpCallCancelPrev(srvCheckDemands, this);
}

void	srvWantSend(server_t *this, struct smbRequest *rq)
{
    fifoAddObject(this->sendRequests, rq);
	this->lowerI->wantSend(this->lowerObj);
}

btu16	srvGetMid(server_t *this)
{
	if(this->nextMid >= 0xfff0)
		this->nextMid = 0;		/* reserve 16 MIDs for spontaneous server msgs */
    return this->nextMid++;
}

/* ------------------------------------------------------------------------- */

static void	srvOplockBreak(server_t *this, smbHeader_t *hdr, btu8 *data, btsize len)
{
btu16	fid;
int		mode, i, l;
share_t	*share;

    if(LOG_MODE_IS_ACTIVE(LOGM_OPLOCK))
        logPrintf(LOGM_OPLOCK, "server: received oplock break\n");
	cifsDecodeOplockBreak(hdr, data, len, &fid, &mode);
	l = arrObjectCount(this->allShares);
	for(i=0;i<l;i++){
		share = arrObjectAtIndex(this->allShares, i);
		if(shrTid(share) == hdr->tid){
			shrOplockBreak(share, fid, mode);
			break;
		}
	}
	if(i >= l && LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
		logPrintf(LOGM_ERROR, "server: oplock break for unknown tree %d\n", hdr->tid);
	}
}

/* ------------------------------------------------------------------------- */

static void	srvRxMessage(void *obj, btu8 *data, btsize len)
{
server_t	*this = obj;
smbHeader_t	hdr;
int			i;

	srvRetain(this);	/* avoid freeing server during data processing */
	smbParseHeader(&hdr, data, len);
    if(LOG_MODE_IS_ACTIVE(LOGM_SERVER_SMB))
        logPrintf(LOGM_SERVER_SMB, "server: received netbios block mid=%d\n", hdr.mid);
	if(hdr.mid == 0xffff){	/* this is a spontaneous server message */
		switch(hdr.command){
		case SMB_COM_LOCKING_ANDX:	/* this must be an oplock break message */
			srvOplockBreak(this, &hdr, data, len);
			break;
		default:
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "server: spontaneous message received: 0x%02x\n", hdr.command);
			break;
		}
	}else{
		for(i=0;i<this->windowSize;i++){
			if(this->requests[i]->requestState != SMBS_IDLE && this->requests[i]->mid == hdr.mid){
				smbRxBlock(this->requests[i], &hdr, data, len);
				break;
			}
		}
		if(i >= this->windowSize && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
        	logPrintf(LOGM_ERROR, "server: reply mid=%d not expected\n", hdr.mid);
	}
	srvRelease(this);
}

/* ------------------------------------------------------------------------- */

static nbmsg_t	*srvGetTxMessage(void *obj)
{
server_t		*this = obj;
nbmsg_t			*block;
smbRequest_t	*smb;

    while((smb = fifoGetObject(this->sendRequests)) != NULL){
        if((block = smbTxBlock(smb)) != NULL){
            if(LOG_MODE_IS_ACTIVE(LOGM_SERVER_SMB))
                logPrintf(LOGM_SERVER_SMB, "server: transmitting netbios block mid=%d\n", smb->mid);
            return block;
        }
    }
    return NULL;
}

/* ------------------------------------------------------------------------- */

static void	srvTxReady(void *obj, btint errcode)
{
	/* ignore this info, nothing to be done */
}

/* ------------------------------------------------------------------------- */

static void	srvConnectStatusChanged(void *obj, btint newStatus)
{
server_t	*this = obj;

	if(newStatus == PSL_NOTCONNECTED && this->cstatus != PSL_NOTCONNECTED){
		srvConnectionLost(this);
	}
}

/* ------------------------------------------------------------------------- */

struct msg	*srvGetMessageBuffer(server_t *this)
{
	return msgNew(this->msgPreInsert);
}

/* ------------------------------------------------------------------------- */

btbool	srvConnectionIsEncrypted(server_t *server)
{
	return pslNetbiosConnectionEncrypted(server->lowerObj);
}

/* ------------------------------------------------------------------------- */

void	srvSetSslPassPhrase(server_t *server, btu8 *passPhrase)
{
	pslNetbiosSetPassPhrase(server->lowerObj, passPhrase);
}

/* ------------------------------------------------------------------------- */

bttime	srvConnectTimeout(server_t *srv)
{
btu32			g = cfgGlobalGeneration();
btint			numEntries;
cfgProperty_t	*path[8];

	if(srv->connectTimeoutGeneration != g){
		numEntries = cifsModServerCfgPath(srv->name, NULL, path, ELEMENTS(path));
		srv->connectTimeout = cfgIntValue(cfgQueryFromArray("connectTimeout", numEntries, path), CONST_CONNECT_TIMEOUT);
		srv->connectTimeoutGeneration = g;
	}
	return srv->connectTimeout;
}

/* ------------------------------------------------------------------------- */

bttime	srvRequestTimeout(server_t *srv)
{
btu32			g = cfgGlobalGeneration();
btint			numEntries;
cfgProperty_t	*path[8];

	if(srv->requestTimeoutGeneration != g){
		numEntries = cifsModServerCfgPath(srv->name, NULL, path, ELEMENTS(path));
		srv->requestTimeout = cfgIntValue(cfgQueryFromArray("requestTimeout", numEntries, path), CONST_REQUEST_TIMEOUT);
		srv->requestTimeoutGeneration = g;
	}
	return srv->requestTimeout;
}

/* ------------------------------------------------------------------------- */

btu8	srvCaseMapping(server_t *srv)
{
btu32			g = cfgGlobalGeneration();
btint			numEntries;
cfgProperty_t	*path[8];

	if(srv->caseMappingGeneration != g){
		numEntries = cifsModServerCfgPath(srv->name, NULL, path, ELEMENTS(path));
		srv->caseMapping = cfgEnumValue(cfgQueryFromArray("caseMapping", numEntries, path), cifsModCasemapEnum, SRV_CASE_NONE);
		srv->caseMappingGeneration = g;
	}
	return srv->caseMapping;
}

/* ------------------------------------------------------------------------- */

btu8	srvExecMapping(server_t *srv)
{
btu32			g = cfgGlobalGeneration();
btint			numEntries;
cfgProperty_t	*path[8];

	if(srv->execMappingGeneration != g){
		numEntries = cifsModServerCfgPath(srv->name, NULL, path, ELEMENTS(path));
		srv->execMapping = cfgEnumValue(cfgQueryFromArray("execMapping", numEntries, path), cifsModExmapEnum, SRV_EXMAP_SYSTEM);
		if(cfgEnumValue(cfgQueryFromArray("execInvert", numEntries, path), cfgYesNoEnum, 0))
			srv->execMapping |= SRV_EXMAP_INVERT;
		srv->execMappingGeneration = g;
	}
	return srv->execMapping;
}

/* ------------------------------------------------------------------------- */

void	srvGuestCredentials(server_t *srv, btu8 **user, btu8 **passwd)
{
btint			numEntries;
cfgProperty_t	*path[8];

	numEntries = cifsModServerCfgPath(srv->name, NULL, path, ELEMENTS(path));
	*user = cfgStringValue(cfgQueryFromArray("guestRemoteUser", numEntries, path), NULL);
	*passwd = cfgStringValue(cfgQueryFromArray("guestPassword", numEntries, path), "");
}

/* ------------------------------------------------------------------------- */
