/*
 * Name: server.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-20
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: server.h,v 1.5 1999/08/01 12:57:14 cs Exp $
 */

#ifndef __server_H_INCLUDED__
#define __server_H_INCLUDED__

#include "basictypes.h"
#include "smbnumbers.h"

/*
General Description:
This module implements the server class. It represents a CIFS server
within the program. Each server object contains a hashtable with references
to all shares connected at this server and a database (also implemented by
a hashtable) of all users logged in to the server. It also manages the context
of outstanding SMB requests.
*/

/* ------------------------------------------------------------------------- */

/* the following defines are used to represent the combination of SMB
 * requets to use, they must match smbModeEnum in server.c:
 # DON'T FORGET TO ADD IT TO srvSetNegotiationResult(), too!!!
 */
#define	SRV_USE_GETATTR_RDIR			0
#define	SRV_USE_GETATTR_PATHINFO		1
#define	SRV_USE_ATTR_UNIX				2
#define	SRV_USE_SETATTR_TRANS2SETFILE	3
#define	SRV_USE_SETATTR_TRANS2SETPATH	4
#define	SRV_USE_SETATTR_SETFILE2		5
#define	SRV_USE_SETATTR_CORETIME		6
#define	SRV_USE_CREATE_OPENX			7
#define	SRV_USE_OPEN_OPENX				8
#define	SRV_USE_READ_READX				9
#define	SRV_USE_READ_OPENREAD			10
#define	SRV_USE_WRITE_WRITEX			11
#define	SRV_USE_WRITE_OPENWRITE			12
#define	SRV_USE_RDIR_UNIX				13
#define	SRV_USE_RDIR_LM					14
#define	SRV_USE_FSINFO_LM				15
#define	SRV_USE_SESSION_SETUP			16
#define	SRV_USE_TCON_ANDX				17
#define	SRV_USE_DIRDATE_EMUL			18
#define	SRV_USE_RDIR_DIRINFO			19
#define	SRV_USE_RDIR_AUTOCLOSE			20
#define	SRV_USE_RDIR_IMPLICIT_RESUME	21
#define	SRV_USE_RDIR_BOTHDIRINFO		22

/* case mapping of path names, must match cifsModCasemapEnum in cifsmod.c: */
#define	SRV_CASE_NONE		0
#define	SRV_CASE_LOWER		1
#define	SRV_CASE_UPPER		2

/* execute mapping definitions, must match cifsModExmapEnum in cifsmod.c: */
#define	SRV_EXMAP_ARCHIVE	0
#define	SRV_EXMAP_SYSTEM	1
#define	SRV_EXMAP_HIDDEN	2
#define	SRV_EXMAP_ON		3
#define	SRV_EXMAP_OFF		4
#define	SRV_EXMAP_INVERT	0x40

/* SSL version definitions, must match sslVersExtNames in cifsmod.c: */
#define	SRV_SSL_V23			0
#define	SRV_SSL_V2			1
#define	SRV_SSL_V3			2
#define	SRV_SSL_TLS1		3

/* ------------------------------------------------------------------------- */

struct smbRequest;
struct psUpInterface;
struct fifo;
struct cifsResNegotiate;
struct cifsUser;
struct share;

typedef struct server{
	btu8				*name;		/* name of this server */
    int					refCount;	/* will be part of collection */
    int					cstatus;	/* whether connection is negotiated */
/* per server information storage */
    struct hashtable	*usersByUnixUid;	/* all authenticated users */
    struct array		*allShares;			/* all shares open at server */
	struct cifsUser		*nullSession;		/* a null user, if available */
	struct share		*ipcShare;			/* tree connected to IPC$ */
/* callback invocation buffer */
    struct procCond		*smbDemands;
    struct procCond		*connectSignal;
/* transport properties */
	void				*lowerObj;		/* next lower object in pstack */
	struct psUpInterface *lowerI;		/* interface of lower object */
	btsize				msgPreInsert;
    struct smbRequest	**requests;
    struct fifo			*sendRequests;	/* smbs requesting transmission */
    btsize				windowSize;
    btu16				nextMid;
/* transmission porperties */
    btu16				txFlags2;
	btu32				capabilities;
    btu8				txFlags;
	btbool				useUnicode;
/* negotiated properties */
	btbool				securityIsUserLevel;
    btbool				canEncrypt;
    btu32				maxXmit;
    btu32				sessionKey;
    int					dialect;
	bti32				serverGmtOffset;	/* add this to server time for GMT */
    btu8				encryptKey[8];
    btu16				encryptKeyLen;
/* configuration constants */
	btu32				cfgProfile;		/* which SMBs are disabled form config */
	btint				maxCachedFiles;
	bttime				attrCacheTime;
	bttime				dirCacheTime;
	btbool				caseSensitive;
	btbool				capitalizeShares;	/* whether share names are capitalized */
	btbool				fakeMountpointDate;	/* fake the modification and access dates of mountpoint */
	btbool				closeForSetattr;	/* files must be closed before attributes are changed */
	btu16				linkModeMask;	/* identifies bits used for faked links */
	btu16				linkMode;		/* status of bits for faked links */
	btbool				linksAreUnicode;
/* SSL options: */
	btbool				sslRequireEncryption;
	btbool				sslRequireServerCert;
	btu8				sslVersion;
	btbool				sslCompatibility;
	btu8				*sslCACertDir;
	btu8				*sslCACertFile;
	btu8				*sslKeyFile;
	btu8				*sslCertFile;
	btu8				*sslCiphers;
/* derived from negotiated properties and configuration: */
	btbool				hadSessionSetup;/* true if reconfigured after session setup */
    btu32				cmdProfile;		/* what command set to use */
	btsize				cachePageSize;	/* size of one cache page in bytes */
	/* profile for unlocked files: */
	bttime				cacheValidUnlocked;	/* how long is a read cache valid if not locked? */
	bttime				closeDelayUnlocked;
	int					readAheadUnlocked;
	btbool				writeThroughUnlocked;
	/* profile for locked files: */
	bttime				closeDelayLocked;
	bttime				closeDelayBatchlock;
	int					readAheadLocked;
	btbool				writeThroughLocked;
/* configuration cache: */
	btu8				caseMapping;	/* one of the SRV_CASE_* defines */
	btu8				execMapping;	/* what to do with execute permission */
	bttime				connectTimeout;
	bttime				requestTimeout;
	btu32				caseMappingGeneration;
	btu32				execMappingGeneration;
	btu32				connectTimeoutGeneration;
	btu32				requestTimeoutGeneration;
}server_t;

typedef struct srvRes{
    btError			err;
	server_t		*server;
}srvRes_t;

/* ------------------------------------------------------------------------- */

server_t	*srvWithName(btu8 *srvName, btbool isLMB);	/* returns retained object */
server_t	*srvNew(btu8 *srvName);
void	srvRetain(void *obj);
void	srvRelease(void *obj);
void	srvSetWindowsize(server_t *this, int newWinSize);
void	srvSetSessionSetupResult(server_t *this, btuni *serverOS, btuni *serverLM);
void	srvSetNegotiationResult(server_t *this, struct cifsResNegotiate *res);
void	srvWantRequest(server_t *this, void *caller);
void	srvRequestStatusChanged(server_t *this, struct smbRequest *rq, int newstat);
void	srvWantSend(server_t *this, struct smbRequest *rq);
btu16	srvGetMid(server_t *this);
void	srvConnectionLost(server_t *this);
void	srvConnect(server_t *this, btint passPhraseUid, void *caller);
void	srvRemoveUser(server_t *this, struct cifsUser *user);
struct cifsUser *srvUserForUid(server_t *this, int unixUid);
void	srvAddShare(server_t *this, struct share *share);
void	srvRemoveShare(server_t *this, struct share *share);
struct msg	*srvGetMessageBuffer(server_t *this);
btbool	srvConnectionIsEncrypted(server_t *server);
void	srvSetSslPassPhrase(server_t *server, btu8 *passPhrase);
bttime	srvConnectTimeout(server_t *srv);
bttime	srvRequestTimeout(server_t *srv);
btu8	srvCaseMapping(server_t *srv);
btu8	srvExecMapping(server_t *srv);
void	srvGuestCredentials(server_t *srv, btu8 **user, btu8 **passwd);

/* ------------------------------------------------------------------------- */

static inline int	srvConnectionStatus(server_t *this)
{
    return this->cstatus;
}

#define	srvAllUsers(srv)	(htabAllObjects((srv)->usersByUnixUid))
/* use define to avoid including hashtable.h here */

static inline btu8	srvOpLock(server_t *this)
{
    return (this->txFlags >> 4) & 6;
}

static inline btsize	srvCachePageSize(server_t *this)
{
	return this->cachePageSize;
}

static inline bttime	srvFileCloseDelay(server_t *server, btint oplockLevel)
{
	switch(oplockLevel){
		default:
		case CIFS_OPLOCK_NONE:		return server->closeDelayUnlocked;
		case CIFS_OPLOCK_L2:
		case CIFS_OPLOCK_EXCLUSIVE:	return server->closeDelayLocked;
		case CIFS_OPLOCK_BATCH:		return server->closeDelayBatchlock;
	}
}

static inline bttime	srvCacheValidDelay(server_t *server, btint oplockLevel)
{
	switch(oplockLevel){
		default:
		case CIFS_OPLOCK_NONE:		return server->cacheValidUnlocked;
		case CIFS_OPLOCK_L2:
		case CIFS_OPLOCK_EXCLUSIVE:
		case CIFS_OPLOCK_BATCH:		return -1;
	}
}

static inline btint	srvReadAheadPages(server_t *server, btint oplockLevel)
{
	switch(oplockLevel){
		default:
		case CIFS_OPLOCK_NONE:		return server->readAheadUnlocked;
		case CIFS_OPLOCK_L2:
		case CIFS_OPLOCK_EXCLUSIVE:
		case CIFS_OPLOCK_BATCH:		return server->readAheadLocked;
	}
}

static inline btbool	srvUseWriteThrough(server_t *server, btint oplockLevel)
{
	switch(oplockLevel){
		default:
		case CIFS_OPLOCK_NONE:		return server->writeThroughUnlocked;
		case CIFS_OPLOCK_L2:
		case CIFS_OPLOCK_EXCLUSIVE:
		case CIFS_OPLOCK_BATCH:		return server->writeThroughLocked;
	}
}

static inline bti32	srvGmtOffset(server_t *this)
{
	return this->serverGmtOffset;
}

static inline int	srvUseRequest(server_t *this, int request)
{
	return (this->cmdProfile & (1 << request)) != 0;
}

static inline btbool	srvCanEncrypt(server_t *this)
{
	return this->canEncrypt;
}

static inline btbool	srvUsesUnicode(server_t *srv)
{
	return srv->useUnicode;
}

static inline int	srvDialect(server_t *this)
{
	return this->dialect;
}

static inline btu8	*srvChallenge(server_t *this)
{
	return this->encryptKey;
}

static inline btu8	*srvName(server_t *this)
{
	return this->name;
}

static inline btbool	srvFakeMountpointDate(server_t *this)
{
	return this->fakeMountpointDate;
}

static inline btbool	srvCapitalizeShares(server_t *this)
{
	return this->capitalizeShares;
}

static inline btbool	srvCloseForSetattr(server_t *this)
{
	return this->closeForSetattr;
}

static inline btsize	srvWindowSize(server_t *server)
{
	return server->windowSize;
}

static inline btbool	srvSslRequireEncryption(server_t *server)
{
	return server->sslRequireEncryption;
}

static inline btu8	srvSslVersion(server_t *server)
{
	return server->sslVersion;
}

static inline btbool	srvSslRequireServerCert(server_t *server)
{
	return server->sslRequireServerCert;
}

static inline btu8	*srvSslCACertDir(server_t *server)
{
	if(server->sslCACertDir != NULL && server->sslCACertDir[0] == 0)
		return NULL;
	return server->sslCACertDir;
}

static inline btu8	*srvSslCACertFile(server_t *server)
{
	if(server->sslCACertFile != NULL && server->sslCACertFile[0] == 0)
		return NULL;
	return server->sslCACertFile;
}

static inline btu8	*srvSslKeyFile(server_t *server)
{
	if(server->sslKeyFile != NULL && server->sslKeyFile[0] == 0)
		return NULL;
	return server->sslKeyFile;
}

static inline btu8	*srvSslCertFile(server_t *server)
{
	if(server->sslCertFile != NULL && server->sslCertFile[0] == 0)
		return NULL;
	return server->sslCertFile;
}

static inline btu8	*srvSslCiphers(server_t *server)
{
	return server->sslCiphers;
}

static inline btbool	srvSslCompatibility(server_t *server)
{
	return server->sslCompatibility;
}

static inline void	srvSetSslRequireEncryption(server_t *server, btbool f)
{
	server->sslRequireEncryption = f;
}

static inline void	srvSetSslRequireServerCert(server_t *server, btbool f)
{
	server->sslRequireServerCert = f;
}

static inline btbool	srvIsUserlevelSecurity(server_t *server)
{
	return server->securityIsUserLevel;
}

static inline struct share	*srvIpcShare(server_t *server)
{
	return server->ipcShare;
}

static inline struct cifsUser	*srvNullSession(server_t *server)
{
	return server->nullSession;
}

static inline btu16   srvLinkMask(server_t *server)
{
	return server->linkModeMask;
}

static inline btu16   srvLinkMode(server_t *server)
{
	return server->linkMode;
}

static inline btbool srvLinksAreUnicode(server_t *server)
{
	return server->linksAreUnicode;
}

static inline bttime	srvDirCacheTime(server_t *server)
{
	return server->dirCacheTime;
}

static inline btbool	srvIsCaseSensitive(server_t *server)
{
	return server->caseSensitive;
}

static inline btint	srvMaxCachedFiles(server_t *server)
{
	return server->maxCachedFiles;
}

static inline bttime	srvAttributesCacheTime(server_t *server)
{
	return server->attrCacheTime;
}

/* ------------------------------------------------------------------------- */

#endif	/* __server_H_INCLUDED__ */
