/*
 * Name: share.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-29
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: share.c,v 1.3 1999/07/22 17:54:09 cs Exp $
 */


#include "stdheaders.h"
#include "unicode.h"
#include "exception.h"
#include "memalloc.h"
#include "file.h"
#include "share.h"
#include "user.h"
#include "blockpool.h"
#include "smb.h"
#include "cifs.h"
#include "server.h"
#include "smbnumbers.h"
#include "logging.h"
#include "process.h"
#include "dircache.h"
#include "datacache.h"
#include "pwhash.h"
#include "pslayer.h"
#include "notification.h"
#include "backend.h"
#include "errcodes.h"
#include "util.h"

/* ------------------------------------------------------------------------- */

typedef struct shrCtx{
    PROC_VARSET;
    share_t		*share;
    btu16		uid;
	cifsUser_t	*pwUser;
	array_t		*openFiles;
	btint		index;
}shrCtx_t;

static blockpool_t	*shrPool;

static void	freePoolBlock(void *block)
{
    poolReturnBlock(shrPool, block);
}

/* ------------------------------------------------------------------------- */

static btint	shareCompare(void *o1, void *o2)
{
share_t	*this = o1, *other = o2;

	return uniStrCmp(this->path, other->path, 0);
}

static btu32	shareHashVal(void *obj)
{
share_t	*this = obj;

	return uniStrHash(this->path, 0);
}

static collectionObjectI_t	sharesByPathInterface = {
	shareCompare,
	shareHashVal,
	NULL,
	NULL,
};

static hashtable_t	*allShares = NULL;

/* ------------------------------------------------------------------------- */

static btint	fileFidCompare(void *obj, void *obj2)
{
    return fileFid(obj) - fileFid(obj2);
}

static btu32	fileFidHashvalue(void *obj)
{
    return fileFid(obj);
}

static collectionObjectI_t	fileByFidInterface = {
    fileFidCompare,
    fileFidHashvalue,
    NULL,	/* no retain/release by collections */
    NULL
};

/* ------------------------------------------------------------------------- */

share_t	*shrNew(btuni *path, struct server *server, btbool isIpcShare)
{
share_t		*volatile this = NULL;

    if(shrPool == NULL){
        shrPool = poolNew(sizeof(shrCtx_t), sizeof(shrCtx_t));
    }
EXC_TRY
    this = memCalloc(sizeof(share_t));
	this->refCount = 1;
    this->path = uniStrDup(path);
	if(srvCapitalizeShares(server))
		uniStrToupper(this->path);
    this->connectSignal = procCondNew();
    this->cifsTid = -1;
    this->service = isIpcShare ? "IPC" : "A:";
    this->openFiles = htabNew(&fileByFidInterface, 32, 75);
    this->lookupCache = lruNew(0);
	this->openNotify = ncNew(8);
	this->invalidateNotify = ncNew(8);
    this->root = fileNew(this, NULL);
	this->server = server;
	srvAddShare(server, this);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "share: allocated new object at 0x%08x\n", (int)this);
EXC_CATCH
	shrFree(this);
	EXC_RETHROW();
EXC_ENDTRY
    return this;
}

share_t	*shrForPath(btuni *path)
{
btuni		service[256], srvName[128];
btu8		srvAscii[128];
server_t	*volatile server;
share_t		*volatile share;

	if(allShares == NULL)
		allShares = htabNew(&sharesByPathInterface, 16, 75);
	utilParseServiceUni(path, srvName, ELEMENTS(srvName), service, ELEMENTS(service));
	path = service;
	share = htabFindObjectEqualTo(allShares, &path);
	if(share != NULL){
		shrRetain(share);
		return share;
	}
	convUniToLocalAscii(srvName, srvAscii, sizeof(srvAscii));
	server = srvWithName(srvAscii, 0);
EXC_TRY
	share = shrNew(service, server, 0);
EXC_CATCH
	srvRelease(server);
	EXC_RETHROW();
EXC_ENDTRY
	htabAddObject(allShares, share);
	return share;
}

/* ------------------------------------------------------------------------- */

void	shrFree(share_t *this)
{
	if(this != NULL){
		if(allShares != NULL)
			htabRemoveObjectEqualTo(allShares, this);
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "share: freeing object at 0x%08x\n", (int)this);
        if(this->root != NULL){
            fileInvalidate(this->root);
			fileRelease(this->root);
			this->root = NULL;
		}
		if(this->server != NULL){
			srvRemoveShare(this->server, this);
			srvRelease(this->server);
			this->server = NULL;
		}
        if(this->path != NULL)
            memFree(this->path);
        if(this->connectSignal != NULL)
            procCondFree(this->connectSignal);
        if(this->openFiles != NULL)
            htabFree(this->openFiles);
        if(this->lookupCache != NULL)
            lruFree(this->lookupCache);
		if(this->openNotify != NULL)
			ncRelease(this->openNotify);
		if(this->invalidateNotify != NULL)
			ncRelease(this->invalidateNotify);
		cacheShareFreed(this);
		memset(this, 0, sizeof(*this));
        memFree(this);
    }
}

/* ------------------------------------------------------------------------- */

void	shrOplockBreak(share_t *this, btu16 fid, int mode)
{
cifsFile_t	f, *file;

	f.fid = fid;
	file = htabFindObjectEqualTo(this->openFiles, &f);
	if(file == NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "share: oplock break for unknown file %d\n", fid);
	}else{
		fileOplockBreak(file, mode);
	}
}

/* ------------------------------------------------------------------------- */

static void	shrConnectProc(void *obj, void *arg)
{
shrCtx_t		*this = obj;
share_t			*share = this->share;
array_t			*allUsers;
int				i, l;
cifsUser_t		*u = NULL;
beVfsRes_t		res;
smbRequest_t	*smb = arg;
cifsRes_t		*rval = arg;
btu8			passwd[64], pwhash[21];
int				pwlen;

    PROC_BEGIN;
	shrRetain(share);
	while(share->cstatus == PSL_DISCONNECTING || share->cstatus == PSL_CONNECTING){
		PROC_WAITFOR_X(1, procCondWait(share->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(share->cstatus == PSL_CONNECTED){
		res.err.code = res.err.class = 0;
		shrRelease(share);
		PROC_RETURN(&res);
	}
    share->cstatus = PSL_CONNECTING;
	if(this->pwUser == NULL){	/* no explicit password user given, take first one we find */
		allUsers = srvAllUsers(share->server);
		l = arrObjectCount(allUsers);
		for(i=0;i<l;i++){
			u = arrObjectAtIndex(allUsers, i);
			if(userConnectionStatus(u) == PSL_CONNECTED)
				break;
		}
		arrFree(allUsers);
		this->pwUser = u;
	}
	if(this->pwUser == NULL || userConnectionStatus(this->pwUser) != PSL_CONNECTED){	/* there's no user set up */
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCconnlost;
		share->cstatus = PSL_NOTCONNECTED;
        goto doReturn;
	}
	this->uid = userSmbUid(this->pwUser);
	if(!srvIsUserlevelSecurity(share->server)){	/* share level security */
		if(!srvCanEncrypt(share->server) && this->pwUser->plainPasswd == NULL){	/* we can't do with this server */
			res.err.class = ERRCLS_CIFS_LOCAL;
			res.err.code = ERR_CIFS_LOCpwencrypt;
			share->cstatus = PSL_NOTCONNECTED;
			goto doReturn;
		}
		PROC_WAITFOR_X(2, srvWantRequest(share->server, this));
	
		if(srvCanEncrypt(share->server)){	/* if encryption is possible, use it */
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "share: tree connect with encrypted password\n");
			memset(pwhash, 0, sizeof(pwhash));
			if(srvDialect(share->server) >= CIFS_DIALECT_NTLM){	/* use MD4 hash */
				memcpy(pwhash, this->pwUser->pwHashNt, 16);
			}else{
				memcpy(pwhash, this->pwUser->pwHashPreNt, 16);
			}
			pwhResponse(pwhash, srvChallenge(share->server), passwd);
			pwlen = 24;
		}else{
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "share: tree connect with plain text password\n");
			/* password must be in plain text here because allowUnencrypted is false for hashed passwords */
			convUniToServerAscii(this->pwUser->plainPasswd, passwd, 64);
			pwlen = strlen(passwd) + 1;	/* length includes terminating 0 */
		}
	}else{	/* user level security, password already given */
		PROC_WAITFOR_X(3, srvWantRequest(share->server, this));
		passwd[0] = 0;
		pwlen = 1;
	}
    PROC_WAITFOR_X(4,
		cifsTreeConnect(smb, this, this->uid, passwd, pwlen, share->path, srvUseRequest(share->server, SRV_USE_TCON_ANDX), share->service);
	)
    if(rval->err.class != 0){
        share->cifsTid = -1;
        share->cstatus = PSL_NOTCONNECTED;
        res.err.class = rval->err.class;
        res.err.code = rval->err.code;
        goto doReturn;
    }
    share->cifsTid = rval->tid;
	cacheNeedPagesize(srvCachePageSize(share->server), share);
    share->cstatus = PSL_CONNECTED;
	res.err.class = res.err.code = 0;
doReturn:
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(share->connectSignal, NULL);
		shrRelease(share);
	);
    PROC_END;
}

void	shrConnect(share_t *this, struct cifsUser *pwUser, void *caller)
{
beVfsRes_t	res;
shrCtx_t	*ctx;

    res.err.class = res.err.code = 0;
    if(this->cstatus == PSL_CONNECTED){
        PROC_IMMEDIATE_RET(caller, &res);
    }
	if(this->server != NULL && srvCapitalizeShares(this->server))
		uniStrToupper(this->path);
    PROC_INIT(ctx, shrConnectProc, caller, shrPool, freePoolBlock);
    ctx->share = this;
	ctx->pwUser = pwUser;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	shrDisconnectProc(void *obj, void *arg)
{
shrCtx_t		*this = obj;
beVfsRes_t		res;
smbRequest_t	*smb = arg;
cifsRes_t		*rval = arg;
cifsFile_t		*p;
btint			i, cnt;

    PROC_BEGIN;
	shrRetain(this->share);
	while(this->share->cstatus == PSL_DISCONNECTING || this->share->cstatus == PSL_CONNECTING){
		PROC_WAITFOR_X(1, procCondWait(this->share->connectSignal, this));
	}	/* wait for connect status to stabilize */
	if(this->share->cstatus == PSL_NOTCONNECTED){
		res.err.code = res.err.class = 0;
		shrRelease(this->share);
		PROC_RETURN(&res);
	}
    this->share->cstatus = PSL_DISCONNECTING;
	this->openFiles = htabAllObjects(this->share->openFiles);
	cnt = arrObjectCount(this->openFiles);
/* we must retain all files because we'll sleep and the files might be freed
 * in the meantime.
 */
	for(i=0;i<cnt;i++)
		fileRetain(arrObjectAtIndex(this->openFiles, i));
	for(this->index=0; this->index<arrObjectCount(this->openFiles); this->index++){
		p = arrObjectAtIndex(this->openFiles, this->index);
		PROC_WAITFOR_X(2, fileSetOpenMode(p, 0, FILE_CLOSED, FILE_OPEN_RW, 0, FILE_OP_NONE, NULL, 0, 0, 0, 0, this));
	}
	cnt = arrObjectCount(this->openFiles);
	for(i=0;i<cnt;i++)
		fileRelease(arrObjectAtIndex(this->openFiles, i));
	arrFree(this->openFiles);
	PROC_WAITFOR_X(3, srvWantRequest(this->share->server, this));
    PROC_WAITFOR_X(4, cifsTreeDisconnect(smb, this, 0, this->share->cifsTid));
/* we might want to invalidate all but the root file here, but why not keep
 * cached files? If the share is freed, all files are invalidated anyway.
 */
    this->share->cstatus = PSL_NOTCONNECTED;
	res.err = rval->err;
    PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(this->share->connectSignal, NULL);
		shrRelease(this->share);
	);
    PROC_END;
}

void	shrDisconnect(share_t *this, void *caller)
{
beVfsRes_t	res;
shrCtx_t	*ctx;

    res.err.class = res.err.code = 0;
    if(this->cstatus == PSL_NOTCONNECTED){
        PROC_IMMEDIATE_RET(caller, &res);
    }
    PROC_INIT(ctx, shrDisconnectProc, caller, shrPool, freePoolBlock);
    ctx->share = this;
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	shrConnectionLost(share_t *this)
{
array_t	*openFiles;
btint	i, cnt;

    this->cstatus = PSL_NOTCONNECTED;
    procCondBroadcast(this->connectSignal, NULL);
	openFiles = htabAllObjects(this->openFiles);
	cnt = arrObjectCount(openFiles);
	for(i=0;i<cnt;i++){
		fileConnectionLost(arrObjectAtIndex(openFiles, i));
	}
	arrFree(openFiles);
/* we might want to invalidate all but the root file here, but why not keep
 * cached files? If the share is freed, all files are invalidated anyway.
 */
}

/* ------------------------------------------------------------------------- */

static void	shrGetAllocationInfoProc(void *obj, void *arg)
{
shrCtx_t				*this = obj;
smbRequest_t			*smb = arg;
beVfsRes_t				res;
server_t				*server = this->share->server;
cifsResQueryFsInfo_t	*rval = arg;

    PROC_BEGIN
    PROC_WAITFOR_X(1, srvWantRequest(server, this));
    if(srvUseRequest(server, SRV_USE_FSINFO_LM)){
        PROC_WAITFOR_X(2, cifsQueryFsInfo(smb, this, this->uid, this->share->cifsTid));
    }else{
        PROC_WAITFOR_X(3, cifsQueryInfoDisk(smb, this, this->uid, this->share->cifsTid));
    }
    if(rval->err.class != 0){
        res.err.class = rval->err.class;
        res.err.code = rval->err.code;
        PROC_RETURN(&res);
    }
    res.err.class = res.err.code = 0;
	res.bytesPerAU = rval->sectorsPerAU * rval->bytesPerSector;
	res.totalAU = rval->totalAU;
	res.availableAU = rval->availableAU;
	res.timeGranularity.seconds = 2;
	res.timeGranularity.nseconds = 0;
	res.features = BE_VFS_IS_CASE_PRESERVING | BE_VFS_CAN_SETTIMES;
	if(srvLinkMask(server) != 0)
		res.features |= BE_VFS_CAN_SYMLINK;
	res.maxFilenameLen = 255;
    PROC_RETURN(&res);
    PROC_END
}

void	shrGetAllocationInfo(share_t *this, struct cifsUser *user, void *caller)
{
shrCtx_t	*ctx;

    PROC_INIT(ctx, shrGetAllocationInfoProc, caller, shrPool, freePoolBlock);
    ctx->share = this;
    ctx->uid = userSmbUid(user);
    PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

void	shrAddToLookupCache(share_t *shr, struct cifsFile *file)
{
btint		maxCached;
cifsFile_t	*f;

	lruAddObject(shr->lookupCache, file);
	maxCached = srvMaxCachedFiles(shr->server);
	while(lruObjectCount(shr->lookupCache) > maxCached){
		f = lruOldestRelativeTo(shr->lookupCache, NULL);
		lruRemoveObject(shr->lookupCache, f);
		fileRelease(f);
	}
}

/* ------------------------------------------------------------------------- */

void	shrRetain(void *obj)
{
share_t	*this = obj;

    this->refCount++;
}

void	shrRelease(void *obj)
{
share_t	*this = obj;

	if(this->refCount == 1 && this->server != NULL)	/* release would free the object */
		shrDisconnect(this, NULL);	/* don't wait for result */
	this->refCount--;
    if(this->refCount == 0){
        shrFree(this);
    }else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */
