/*
 * Name: share.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-24
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: share.h,v 1.2 1999/07/22 17:54:10 cs Exp $
 */

#ifndef __share_H_INCLUDED__
#define __share_H_INCLUDED__

#include "basictypes.h"
#include "lru.h"

/*
General Description:
This module implements the "share" class. It represents a share mounted
from a remote server. It contains hashtables of all known files (by nfs
file handle), all open files (by file ID) and all local mountpoints of
this share. Furthermore it contains LRU-Tables of all known files, the open
files and the data and directory caches. Each share is related to a server.
*/

/* ------------------------------------------------------------------------- */

struct server;
struct cifsFile;
struct cifsUser;

typedef struct share{
    btuni				*path;			/* UNC share rep */
    struct server		*server;		/* server responsible for this share */
    int					refCount;		/* will be part of collection */
	btu8				*service;		/* either "A:" for disks or "IPC" */
    btu16				cifsTid;		/* tree ID at the server */
	struct hashtable	*openFiles;		/* files referenced by file ID */
    lru_t				*lookupCache;	/* lru-list of pre-looked up files */
    struct cifsFile		*root;			/* root directory */
    int					cstatus;		/* connection status */
    struct procCond		*connectSignal;	/* only one connect at a time */
	struct notificationCenter	*openNotify;
	struct notificationCenter	*invalidateNotify;
}share_t;

/* ------------------------------------------------------------------------- */

share_t	*shrForPath(btuni *path);/* returns retained object */
share_t	*shrNew(btuni *path, struct server *server, btbool isIpcShare); /* expects that server is retained, creates with refcount=1 */
void	shrFree(share_t *this);
void	shrRetain(void *obj);
void	shrRelease(void *obj);
void	shrGetAllocationInfo(share_t *this, struct cifsUser *user, void *caller);
void	shrOplockBreak(share_t *this, btu16 fid, int mode);
void	shrConnect(share_t *this, struct cifsUser *pwUser, void *caller);
void	shrDisconnect(share_t *this, void *caller);
void	shrConnectionLost(share_t *this);
void	shrAddToLookupCache(share_t *shr, struct cifsFile *file); /* file is retained, will be released if it falls out of cache */

/* ------------------------------------------------------------------------- */

static inline struct notificationCenter	*shrOpenNc(share_t *shr)
{
	return shr->openNotify;
}

static inline struct notificationCenter	*shrInvalidateNc(share_t *shr)
{
	return shr->invalidateNotify;
}

static inline int	shrConnectionStatus(share_t *this)
{
    return this->cstatus;
}

static inline struct server	*shrServer(share_t *this)
{
    return this->server;
}

static inline btu16	shrTid(share_t *this)
{
    return this->cifsTid;
}

static inline btuni	*shrPath(share_t *this)
{
	return this->path;
}

static inline struct cifsFile	*shrRootFile(share_t *this)
{
	return this->root;
}

/* ------------------------------------------------------------------------- */

/* The following are macros in order to avoid including various headers here */

#define	shrAddOpenFile(shr, file)	(htabAddObject((shr)->openFiles, (file)))

#define	shrRemoveOpenFile(shr, file)	(htabRemoveObjectEqualTo((shr)->openFiles, (file)))

#define	shrAddFile(this, file)					\
	do{											\
		if(fileOpenMode(file) != FILE_CLOSED){	\
			shrAddOpenFile((this), file);		\
		}										\
	}while(0)

#define	shrRemoveFile(shr, file)							\
	do{														\
		htabRemoveObjectEqualTo((shr)->openFiles, (file));	\
	}while(0)

#define	shrRemoveFromLookupCache(shr, file)					\
	do{														\
		struct cifsFile	*__file = (file);					\
		if(__file->lookupLru.prevnext != NULL){	/* file is in LRU cache */\
			share_t			*__shr = (shr);					\
			lruRemoveObject(__shr->lookupCache, __file);	\
			fileRelease(__file);							\
		}													\
	}while(0)

#define	shrFileBecameInvalid(shr, file)							\
	do{															\
		share_t			*__shr1 = (shr);						\
		struct cifsFile	*__file1 = (file);						\
		shrRemoveFromLookupCache(__shr1, __file1);				\
		ncPostNotification((__shr1)->invalidateNotify, __file1);\
	}while(0)

#define	shrFileOpenModeChanged(shr, file)	(ncPostNotification((shr)->openNotify, (file)))

/* ------------------------------------------------------------------------- */

#endif	/* __share_H_INCLUDED__ */
