/*
 * Name: smb.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-18
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: smb.c,v 1.6 1999/08/01 12:57:15 cs Exp $
 */


#include "stdheaders.h"
#include "exception.h"
#include "memalloc.h"
#include "logging.h"
#include "smb.h"
#include "server.h"
#include "smbnumbers.h"
#include "statemachine.h"
#include "unicode.h"
#include "cifsconst.h"
#include "errcodes.h"

/*#define	REPORT_DURATION*/

/* ------------------------------------------------------------------------- */

EXC_DEFINE(smbExcOverflow);

/* ------------------------------------------------------------------------- */

static void	smbTimeout(void *ref, void *arg);

/* ------------------------------------------------------------------------- */

smbRequest_t	*smbRequestNew(struct server *server)
{
smbRequest_t	*this = memCalloc(sizeof(smbRequest_t));

	this->requestState = SMBS_IDLE;
    this->server = server;
	return this;
}

void	smbRequestFree(smbRequest_t *this)
{
	if(this != NULL){
		elpCancelCallAt(smbTimeout, this);
		if(this->rxBufferAllocated > 0)
			memFree(this->rxBuffer);
		memset(this, 0, sizeof(*this));
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

#define	EVNT_TX_START				0
#define	EVNT_SMB_SENT				1
#define	EVNT_RX_RESPONSE			2
#define	EVNT_RX_TIMEOUT				3
#define	EVNT_RX_INTERIM				4

#define	SM_STATE	this->requestState

static void	smbStateMachine(smbRequest_t *this, int event)
{
btbool	stateChanged = 0;

#ifdef REPORT_DURATION
	switch(event){
	case EVNT_TX_START:
		this->startTime = elpNow();
		break;
	case EVNT_SMB_SENT:
		{
			bttime			duration;
			static bttime	maxDuration;
			duration = elpNow() - this->startTime;
			if(duration > maxDuration){
				maxDuration = duration;
				if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
					logPrintf(LOGM_INFO, "smb: maximum request delay: %d ms\n", maxDuration);
			}
		}
		break;
	case EVNT_RX_RESPONSE:
		{
			bttime			duration;
			static bttime	maxDuration;
			duration = elpNow() - this->startTime;
			if(duration > maxDuration){
				maxDuration = duration;
				if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
					logPrintf(LOGM_INFO, "smb: maximum request duration: %d ms\n", maxDuration);
			}
		}
		break;
	}
#endif

    if(event == EVNT_TX_START && this->requestState != SMBS_IDLE){
        logPrintf(LOGM_ERROR, "smbs: start event in state %d!!!\n", this->requestState);
        this->requestState = SMBS_IDLE;	/* bruteforce idle */
    }
	SM_BEGIN_STATEMACHINE(
		logPrintf(LOGM_ERROR, "smb: reached illegal state %d\n", SM_STATE);
		SM_STATE = 0;
		elpCancelCallAt(smbTimeout, this);
		this->txSmb = this->rxInitialized = 0;
		this->inSmb = this->inParameters = this->inData = 0;
		srvRequestStatusChanged(this->server, this, this->requestState);
		return;
	)
	SM_DEFINE_STATE(SMBS_IDLE)
		SM_ENTRY_CODE
			elpCancelCallAt(smbTimeout, this);
			this->txSmb = this->rxInitialized = 0;
			this->inSmb = this->inParameters = this->inData = 0;
		SM_EXIT_CODE
		SM_TRANSITIONS
			if(event == EVNT_TX_START){
				SM_GOTO(SMBS_WAIT_TX1);
			}
	SM_DEFINE_STATE(SMBS_WAIT_TX1)
		SM_ENTRY_CODE
            srvWantSend(this->server, this);
		SM_EXIT_CODE
		SM_TRANSITIONS
			if(event == EVNT_SMB_SENT){
				if(this->txSmb >= this->currentSmb){
					if(this->callback == NULL){
						SM_GOTO(SMBS_IDLE);
					}else{
						SM_GOTO(SMBS_WAIT_RESPONSE);
					}
				}else{
					SM_GOTO(SMBS_WAIT_INTERIM);
				}
			}else if(event == EVNT_RX_RESPONSE){	/* connection lost */
				SM_GOTO(SMBS_IDLE);
			}
	SM_DEFINE_STATE(SMBS_WAIT_INTERIM)
		SM_ENTRY_CODE
			elpCallAtCancelPrev(elpNow() + srvRequestTimeout(this->server), smbTimeout, this);
		SM_EXIT_CODE
		SM_TRANSITIONS
			if(event == EVNT_RX_TIMEOUT){
				SM_GOTO(SMBS_IDLE);
			}else if(event == EVNT_RX_INTERIM){
				SM_GOTO(SMBS_WAIT_TXN);
			}else if(event == EVNT_RX_RESPONSE){	/* connection lost */
				SM_GOTO(SMBS_IDLE);
			}
	SM_DEFINE_STATE(SMBS_WAIT_TXN)
		SM_ENTRY_CODE
            srvWantSend(this->server, this);
		SM_EXIT_CODE
		SM_TRANSITIONS
			if(event == EVNT_SMB_SENT){
				if(this->txSmb >= this->currentSmb){
					if(this->callback == NULL){
						SM_GOTO(SMBS_IDLE);
					}else{
						SM_GOTO(SMBS_WAIT_RESPONSE);
					}
				}
			}else if(event == EVNT_RX_RESPONSE){	/* connection lost */
				SM_GOTO(SMBS_IDLE);
			}
	SM_DEFINE_STATE(SMBS_WAIT_RESPONSE)
		SM_ENTRY_CODE
			elpCallAtCancelPrev(elpNow() + srvRequestTimeout(this->server), smbTimeout, this);
		SM_EXIT_CODE
		SM_TRANSITIONS
			if(event == EVNT_RX_TIMEOUT){
				SM_GOTO(SMBS_IDLE);
			}else if(event == EVNT_RX_RESPONSE){
				SM_GOTO(SMBS_IDLE);
			}
	SM_END_STATEMACHINE(stateChanged);
    if(stateChanged){
		srvRequestStatusChanged(this->server, this, this->requestState);
    }
}

/* ------------------------------------------------------------------------- */

static void	smbTimeout(void *ref, void *arg)
{
smbRequest_t	*this = ref;
    
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "smb: timeout for mid=%d, disconnecting\n", this->mid);
	srvRetain(this->server);	/* avoid freeing from callback */
	srvConnectionLost(this->server);
	srvRelease(this->server);
#if 0	/* the following should be done from smbConnectionLost() */
    this->rxHeader = NULL;		/* inidcation for local timeout */
	srvRetain(this->server);	/* avoid freeing from callback */
	this->callback(this->callbackRef, this);
	smbStateMachine(ref, EVNT_RX_TIMEOUT);
	srvRelease(this->server);
#endif
}

nbmsg_t	*smbTxBlock(smbRequest_t *this)
{
nbmsg_t	*msg;

    if(this->requestState != SMBS_WAIT_TX1 && this->requestState != SMBS_WAIT_TXN)
        return NULL;
    if(this->txSmb >= this->currentSmb){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "smb: txSmb (%d) out of range (%d)\n", this->txSmb, this->currentSmb);
        return NULL;
    }
	msg = this->smbs[this->txSmb].block;
	this->smbs[this->txSmb++].block = NULL;
    smbStateMachine(this, EVNT_SMB_SENT);
    return msg;	/* receiver must release msg */
}

/* ------------------------------------------------------------------------- */
/* --------------------------- transmitting data --------------------------- */
/* ------------------------------------------------------------------------- */

void	smbAddVariable(smbRequest_t *this, smbVar_t *p, int size, int scaling)
{
	_smbMakeSpace(this, size);
	p->position = msgSize(this->currentMsg);
	p->size = size;
	p->scaling = scaling;
	msgAddSpace(this->currentMsg, size);
}

static inline void	invalidVariable(smbVar_t *p)
{
	p->position = -1;
}

void	smbWriteVariable(smbRequest_t *this, smbVar_t *p, btu32 value)
{
int		i;

	if(p->position >= 0){
		value >>= p->scaling;
		for(i=0;i<p->size;i++){
			msgSetByte(this->currentMsg, p->position + i, value);
			value >>= 8;
		}
	}
}

static void	writeSmbHeader(smbRequest_t *this, btu8 cmd)
{
	smbAddBlock(this, "\377SMB", 4);
	smbAddByte(this, cmd);
	smbAddLong(this, 0);
	smbAddByte(this, this->server->txFlags);
	if(this->server->useUnicode)
		smbAddWord(this, this->server->txFlags2 | CIFS_FLAG2_UNICODE);
	else
		smbAddWord(this, this->server->txFlags2);
	smbAddWord(this, this->pid >> 16);
	smbAddLong(this, 0);
	smbAddLong(this, 0);
	smbAddWord(this, 0);
	smbAddWord(this, this->tid);
	smbAddWord(this, this->pid);
	smbAddWord(this, this->uid);
	smbAddWord(this, this->mid);
}

/* ------------------------------------------------------------------------- */

void	smbAddUniFromAscii(smbRequest_t *this, btu8 *s)
{
btuni	c;

	smbAddAlignment(this, 2);	/* align to word boundary */
	_smbMakeSpace(this, 2 * (strlen(s) + 1));
	while(*s != 0){
		c = convCharLocalAsciiToUni(&s);
		msgAddByteUnchecked(this->currentMsg, c);
		msgAddByteUnchecked(this->currentMsg, c >> 8);
	}
	msgAddByteUnchecked(this->currentMsg, 0);
	msgAddByteUnchecked(this->currentMsg, 0);
}

void	smbAddUniFromUni(smbRequest_t *this, btuni *s)
{
	smbAddAlignment(this, 2);	/* align to word boundary */
	do{
		smbAddWord(this, *s);
	}while(*s++ != 0);
}

void	smbAddAsciiFromUni(smbRequest_t *this, btuni *s)
{
btu8	*p;

	while(*s != 0){
		_smbMakeSpace(this, UNI_MAX_ASCIIMAP);
		p = this->currentMsg->data + this->currentMsg->size;
		convCharUniToServerAscii(*s++, &p);
		this->currentMsg->size = p - this->currentMsg->data;
		/* this breaks encapsulation, but it should be faster */
	}
	smbAddByte(this, 0);
}

void	smbAddStringFromAscii(smbRequest_t *this, btu8 *s)
{
	if(this->server->useUnicode){
		smbAddUniFromAscii(this, s);
	}else{
    	smbAddAsciiFromAscii(this, s);
	}
}

void	smbAddStringFromUni(smbRequest_t *this, btuni *s)
{
	if(this->server->useUnicode){
		smbAddUniFromUni(this, s);
	}else{
		smbAddAsciiFromUni(this, s);
	}
}

/* ------------------------------------------------------------------------- */

static void	endSmb(smbRequest_t *this);
static void	endParameters(smbRequest_t *this);
static void	endData(smbRequest_t *this);

/* ------------------------------------------------------------------------- */

void	smbStartCommand(smbRequest_t *this, btu16 uid, btu32 pid, btu16 tid, smbCallback_t callback, void *ref)
{
int		i;

	for(i=0;i<this->currentSmb;i++){
		if(this->smbs[i].block != NULL){
			msgRelease(this->smbs[i].block);
			this->smbs[i].block = NULL;
		}
	}
	this->currentSmb = 0;
	this->parameterByteCount = 0;
	this->dataByteCount = 0;
	this->inSmb = this->inParameters = this->inData = 0;
    this->uid = uid;
    this->pid = pid;
    this->tid = tid;
    this->mid = srvGetMid(this->server);
    this->callback = callback;
    this->callbackRef = ref;
	if(LOG_MODE_IS_ACTIVE(LOGM_CIFSTRACE))
		logPrintf(LOGM_CIFSTRACE, "smb: startCommand: %d\n", this->mid);
}

void	smbEndCommand(smbRequest_t *this)
{
int		i;

	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
	for(i=0;i<this->currentSmb;i++){
		smbWriteVariable(this, &this->smbs[i].totalParameters, this->parameterByteCount);
		smbWriteVariable(this, &this->smbs[i].totalData, this->dataByteCount);
	}
    this->txSmb = 0;
    smbStateMachine(this, EVNT_TX_START);
}

/* ------------------------------------------------------------------------- */

void	smbStartNormalSmb(smbRequest_t *this, btu8 cmd)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
    this->inSmb = 1;
	this->currentMsg = this->smbs[this->currentSmb].block = srvGetMessageBuffer(this->server);
	smbAddAlignment(this, 4);
	writeSmbHeader(this, cmd);
	invalidVariable(&this->smbs[this->currentSmb].totalParameters);
	invalidVariable(&this->smbs[this->currentSmb].totalData);
	invalidVariable(&this->parameterCount);
	invalidVariable(&this->parameterOffset);
	invalidVariable(&this->dataCount);
	invalidVariable(&this->dataOffset);
	invalidVariable(&this->metaDataCount);
	invalidVariable(&this->andXOffset);
	this->prefixParameters = this->prefixData = 1;
	this->alignParameters = this->alignData = 0;
}

void  smbStartTransSmb(smbRequest_t *this, btu8 *tactName, btu16 maxRxParam, btu16 maxRxData, btu32 timeout)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
	this->inSmb = 1;
	this->currentMsg = this->smbs[this->currentSmb].block = srvGetMessageBuffer(this->server);
	smbAddAlignment(this, 4);
	writeSmbHeader(this, SMB_COM_TRANSACTION);
	smbAddByte(this, 14);	/* number of parameter words */
	smbAddVariable(this, &this->smbs[this->currentSmb].totalParameters, 2, 0);
	smbAddVariable(this, &this->smbs[this->currentSmb].totalData, 2, 0);
	smbAddWord(this, maxRxParam);
	smbAddWord(this, maxRxData);
	smbAddWord(this, 0);	/* max rx setup + reserved */
	smbAddWord(this, 0);	/* flags */
	smbAddLong(this, timeout);
	smbAddWord(this, 0);	/* reserved 2 */
	smbAddVariable(this, &this->parameterCount, 2, 0);
	smbAddVariable(this, &this->parameterOffset, 2, 0);
	smbAddVariable(this, &this->dataCount, 2, 0);
	smbAddVariable(this, &this->dataOffset, 2, 0);
	smbAddWord(this, 0);	/* setup count + reserved 3 */
	smbAddVariable(this, &this->metaDataCount, 2, 0);
	smbAddStringFromAscii(this, tactName);
	this->metaDataBase = msgSize(this->currentMsg);
	this->prefixParameters = this->prefixData = 0;
	this->alignParameters = this->alignData = 4;
	invalidVariable(&this->andXOffset);
}


void	smbStartTrans2Smb(smbRequest_t *this, btu16 cmd, btu16 maxRxParam, btu16 maxRxData, btu32 timeout)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
    this->inSmb = 1;
	this->currentMsg = this->smbs[this->currentSmb].block = srvGetMessageBuffer(this->server);
	smbAddAlignment(this, 4);
	writeSmbHeader(this, SMB_COM_TRANSACTION2);
	smbAddByte(this, 15);	/* number of parameter words */
	smbAddVariable(this, &this->smbs[this->currentSmb].totalParameters, 2, 0);
	smbAddVariable(this, &this->smbs[this->currentSmb].totalData, 2, 0);
	smbAddWord(this, maxRxParam);
	smbAddWord(this, maxRxData);
	smbAddWord(this, 0);	/* max rx setup + reserved */
	smbAddWord(this, 0);	/* flags */
	smbAddLong(this, timeout);
	smbAddWord(this, 0);	/* reserved 2 */
	smbAddVariable(this, &this->parameterCount, 2, 0);
	smbAddVariable(this, &this->parameterOffset, 2, 0);
	smbAddVariable(this, &this->dataCount, 2, 0);
	smbAddVariable(this, &this->dataOffset, 2, 0);
	smbAddWord(this, 1);	/* setup count + reserved 3 */
	smbAddWord(this, cmd);	/* setup data */
	smbAddVariable(this, &this->metaDataCount, 2, 0);
	this->metaDataBase = msgSize(this->currentMsg);
	smbAddStringFromAscii(this, "");
	this->prefixParameters = this->prefixData = 0;
	this->alignParameters = this->alignData = 4;
	invalidVariable(&this->andXOffset);
}

void	smbStartTrans2SecSmb(smbRequest_t *this, btu16 fid)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
    this->inSmb = 1;
	this->currentMsg = this->smbs[this->currentSmb].block = srvGetMessageBuffer(this->server);
	smbAddAlignment(this, 4);
	writeSmbHeader(this, SMB_COM_TRANSACTION_SECONDARY);
	smbAddByte(this, 8);	/* number of parameter words */
	smbAddVariable(this, &this->smbs[this->currentSmb].totalParameters, 2, 0);
	smbAddVariable(this, &this->smbs[this->currentSmb].totalData, 2, 0);
	smbAddVariable(this, &this->parameterCount, 2, 0);
	smbAddVariable(this, &this->parameterOffset, 2, 0);
	smbAddWord(this, this->parameterByteCount);	/* parameter displacement */
	smbAddVariable(this, &this->dataCount, 2, 0);
	smbAddVariable(this, &this->dataOffset, 2, 0);
	smbAddWord(this, this->dataByteCount);	/* data displacement */
	smbAddWord(this, fid);
	smbAddVariable(this, &this->metaDataCount, 2, 0);
	this->metaDataBase = msgSize(this->currentMsg);
	this->prefixParameters = this->prefixData = 0;
	this->alignParameters = this->alignData = 4;
	invalidVariable(&this->andXOffset);
}

void	smbStartNtTransSmb(smbRequest_t *this, btu16 cmd, btu32 maxRxParam, btu32 maxRxData)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
    this->inSmb = 1;
	this->currentMsg = this->smbs[this->currentSmb].block = srvGetMessageBuffer(this->server);
	smbAddAlignment(this, 4);
	writeSmbHeader(this, SMB_COM_NT_TRANSACT);
	smbAddByte(this, 20);	/* number of parameter words */
	smbAddByte(this, 0);	/* max Rx setup words */
	smbAddWord(this, 0);	/* reserved */
	smbAddVariable(this, &this->smbs[this->currentSmb].totalParameters, 4, 0);
	smbAddVariable(this, &this->smbs[this->currentSmb].totalData, 4, 0);
	smbAddLong(this, maxRxParam);
	smbAddLong(this, maxRxData);
	smbAddVariable(this, &this->parameterCount, 4, 0);
	smbAddVariable(this, &this->parameterOffset, 4, 0);
	smbAddVariable(this, &this->dataCount, 4, 0);
	smbAddVariable(this, &this->dataOffset, 4, 0);
	smbAddByte(this, 0);	/* setup count */
	smbAddWord(this, cmd);
	smbAddByte(this, 0);	/* buffer[1] (??) */
	/* setup words */
	smbAddVariable(this, &this->metaDataCount, 2, 0);
	this->metaDataBase = msgSize(this->currentMsg);
	this->prefixParameters = this->prefixData = 0;
	this->alignParameters = this->alignData = 4;
	invalidVariable(&this->andXOffset);
}

void	smbStartNtTransSecSmb(smbRequest_t *this)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
	if(this->inSmb)			endSmb(this);
    this->inSmb = 1;
	this->currentMsg = this->smbs[this->currentSmb].block = srvGetMessageBuffer(this->server);
	smbAddAlignment(this, 4);
	writeSmbHeader(this, SMB_COM_NT_TRANSACT_SECONDARY);
	smbAddWord(this, 18);	/* number of parameter words + reserved */
	smbAddWord(this, 0);	/* reserved */
	smbAddVariable(this, &this->smbs[this->currentSmb].totalParameters, 4, 0);
	smbAddVariable(this, &this->smbs[this->currentSmb].totalData, 4, 0);
	smbAddVariable(this, &this->parameterCount, 4, 0);
	smbAddVariable(this, &this->parameterOffset, 4, 0);
	smbAddLong(this, this->parameterByteCount);	/* parameter displacement */
	smbAddVariable(this, &this->dataCount, 4, 0);
	smbAddVariable(this, &this->dataOffset, 4, 0);
	smbAddLong(this, this->dataByteCount);	/* data displacement */
	smbAddByte(this, 0);	/* reserved1 */
	smbAddVariable(this, &this->metaDataCount, 2, 0);
	this->metaDataBase = msgSize(this->currentMsg);
	this->prefixParameters = this->prefixData = 0;
	this->alignParameters = this->alignData = 4;
	invalidVariable(&this->andXOffset);
}

static void	endSmb(smbRequest_t *this)
{
    this->inSmb = 0;
	smbWriteVariable(this, &this->metaDataCount, msgSize(this->currentMsg) - this->metaDataBase);
    this->currentSmb++;
}

/* ------------------------------------------------------------------------- */

void	smbStartParameters(smbRequest_t *this, int xCmd)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
    this->inParameters = 1;
	smbWriteVariable(this, &this->andXOffset, msgSize(this->currentMsg));
	if(this->prefixParameters){
		smbAddVariable(this, &this->parameterCount, 1, 1);
	}
	if(this->alignParameters > 0)
		smbAddAlignment(this, this->alignParameters);
	smbWriteVariable(this, &this->parameterOffset, msgSize(this->currentMsg));
	this->parameterBase = msgSize(this->currentMsg);
	if(xCmd < 0){
		invalidVariable(&this->andXOffset);
	}else{
		smbAddWord(this, xCmd);	/* andX cmd + reserved */
		smbAddVariable(this, &this->andXOffset, 2, 0);
	}
}

static void	endParameters(smbRequest_t *this)
{
btsize	parcount;

    this->inParameters = 0;
	parcount = msgSize(this->currentMsg) - this->parameterBase;
	this->parameterByteCount += parcount;
	smbWriteVariable(this, &this->parameterCount, parcount);	
}

/* ------------------------------------------------------------------------- */

void	smbStartData(smbRequest_t *this)
{
	if(this->inData)		endData(this);
	if(this->inParameters)	endParameters(this);
    this->inData = 1;
	if(this->prefixData){
		smbAddVariable(this, &this->dataCount, 2, 0);
	}
    if(this->alignData > 0)
		smbAddAlignment(this, this->alignData);
	smbWriteVariable(this, &this->dataOffset, msgSize(this->currentMsg));
	this->dataBase = msgSize(this->currentMsg);
}

static void	endData(smbRequest_t *this)
{
btsize	dcount;

    this->inData = 0;
	dcount = msgSize(this->currentMsg) - this->dataBase;
	this->dataByteCount += dcount;
	smbWriteVariable(this, &this->dataCount, dcount);
	if(dcount == 0){
		smbWriteVariable(this, &this->dataOffset, 0);
	}
}

/* ------------------------------------------------------------------------- */
/* ---------------------------- receiving data ----------------------------- */
/* ------------------------------------------------------------------------- */

void	smbParseHeader(smbHeader_t *hdr, btu8 *block, btsize len)
{
btu8	*p = block;
btu16	pidHigh;

	if(memcmp(p, "\377SMB", 4) != 0){
        goto bogusDataException;
	}
	p += 4;
	hdr->command = *p++;
	hdr->err.class = *p++;
	p++;
	hdr->err.code = smbGetWord(&p);
	hdr->flags = *p++;
	hdr->flags2 = smbGetWord(&p);
	pidHigh = smbGetWord(&p);
	p += 10;
	hdr->tid = smbGetWord(&p);
	hdr->pid = (btu32)pidHigh << 16 | smbGetWord(&p);
	hdr->uid = smbGetWord(&p);
	hdr->mid = smbGetWord(&p);
	hdr->parameterWords = *p++;
	hdr->parameterOffset = p - block;
	p += 2 * hdr->parameterWords;
	if(p - block >= len){
        goto bogusDataException;
	}
	p += 2;	/* ignore dataBytes entry, we know it (better) from block size */
	hdr->dataOffset = p - block;
	hdr->dataBytes = len - hdr->dataOffset;
    return;
bogusDataException:
    hdr->err.code = ERR_CIFS_LOCbogusdata;
    hdr->err.class = ERRCLS_CIFS_LOCAL;
}

/* ------------------------------------------------------------------------- */

void	smbAdvanceToAndx(smbRequest_t *this)
{
btu8	xcmd, *p;
btu16	xoffset;

    p = this->rxBuffer + this->rxParameterBase;
    xcmd = *p++;
    if(xcmd != 0xff){	/* advance by one andX command frame */
        p++;	/* reserved */
        xoffset = smbGetWord(&p);
        p = this->rxBuffer + xoffset;
        this->rxParameterBytes = *p++ * 2;
        this->rxParameterBase = p - this->rxBuffer;
        p += this->rxParameterBytes;
        this->rxDataBytes = smbGetWord(&p);
        this->rxDataBase = p - this->rxBuffer;
    }
}

/* ------------------------------------------------------------------------- */

void	smbRxBlock(smbRequest_t *this, smbHeader_t *hdr, btu8 *block, btsize len)
{
btu32	paramOffset = 0, paramLen = 0, totalParam = 0, dataOffset = 0;
btu32	dataLen = 0, totalData = 0, paramDisplacement = 0, dataDisplacement = 0;
btu8	*p = block;
btsize	bufferNeeded;
btbool	isTransact = 1;

	this->rxHeader = hdr;
	switch(hdr->command){
	case SMB_COM_TRANSACTION2:
	case SMB_COM_TRANSACTION2_SECONDARY:
	case SMB_COM_TRANSACTION:
	case SMB_COM_TRANSACTION_SECONDARY:
		if(hdr->parameterWords == 0 && hdr->dataBytes == 0){
			this->rxParameterBytes = this->rxDataBytes = this->rxDataBase = 0;
			smbStateMachine(this, EVNT_RX_INTERIM);
			break;
		}
		p = block + hdr->parameterOffset;
		totalParam = smbGetWord(&p);
		totalData = smbGetWord(&p);
		p += 2;	/* reserved */
		paramLen = smbGetWord(&p);
		paramOffset = smbGetWord(&p);
		paramDisplacement = smbGetWord(&p);
		dataLen = smbGetWord(&p);
		dataOffset = smbGetWord(&p);
		dataDisplacement = smbGetWord(&p);
		break;
	case SMB_COM_NT_TRANSACT:
	case SMB_COM_NT_TRANSACT_SECONDARY:
		if(hdr->parameterWords == 0 && hdr->dataBytes == 0){
			this->rxParameterBytes = this->rxDataBytes = this->rxDataBase = 0;
			smbStateMachine(this, EVNT_RX_INTERIM);
			break;
		}
		p = block + hdr->parameterOffset + 3;
		totalParam = smbGetLong(&p);
		totalData = smbGetLong(&p);
		paramLen = smbGetLong(&p);
		paramOffset = smbGetLong(&p);
		paramDisplacement = smbGetLong(&p);
		dataLen = smbGetLong(&p);
		dataOffset = smbGetLong(&p);
		dataDisplacement = smbGetLong(&p);
		break;
	default:
		isTransact = 0;
		if(this->rxBufferAllocated > 0)
			memFree(this->rxBuffer);
		this->rxBuffer = block;
		this->rxBufferLimit = block + len;
		this->rxBufferAllocated = 0;
        this->rxParameterBase = hdr->parameterOffset;
		this->rxParameterBytes = (int)hdr->parameterWords * 2;
		this->rxDataBase = hdr->dataOffset;
		this->rxDataBytes = hdr->dataBytes;
        this->callback(this->callbackRef, this);
		smbStateMachine(this, EVNT_RX_RESPONSE);
	}
	if(isTransact){
		if(p - block > len || paramOffset + paramLen > len || dataOffset + dataLen > len){
            this->haveAllData = this->haveAllParam = 1;
            goto bogusDataException;
		}
		if(!this->rxInitialized){
			bufferNeeded = totalParam + totalData;
			if(this->rxBufferAllocated < bufferNeeded){
				if(this->rxBufferAllocated > 0){
					memFree(this->rxBuffer);
				}
				this->rxBufferAllocated = 0;
				this->rxBuffer = memAlloc(bufferNeeded + 2);	/* leave 2 bytes for 0-termination of Unicode strings in cifs.c */
				this->rxBufferAllocated = bufferNeeded;
				this->rxBufferLimit = this->rxBuffer + bufferNeeded;
			}
            this->rxParameterBase = 0;
			this->rxDataBase = totalParam;
			this->haveAllData = this->haveAllParam = 0;
			this->rxInitialized = 1;
		}else if(totalParam > this->rxDataBase || totalData + this->rxDataBase > this->rxBufferAllocated){
            goto bogusDataException;
		}
		this->rxParameterBytes = totalParam;
		this->rxDataBytes = totalData;
		if(paramDisplacement + paramLen > totalParam || dataDisplacement + dataLen > totalData){
            goto bogusDataException;
		}
		memcpy(this->rxBuffer + paramDisplacement, block + paramOffset, paramLen);
		if(paramDisplacement + paramLen == totalParam)
			this->haveAllParam = 1;
		memcpy(this->rxBuffer + this->rxDataBase + dataDisplacement, block + dataOffset, dataLen);
		if(dataDisplacement + dataLen == totalData)
			this->haveAllData = 1;
bogusDataException:
		if(this->haveAllParam && this->haveAllData){
			this->rxBufferLimit = this->rxBuffer + this->rxDataBase + totalData;
        	this->callback(this->callbackRef, this);
			smbStateMachine(this, EVNT_RX_RESPONSE);
		}
	}
}

/* ------------------------------------------------------------------------- */

void	smbConnectionLost(smbRequest_t *this)
{
	this->rxHeader = NULL;
    if(this->rxBufferAllocated > 0)
        memFree(this->rxBuffer);
    this->rxBuffer = NULL;
    this->rxBufferAllocated = 0;
    this->rxParameterBase = 0;
    this->rxParameterBytes = 0;
    this->rxDataBase = 0;
    this->rxDataBytes = 0;
    this->callback(this->callbackRef, this);
    smbStateMachine(this, EVNT_RX_RESPONSE);
}

/* ------------------------------------------------------------------------- */

