/*
 * Name: smb.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-18
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: smb.h,v 1.1.1.1 1999/04/12 13:56:35 cs Exp $
 */

#ifndef __SMB_H_INCLUDED__
#define __SMB_H_INCLUDED__

#include "basictypes.h"
#include "evntloop.h"
#include "fifo.h"
#include "message.h"

/*
General Description:
This module implements the smbRequest object, which is the program's internal
representation of an SMB request (including the status of a pending request)
and the basic mashalling and unmarshalling primitives. For each server
there is a maximum number of smbRequests (defined by the multiplex count sent
by the server). It also contains the marshalling and unmarshalling of SMB
headers for normal SMB requests, batched requests and the implementation of
the transaction style subprotocols.
*/

/* ------------------------------------------------------------------------- */

#define	MAX_TX_SMBS	16

struct smbRequest;

typedef void	(*smbCallback_t)(void *ref, struct smbRequest *caller);

typedef struct smbVar{
	btsize	position;	/* absolute offset in txBuffer (< 0 -> invalid)*/
	int		size;		/* parameter size in bytes */
	int		scaling;	/* binary logarithm of scaling factor */
}smbVar_t;

typedef struct smbHeader{
	btu8		command;
    btError		err;
	btu8		flags;
	btu16		flags2;
	btu16		tid;
	btu32		pid;
	btu16		uid;
	btu16		mid;
	btu8		parameterWords;
	btu16		dataBytes;
	btu16		parameterOffset;
	btu16		dataOffset;
}smbHeader_t;

typedef struct smbRequest{
    fifoObject_t	fifoPtr;		/* we are subclass of fifoObject_t */
	int				requestState;
	smbCallback_t	callback;
    void			*callbackRef;
    void			*usedByCifs2;	/* storage space used by cifs module */
    int				usedByCifs3;
	struct server	*server;		/* needed for maxXmit, capabilities, midCounter, ... */

/* instance variables for transmitting data: */
	btu16			uid;
	btu32			pid;
	btu16			tid;
	btu16			mid;
	struct{
		struct msg	*block;
		smbVar_t	totalParameters;
		smbVar_t	totalData;
	}				smbs[MAX_TX_SMBS];
	struct msg		*currentMsg;	/* msg block of current SMB */
	int				currentSmb;		/* currently marshalled SMB */
	int				txSmb;			/* currently transmitted SMB */

	btsize			parameterBase;	/* index where parameters start */
	btsize			dataBase;		/* index where data starts */
	btsize			metaDataBase;	/* for TRANSxxx commands: index where TRANS content starts */

	btsize			parameterByteCount;	/* total parameter bytes so far */
	btsize			dataByteCount;		/* total data bytes so far */
	btu8			alignParameters;	/* alignment necessary for parameters */
	btu8			alignData;			/* alignment necessary for data */
	btbool			prefixParameters;	/* should we prefix parameters with count? */
	btbool			prefixData;			/* should we prefix data with count? */

	smbVar_t		parameterOffset;
	smbVar_t		parameterCount;
	smbVar_t		dataOffset;
	smbVar_t		dataCount;
	smbVar_t		metaDataCount;
	smbVar_t		andXOffset;

	btbool			inSmb;
	btbool			inParameters;
	btbool			inData;
	
	bttime			startTime;			/* when this SMB has been transmitted */

/* instance variables for receiving data: */
	smbHeader_t		*rxHeader;
	btu8			*rxBuffer;
	btsize			rxBufferAllocated;	/* number of bytes allocated */
	btu8			*rxBufferLimit;		/* points to first byte behind buffer */

	btbool			rxInitialized;		/* true if following values are valid */
    btsize			rxParameterBase;	/* base of parameters in rxBuffer */
	btsize			rxParameterBytes;	/* number of parameter bytes */
	btsize			rxDataBase;			/* rx data start in rxBuffer */
	btsize			rxDataBytes;		/* number of data bytes */
	
	btbool			haveAllData;
	btbool			haveAllParam;
}smbRequest_t;


/* ------------------------------------------------------------------------- */

EXC_DECLARE(smbExcOverflow);

#define	SMBS_IDLE					0
#define	SMBS_WAIT_TX1				1
#define	SMBS_WAIT_INTERIM			2
#define	SMBS_WAIT_TXN				3
#define	SMBS_WAIT_RESPONSE			4

/* ------------------------------------------------------------------------- */

static inline void	_smbMakeSpace(smbRequest_t *this, btsize bytesToAdd)
{
	msgEnsureSpace(this->currentMsg, bytesToAdd);
}

static inline void	smbAddByte(smbRequest_t *this, btu8 x)
{
	msgAddByte(this->currentMsg, x);
}

static inline void	smbAddWord(smbRequest_t *this, btu16 x)
{
	_smbMakeSpace(this, 2);
	msgAddByteUnchecked(this->currentMsg, x);
	msgAddByteUnchecked(this->currentMsg, x >> 8);
}

static inline void	smbAddLong(smbRequest_t *this, btu32 x)
{
	_smbMakeSpace(this, 4);
	msgAddByteUnchecked(this->currentMsg, x);
	msgAddByteUnchecked(this->currentMsg, x >> 8);
	msgAddByteUnchecked(this->currentMsg, x >> 16);
	msgAddByteUnchecked(this->currentMsg, x >> 24);
}

static inline void	smbAddLarge(smbRequest_t *this, btu64 x)
{
	_smbMakeSpace(this, 8);
	msgAddByteUnchecked(this->currentMsg, x);
	msgAddByteUnchecked(this->currentMsg, x >> 8);
	msgAddByteUnchecked(this->currentMsg, x >> 16);
	msgAddByteUnchecked(this->currentMsg, x >> 24);
	msgAddByteUnchecked(this->currentMsg, x >> 32);
	msgAddByteUnchecked(this->currentMsg, x >> 40);
	msgAddByteUnchecked(this->currentMsg, x >> 48);
	msgAddByteUnchecked(this->currentMsg, x >> 56);
}

static inline void	smbAddBlock(smbRequest_t *this, btu8 *data, btsize len)
{
	msgAddBlock(this->currentMsg, data, len);
}

static inline void	smbAddAlignment(smbRequest_t *this, int bytes)
{
int		align = (bytes - (msgSize(this->currentMsg) % bytes)) % bytes;

	msgAddSpace(this->currentMsg, align);
}

static inline void	smbAddAsciiFromAscii(smbRequest_t *this, btu8 *s)
{
    smbAddBlock(this, (btu8 *)s, strlen(s) + 1);
}

/* ------------------------------------------------------------------------- */

void	smbAddStringFromAscii(smbRequest_t *this, btu8 *s);
void	smbAddStringFromUni(smbRequest_t *this, btuni *s);
void	smbAddUniFromAscii(smbRequest_t *this, btu8 *s);
void	smbAddUniFromUni(smbRequest_t *this, btuni *s);
void	smbAddAsciiFromUni(smbRequest_t *this, btuni *s);

void	smbAddVariable(smbRequest_t *this, smbVar_t *p, int size, int scaling);
void	smbWriteVariable(smbRequest_t *this, smbVar_t *p, btu32 value);

/* ------------------------------------------------------------------------- */

static inline btsize	smbCurrentOffset(smbRequest_t *this)
{
	return msgSize(this->currentMsg);
}

/* ------------------------------------------------------------------------- */

smbRequest_t	*smbRequestNew(struct server *server);
void	smbRequestFree(smbRequest_t *this);
void	smbStartCommand(smbRequest_t *this, btu16 uid, btu32 pid, btu16 tid, smbCallback_t callback, void *ref);
void	smbEndCommand(smbRequest_t *this);
void	smbStartNormalSmb(smbRequest_t *this, btu8 cmd);
void	smbStartTrans2Smb(smbRequest_t *this, btu16 cmd, btu16 maxRxParam, btu16 maxRxData, btu32 timeout);
void	smbStartTrans2SecSmb(smbRequest_t *this, btu16 fid);
void	smbStartTransSmb(smbRequest_t *this, btu8 *tactName, btu16 maxRxParam, btu16 maxRxData, btu32 timeout);
void	smbStartNtTransSmb(smbRequest_t *this, btu16 cmd, btu32 maxRxParam, btu32 maxRxData);
void	smbStartNtTransSecSmb(smbRequest_t *this);
void	smbStartParameters(smbRequest_t *this, int xCmd);
void	smbStartData(smbRequest_t *this);
void	smbParseHeader(smbHeader_t *hdr, btu8 *block, btsize len);
void	smbAdvanceToAndx(smbRequest_t *this);
void	smbRxBlock(smbRequest_t *this, smbHeader_t *hdr, btu8 *block, btsize len);
void	smbConnectionLost(smbRequest_t *this);
nbmsg_t	*smbTxBlock(smbRequest_t *this);

static inline btu8	smbAndxCommand(smbRequest_t *this)
{
    if(this->rxParameterBytes < 1)
        return 0xff;
    return this->rxBuffer[this->rxParameterBase];
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static inline btu16	smbGetWord(btu8 **raw)
{
btu16	x;

	x = (*raw)[0] | ((btu16)(*raw)[1] << 8);
	*raw += 2;
	return x;
}

static inline btu32	smbGetLong(btu8 **raw)
{
btu32	x;

	x = (*raw)[0] | ((btu32)(*raw)[1] << 8) | ((btu32)(*raw)[2] << 16)
				| ((btu32)(*raw)[3] << 24);
	*raw += 4;
	return x;
}

static inline btu64	smbGetLarge(btu8 **raw)
{
btu64	x;

	x = (*raw)[0] | ((btu64)(*raw)[1] << 8) | ((btu64)(*raw)[2] << 16)
				| ((btu64)(*raw)[3] << 24) | ((btu64)(*raw)[4] << 32)
				| ((btu64)(*raw)[5] << 40) | ((btu64)(*raw)[6] << 48)
				| ((btu64)(*raw)[7] << 56);
	*raw += 8;
	return x;
}

/* Unicode strings must be byte-swapped on big endian machines. The
 * following routine does this in-place.
 */
static inline btuni *smbGetUniString(btu8 **raw)
{
btuni	*p, *base, c;

	if(((int)(*raw) & 1) != 0)
		(*raw)++;
	p = base = (btuni *)*raw;
	do{
		c = *p++ = smbGetWord(raw);
	}while(c != 0);
	return base;
}

/* ------------------------------------------------------------------------- */

#endif	/* __SMB_H_INCLUDED__ */
