/*
 * Name: uirep.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-13
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: uirep.c,v 1.5 1999/07/05 16:31:06 cs Exp $
 */

#include "stdheaders.h"
#include "uirep.h"
#include "process.h"
#include "blockpool.h"
#include "mainmodule.h"
#include "guimanager.h"
#include "memalloc.h"
#include "array.h"
#include "ipcserver.h"
#include "module.h"
#include "util.h"
#include "../../uiLib/maininterface.h"
#include "errcodes.h"
#include "ipc.h"
#include "cifsmod.h"

/* ------------------------------------------------------------------------- */

typedef struct loginPanelData{
	btu8	*remoteUser;
	btu8	*password;
	btbool	allowUnencrypted;
	btbool	savePasswd;
	btbool	saveAsDefault;
}loginPanelData_t;

typedef struct uirepCtx{
	PROC_VARSET;
	btint		uid;
	btu8		scratch[128], host[128];
	uirepRes_t	res;
	array_t		*clients;
	btint		index;
	guiClass_t	guiClass;
	ipcSrvClient_t	*client;
}uirepCtx_t;

static blockpool_t *uirepPool = NULL;

static void	poolFreeBlock(void *block)
{
    poolReturnBlock(uirepPool, block);
}

/* ------------------------------------------------------------------------- */


#define	PANEL_OFFSET(entry)	(void *)&(((uirepRes_t *)NULL)->entry)

#define	ALLOW_NULL	0
#define	FLG_0		0

static guiEntry_t	passPhraseEntries[] = {
	{	MAIN_GUI_ENTRY_PASSWORD,	"PEM Pass Phrase",
		NULL,	"cifsDialogPassPhrase", MAIN_GUI_PERM_ANYBODY, 
		FLG_0, ALLOW_NULL,	PANEL_OFFSET(cifsPasswd)
	},
};

static guiEntry_t	loginPanelEntries[] = {
	{	MAIN_GUI_ENTRY_STRING,		"Remote Username",
		NULL,	"cifsDialogLoginRemoteUser", MAIN_GUI_PERM_ANYBODY, 
		FLG_0, ALLOW_NULL,	PANEL_OFFSET(cifsUsername)
	},
	{	MAIN_GUI_ENTRY_PASSWORD,	"Password",
		NULL,	"cifsDialogLoginPassword", MAIN_GUI_PERM_ANYBODY, 
		MAIN_GUI_ENTRY_FLG_DEFAULT_RESPONDER,
		ALLOW_NULL,	PANEL_OFFSET(cifsPasswd)
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Allow sending Password unencrypted",
		NULL,	"cifsDialogLoginAllowUnencrypted", MAIN_GUI_PERM_ANYBODY, 
		FLG_0, ALLOW_NULL,	PANEL_OFFSET(allowUnencrypted)
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Store Username and Password",
		NULL,	"cifsDialogLoginStore", MAIN_GUI_PERM_ANYBODY, 
		FLG_0, ALLOW_NULL,	PANEL_OFFSET(storePasswd)
	},
	{	MAIN_GUI_ENTRY_BOOL,	"Store Credentials as Default",
		NULL,	"cifsDialogLoginStoreDefault", MAIN_GUI_PERM_ANYBODY, 
		FLG_0, ALLOW_NULL,	PANEL_OFFSET(isDefault)
	},
};


static guiButton_t	panelButtons[] = {
	{	"Cancel",			MAIN_GUI_METHOD_CANCEL,
		0,	/* parameter */	MAIN_GUI_BUTTON_CLOSE_DIALOG | MAIN_GUI_BUTTON_EQUALS_CLOSEBOX,
		"cifsDialogLoginButtonCancel",
	},
	{	"Revert",			MAIN_GUI_METHOD_GET,
		0,	/* parameter */	MAIN_GUI_BUTTON_ACTIVE_IF_DIRTY,	/* flags */
		"cifsDialogLoginButtonRevert",
	},
	{	"Login",			MAIN_GUI_METHOD_SET,
		0,	/* parameter */	MAIN_GUI_BUTTON_CLOSE_DIALOG | MAIN_GUI_BUTTON_DEFAULT,	/* flags */
		"cifsDialogLoginButtonLogin",
	},
};

/* ------------------------------------------------------------------------- */

static void	uirepGetPassPhraseProc(void *obj, void *arg)
{
uirepCtx_t		*this = obj;
btint			i;
guimanRes_t		*rval = arg;
ipcMsg_t		*msg = arg;
btu8			*p;

	PROC_BEGIN;
	this->guiClass.type = MAIN_GUI_CLASS_PLAINCFG;
	this->guiClass.name = this->scratch;
	this->guiClass.entries = arrNew(NULL, 8);
	for(i=0;i<ELEMENTS(passPhraseEntries);i++)
		arrAddObject(this->guiClass.entries, &passPhraseEntries[i]);
	this->guiClass.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(panelButtons);i++)
		arrAddObject(this->guiClass.buttons, &panelButtons[i]);
	this->clients = ipcClientServicesForUid(mainModule->ipcModule, this->uid);
	for(this->index=0;this->index<arrObjectCount(this->clients);this->index++){
		this->client = arrObjectAtIndex(this->clients, this->index);
		if(ipcSrvClientIsInvalid(this->client))
			continue;
		/* first try the specific method for the commandline interface */
		memset(&this->res, 0, sizeof(this->res));
		msg = ipcMsgNew();
		ipcAddInt(msg, 0);	/* size */
		ipcAddInt(msg, 0);	/* protocol version */
		ipcAddInt(msg, 0);	/* sequence number */
		ipcAddInt(msg, MAIN_IPCCFN_ASK_DATA);
		ipcAddInt(msg, 0);	/* uid */
		ipcAddInt(msg, 0);	/* gid */
		ipcAddAscii(msg, this->scratch);	/* prompt */
		ipcAddInt(msg, 0);	/* without echo */
		PROC_WAITFOR_X(1, ipcSrvRPC(cifsFileModule->ipcModule, this->client, msg, this));
		if(msg == NULL){	/* client died */
			this->res.err.class = ERRCLS_FRAMEWORK;
			this->res.err.code = ERR_FRWclientcomm;
			goto doReturn;
		}
		ipcGetInt(msg);	/* uid */
		ipcGetInt(msg);	/* gid */
		this->res.err.class = ipcGetInt(msg);
		this->res.err.code = ipcGetInt(msg);
		if(this->res.err.class == 0){
			p = ipcGetAscii(msg);	/* pass phrase */
			if(p == NULL)
				goto doReturn;	/* cancel */
			this->res.isOk = 1;
			this->res.cifsPasswd = utilStrDup(p);
			goto doReturn;
		}
		/* if cmdline interface failed, try generic GUI method */
		memset(&this->res, 0, sizeof(this->res));
		PROC_WAITFOR(guimanRunDialog(&this->guiClass, this->client, &this->res, 1, this));
		if(rval->err.class == ERRCLS_FRAMEWORK){
			if(rval->err.code == ERR_FRWipcnotimpl){
				if(this->res.cifsPasswd != NULL)
					memFree(this->res.cifsPasswd);
				continue;	/* try next client */
			}
		}
		if(rval->err.class != 0){
			this->res.err = rval->err;
			goto doReturn;
		}else{
			this->res.err.class = this->res.err.code = 0;
			this->res.isOk = rval->didSet;
			goto doReturn;
		}
	}
doReturn:
	arrFree(this->guiClass.entries);
	arrFree(this->guiClass.buttons);
	this->res.client = this->client;
	PROC_RETURN_AND_CLEANUP(&this->res,
		arrFree(this->clients);
		if(this->res.cifsPasswd != NULL){
			memset(this->res.cifsPasswd, 0, strlen(this->res.cifsPasswd));
			memFree(this->res.cifsPasswd);
		}
	);
	PROC_END;
}

void	uirepGetPassPhrase(btint uid, btu8 *srvName, void *caller)
{
uirepCtx_t	*ctx;

	if(uirepPool == NULL)
		uirepPool = poolNew(sizeof(uirepCtx_t), sizeof(uirepCtx_t));
	PROC_INIT(ctx, uirepGetPassPhraseProc, caller, uirepPool, poolFreeBlock);
	ctx->uid = uid;
	if(strlen(srvName) > ELEMENTS(ctx->scratch) - 20)
		srvName[ELEMENTS(ctx->scratch) - 20] = 0;	/* emergency brake */
	sprintf(ctx->scratch, "Pass Phrase for Server %s: ", srvName);
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	uirepGetCredentialsProc(void *obj, void *arg)
{
uirepCtx_t		*this = obj;
btint			i;
guimanRes_t		*rval = arg;
struct passwd	*pw;
btu8			*p;
ipcMsg_t		*msg = arg;

	PROC_BEGIN;
	this->guiClass.type = MAIN_GUI_CLASS_PLAINCFG;
	this->guiClass.name = this->scratch;
	this->guiClass.entries = arrNew(NULL, 8);
	for(i=0;i<ELEMENTS(loginPanelEntries);i++)
		arrAddObject(this->guiClass.entries, &loginPanelEntries[i]);
	this->guiClass.buttons = arrNew(NULL, 4);
	for(i=0;i<ELEMENTS(panelButtons);i++)
		arrAddObject(this->guiClass.buttons, &panelButtons[i]);
	this->clients = ipcClientServicesForUid(mainModule->ipcModule, this->uid);
	for(this->index=0;this->index<arrObjectCount(this->clients);this->index++){
		this->client = arrObjectAtIndex(this->clients, this->index);
		if(ipcSrvClientIsInvalid(this->client))
			continue;
		/* first try the specific method for the commandline interface */
		memset(&this->res, 0, sizeof(this->res));
		msg = ipcMsgNew();
		ipcAddInt(msg, 0);	/* size */
		ipcAddInt(msg, 0);	/* protocol version */
		ipcAddInt(msg, 0);	/* sequence number */
		ipcAddInt(msg, MAIN_IPCCFN_ASK_PASSWORD);
		ipcAddInt(msg, 0);	/* uid */
		ipcAddInt(msg, 0);	/* gid */
		ipcAddAscii(msg, this->host);	/* hostname */
		PROC_WAITFOR(ipcSrvRPC(cifsFileModule->ipcModule, this->client, msg, this));
		if(msg == NULL){	/* client died */
			this->res.err.class = ERRCLS_FRAMEWORK;
			this->res.err.code = ERR_FRWclientcomm;
			goto doReturn;
		}
		ipcGetInt(msg);	/* uid */
		ipcGetInt(msg);	/* gid */
		this->res.err.class = ipcGetInt(msg);
		this->res.err.code = ipcGetInt(msg);
		if(this->res.err.class == 0){
			p = ipcGetAscii(msg);	/* username */
			if(p == NULL)
				goto doReturn;	/* cancel */
			this->res.cifsUsername = utilStrDup(p);
			p = ipcGetAscii(msg);	/* passwd */
			if(p == NULL)
				goto doReturn;	/* cancel */
			this->res.cifsPasswd = utilStrDup(p);
			this->res.allowUnencrypted = ipcGetInt(msg);
			this->res.isOk = 1;
			goto doReturn;
		}
		if(this->res.err.class == ERRCLS_FRAMEWORK && this->res.err.code == ERR_FRWusrabort)
			goto doReturn;
		/* if cmdline interface failed, try generic GUI method */
		memset(&this->res, 0, sizeof(this->res));
		pw = getpwuid(this->uid);
		if(pw != NULL)
			this->res.cifsUsername = utilStrDup(pw->pw_name);
		else
			this->res.cifsUsername = utilStrDup("");
		this->res.cifsPasswd = utilStrDup("");
		PROC_WAITFOR(guimanRunDialog(&this->guiClass, this->client, &this->res, 1, this));
		if(rval->err.class == ERRCLS_FRAMEWORK){
			if(rval->err.code == ERR_FRWipcnotimpl){
				if(this->res.cifsUsername != NULL){
					memFree(this->res.cifsUsername);
					this->res.cifsUsername = NULL;
				}
				if(this->res.cifsPasswd != NULL){
					memFree(this->res.cifsPasswd);
					this->res.cifsPasswd = NULL;
				}
				continue;	/* try next client */
			}
		}
		if(rval->err.class != 0){
			this->res.err = rval->err;
		}else{
			this->res.err.class = this->res.err.code = 0;
			this->res.isOk = rval->didSet;
		}
		goto doReturn;
	}
	/* if the loop runs through: implicit result = OK, user cancelled */
doReturn:
	arrFree(this->guiClass.entries);
	arrFree(this->guiClass.buttons);
	this->res.client = this->client;
	PROC_RETURN_AND_CLEANUP(&this->res,
		arrFree(this->clients);
		if(this->res.cifsUsername != NULL)
			memFree(this->res.cifsUsername);
		if(this->res.cifsPasswd != NULL){
			memset(this->res.cifsPasswd, 0, strlen(this->res.cifsPasswd));
			memFree(this->res.cifsPasswd);
		}
	);
	PROC_END;
}

void	uirepGetCredentials(btint uid, btu8 *host, void *caller)
{
uirepCtx_t	*ctx;

	if(uirepPool == NULL)
		uirepPool = poolNew(sizeof(uirepCtx_t), sizeof(uirepCtx_t));
	PROC_INIT(ctx, uirepGetCredentialsProc, caller, uirepPool, poolFreeBlock);
	ctx->uid = uid;
	if(strlen(host) > ELEMENTS(ctx->scratch) - 20)
		host[ELEMENTS(ctx->scratch) - 20] = 0;	/* emergency brake */
	sprintf(ctx->scratch, "Login on Host %s", host);
	strcpy(ctx->host, host);
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */
