/*
 * Name: userdb.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-11
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: userdb.c,v 1.3 1999/07/22 17:54:13 cs Exp $
 */


#include "stdheaders.h"
#include "rndtemp.h"
#include "pwhash.h"
#include "userdb.h"
#include "array.h"
#include "exception.h"
#include "memalloc.h"
#include "unicode.h"
#include "logging.h"
#include "user.h"
#include "server.h"
#include "uidsecure.h"
#include "util.h"
#include "system.h"
#include "ipc.h"
#include "mainmodule.h"
#include "cifsmod.h"

#define	UDB_VERSION		2	/* this version */

#define	ETYPE_EOF		0	/* EOF entity type */
#define	ETYPE_HDR		1	/* file header */
#define	ETYPE_LICENSE	2	/* license string */
#define	ETYPE_USER		3	/* stored user-data and password */

#define	UDB_RELEASE(x)	if((x) != NULL){memFree(x); (x) = NULL;}

/* ------------------------------------------------------------------------- */

typedef struct fileHeader{
	btu8	rnumber;
	btu32	progVersion;
	btu32	userRecLen;
}fileHeader_t;

/* ------------------------------------------------------------------------- */

static void	*udbReadEntity(int fd, btsize *len, btu32 *type)
{
btu32		hdr[2];
int			rval;
static char	dummy[1];
void		*data;

	if((rval = read(fd, hdr, sizeof(hdr))) != sizeof(hdr)){
		if(rval == 0){	/* EOF */
			*len = 0;
			*type = ETYPE_EOF;
			return dummy;
		}
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			if(rval < 0){
				logPrintf(LOGM_ERROR, "userdb: error reading entity header: [%d] %s\n", errno, sysStrerror(errno));
			}else{
				logPrintf(LOGM_ERROR, "userdb: only %d bytes of entity header read\n", rval);
			}
		}
		return NULL;
	}
	*type = hdr[0];
	*len = hdr[1];
	data = memAlloc(*len);
	if((rval = read(fd, data, *len)) != *len){
		memFree(data);
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR)){
			if(rval < 0){
				logPrintf(LOGM_ERROR, "userdb: error reading entity data: [%d] %s\n", errno, sysStrerror(errno));
			}else{
				logPrintf(LOGM_ERROR, "userdb: only %d bytes of entity data read\n", rval);
			}
		}
		return NULL;
	}
	return data;
}

static btbool	udbWriteEntity(int fd, btsize len, void *data, btu32 type)
{
btu32	hdr[2];

	hdr[0] = type;
	hdr[1] = len;
	if(write(fd, hdr, sizeof(hdr)) != sizeof(hdr)){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: error writing entity header: [%d] %s\n", errno, sysStrerror(errno));
		return 0;
	}
	if(write(fd, data, len) != len){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: error writing %d bytes entity data: [%d] %s\n", len, errno, sysStrerror(errno));
		return 0;
	}
	return 1;
}

/* ------------------------------------------------------------------------- */

static inline btu64	ipow(btu64 x, int exponent)
{
btu64	res = 1;
int		i;

	for(i=0;i<exponent;i++)
		res *= x;
	return res;
}

/* ------------------------------------------------------------------------- */

void	udbCryptBlock(btu8 rblock[8], void *data, btsize len)
{
btu8	hash[8], *p = data;
btu64	*rbase = (btu64 *)rblock;
btsize	thisLen, i;

	while(len > 0){
		pwhBlockHash(rblock, hash);
		*rbase = *rbase * *rbase - 1;
		thisLen = len > 8 ? 8 : len;
		for(i=0;i<thisLen;i++)
			*p++ ^= hash[i];
		len -= thisLen;
	}
}

/* ------------------------------------------------------------------------- */

static void	udbEncodeInt(btu32 x, btu8 **stream)
{
btu8	*p = *stream;

	*p++ = x;
	*p++ = x >> 8;
	*p++ = x >> 16;
	*p++ = x >> 24;
	*stream = p;
}

static btu32	udbDecodeInt(btu8 **stream)
{
btu8	*p = *stream;
btu32	x;

	x = *p++;
	x |= *p++ << 8;
	x |= *p++ << 16;
	x |= *p++ << 24;
	*stream = p;
	return x;
}

/* ------------------------------------------------------------------------- */

static btsize	udbAsciiSize(btu8 *string)
{
	if(string == NULL)
		return 4;
	else
		return 5 + strlen(string);
}

static void	udbEncodeAscii(btu8 *string, btu8 **stream)
{
btsize	l;

	if(string == NULL){
		udbEncodeInt(0, stream);
	}else{
		l = strlen(string) + 1;
		udbEncodeInt(l, stream);
		memcpy(*stream, string, l);
		*stream += l;
	}
}

static btu8	*udbDecodeAscii(btu8 **stream)
{
btsize	l;
btu8	*s;

	l = udbDecodeInt(stream);
	if(l == 0)
		return NULL;
	s = *stream;
	*stream += l;
	return s;
}

/* ------------------------------------------------------------------------- */

static btsize	udbUniSize(btuni *string)
{
	if(string == NULL)
		return 4;
	else
		return 6 + 2 * uniStrLen(string);
}

static void	udbEncodeUni(btuni *string, btu8 **stream)
{
btsize	l;

	if(string == NULL){
		udbEncodeInt(0, stream);
	}else{
		l = 2 * (uniStrLen(string) + 1);
		udbEncodeInt(l, stream);
		memcpy(*stream, string, l);
		*stream += l;
	}
}

static btuni	*udbDecodeUni(btu8 **stream)
{
btsize	l;
btuni	*s;

	l = udbDecodeInt(stream);
	if(l == 0)
		return NULL;
	s = (btuni *)*stream;
	*stream += l;
	return s;
}

/* ------------------------------------------------------------------------- */

static void	udbEncodeBlock(void *data, btsize len, btu8 **stream)
{
	memcpy(*stream, data, len);
	*stream += len;
}

static void	*udbDecodeBlock(btsize len, btu8 **stream)
{
btu8	*p = *stream;

	*stream += len;
	return p;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct{
	btu8	rnumber;
	btu32	version;
	btu32	testconst;
}udbFileHeader_t;

static btbool	udbEncodeHeader(int fd, btu8 *rblock)
{
udbFileHeader_t	hdr;

	memset(&hdr, 0, sizeof(hdr));
	hdr.rnumber = rblock[7];
	hdr.version = UDB_VERSION;
	hdr.testconst = 0x12345678;
	udbCryptBlock(rblock, &hdr.testconst, sizeof(hdr.testconst));
	return udbWriteEntity(fd, sizeof(hdr), &hdr, ETYPE_HDR);
}

static void	*udbDecodeHeader(btu8 *entity, btsize len, btu8 *rblock)
{
udbFileHeader_t	*hdr = (void *)entity;
btu8			rnum = rblock[7];
static char		dummy;

	if(len != sizeof(*hdr) || hdr->rnumber != rnum || hdr->version != UDB_VERSION){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: database file is incompatible\n");
		return NULL;
	}
	udbCryptBlock(rblock, &hdr->testconst, sizeof(hdr->testconst));
	if(hdr->testconst != 0x12345678){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: database file is incompatible\n");
		return NULL;
	}
	return &dummy;	/* dummy non-null pointer */
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static btbool	udbEncodeLicense(btu8 licenseString[12], int fd, btu8 *rblock)
{
btu8	buffer[12];

	memcpy(buffer, licenseString, 12);
	udbCryptBlock(rblock, buffer, 12);
	return udbWriteEntity(fd, 12, buffer, ETYPE_LICENSE);
}

static void	*udbDecodeLicense(btu8 *entity, btsize len, btu8 *rblock)
{
	if(len != 12)
		return NULL;
	udbCryptBlock(rblock, entity, 12);
	return entity;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static btbool	udbEncodeUserEntry(void *element, int fd, btu8 *rblock)
{
userRecord_t	*this = element;
btu8			*entity, *p;
btbool			success;
btu8			buffer[16];

	p = entity = memAlloc(36 + udbAsciiSize(this->serverName) + udbUniSize(this->cifsName) + udbUniSize(this->plainPasswd));
	memcpy(buffer, this->preNtHash, 16);
	udbCryptBlock(rblock, buffer, 16);
	udbEncodeBlock(buffer, 16, &p);
	memcpy(buffer, this->ntHash, 16);
	udbCryptBlock(rblock, buffer, 16);
	udbEncodeBlock(buffer, 16, &p);
	udbEncodeInt(this->unixUid, &p);
	udbEncodeUni(this->cifsName, &p);
	udbEncodeUni(this->plainPasswd, &p);
	udbEncodeAscii(this->serverName, &p);
	success = udbWriteEntity(fd, p - entity, entity, ETYPE_USER);
	memFree(entity);
	return success;
}


static void	udbUserEntryFree(userRecord_t *this)
{
	if(this != NULL){
		UDB_RELEASE(this->serverName);
		UDB_RELEASE(this->cifsName);
		UDB_RELEASE(this->plainPasswd);
		memFree(this);
	}
}

static userRecord_t	*udbUserEntryNew(btu32 volatile uid, btu8 *serverName, btuni *cifsName, btu8 preNtHash[16], btu8 ntHash[16], btuni *plainPasswd)
{
userRecord_t	*volatile rec;

EXC_TRY
	rec = memCalloc(sizeof(userRecord_t));
	rec->unixUid = uid;
	rec->serverName = utilStrDup(serverName);
	rec->cifsName = uniStrDup(cifsName);
	memcpy(rec->preNtHash, preNtHash, 16);
	memcpy(rec->ntHash, ntHash, 16);
	if(plainPasswd != NULL)
		rec->plainPasswd = uniStrDup(plainPasswd);
EXC_CATCH
	udbUserEntryFree(rec);
	EXC_RETHROW();
EXC_ENDTRY
	return rec;
}

static userRecord_t	*udbDecodeUserEntry(btu8 *entity, btsize len, btu8 *rblock)
{
btu8			*serverName, *preNtHash, *ntHash, *p = entity;
btuni			*cifsName, *plainPasswd;
btint			unixUid;

	preNtHash = udbDecodeBlock(16, &p);
	if(p - entity > len)
		return NULL;
	ntHash = udbDecodeBlock(16, &p);
	if(p - entity > len)
		return NULL;
	unixUid = udbDecodeInt(&p);
	if(p - entity > len)
		return NULL;
	cifsName = udbDecodeUni(&p);
	if(p - entity > len)
		return NULL;
	plainPasswd = udbDecodeUni(&p);
	if(p - entity > len)
		return NULL;
	serverName = udbDecodeAscii(&p);
	if(p - entity > len)
		return NULL;
	udbCryptBlock(rblock, preNtHash, 16);
	udbCryptBlock(rblock, ntHash, 16);
	return udbUserEntryNew(unixUid, serverName, cifsName, preNtHash, ntHash, plainPasswd);
}

static void	udbUserEntryRetain(void *obj)
{
userRecord_t	*this = obj;

	this->refCount++;
}

static void	udbUserEntryRelease(void *obj)
{
userRecord_t	*this = obj;

	this->refCount--;
	if(this->refCount == 0){
		udbUserEntryFree(this);
	}else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

static btu32	udbUserEntryHash(void *obj)
{
userRecord_t	*this = obj;

	return this->unixUid;
}

static btint	udbUserEntryCompare(void *obj, void *otherObj)
{
userRecord_t	*this = obj, *other = otherObj;
btint			rval;

	rval = this->unixUid - other->unixUid;
	if(rval != 0)
		return rval;
	return utilStrCmp(this->serverName, other->serverName, 0);
}

static collectionObjectI_t	udbUserEntryInterface = {
	udbUserEntryCompare,
	udbUserEntryHash,
	udbUserEntryRetain,
	udbUserEntryRelease
};

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static udbData_t	*udbReadFile(btu8 *path)
{
int				storederr, fd = -1;
btu64			rbase = RND_NUMBER;
btu8			*rblock = (btu8 *)&rbase;
struct stat		sbuf;
void			*element, *entity = NULL;
btsize			len;
btu32			type;
udbData_t		*content = NULL;

	uidsecBecomeRoot();
	fd = open(path, O_RDONLY, 0);
	storederr = errno;
	uidsecBecomeUser();
	if(fd < 0){
		if(storederr != ENOENT && LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: cannot open file %s: [%d] %s\n", path, storederr, sysStrerror(storederr));
		goto errorOccured;
	}
#ifdef INSECURE_USERDB
	memset(&sbuf, 0, sizeof(sbuf));
#else
	if(fstat(fd, &sbuf) != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: cannot stat file %s: [%d] %s\n", path, errno, sysStrerror(errno));
		memset(&sbuf, 0, sizeof(sbuf));
	}
#endif
	content = memCalloc(sizeof(udbData_t));
	content->users = htabNew(&udbUserEntryInterface, 16, 75);
	rbase += 89 * ipow(sbuf.st_ino, 4) + 13 * ipow(sbuf.st_ino, 3) + ipow(sbuf.st_ino, 2) + sbuf.st_ino;
	for(;;){
		if((entity = udbReadEntity(fd, &len, &type)) == NULL)
			goto errorOccured;
		if(type == ETYPE_EOF)
			break;
		switch(type){
		case ETYPE_HDR:
			if((element = udbDecodeHeader(entity, len, rblock)) == NULL)
				goto errorOccured;
			break;
		case ETYPE_LICENSE:
			if((element = udbDecodeLicense(entity, len, rblock)) == NULL)
				goto errorOccured;
			memcpy(content->licenseString, element, sizeof(content->licenseString));
			break;
		case ETYPE_USER:
			if((element = udbDecodeUserEntry(entity, len, rblock)) == NULL)
				goto errorOccured;
			htabAddObject(content->users, element);
			break;
		default:
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "userdb: invalid entity type %d found\n", type);
			goto errorOccured;
		}
		memFree(entity);
		entity = NULL;
	}
	return content;
errorOccured:
	if(entity != NULL)
		memFree(entity);
	if(fd >= 0)
		close(fd);
	if(content != NULL){
		if(content->users != NULL)
			htabFree(content->users);
		memFree(content);
	}
	return NULL;
}

/* ------------------------------------------------------------------------- */

typedef btbool	(*encoder_t)(void *element, int fd, btu8 *rblock);

static btbool	udbWriteArray(int fd, array_t *arr, encoder_t encoder, btu8 *rblock)
{
btsize	i, l;

	l = arrObjectCount(arr);
	for(i=0;i<l;i++){
		if(!encoder(arrObjectAtIndex(arr, i), fd, rblock)){
			return 0;
		}
	}
	return 1;
}

/* ------------------------------------------------------------------------- */

static btbool	udbWriteFile(userdb_t *this)
{
btu8			bakfile[MAXPATHLEN];
int				storederr, fd = -1;
btu64			rbase = RND_NUMBER;
btu8			*rblock = (btu8 *)&rbase;
struct stat		sbuf;
array_t			*arr;

	if(!this->dirty)
		return 1;
	mainModGuiUpdate(cifsLoginListObject);
	sprintf(bakfile, "%s.bak", this->file);
	uidsecBecomeRoot();
	unlink(bakfile);				/* remove old backup files, who knows what rename semantics we have... */
	rename(this->file, bakfile);	/* make backup of old file, ignore errors */
	fd = creat(this->file, 0600);
	storederr = errno;
	uidsecBecomeUser();
	if(fd < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: cannot create file %s: [%d] %s\n", this->file, storederr, sysStrerror(storederr));
		goto errorOccured;
	}
#ifdef INSECURE_USERDB
	memset(&sbuf, 0, sizeof(sbuf));
#else
	if(fstat(fd, &sbuf) != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: cannot stat file %s: [%d] %s\n", this->file, errno, sysStrerror(errno));
		memset(&sbuf, 0, sizeof(sbuf));
	}
#endif
	rbase += 89 * ipow(sbuf.st_ino, 4) + 13 * ipow(sbuf.st_ino, 3) + ipow(sbuf.st_ino, 2) + sbuf.st_ino;
	if(!udbEncodeHeader(fd, rblock))
		goto errorOccured;
	if(!udbEncodeLicense(this->data->licenseString, fd, rblock))
		goto errorOccured;
	arr = htabAllObjects(this->data->users);
	if(!udbWriteArray(fd, arr, udbEncodeUserEntry, rblock)){
		arrFree(arr);
		goto errorOccured;
	}
	arrFree(arr);
	if(close(fd) < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "userdb: error closing file %s: [%d] %s\n", this->file, errno, sysStrerror(errno));
		goto errorOccured;
	}
	uidsecBecomeRoot();
	unlink(bakfile);				/* remove backup file if everything went OK */
	uidsecBecomeUser();
	this->dirty = 0;
	return 1;
errorOccured:
	if(fd >= 0)
		close(fd);
	uidsecBecomeRoot();
	unlink(this->file);				/* remove our attempt to save */
	rename(bakfile, this->file);	/* try to restore backup */
	uidsecBecomeUser();
	return 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

userdb_t	*udbNew(btu8 *file)
{
userdb_t	*volatile this = NULL;
udbData_t	*content;

EXC_TRY
	this = memCalloc(sizeof(userdb_t));
	this->file = utilStrDup(file);
	this->data = udbReadFile(file);
	if(this->data == NULL){	/* file could not be read */
		content = memCalloc(sizeof(udbData_t));
		this->data = content;
		content->users = htabNew(&udbUserEntryInterface, 16, 75);
	}
EXC_CATCH
	udbFree(this);
	EXC_RETHROW();
EXC_ENDTRY
	return this;
}

void	udbFree(userdb_t *this)
{
	if(this != NULL){
		if(this->file != NULL)
			memFree(this->file);
		if(this->data != NULL){
			if(this->data->users != NULL)
				htabFree(this->data->users);
			memFree(this->data);
		}
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

void	udbTransact(userdb_t *this)
{
	this->commitCount++;
}

void	udbCommit(userdb_t *this)
{
	if(--this->commitCount <= 0)
		udbWriteFile(this);
}

/* ------------------------------------------------------------------------- */

userRecord_t	*udbQueryUser(userdb_t *this, btu32 unixUid, btu8 *serverName)
{
userRecord_t	*rec, r;

	r.unixUid = unixUid;
	r.serverName = serverName;
	rec = htabFindObjectEqualTo(this->data->users, &r);
	return rec;
}

void	udbAddUser(userdb_t *this, btu8 *srvName, struct cifsUser *user)
{
userRecord_t	*rec;

	rec = udbUserEntryNew(user->unixUid, srvName, user->smbUsername, user->pwHashPreNt, user->pwHashNt, user->plainPasswd);
	if(rec != NULL){
		htabAddObject(this->data->users, rec);
		this->dirty = 1;
		if(this->commitCount <= 0)
			udbWriteFile(this);
	}
}

void	udbRemoveUser(userdb_t *this, btu32 unixUid, btu8 *serverName)
{
userRecord_t	r;

	r.unixUid = unixUid;
	r.serverName = serverName;
	htabRemoveObjectEqualTo(this->data->users, &r);
	this->dirty = 1;
	if(this->commitCount <= 0)
		udbWriteFile(this);
}

void	udbRemoveAllUsers(userdb_t *this)
{
	htabFree(this->data->users);
	this->data->users = htabNew(&udbUserEntryInterface, 16, 75);
	this->dirty = 1;
	if(this->commitCount <= 0)
		udbWriteFile(this);
}

/* ------------------------------------------------------------------------- */

void	udbSetLicenseString(userdb_t *this, btu8 licenseString[12])
{
	memcpy(this->data->licenseString, licenseString, 12);
	this->dirty = 1;
	if(this->commitCount <= 0)
		udbWriteFile(this);
}

/* ------------------------------------------------------------------------- */

