/*
 * Name: userdb.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-11
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: userdb.h,v 1.3 1999/07/22 17:54:13 cs Exp $
 */

#ifndef __userdb_H_INCLUDED__
#define __userdb_H_INCLUDED__

#include "basictypes.h"
#include "hashtable.h"
#include "array.h"

/*
General Description:
This module implements some persistent classes. Any update is immediately
synchronized to the database file (unless several updates are enclosed in
'udbTransact()' and 'udbCommit()' calls). Sensitive information like the
license key and user passwords are stored in encrypted format.
There are now two tables stored in the database:
1.) The license key
2.) User passwords

Remarks regarding the encryption of passwords:
I's necessary to store both kinds of encrypted SMB passwords (with md4 hash
and with des hash). These encryptions can not easily be inverted, making it
hard to obtain the plain text passwords from the SMB encrypted passwords.
This is, however, not necessary to do evil things with the passwords. The
SMB encrypted passwords alone suffice to make a connect to an SMB server.
To protect these passwords further, they are encrypted one more time before
they are stored. This "second encryption" must, however, be easily reversible.
Although I have done my best to obscure the way to do this, it is still
possible that a hacker steals the user database file (and some additional
information, the file alone is not sufficient) and decrypts your SMB encrypted
passwords.
The conclusion of all this:
Use stored passwords only if the people who have access to your machine are
known or at least limited. Don't store passwords on a publicly available
machine like a university guest machine etc. If it's your private workplace
machine you should be safe.
*/

/* ------------------------------------------------------------------------- */

typedef struct userRecord{
	bti32	refCount;
	btu32	unixUid;
	btu8	*serverName;	/* ".default" for default credentials */
	btuni	*cifsName;		/* user name at server */
	btu8	preNtHash[16];
	btu8	ntHash[16];
	btuni	*plainPasswd;
}userRecord_t;

typedef struct udbData{
	hashtable_t	*users;		/* elements: userRecord_t hashed by uid+server */
	btu8		licenseString[12];
}udbData_t;

typedef struct userdb{
	btu8		*file;
	udbData_t	*data;
	bti32		commitCount;
	btbool		dirty;
}userdb_t;

struct cifsUser;

/* ------------------------------------------------------------------------- */

userdb_t	*udbNew(btu8 *file);
void	udbFree(userdb_t *this);
userRecord_t	*udbQueryUser(userdb_t *this, btu32 unixUid, btu8 *serverName);
void	udbTransact(userdb_t *this);
void	udbCommit(userdb_t *this);

void	udbAddUser(userdb_t *this, btu8 *srvName, struct cifsUser *user);
void	udbRemoveUser(userdb_t *this, btu32 unixUid, btu8 *serverName);
void	udbRemoveAllUsers(userdb_t *this);

void	udbSetLicenseString(userdb_t *this, btu8 licenseString[12]);

/* ------------------------------------------------------------------------- */

static inline btu8	*udbLicenseString(userdb_t *this)
{
	return this->data->licenseString;
}

static inline array_t	*udbAllRecords(userdb_t *this)
{
	return htabAllObjects(this->data->users);
}

static inline btu8	*udbrecServer(userRecord_t *this)
{
	return this->serverName;
}

static inline btuni	*udbrecCifsUser(userRecord_t *this)
{
	return this->cifsName;
}

static inline btu32	udbrecUnixUid(userRecord_t *this)
{
	return this->unixUid;
}

/* ------------------------------------------------------------------------- */

#endif	/* __userdb_H_INCLUDED__ */
