/*
 * Name: evntloop.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-12
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: evntloop.h,v 1.1.1.1 1999/04/12 13:56:37 cs Exp $
 */

#ifndef __EVNTLOOP_H_INCLUDED__
#define __EVNTLOOP_H_INCLUDED__

#include "basictypes.h"
#include "exception.h"

/*
General Description:
This module implements the mainloop of the program. For every event
you can schedule a function that is called when the event occures.
*/

typedef struct elpFile{
    int	mode;
    int	fd;
}elpFile_t;

typedef struct elpPid{
    int	pid;
    int	rval;
}elpPid_t;

#define	ELP_FDMODE_READ		0
#define	ELP_FDMODE_WRITE	1
#define	ELP_FDMODE_EXCEPT	2
#define	ELP_FDMODES			3

EXC_DECLARE(elpExcBadArgument);
EXC_DECLARE(elpExcFdAlreadyUsed);

/* ------------------------------------------------------------------------- */
void	elpInit(void);
/* This function must be called during system initialisation before any other
 * function of this modle may be called.
 */
/* ------------------------------------------------------------------------- */
extern bttime	elpCurrentTime;
extern btbool	elpCurrentTimeValid;
void			elpMakeCurrentTimeValid(void);

static inline bttime	elpNow(void)
{
	if(!elpCurrentTimeValid){
		elpMakeCurrentTimeValid();
	}
	return elpCurrentTime;
}
/* Returns the current absolute time in milliseconds. It is the basis for
 * giving absolute times in elpCallAt().
 */
/* ------------------------------------------------------------------------- */
extern bti64	elpCurrentTime64;
extern btbool	elpCurrentTime64Valid;
void			elpMakeCurrentTime64Valid(void);

static inline bti64	elpNow64(void)
{
	if(!elpCurrentTime64Valid){
		elpMakeCurrentTime64Valid();
	}
	return elpCurrentTime64;
}

/* ------------------------------------------------------------------------- */
void	elpRegisterForFd(int fd, int mode, btcallback callb, void *ref);
/* This function registers the function 'callb' to be called with argument
 * 'ref', 'fd' and 'mode' if the file descriptor 'fd' becomes ready for mode
 * 'mode'. 'mode' may be one of 'ELP_FDMODE_READ', 'ELP_FDMODE_WRITE' or
 * 'ELP_FDMODE_EXCEPT'. If 'callb' has already been registered with the same
 * arguments, the call is ignored. If it has been registered with different
 * arguments, the exception 'elpExcFdAlreadyUsed' is thrown. The function stays
 * registered until 'elpCancelForFd()' is called.
 */
/* ------------------------------------------------------------------------- */
void	elpCancelForFd(int fd, int mode);
/* This function is the opposite of 'elpRegisterForFd()': It is called to
 * cancel the registration of any function for a given 'fd' and 'mode'. If
 * 'mode' is '-1', 'fd' is cancelled for all modes.
 */
/* ------------------------------------------------------------------------- */
void	elpRegisterForPid(int pid, btcallback cb, void *ref);
/* This function registers the function 'cb' to be called when the process
 * 'pid' terminates.
 */
/* ------------------------------------------------------------------------- */
void	elpCancelForPid(int pid);
/* This function is the opposite of 'elpRegisterForPid()'. It deletes the
 * registration of any callback function for the process 'pid'.
 */
/* ------------------------------------------------------------------------- */
/*void	elpCallAt(bttime when, btcallback callback, void *ref);*/
/* currently not implemented because not needed */
/* ------------------------------------------------------------------------- */
btbool	elpCallAtCancelPrev(bttime when, btcallback callback, void *ref);
/* This function registers 'callback' to be called with argument 'ref' at the
 * time when 'elpNow()' returns the value 'when' (which is an absolute time
 * in milliseconds). If the same pair of 'callback' and 'ref' has already been
 * registered for an other time, that other call is canceled.
 * Returns: 1, if a previous call has been cancelled, 0 otherwise.
 */
/* ------------------------------------------------------------------------- */
void	elpCancelCallAt(btcallback callback, void *ref);
/* This function cancels a delayed call registered with
 * 'elpCallAtCancelPrev()'. The arguments 'callback' and 'ref' have to match
 * those given in 'elpCallAtCancelPrev()'.
 */
/* ------------------------------------------------------------------------- */
void	elpCallCancelPrev(btcallback callback, void *ref);
/* This function registers 'callback' to be called with argument 'ref'
 * immediately from the next eventloop cycle. if the pair 'callback' and
 * 'ref' has already been registered, the call is ignored.
 */
/* ------------------------------------------------------------------------- */
void	elpCancelCall(btcallback callback, void *ref);
/* This function cancels a callback scheduled by 'elpCallCancelPrev()' or
 * 'elpCall()'.
 */
/* ------------------------------------------------------------------------- */
void	elpRun(void);
/* This function starts the eventloop. It does not return until
 * 'elpTerminate()' is called. You should register at least one function with
 * one of the elpXXX()-functions before you run the eventloop (it won't call
 * anything and will never return, otherwise).
 */
/* ------------------------------------------------------------------------- */
void	elpTerminate(void);
/* If this function is called (after a call to 'elpRun()'), the eventloop
 * terminates after the current cycle.
 */
/* ------------------------------------------------------------------------- */

#endif	/* __EVNTLOOP_H_INCLUDED__ */
