/*
 * Name: ipc.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-12
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: ipc.c,v 1.1.1.1 1999/04/12 13:56:38 cs Exp $
 */


#include "stdheaders.h"
#include "ipc.h"
#include "memalloc.h"

/* ------------------------------------------------------------------------- */

ipcMsg_t	*ipcMsgNew(void)
{
ipcMsg_t	*this = NULL;

	this = memCalloc(sizeof(ipcMsg_t));
	this->allocated = 16;
	this->buffer = memAlloc(this->allocated);
	return	this;
}

void	ipcMsgFree(ipcMsg_t *this)
{
	if(this != NULL){
		if(this->buffer != NULL)
			memFree(this->buffer);
		memFree(this);
	}
}

void	ipcMsgClear(ipcMsg_t *this)
{
	this->size = 0;
	this->index = 0;
}

void	ipcMsgEnsureSize(ipcMsg_t *this, btsize bytesNeeded)
{
btsize	newAlloc;

	if(bytesNeeded > this->allocated){
		newAlloc = this->allocated * 2;
		while(newAlloc < bytesNeeded){
			newAlloc *= 2;
		}
		this->buffer = memRealloc(this->buffer, newAlloc);
		this->allocated = newAlloc;
	}
}

static inline void	ipcMsgMakeSpace(ipcMsg_t *this, btsize bytesNeeded)
{
	bytesNeeded += this->size;
	if(bytesNeeded > this->allocated){
		ipcMsgEnsureSize(this, bytesNeeded);
	}
}

void	ipcAddInt(ipcMsg_t *this, btu32 val)
{
	ipcMsgMakeSpace(this, 4);
	this->buffer[this->size++] = val;
	this->buffer[this->size++] = val >> 8;
	this->buffer[this->size++] = val >> 16;
	this->buffer[this->size++] = val >> 24;
}

void	ipcAddBlock(ipcMsg_t *this, btu8 *block, btsize len)
{
	ipcMsgMakeSpace(this, len + 4);
	if(block != NULL){
		ipcAddInt(this, len);
		memcpy(&this->buffer[this->size], block, len);
		this->size += len;
	}else
		ipcAddInt(this, -1);
}

void	ipcAddAscii(ipcMsg_t *this, btu8 *asciiString)
{
btsize	len = 0;

	if(asciiString != NULL)
		len = strlen(asciiString) + 1;
	ipcAddBlock(this, asciiString, len);
}


btu32	ipcGetInt(ipcMsg_t *this)
{
btu32	x;

	if(this->index + 4 > this->size){
		return 0;	/* error condition */
	}
	x = this->buffer[this->index++];
	x |= (btu32)this->buffer[this->index++] << 8;
	x |= (btu32)this->buffer[this->index++] << 16;
	x |= (btu32)this->buffer[this->index++] << 24;
	return x;
}

btsize	ipcGetBlock(ipcMsg_t *this, btu8 **block)
{
btsize	len;

	len = ipcGetInt(this);
	if(len == (btsize)-1){	/* null block */
		*block = NULL;
		len = 0;
	}else{
		if(this->index + len > this->size){
			len = this->size - this->index;
		}
		*block = &this->buffer[this->index];
		this->index += len;
	}
	return len;
}

btu8	*ipcGetAscii(ipcMsg_t *this)
{
btu8	*s;

	if(ipcGetBlock(this, &s) == 0 && s != NULL)
		return "";
	return s;
}


/* ------------------------------------------------------------------------- */

