/*
 * Name: scanner.c
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1996-10-04
 * Tabsize: 4
 * Copyright: (c) 1996, 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: scanner.c,v 1.2 1999/08/01 12:57:21 cs Exp $
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include "scanner.h"
#include "memalloc.h"

/* ------------------------------------------------------------------------- */

static inline btbool	is_alphanum(btu8 c)
{
	if(isalnum(c)){
		return 1;
	}else if(c == '_'){
		return 1;
	}else{
		return 0;
	}
}

/* ------------------------------------------------------------------------- */

scanner_t	*scannerNew(int (*getc_prog)(), void *arg)
{
scanner_t	*volatile p;

EXC_TRY
	p = memCalloc(sizeof(struct scanner));
	p->getc_prog = getc_prog;
	p->arg = arg;
	p->ungetc_val = 0;
	p->ungetc_valid = 0;
	p->allow_crem = p->allow_cpprem = p->allow_shrem = 1;
	p->buffer_len = 256;
	p->token_buffer = memAlloc(p->buffer_len);
	p->buffer_i = 0;
	p->line = 1;
	p->unget_token = 0;
	p->unget_valid = 0;
	p->nl_is_token = 0;
EXC_CATCH
	scannerFree(p);
	EXC_RETHROW();
EXC_ENDTRY
	return p;
}

/* ------------------------------------------------------------------------- */

void	scannerFree(scanner_t *this)
{
	if(this != NULL){
		if(this->token_buffer != NULL)
			memFree(this->token_buffer);
		memFree(this);
	}
}

/* ------------------------------------------------------------------------- */

void	scannerSetAllowRem(scanner_t *this, btbool crem, btbool cpprem, btbool shrem)
{
	this->allow_crem = crem;
	this->allow_cpprem = cpprem;
	this->allow_shrem = shrem;
}

/* ------------------------------------------------------------------------- */

void	scannerSetGetc(scanner_t *this, int (*getc_prog)(void *arg), void *arg)
{
	this->getc_prog = getc_prog;
	this->arg = arg;
}

/* ------------------------------------------------------------------------- */

void	scannerSetNlIsToken(scanner_t *this, btbool is_token)
{
	this->nl_is_token = is_token;
}

/* ------------------------------------------------------------------------- */

static inline int	scanner_getc(scanner_t *this)
{
int		c;

	if(this->ungetc_valid){
		this->ungetc_valid = 0;
		return this->ungetc_val;
	}else{
		c = (this->getc_prog)(this->arg);
		if(c == '\n')
			this->line++;
		return c;
	}
}

/* ------------------------------------------------------------------------- */

static inline void	scanner_ungetc(scanner_t *this, int c)
{
	this->ungetc_val = c;
	this->ungetc_valid = 1;
}

/* ------------------------------------------------------------------------- */

static inline void	buffer_putc(scanner_t *this, btu8 c)
{
	if(this->buffer_i >= this->buffer_len){
		this->buffer_len *= 2;
		this->token_buffer = memRealloc(this->token_buffer, this->buffer_len);
	}
	this->token_buffer[this->buffer_i++] = c;
}

/* ------------------------------------------------------------------------- */

static void	scan_alnum(scanner_t *this)
{
int	c;

	for(;;){
		if((c = scanner_getc(this)) < 0){
			break;
		}
		if(is_alphanum(c)){
			buffer_putc(this, c);
		}else{
			scanner_ungetc(this, c);
			break;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	scan_string(scanner_t *this, btu8 termchar)
{
int				c, n, i;
btbool			escmode = 0;

	for(;;){
		if((c = scanner_getc(this)) < 0){
			break;
		}
		if(escmode){
			switch(c){
				case 'a':	buffer_putc(this, '\a');	break;
				case 'e':	buffer_putc(this, '\33');	break;
				case 'f':	buffer_putc(this, '\f');	break;
				case 'n':	buffer_putc(this, '\n');	break;
				case 'r':	buffer_putc(this, '\r');	break;
				case 't':	buffer_putc(this, '\t');	break;
				case 'b':	buffer_putc(this, '\b');	break;
				case 'v':	buffer_putc(this, '\v');	break;
				default:
					if(isdigit(c)){
						n = c - '0';
						for(i=0;i<2;i++){
							c = scanner_getc(this);
							if(isdigit(c)){
								n = n * 8 + c - '0';
							}else{
								scanner_ungetc(this, c);
								break;
							}
						}
						buffer_putc(this, n);
					}else{
						buffer_putc(this, c);
					}
			}
			escmode = 0;
		}else{
			if(c == termchar){
				break;
			}else if(c == '\\'){
				escmode = 1;
			}else{
				buffer_putc(this, c);
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	scan_cremark(scanner_t *this)
{
int	c;

	for(;;){
		if((c = scanner_getc(this)) < 0){
			break;
		}else if(c == '*'){
			c = scanner_getc(this);
			if(c == '/'){
				break;
			}else{
				scanner_ungetc(this, c);
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	scan_til_eol(scanner_t *this)
{
int	c;

	for(;;){
		c = scanner_getc(this);
		if(c < 0 || c == '\n'){
			break;
		}
	}
}

/* ------------------------------------------------------------------------- */

int		scannerGetLineNumber(scanner_t *this)
{
	return this->line;
}

/* ------------------------------------------------------------------------- */

int		scannerGetToken(scanner_t *this, btu8 **value)
{
int		c, type = 0;

	if(this->unget_valid){
		this->unget_valid = 0;
		*value = this->token_buffer;
		return this->unget_token;
	}
	this->buffer_i = 0;		/* clear buffer */
	while(type == 0){
		c = scanner_getc(this);
		if(!isspace(c) || (this->nl_is_token && c == '\n')){
			if(is_alphanum(c)){
				buffer_putc(this, c);
				scan_alnum(this);
				type = SCAN_TOKEN_ALNUM;
			}else if(c < 0){
				type = SCAN_TOKEN_EOF;
			}else{
				switch(c){
				case '\"':
					scan_string(this, '\"');
					type = SCAN_TOKEN_STRING;
					break;
				case '\'':
					scan_string(this, '\'');
					type = SCAN_TOKEN_CHAR;
					break;
				case '/':
					c = scanner_getc(this);
					if(c == '*' && this->allow_crem){
						scan_cremark(this);
					}else if(c == '/' && this->allow_cpprem){
						scan_til_eol(this);
					}else{
						scanner_ungetc(this, c);
						buffer_putc(this, '/');
						type = SCAN_TOKEN_OPERATOR;
					}
					break;
				case '#':
					if(this->allow_shrem){
						scan_til_eol(this);
					}else{
						buffer_putc(this, '#');
						type = SCAN_TOKEN_OPERATOR;
					}
					break;
				default:
					buffer_putc(this, c);
					type = SCAN_TOKEN_OPERATOR;
					break;
				}
			}
		}
	}
	buffer_putc(this, 0);
	*value = this->token_buffer;
	return type;
}

/* ------------------------------------------------------------------------- */

void	scannerUngetToken(scanner_t *this, int token)
{
	this->unget_token = token;
	this->unget_valid = 1;
}

/* ------------------------------------------------------------------------- */

#ifdef TEST_SCANNER

int	main(void)
{
scanner_t	*s = scannerNew(fgetc, stdin);
btu8		*text;
int			token;

	for(;;){
		token = scannerGetToken(s, &text);
		if(token == SCAN_TOKEN_EOF)
			break;
		printf("scanned token type %d: ->%s<-\n", token, text);
	}
	return 0;
}

#endif

/* ------------------------------------------------------------------------- */

