/*
 * Name: convert.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-10-22
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: unicode.c,v 1.1.1.1 1999/04/12 13:56:38 cs Exp $
 */


#include "stdheaders.h"
#include "memalloc.h"
#include "unicode.h"
#include "logging.h"
#include "util.h"

/* ------------------------------------------------------------------------- */

typedef struct	uniTable{
	btuni			codeBase;
	btsize			codeEnd;	/* first char that is not part of table */
	uniProperty_t	*tableData;
}uniTable_t;

typedef void	*asciiToUniEntry_t;

/* ------------------------------------------------------------------------- */

static btuni	codepage437[128] = {	/* this is the default codepage for the server */
	0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
	0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
	0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
	0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
	0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
	0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
	0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
	0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
	0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x255e, 0x255f,
	0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256c, 0x2567,
	0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256b,
	0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
	0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x00b5, 0x03c4,
	0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
	0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
	0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0,
};

/* ------------------------------------------------------------------------- */
/* ------------------------ unicode initialisation ------------------------- */
/* ------------------------------------------------------------------------- */

static btuni	serverUniErrChar, clientUniErrChar;
static btu8		serverAsciiErrChar[UNI_MAX_ASCIIMAP];
static btu8		clientAsciiErrChar[UNI_MAX_ASCIIMAP];
static btsize	serverAsciiErrLen, clientAsciiErrLen;

static uniTable_t			*uniTables;		/* tables representing unicode entities */
static btsize				uniTableCount;	/* number of tables for representation */
static asciiToUniEntry_t	serverDecode[256];	/* ascii to unicode conversion table for server */
static asciiToUniEntry_t	clientDecode[256];	/* ascii to unicode conversion table for client */

/* ------------------------------------------------------------------------- */

static inline btbool	isTouched(uniProperty_t *uniTable, btsize index)
{
	if(uniTable[index].serverRep[0] != 0)
		return 1;
	if(uniTable[index].clientRep[0] != 0)
		return 1;
	if(uniTable[index].upcase != index)
		return 1;
	if(uniTable[index].lowcase != index)
		return 1;
	return 0;
}

struct gap{
	btsize	base;
	btsize	count;
};

static int gapCompare(const void *e1, const void *e2)
{
const struct gap	*g1 = e1, *g2 = e2;

	if(g1->base < g2->base)
		return -1;
	if(g1->base > g2->base)
		return 1;
	return 0;
}

static void	compressTable(uniProperty_t *uniTable, btsize maxBlocks)
{
btsize		base = 0, count = 0, i, j, k;
struct gap	*gaps;

	if(uniTables != NULL){
		for(i=0;i<uniTableCount;i++){
			if(uniTables[i].tableData != NULL)
				memFree(uniTables[i].tableData);
		}
		memFree(uniTables);
	}
	uniTables = memCalloc(maxBlocks * sizeof(uniTable_t));
	gaps = memCalloc((maxBlocks + 1) * sizeof(*gaps));
	for(i=1;i<maxBlocks+1;i++)	/* leave one null-gap at the beginning */
		gaps[i].base = 65536;	/* dummy-gaps that will be sorted to the end */
	for(i=0;i<65536;i++){
		if(isTouched(uniTable, i)){
			if(count > 0){
				for(j=0;j<maxBlocks;j++){
					/* Insert before next smaller gap, throwing out the smallest gaps
					 * at the end of the list. Only replace null-gap at j==0 (the
					 * one with base=0, count=0) with a larger base=0 gap.
					 */
					if(count > gaps[j].count && (j != 0 || base == 0)){
						for(k=maxBlocks-1;k>j;k--){
							gaps[k] = gaps[k-1];
						}
						gaps[j].count = count;
						gaps[j].base = base;
						break;
					}
				}
			}
			base = i + 1;
			count = 0;
		}else{
			count++;
		}
	}
	gaps[maxBlocks].base = base;	/* last gap starts one behind the last touched char */
	gaps[maxBlocks].count = 65536 - base;
	qsort(gaps, maxBlocks + 1, sizeof(*gaps), gapCompare);
	i=0;
	for(j=1;j<maxBlocks+1 && i<maxBlocks;j++){
		uniTables[i].codeBase = gaps[j-1].base + gaps[j-1].count;
		uniTables[i].codeEnd = gaps[j].base;
		if(uniTables[i].codeEnd - uniTables[i].codeBase > 0)
			i++;
	}
	uniTableCount = i;
	memFree(gaps);
	for(i=0;i<uniTableCount;i++){
		count = (uniTables[i].codeEnd - uniTables[i].codeBase) * sizeof(uniProperty_t);
		uniTables[i].tableData = memAlloc(count);
		memcpy(uniTables[i].tableData, &uniTable[uniTables[i].codeBase], count);
	}
}

static inline void	reverseTableInit(asciiToUniEntry_t *table, btbool isServer)
{
btsize	i;

	for(i=0;i<256;i++){
		table[i] = (void *)(((isServer ? serverUniErrChar : clientUniErrChar) << 1) + 1);
	}
}

static void	freeReverseTable(asciiToUniEntry_t *table, btbool freeRoot)
{
int		i;

	for(i=0;i<256;i++){
		if(table[i] != NULL && ((int)table[i] & 1) == 0){
			freeReverseTable(table[i], 1);
			table[i] = NULL;
		}
	}
	if(freeRoot)
		memFree(table);
}

static void	calcReverseTables(btbool isServer, asciiToUniEntry_t *rootTable)
{
btsize				i;
btu8				buffer[UNI_MAX_ASCIIMAP], *p, *q;
asciiToUniEntry_t	*table;
btbool				mapOK;

	freeReverseTable(rootTable, 0);
	reverseTableInit(rootTable, isServer);
	for(i=0;i<65536;i++){
		q = buffer;
		if(isServer)
			mapOK = convCharUniToServerAscii(i, &q);
		else
			mapOK = convCharUniToLocalAscii(i, &q);
		if(mapOK && q - buffer > 0){
			table = rootTable;
			for(p=buffer;p<q-1;p++){
				if(((int)(table[*p]) & 1) != 0){
					table[*p] = memAlloc(256 * sizeof(asciiToUniEntry_t));
					reverseTableInit(table[*p], isServer);
				}
				table = table[*p];
			}
			if(((int)(table[*p]) & 1) != 0){
			/* don't overwrite buffers with single characters! */
				table[*p] = (void *)((i << 1) + 1);
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	trivialSetup(uniProperty_t *uniTable, btsize size, btbool withCp437)
{
btsize			i;

	if(withCp437){
		for(i=0;i<128;i++){	/* server is initialized to codepage 437 */
			if(codepage437[i] < size){
				uniTable[codepage437[i]].serverRep[0] = i + 128;
				uniTable[codepage437[i]].serverRep[1] = 0;
			}
		}
	}
	for(i=0;i<size;i++){
		if((i >= 0x61 && i <= 0x7a) || (i >= 0xe0 && i <= 0xfe && i != 0xf7))
			uniTable[i].upcase = i - 0x20;
		else
			uniTable[i].upcase = i;
		if((i >= 0x41 && i <= 0x5a) || (i >= 0xc0 && i <= 0xde && i != 0xd7))
			uniTable[i].lowcase = i + 0x20;
		else
			uniTable[i].lowcase = i;
		if(i < 256){
			uniTable[i].clientRep[0] = i;
			uniTable[i].clientRep[1] = 0;
		}
		if(i < 128){
			uniTable[i].serverRep[0] = i;
			uniTable[i].serverRep[1] = 0;
		}
	}
}

/* ------------------------------------------------------------------------- */

void	uniTrivialInit(void)
{
uniProperty_t	*uniTable = NULL;
btsize			i;

	if(uniTables != NULL){
		for(i=0;i<uniTableCount;i++){
			if(uniTables[i].tableData != NULL){
				memFree(uniTables[i].tableData);
				uniTables[i].tableData = NULL;
			}
		}
	}else{
		uniTables = memCalloc(sizeof(uniTable_t));
	}
	uniTableCount = 1;
	/* set up some reasonable defaults */
	uniTables->codeBase = 0;
	uniTables->codeEnd = 256;
	uniTables->tableData = uniTable = memCalloc(256 * sizeof(uniProperty_t));
	trivialSetup(uniTable, 256, 1);
	serverUniErrChar = clientUniErrChar = serverAsciiErrChar[0] = clientAsciiErrChar[0] = '?';
	serverAsciiErrLen = clientAsciiErrLen = 1;
	calcReverseTables(0, clientDecode);
	calcReverseTables(1, serverDecode);
}

/* ------------------------------------------------------------------------- */

void	uniInit(btu8 *caseConvertPath, btu8 *serverMapPath, btu8 *clientMapPath, btsize maxBlocks)
{
uniProperty_t	*uniTable = NULL;

	uniTable = memCalloc(65536 * sizeof(uniProperty_t));
	trivialSetup(uniTable, 65536, 0);
	if(caseConvertPath != NULL && !uniParseCaseConvertFile(caseConvertPath, uniTable))
		goto errorOccured;
	if(serverMapPath != NULL && !uniParseMappingFile(serverMapPath, 1, uniTable, &serverUniErrChar, serverAsciiErrChar, &serverAsciiErrLen))
		goto errorOccured;
	if(clientMapPath != NULL && !uniParseMappingFile(clientMapPath, 0, uniTable, &clientUniErrChar, clientAsciiErrChar, &clientAsciiErrLen))
		goto errorOccured;
	compressTable(uniTable, maxBlocks);
	memFree(uniTable);
	calcReverseTables(0, clientDecode);
	calcReverseTables(1, serverDecode);
	return;
errorOccured:
	memFree(uniTable);
	if(LOG_MODE_IS_ACTIVE(LOGM_WARN))
		logPrintf(LOGM_WARN, "unicode: error parsing tables, using defaults\n");
	uniTrivialInit();
}

/* ------------------------------------------------------------------------- */
/* -------------------------- unicode operations --------------------------- */
/* ------------------------------------------------------------------------- */

btuni	*uniStrDup(btuni *s)
{
btsize	len;
btuni	*new;

	if(s == NULL)
		return NULL;
	len = uniStrLen(s) + 1;
    new = memAlloc(len * sizeof(btuni));
    memcpy(new, s, len * sizeof(btuni));
    return new;
}

void	uniStrCpy(btuni *dest, btuni *src, btsize arrSize)
{
int		i;

	arrSize--;
	for(i=0; src[i] != 0 && i < arrSize; i++){
		dest[i] = src[i];
	}
	dest[i] = 0;
}

void	asciiStrCpy(btu8 *dest, btu8 *src, btsize arrSize)
{
int		i;

	arrSize--;
	for(i=0; src[i] != 0 && i < arrSize; i++){
		dest[i] = src[i];
	}
	dest[i] = 0;
}

void	uniStrAddHash(btu32 *crcp, btuni *s, btbool isCaseSensitive)
{
btu32	crc = *crcp;
btuni	c;

	if(isCaseSensitive){
		while(*s){
			c = *s;
			crc = utilCrc32table[(crc ^ c) & 0xff] ^ (crc >> 8);
			s++;
		}
	}else{
		while(*s){
			c = uniCharToupper(*s);
			crc = utilCrc32table[(crc ^ c) & 0xff] ^ (crc >> 8);
			s++;
		}
	}
	*crcp = crc;
}

static inline uniProperty_t	*uniLookup(btuni uniChar)
{
btsize	i;

	for(i=0;i<uniTableCount;i++){
		if(uniChar >= uniTables[i].codeBase && uniChar < uniTables[i].codeEnd){
			return &uniTables[i].tableData[uniChar - uniTables[i].codeBase];
		}
	}
	return NULL;
}

btuni	uniCharToupper(btuni c)
{
uniProperty_t	*entry;

	if((entry = uniLookup(c)) == NULL)
		return c;
	return entry->upcase;
}

btuni	uniCharTolower(btuni c)
{
uniProperty_t	*entry;

	if((entry = uniLookup(c)) == NULL)
		return c;
	return entry->lowcase;
}

btu8	asciiCharToupper(btu8 c)
{
	return uniCharToupper(c);	/* this is not strictly correct... */
}

btu8	asciiCharTolower(btu8 c)
{
	return uniCharTolower(c);	/* this is not strictly correct... */
}

/* ------------------------------------------------------------------------- */
/* --------------------------- basic conversion ---------------------------- */
/* ------------------------------------------------------------------------- */

static inline btuni	convCharAnyToUni(btu8 **s, asciiToUniEntry_t *table)
{
register void	*p;
register btu8	c;

	for(;;){
		c = *(*s)++;
		if(c == 0)		/* always map terminating 0 to 0 */
			return 0;
		p = table[c];
		if(((int)p & 1) != 0){
			return (int)p >> 1;
		}else{
			table = p;
		}
	}
}

btuni	convCharLocalAsciiToUni(btu8 **string)
{
	return convCharAnyToUni(string, clientDecode);
}

btuni	convCharServerAsciiToUni(btu8 **string)
{
	return convCharAnyToUni(string, serverDecode);
}

btbool	convCharUniToLocalAscii(btuni c, btu8 **string)
{
uniProperty_t	*entry;
btsize			i;

	if((entry = uniLookup(c)) == NULL || entry->clientRep[0] == 0){
		memcpy(*string, clientAsciiErrChar, clientAsciiErrLen);
		*string += clientAsciiErrLen;
		return 0;	/* mapping had an error */
	}else{
		for(i=0;i<UNI_MAX_ASCIIMAP;i++){
			if(entry->clientRep[i] == 0)
				break;
			*(*string)++ = entry->clientRep[i];
		}
		return 1;	/* mapping was OK */
	}
}

btbool	convCharUniToServerAscii(btuni c, btu8 **string)
{
uniProperty_t	*entry;
btsize			i;

	if((entry = uniLookup(c)) == NULL || entry->serverRep[0] == 0){
		memcpy(*string, serverAsciiErrChar, serverAsciiErrLen);
		*string += serverAsciiErrLen;
		return 0;	/* mapping had an error */
	}else{
		for(i=0;i<UNI_MAX_ASCIIMAP;i++){
			if(entry->serverRep[i] == 0)
				break;
			*(*string)++ = entry->serverRep[i];
		}
		return 1;	/* mapping was OK */
	}
}

/* ------------------------------------------------------------------------- */
/* -------------------------- derived conversion --------------------------- */
/* ------------------------------------------------------------------------- */

void	convLocalAsciiToUni(btu8 *ascii, btuni *uni, btsize arrSize)
{
int		i;

	arrSize--;	/* reserve one char for terminating null */
	for(i=0; *ascii != 0 && i < arrSize; i++){
		uni[i] = convCharLocalAsciiToUni(&ascii);
	}
	uni[i] = 0;
}

void	convServerAsciiToUni(btu8 *ascii, btuni *uni, btsize arrSize)
{
int		i;

	arrSize--;	/* reserve one char for terminating null */
	for(i=0; *ascii != 0 && i < arrSize; i++){
		uni[i] = convCharServerAsciiToUni(&ascii);
	}
	uni[i] = 0;
}

void	convUniToLocalAscii(btuni *uni, btu8 *ascii, btsize arrSize)
{
int		i;

	arrSize -= UNI_MAX_ASCIIMAP;
	for(i=0; uni[i] != 0 && i < arrSize; i++){
		convCharUniToLocalAscii(uni[i], &ascii);
	}
	*ascii = 0;
}

void	convUniToServerAscii(btuni *uni, btu8 *ascii, btsize arrSize)
{
int		i;

	arrSize -= UNI_MAX_ASCIIMAP;
	for(i=0; uni[i] != 0 && i < arrSize; i++){
		convCharUniToServerAscii(uni[i], &ascii);
	}
	*ascii = 0;
}

/* ------------------------------------------------------------------------- */

btu8	*convUniToAscii(btuni *uni)
{
static btu8	rval[512];	/* this should be sufficient for debug prints */

	convUniToLocalAscii(uni, rval, sizeof(rval));
	return rval;
}

/* ------------------------------------------------------------------------- */
