/*
 * Name: nfs2.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-04
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfs2.c,v 1.7 1999/07/26 20:06:59 cs Exp $
 */

/*
General Description:
This is the NFS2 frontend module. It implements the frontend interface.
*/

#include "stdheaders.h"
#include "memalloc.h"
#include "hashtable.h"
#include "logging.h"
#include "system.h"
#include "sunrpc.h"
#include "blockpool.h"
#include "process.h"
#include "uidsecure.h"
#include "util.h"
#include "nfs2const.h"
#include "main.h"
#include "cfgparse.h"
#include "evntloop.h"
#include "nfsfileadapter.h"
#include "notification.h"
#include "frontend.h"
#include "backend.h"
#include "errcodes.h"
#include "unicode.h"
#include "module.h"
#include "mainmodule.h"
#include "../../uiLib/maininterface.h"
#include "mainconst.h"


/* ------------------------------------------------------------------------- */

typedef struct backendImp{
	btint	(*vfsConfigPath)(void *obj, struct cfgProperty **path, btint maxEntries);
	void	(*vfsRetain)(void *obj);
	void	(*vfsRelease)(void *obj);
	btuni	*(*vfsPath)(void *obj);
	void	(*userAuthenticate)(void *vfs, btint unixUid, void *caller);
	void	(*userRetain)(void *obj);
	void	(*userRelease)(void *obj);
}backendImp_t;

typedef struct nfs2{
	btu32					mountId;		/* unique ID for this mount, first entry for hashtable search */
	btu8					*mountpoint;
	backendImp_t			backend;
	struct module			*backendMod;
	void					*vfs;
	struct nfsFileAdapter	*adapter;

	btint					refCount;
	btint					mountedBy;
	btbool					isFakeMode;
	btbool					isReadonly;
	btbool					isMounted;
}nfs2_t;

/* ------------------------------------------------------------------------- */

typedef struct nfsCtx{
    PROC_VARSET;
    nfs2_t			*nfsMp;
	srpcMsg_t		*msg;
	btint			retryCount;
	void			*user;
/* context is cleared only until here */
	bttime			startTime;
    union {
        nfs_fh		getattr;
        sattrargs	setattr;
        diropargs	lookup;
        nfs_fh		readlink;
        readargs	read;
        writeargs	write;
        createargs	create;
        diropargs	remove;
        renameargs	rename;
        linkargs	link;
        symlinkargs	symlink;
        createargs	mkdir;
        diropargs	rmdir;
        readdirargs	readdir;
        nfs_fh		statfs;
    }				argument;

/* reply context: */
	srpcRequestHeader_t	requestHdr;
    struct sockaddr_in	remoteAddr;
    int					raddrLen;
}nfsCtx_t;

static blockpool_t	*nfsPool = NULL;

static void	freeNfsBlock(void *block)
{
    poolReturnBlock(nfsPool, block);
}

/* ------------------------------------------------------------------------- */

btint	nfs2OperationsRunning = 0;
/* nfs2OperationsRunning is used for debugging purposes. In idle mode, it
 * should go back to 0. If not, an operation got stuck, probably due to an
 * internal deadlock. The state of this variable can be checked by means
 * of the debugger.
 */


static btint				nfsSock = -1;	/* socket for NFS RPC */
static struct sockaddr_in	nfsSockAddr;
static struct hashtable		*runningRequests;
static btu16				mountCounter;
static btu16				incarnation;
static struct hashtable		*mountsById;

/* ------------------------------------------------------------------------- */

static void	nfsMpRetain(void *obj);
static void	nfsMpRelease(void *obj);
static void	nfsMpFree(nfs2_t *mp);
static btint	nfs2ConfigPath(nfs2_t *mp, struct cfgProperty **path, btint maxEntries);

/* ------------------------------------------------------------------------- */
/* ----------------- printing of argument and result types ----------------- */
/* ------------------------------------------------------------------------- */

static void	printFilename(filename *arg)
{
    logPrintf(LOGM_NFSTRACE, "filename: %s\n", *arg);
}

static void	printNfs_fh(void *arg)
{
btu32	*fh = arg;

    logPrintf(LOGM_NFSTRACE, "nfs_fh: 0x%08x 0x%08x 0x%08x 0x%08x\n", fh[0], fh[1], fh[2], fh[3]);
}

static void	printDiropargs(diropargs *arg)
{
    printNfs_fh(&arg->dir);
    printFilename(&arg->name);
}

static void	printNfstime(nfstime *arg)
{
time_t	tm = arg->seconds;
    
    logPrintf(LOGM_NFSTRACE, "nfstime: %s", ctime(&tm));
}

static void	printSattr(sattr *arg)
{
    logPrintf(LOGM_NFSTRACE, "sattr: mode=0%011o uid=%d gid=%d size=%d\n", arg->mode, arg->uid, arg->gid, arg->size);
    printNfstime(&arg->atime);
    printNfstime(&arg->mtime);
}

static void	printCreateargs(createargs *arg)
{
    printDiropargs(&arg->where);
    printSattr(&arg->attributes);
}

static void	printLinkargs(linkargs *arg)
{
    printNfs_fh(&arg->from);
    printDiropargs(&arg->to);
}

static void	printReadargs(readargs *arg)
{
    printNfs_fh(&arg->file);
    logPrintf(LOGM_NFSTRACE, "offset=%d count=%d totalcount=%d\n", arg->offset, arg->count, arg->totalcount);
}

static void	printReaddirargs(readdirargs *arg)
{
    printNfs_fh(&arg->dir);
    logPrintf(LOGM_NFSTRACE, "cookie=%d count=%d\n", ntohl(*(btu32 *)arg->cookie), arg->count);
}

static void	printRenameargs(renameargs *arg)
{
    printDiropargs(&arg->from);
    printDiropargs(&arg->to);
}

static void	printSattrargs(sattrargs *arg)
{
    printNfs_fh(&arg->file);
    printSattr(&arg->attributes);
}

static void	printSymlinkargs(symlinkargs *arg)
{
    printDiropargs(&arg->from);
    printFilename(&arg->to);
    printSattr(&arg->attributes);
}

static void	printWriteargs(writeargs *arg)
{
    printNfs_fh(&arg->file);
    logPrintf(LOGM_NFSTRACE, "beginoffset=%d offset=%d totalcount=%d data_len=%d\n", arg->beginoffset, arg->offset, arg->totalcount, arg->data.data_len);
}

static void	printVoid(void *arg)
{
    logPrintf(LOGM_NFSTRACE, "none\n");
}

static void	printNfsstat(nfsstat *arg)
{
char	*e, err[256];
    
    switch(*arg){
        case NFS_OK:			e = "NFS_OK";			break;
        case NFSERR_PERM:		e = "NFSERR_PERM";		break;
        case NFSERR_NOENT:		e = "NFSERR_NOENT";		break;
        case NFSERR_IO:			e = "NFSERR_IO";		break;
        case NFSERR_NXIO:		e = "NFSERR_NXIO";		break;
        case NFSERR_ACCES:		e = "NFSERR_ACCES";		break;
        case NFSERR_EXIST:		e = "NFSERR_EXIST";		break;
        case NFSERR_NODEV:		e = "NFSERR_NODEV";		break;
        case NFSERR_NOTDIR:		e = "NFSERR_NOTDIR";	break;
        case NFSERR_ISDIR:		e = "NFSERR_ISDIR";		break;
        case NFSERR_FBIG:		e = "NFSERR_FBIG";		break;
        case NFSERR_NOSPC:		e = "NFSERR_NOSPC";		break;
        case NFSERR_ROFS:		e = "NFSERR_ROFS";		break;
        case NFSERR_NAMETOOLONG:e = "NFSERR_NAMETOOLONG";break;
        case NFSERR_NOTEMPTY:	e = "NFSERR_NOTEMPTY";	break;
        case NFSERR_DQUOT:		e = "NFSERR_DQUOT";		break;
        case NFSERR_STALE:		e = "NFSERR_STALE";		break;
        case NFSERR_WFLUSH:		e = "NFSERR_WFLUSH";	break;
        default:	e = err;	sprintf(err, "[%d]", *arg);	break;
    }
    logPrintf(LOGM_NFSTRACE, "nfsstat: %s\n", e);
}

static void	printFattr(fattr *arg)
{
    logPrintf(LOGM_NFSTRACE, "fattr: type=%d mode=0%011o uid=%d gid=%d size=%d\n", arg->type, arg->mode, arg->uid, arg->gid, arg->size);
    logPrintf(LOGM_NFSTRACE, "       nlink=%d blocksize=%d rdev=0x%x blocks=%d fsid=%d fileid=%d\n", arg->nlink, arg->blocksize, arg->rdev, arg->blocks, arg->fsid, arg->fileid);
    printNfstime(&arg->atime);
    printNfstime(&arg->mtime);
    printNfstime(&arg->ctime);
}

static void	printAttrstat(attrstat *arg)
{
    printNfsstat(&arg->status);
    if(arg->status == NFS_OK){
    	printFattr(&arg->attrstat_u.attributes);
    }
}

static void	printDiropres(diropres *arg)
{
    printNfsstat(&arg->status);
    if(arg->status == NFS_OK){
    	printNfs_fh(&arg->diropres_u.diropres.file);
    	printFattr(&arg->diropres_u.diropres.attributes);
    }
}

static void	printEntry(entry *arg)
{
    if(arg != NULL){
        logPrintf(LOGM_NFSTRACE, "entry: fileid=%d cookie=%d\n", arg->fileid, ntohl(*(btu32 *)arg->cookie));
        printFilename(&arg->name);
        printEntry(arg->nextentry);
    }
}

static void	printReaddirres(readdirres *arg)
{
    printNfsstat(&arg->status);
    if(arg->status == NFS_OK){
        printEntry(arg->readdirres_u.reply.entries);
    	logPrintf(LOGM_NFSTRACE, "eof: %d\n", arg->readdirres_u.reply.eof);
    }
}

static void	printReadlinkres(readlinkres *arg)
{
    printNfsstat(&arg->status);
    if(arg->status == NFS_OK){
    	printFilename(&arg->readlinkres_u.data);
    }
}

static void	printReadres(readres *arg)
{
    printNfsstat(&arg->status);
    if(arg->status == NFS_OK){
    	printFattr(&arg->readres_u.reply.attributes);
    	logPrintf(LOGM_NFSTRACE, "data_len=%d\n", arg->readres_u.reply.data.data_len);
    }
}

static void	printStatfsres(statfsres *arg)
{
    printNfsstat(&arg->status);
    if(arg->status == NFS_OK){
    	logPrintf(LOGM_NFSTRACE, "statfsokres: tsize=%d bsize=%d blocks=%d bfree=%d bavail=%d\n", arg->statfsres_u.reply.tsize, arg->statfsres_u.reply.bsize, arg->statfsres_u.reply.blocks, arg->statfsres_u.reply.bfree, arg->statfsres_u.reply.bavail);
    }
}

/* ------------------------------------------------------------------------- */
/* -------------------------- function dispatcher -------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct nfsProc {
	void	(*encodeResult)();	/* result marshalling routine */
	void	(*decodeArg)();		/* argument unmarshalling routine */
    char	*name;				/* function name */
    void	(*printArg)(void *);/* print argument for debug */
    void	(*printRes)(void *);/* print result for debug */
}nfsProc_t;

#define NFSPROC_ENTRY(resType, function, argType) {	\
	srpcEncode##resType,		\
	srpcDecode##argType,		\
	#function,					\
	(void *)print##argType,		\
	(void *)print##resType		\
}

static void	srpcEncodeVoid(){}
static void	srpcDecodeVoid(){}

static nfsProc_t	dispatchTable[] = {
	NFSPROC_ENTRY(Void, nfsNull, Void),					/* NULL */
	NFSPROC_ENTRY(Attrstat, nfsGetattr, Nfs_fh),		/* GETATTR */
	NFSPROC_ENTRY(Attrstat, nfsSetattr, Sattrargs),		/* SETATTR */
	NFSPROC_ENTRY(Void, nfsRoot, Void),					/* ROOT */
	NFSPROC_ENTRY(Diropres, nfsLookup, Diropargs),		/* LOOKUP */
	NFSPROC_ENTRY(Readlinkres, nfsReadlink, Nfs_fh),	/* READLINK */
	NFSPROC_ENTRY(Readres, nfsRead, Readargs),			/* READ */
	NFSPROC_ENTRY(Void, nfsWritecache, Void),			/* WRITECACHE */
	NFSPROC_ENTRY(Attrstat, nfsWrite, Writeargs),		/* WRITE */
	NFSPROC_ENTRY(Diropres, nfsCreate, Createargs),		/* CREATE */
	NFSPROC_ENTRY(Nfsstat, nfsRemove, Diropargs),		/* REMOVE */
	NFSPROC_ENTRY(Nfsstat, nfsRename, Renameargs),		/* RENAME */
	NFSPROC_ENTRY(Nfsstat, nfsLink, Linkargs),			/* LINK */
	NFSPROC_ENTRY(Nfsstat, nfsSymlink, Symlinkargs),	/* SYMLINK */
	NFSPROC_ENTRY(Diropres, nfsMkdir, Createargs),		/* MKDIR */
	NFSPROC_ENTRY(Nfsstat, nfsRmdir, Diropargs),		/* RMDIR */
	NFSPROC_ENTRY(Readdirres, nfsReaddir, Readdirargs),	/* READDIR */
	NFSPROC_ENTRY(Statfsres, nfsStatfs, Nfs_fh),		/* STATFS */
};

/* ------------------------------------------------------------------------- */

void	*fakedReply(nfsCtx_t *ctx)
{
static nfsstat		errorRes = NFSERR_ACCES;
static attrstat		attrRes;
static statfsres	statfsRes;
nfs_fh				rootFh;
btinum				inum = 0;

	switch(ctx->requestHdr.proc){
	case NFSPROC_GETATTR:
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
			logPrintf(LOGM_DEBUG, "nfs: faking getattr reply\n");
		memset(&attrRes, 0, sizeof(attrRes));
		if(ctx->nfsMp->adapter != NULL){
			inum = nfsFaRootHandle(ctx->nfsMp->adapter, &rootFh, ctx->nfsMp->mountId);
		}
		if(ctx->nfsMp->adapter != NULL && memcmp(&rootFh, &ctx->argument.getattr, sizeof(rootFh)) == 0){	/* getattr for root dir */
			attrRes.attrstat_u.attributes.type = NFDIR;
			attrRes.attrstat_u.attributes.mode = NFSMODE_DIR | 0444;
			attrRes.attrstat_u.attributes.nlink = 2;
			attrRes.attrstat_u.attributes.blocksize = NFS2CONST_BLOCKSIZE;
			attrRes.attrstat_u.attributes.size = 2 * NFS2CONST_BLOCKSIZE;
			attrRes.attrstat_u.attributes.blocks = 2;
			attrRes.attrstat_u.attributes.fileid = inum;
			attrRes.status = NFS_OK;
		}else{
			attrRes.status = NFSERR_ACCES;
		}
		return &attrRes;
	case NFSPROC_STATFS:
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
			logPrintf(LOGM_DEBUG, "nfs: faking statfs reply\n");
		memset(&statfsRes, 0, sizeof(statfsRes));
		statfsRes.statfsres_u.reply.tsize = NFS2CONST_TRANSFERSIZE;
		statfsRes.statfsres_u.reply.bsize = 32768;
		statfsRes.statfsres_u.reply.blocks = 1;
		statfsRes.statfsres_u.reply.bfree = 0;
		statfsRes.statfsres_u.reply.bavail = 0;
		return &statfsRes;
	default:
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG) || LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE))
			logPrintf(LOGM_DEBUG, "nfs: faking error reply\n");
		return &errorRes;
	}
}

/* ------------------------------------------------------------------------- */

static void	sendreply(nfsCtx_t *this)
{
	if(sendto(nfsSock, this->msg->data, this->msg->size, 0, (struct sockaddr *)&this->remoteAddr, this->raddrLen) < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs: error sending reply: [%d] %s\n", errno, sysStrerror(errno));
	}
}

void	handleRequestProc(void *ctx, void *arg)
{
nfsCtx_t	*this = ctx;
nfsstat		*volatile res = arg;
nfsstat		dummyRes;
beVfsRes_t	*rval = arg;

    PROC_BEGIN
	nfsMpRetain(this->nfsMp);
    if(this->nfsMp->vfs == NULL || this->nfsMp->isFakeMode){
        res = fakedReply(this);
	}else{
		nfs2OperationsRunning++;
		for(this->retryCount = 0; this->retryCount < 2; this->retryCount++){
			/* try to authenticate user before we do anything */
			PROC_WAITFOR_X(1, this->nfsMp->backend.userAuthenticate(this->nfsMp->vfs, this->requestHdr.credUid, this));
			if(rval->err.class == ERRCLS_CIFS_LOCAL && rval->err.code == ERR_CIFS_LOCconnlost){
/* No use to retry if connecting the user failed, because there are two possibilities:
 * 1.: The user was already connected. In this case userAuthenticate() is a no-op.
 * 2.: The user was not connected. In this case the connection to the server has been lost
 *     before or has never been established; it's an initial connect attempt which should
 *     not be retried.
 */
				dummyRes = NFSERR_IO;
				res = &dummyRes;
				break;
			}
			if(rval->err.class != 0){	/* error during authentication */
				dummyRes = NFSERR_ACCES;
				res = &dummyRes;
				break;
			}
			this->user = rval->user;
			this->nfsMp->backend.userRetain(this->user);
			PROC_WAITFOR_X(2, nfsFaExecuteRequest(this->nfsMp->adapter, this->requestHdr.proc, &this->argument, this->user, this));
			this->nfsMp->backend.userRelease(this->user);
			if((btint)*res != -1)
				break;	/* retry if result == -1 which means: connection lost */
			/* get rid of stack: */
			PROC_WAITFOR_X(3, elpCallCancelPrev(handleRequestProc, this));
		}
		if(this->retryCount >= 2){	/* even retry failed */
			dummyRes = NFSERR_IO;
			res = &dummyRes;
		}
		nfs2OperationsRunning--;
	}
    if(LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE)){
        if(res != NULL)
			dispatchTable[this->requestHdr.proc].printRes(res);
        logPrintf(LOGM_NFSTRACE, "*** nfs return[%d].%s\n", this->requestHdr.xid, res == NULL ? " (NULL)" : "");
    }
    if(res == NULL){
		dummyRes = NFSERR_IO;
        res = &dummyRes;	/* we *must* send a reply! */
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG)){
		bttime			duration;
		static bttime	maxDuration;
		duration = elpNow() - this->startTime;
		if(duration > maxDuration){
			maxDuration = duration;
			logPrintf(LOGM_DEBUG, "nfs: maximum request duration: %d ms\n", maxDuration);
		}
	}
	/* we use the same srpcMsg object for call and reply data. This implies
	 * that the result *must not* reference any argument data.
	 */
	this->msg->size = this->msg->readIndex = 0;	/* reset message */
EXC_TRY
	srpcEncodeRpcHeader(this->msg, this->requestHdr.xid, 0, 0);
	dispatchTable[this->requestHdr.proc].encodeResult(this->msg, res);
EXC_CATCH
	this->msg->size = this->msg->readIndex = 0;	/* reset message */
	srpcEncodeRpcHeader(this->msg, this->requestHdr.xid, 0, 4);	/* GARBAGE_ARGS */
EXC_ENDTRY
	sendreply(this);	/* just send for the risk that we block or lose a packet */
	srpcMsgFree(this->msg);
	htabRemoveObjectEqualTo(runningRequests, this);
	nfsMpRelease(this->nfsMp);
    PROC_END
}

/* ------------------------------------------------------------------------- */

static void	sockReadable(void *dummyObj, void *dummyArg)
{
nfsCtx_t	*volatile ctx = NULL;
nfsProc_t	*proc;
btu32		mountId;
nfsstat		dummyRes;

EXC_TRY
    PROC_INIT(ctx, handleRequestProc, NULL, nfsPool, freeNfsBlock);
	ctx->msg = srpcMsgNew();
	ctx->raddrLen = sizeof(ctx->remoteAddr);
	ctx->msg->size = recvfrom(nfsSock, ctx->msg->data, SRPC_MAX_BUFFER, 0, (struct sockaddr *)&ctx->remoteAddr, &ctx->raddrLen);
	if(ctx->msg->size < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs: error receiving: [%d] %s\n", errno, sysStrerror(errno));
		srpcMsgFree(ctx->msg);
		freeNfsBlock(ctx);
		return;
	}
	srpcDecodeRpcHeader(ctx->msg, &ctx->requestHdr);
	if(ctx->requestHdr.rpcvers != 2){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 1, 0);	/* RPC_MISMATCH */
	}else if(ctx->requestHdr.prog != NFS_PROGRAM){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 1);	/* PROG_UNAVAIL */
	}else if(ctx->requestHdr.vers != NFS_VERSION){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 2);	/* PROG_MISMATCH */
	}else if(ctx->requestHdr.proc > NFSPROC_STATFS){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 3);	/* PROC_UNAVAIL */
	}else if(ctx->requestHdr.credFlavor != AUTH_UNIX){
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 1, 1);	/* AUTH_ERROR */
	}else{
		if(htabFindObjectEqualTo(runningRequests, ctx) != NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				logPrintf(LOGM_DEBUG, "*** nfs: detected RPC retry for xid=%d\n", ctx->requestHdr.xid);
			srpcMsgFree(ctx->msg);
			freeNfsBlock(ctx);
			return;	/* just ignore the message */
		}
/* By coincidence (?), all NFS argument structures have an nfs_fh as the first argument.
 * Use this coincidence to find the mountpoint related to the file handle.
 */
		proc = &dispatchTable[ctx->requestHdr.proc];
		proc->decodeArg(ctx->msg, &ctx->argument);
 		mountId = ((btu32 *)&ctx->argument.getattr)[7];
		ctx->nfsMp = htabFindObjectEqualTo(mountsById, &mountId);
		if(ctx->nfsMp == NULL){	/* must be a stale mount, reply with NFSERR_STALE */
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "nfs2: received request for stale mount id=0x%08x xid=%d\n", mountId, ctx->requestHdr.xid);
			ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
			srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 0);
			dummyRes = NFSERR_STALE;
			srpcEncodeNfsstat(ctx->msg, &dummyRes);
		}else{
			htabAddObject(runningRequests, ctx);
			if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
				ctx->startTime = elpNow();
			if(LOG_MODE_IS_ACTIVE(LOGM_NFSTRACE)){
				logPrintf(LOGM_NFSTRACE, "*** nfs command[%d]: %s\n", ctx->requestHdr.xid, proc->name);
				proc->printArg(&ctx->argument);
			}
			PROC_START(ctx, NULL);
			return;
		}
	}
EXC_CATCH
	if(EXC_ISX(srpcExcDecodeOverflow)){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs: error unmarshalling arguments for proc=%d xid=%d\n", ctx->requestHdr.proc, ctx->requestHdr.xid);
		ctx->msg->size = ctx->msg->readIndex = 0;	/* reset message */
		srpcEncodeRpcHeader(ctx->msg, ctx->requestHdr.xid, 0, 4);	/* GARBAGE_ARGS */
		sendreply(ctx);
	}
EXC_ENDTRY
	if(ctx != NULL){
		if(ctx->msg != NULL){
			sendreply(ctx);
			srpcMsgFree(ctx->msg);
		}
		freeNfsBlock(ctx);
	}
}

/* ------------------------------------------------------------------------- */

static btint	ctxXidCompare(void *obj, void *otherObj)
{
nfsCtx_t	*this = obj, *other = otherObj;

	return this->requestHdr.xid - other->requestHdr.xid;
}

static btu32	ctxHashval(void *obj)
{
nfsCtx_t	*ctx = obj;

	return ctx->requestHdr.xid ^ ctx->requestHdr.xid >> 24;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------- mounting -------------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct mntCtx{
	PROC_VARSET;
	nfs2_t		*nfsMp;
	btint		mountingUid;
	btbool		readOnly;
}mntCtx_t;

static blockpool_t	*mntPool = NULL;

static void	freeMntBlock(void *block)
{
    poolReturnBlock(mntPool, block);
}

static procCond_t	*mntSignal = NULL;
static btbool		mntRunning = 0;

/* ------------------------------------------------------------------------- */

static void	setSockBuffer(int fd, btbool isTx, btsize size)
{
int		optname = isTx ? SO_SNDBUF : SO_RCVBUF;
int		bufSize, defaultSize = 0, optlen;

	optlen = sizeof(int);
	if(getsockopt(fd, SOL_SOCKET, optname, (char *)&defaultSize, &optlen) != 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs: getsockopt(): [%d] %s\n", errno, sysStrerror(errno));
		return;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
		logPrintf(LOGM_DEBUG, "nfs: set %s buffer size: want=%d, default=%d\n", isTx ? "TX" : "RX", size, defaultSize);
	for(bufSize = size; bufSize > 8; bufSize -= bufSize / 8){
		if(setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&bufSize, sizeof(int)) == 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_INFO))
				logPrintf(LOGM_INFO, "nfs: set %s buffer to %d bytes\n", isTx ? "TX" : "RX", bufSize);
			return;
		}
		if(bufSize < defaultSize)	/* already less than default, does still not work */
			break;
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
		logPrintf(LOGM_ERROR, "nfs: setsockopt(%s, %d): [%d] %s\n", isTx ? "txbuf" : "rxbuf", bufSize, errno, sysStrerror(errno));
}

/* ------------------------------------------------------------------------- */

static collectionObjectI_t	contextByXidInterface = {
	ctxXidCompare,
	ctxHashval,
	NULL,
	NULL
};

static btint	createNfsSocket(void)
{
cfgProperty_t	*cfgPath[4];
btint			rval, numEntries, wantPort, storedErrno;
btsize			socketBuffer;
btint			sockAddrlen = sizeof(nfsSockAddr);

	if(nfsSock != -1)
		return 0;
	uidsecBecomeRoot();
	nfsSock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	storedErrno = errno;
	uidsecBecomeUser();
	errno = storedErrno;
    if(nfsSock < 0)
		goto errorOccured;
	numEntries = nfs2ConfigPath(NULL, cfgPath, sizeof(cfgPath)/sizeof(cfgPath[0]));
	socketBuffer = cfgIntValue(cfgQueryFromArray("nfsSockRxBuf", numEntries, cfgPath), 0x18000);
	if(socketBuffer > 0)	/* increase rx buffer to avoid nfs retries on writes */
		setSockBuffer(nfsSock, 0, socketBuffer);
	socketBuffer = cfgIntValue(cfgQueryFromArray("nfsSockTxBuf", numEntries, cfgPath), 0x18000);
	if(socketBuffer > 0)	/* allow changing the tx buffer of the socket, usually not needed */
		setSockBuffer(nfsSock, 1, socketBuffer);
	wantPort = cfgIntValue(cfgQueryFromArray("preferredPort", numEntries, cfgPath), 991);

    memset(&nfsSockAddr, 0, sockAddrlen);
	nfsSockAddr.sin_family = AF_INET;
	nfsSockAddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	nfsSockAddr.sin_port = htons(wantPort);

	uidsecBecomeRoot();
	rval = bind(nfsSock, (struct sockaddr *)&nfsSockAddr, sockAddrlen);
	storedErrno = errno;
	uidsecBecomeUser();
	if(rval < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs2: can\'t bind to preferred port %d, trying free port ([%d] %s)\n", wantPort, storedErrno, sysStrerror(storedErrno));
		nfsSockAddr.sin_port = 0;	/* if this port fails, try the any-port */
		if(bind(nfsSock, (struct sockaddr *)&nfsSockAddr, sockAddrlen) < 0)
			goto errorOccured;
	}
    if(getsockname(nfsSock, (struct sockaddr *)&nfsSockAddr, (void *)&sockAddrlen) < 0)
        goto errorOccured;
    elpRegisterForFd(nfsSock, ELP_FDMODE_READ, sockReadable, NULL);	/* nfs request handler */
	return 0;
errorOccured:
	if(nfsSock != -1){
		storedErrno = errno;
		close(nfsSock);
		nfsSock = -1;
		errno = storedErrno;
	}
	return -1;
}

/* ------------------------------------------------------------------------- */

void	nfsMountProc(void *obj, void *arg)
{
mntCtx_t			*this = obj;
nfs2_t				*mp = this->nfsMp;
elpPid_t			*rval = arg;
feRes_t				res;
btint				nfsTimeout, nfsRetransmit, len, numEntries, child, kernelCacheTime;
btint				nfsTransferSize;
nfs_fh				rootFh;
char				nfsService[1024], *p;
btu8				*nfsHost;
cfgProperty_t		*cfgPath[10];

	PROC_BEGIN;
	nfsMpRetain(mp);
	while(mntRunning){	/* restrict Sharity to one running mount/umount */
		PROC_WAITFOR_X(1, procCondWait(mntSignal, this));
	}
	mntRunning = 1;
	if(mp->isMounted){	/* already mounted */
		res.err.class = ERRCLS_UNIX;
		res.err.code = EBUSY;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfsMount: [%d] %s\n", res.err.code, sysStrerror(res.err.code));
		goto doReturn;
	}
#if 0
/* The permission check must not be done in the server because it requires
 * the stat() system call. If stat() is called, it may be called on a
 * mounted directory served by Sharity, which results in a deadlock.
 * Since the cifsmount utility must be trusted to supply the correct
 * user id anyway, the check can equally well be done in that utility.
 */
	if(!mountPermissionsOk(mp, mountingUid)){
        errno = EPERM;	/* Not owner */
		goto errorOccured;
	}
#endif
	if(nfsSock == -1){
		res.err.class = ERRCLS_UNIX;
		res.err.code = EBADF;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfsMount: could not create socket during startup.\n");
		goto doReturn;
	}
	numEntries = nfs2ConfigPath(mp, cfgPath, sizeof(cfgPath)/sizeof(cfgPath[0]));
	nfsHost = cfgStringValue(cfgQueryFromArray("nfsHost", numEntries, cfgPath), "localhost");
	nfsTimeout = cfgIntValue(cfgQueryFromArray("nfsTimeout", numEntries, cfgPath), 6);
	nfsRetransmit = cfgIntValue(cfgQueryFromArray("nfsRetransmit", numEntries, cfgPath), 7);
	nfsTransferSize = cfgIntValue(cfgQueryFromArray("nfsTransferSize", numEntries, cfgPath), 8192);
	kernelCacheTime = cfgIntValue(cfgQueryFromArray("nfsKernelCacheTime", numEntries, cfgPath), 0);

	if(nfsTransferSize > 8192){		/* limit to useful range */
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs: nfsTransferSize %d out of range, limiting to 8192\n", nfsTransferSize);
		nfsTransferSize = 8192;
	}else if(nfsTransferSize < 1024){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs: nfsTransferSize %d out of range, setting to 1024\n", nfsTransferSize);
		nfsTransferSize = 1024;
	}
	/* Emergency brakes to avoid buffer overflow attacks: */
	if(strlen(nfsHost) > 500)
		nfsHost[500] = 0;
	if(strlen(mp->backendMod->name) > 500)
		mp->backendMod->name[500] = 0;
	sprintf(nfsService, "%s:(%s)", nfsHost, mp->backendMod->name);
	len = strlen(nfsService);
	convUniToLocalAscii(mp->backend.vfsPath(mp->vfs), &nfsService[len], sizeof(nfsService)-len);
	/* Replace whitespace in nfsService with underscore. At least Solaris can't
	 * tolerate whitespace in the mtab entry.
	 */
	for(p=nfsService;*p!=0;p++){
		if(*p == ' ' || *p == '\t')
			*p = '_';
	}
	mp->isReadonly = this->readOnly;
	nfsFaRootHandle(mp->adapter, &rootFh, mp->mountId);
/* we must fork because there is no non-blocking mount() */
    if((child = fork()) == 0){	/* we are the child */
		btint	storederr, mountRval;
		mainSignalHandlers(1);
		uidsecBecomeRoot();
		mountRval = sysMount(mp->mountpoint, &rootFh, nfsSock, &nfsSockAddr, nfsHost, &nfsService[len], nfsService, nfsTimeout, nfsTransferSize, this->readOnly, nfsRetransmit, kernelCacheTime);
		storederr = errno;
		uidsecBecomeUser();
        if(mountRval < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "nfs: sysMount(): [%d] %s\n", storederr, sysStrerror(storederr));
			exit(storederr);
        }
        exit(0);
    }
	mp->mountedBy = this->mountingUid;
    PROC_WAITFOR_X(2, elpRegisterForPid(child, nfsMountProc, this));
	res.err.class = rval->rval != 0 ? ERRCLS_UNIX : ERRCLS_OK;
    res.err.code = rval->rval;
	if(res.err.class == 0){
		mp->isMounted = 1;
	}
doReturn:
	mntRunning = 0;
	PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(mntSignal, NULL);
		nfsMpRelease(mp);
	);
	PROC_END;
}

static void	nfsMount(void *obj, btint mountingUid, btbool readOnly, void *caller)
{
mntCtx_t	*ctx;

    PROC_INIT(ctx, nfsMountProc, caller, mntPool, freeMntBlock);
	ctx->nfsMp = obj;
	ctx->mountingUid = mountingUid;
	ctx->readOnly = readOnly;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static void	nfsUmountProc(void *obj, void *arg)
{
mntCtx_t	*this = obj;
nfs2_t		*mp = this->nfsMp;
elpPid_t	*rval = arg;
btint		child;
feRes_t		res;

	PROC_BEGIN;
	nfsMpRetain(mp);
	while(mntRunning){	/* serialize Sharity's mounts and umounts */
		PROC_WAITFOR_X(1, procCondWait(mntSignal, this));
	}
	mntRunning = 1;
	if(this->mountingUid != 0 && this->mountingUid != mp->mountedBy){	/* not permitted */
		res.err.class = ERRCLS_UNIX;
        res.err.code = EPERM;
		goto errorOccured;
	}
	if(!mp->isMounted){	/* not mounted */
		res.err.class = ERRCLS_UNIX;
		res.err.code = ENOENT;
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfsUmount(): not mounted\n");
		goto errorOccured;
	}
    if((child = fork()) == 0){	/* we are the child */
		btint	storederr, umountRval;
		mainSignalHandlers(1);
		uidsecBecomeRoot();
		umountRval = sysUmount(mp->mountpoint);
		storederr = errno;
		uidsecBecomeUser();
        if(umountRval < 0){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
            	logPrintf(LOGM_ERROR, "nfs: sysUmount(): [%d] %s\n", storederr, sysStrerror(storederr));
        	exit(storederr);
        }
        exit(0);
    }
	PROC_WAITFOR_X(2, elpRegisterForPid(child, nfsUmountProc, this));
	if(rval->rval == 0 || rval->rval == ENOENT){	/* unmount succeeded or not mounted */
		mp->isMounted = 0;
	}
	res.err.class = rval->rval != 0 ? ERRCLS_UNIX : ERRCLS_OK;
    res.err.code = rval->rval;
errorOccured:
	mntRunning = 0;
	PROC_RETURN_AND_CLEANUP(&res,
		procCondBroadcast(mntSignal, NULL);
		nfsMpRelease(mp);
	);
	PROC_END;
}

static void	nfsUmount(nfs2_t *mp, btint uid, void *caller)
{
mntCtx_t	*ctx;

    PROC_INIT(ctx, nfsUmountProc, caller, mntPool, freeMntBlock);
	ctx->nfsMp = mp;
	ctx->mountingUid = uid;
	PROC_START(ctx, NULL);
}

/* ------------------------------------------------------------------------- */

static inline btbool	nfsMpIsMounted(nfs2_t *mp)
{
    return mp->isMounted;
}

/* ------------------------------------------------------------------------- */
/* --------------------------- object allocation --------------------------- */
/* ------------------------------------------------------------------------- */

static void	nfsMpRetain(void *obj)
{
nfs2_t	*this = obj;

    this->refCount++;
}

static void	nfsMpRelease(void *obj)
{
nfs2_t	*this = obj;

    if(--this->refCount == 0){
        nfsMpFree(this);
    }else if(this->refCount < 0){
		DEBUG_TRAP();
	}
}

static btint	nfsMpCompare(void *mp1, void *mp2)
{
nfs2_t	*this = mp1, *other = mp2;

	return this->mountId - other->mountId;
}

static btu32	nfsMpHashVal(void *mp)
{
nfs2_t	*this = mp;

	return this->mountId;
}

/* ------------------------------------------------------------------------- */

static nfs2_t	*nfsMpNew(btu8 *mountpoint)
{
nfs2_t		*volatile mp = NULL;
btint		i;
    
EXC_TRY
    mp = memCalloc(sizeof(nfs2_t));
	mp->refCount = 1;
    mp->mountpoint = utilStrDup(mountpoint);
	i = 0;
	do{
		mp->mountId = ((btu32)incarnation << 16) | mountCounter++;
	}while(htabFindObjectEqualTo(mountsById, mp) != NULL && i++ < 65536);
	if(i >= 65536){
		EXC_THROW_MSG(excInternalError, "nfs2: maximum number of mountpoints exceeded");
	}
	htabAddObject(mountsById, mp);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "nfsMp: allocated new object at 0x%08x with id=0x%08x\n", (int)mp, mp->mountId);
EXC_CATCH
	nfsMpFree(mp);
	EXC_RETHROW();
EXC_ENDTRY
    return mp;
}

/* ------------------------------------------------------------------------- */

static void	nfsMpFree(nfs2_t *mp)
{
    if(mp != NULL){
        if(nfsMpIsMounted(mp)){
            EXC_THROW_MSG(excInternalError, "nfs: freeing mounted nfsMp");
        }
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "nfsMp: freeing object at 0x%08x\n", (int)mp);
		htabRemoveObjectEqualTo(mountsById, mp);
		if(mp->adapter != NULL){
			nfsFaRelease(mp->adapter);
			mp->adapter = NULL;
		}
		if(mp->vfs != NULL){
			mp->backend.vfsRelease(mp->vfs);
			mp->vfs = NULL;
			mp->backendMod = NULL;
		}
        if(mp->mountpoint != NULL){
            memFree(mp->mountpoint);
		}
        memFree(mp);
    }
}

/* ------------------------------------------------------------------------- */

static void	nfsMpSetFakemode(nfs2_t *this, btbool fakeMode)
{
	this->isFakeMode = fakeMode;
}

/* ------------------------------------------------------------------------- */

static void	nfsMpSetBackend(nfs2_t *mp, struct module *backendMod, void *backendObj)
{
cfgProperty_t	*cfgPath[10];
btint			numEntries;

	if(mp->vfs != NULL)
		DEBUG_TRAP();
	mp->backendMod = backendMod;
	mp->vfs = backendObj;
	mp->backend.userAuthenticate = backendMod->implementationOf(BE_USER_AUTHENTICATE);
	mp->backend.userRetain = backendMod->implementationOf(BE_USER_RETAIN);
	mp->backend.userRelease = backendMod->implementationOf(BE_USER_RELEASE);
	mp->backend.vfsRetain = backendMod->implementationOf(BE_VFS_RETAIN);
	mp->backend.vfsRelease = backendMod->implementationOf(BE_VFS_RELEASE);
	mp->backend.vfsConfigPath = backendMod->implementationOf(BE_CONFIG_PATH);
	mp->backend.vfsPath = backendMod->implementationOf(BE_VFS_PATH);
	mp->backend.vfsRetain(mp->vfs);
	numEntries = nfs2ConfigPath(mp, cfgPath, sizeof(cfgPath)/sizeof(cfgPath[0]));
	mp->adapter = nfsFaForBackend(backendMod, backendObj, cfgPath, numEntries);
}

/* ------------------------------------------------------------------------- */

static void	nfsMpGetBackend(nfs2_t *mp, struct module **backendMod, void **backendObj)
{
	*backendMod = mp->backendMod;
	*backendObj = mp->vfs;
}

/* ------------------------------------------------------------------------- */

static btbool	nfsMpIsReadonly(nfs2_t *mp)
{
	return mp->isReadonly;
}

/* ------------------------------------------------------------------------- */

static btu8	*nfsMpMountpoint(nfs2_t *mp)
{
	return mp->mountpoint;
}

/* ------------------------------------------------------------------------- */

static btint	nfsMpMountedBy(nfs2_t *mp)
{
	return mp->mountedBy;
}

/* ------------------------------------------------------------------------- */

static btint	nfs2ConfigPath(nfs2_t *mp, struct cfgProperty **path, btint maxEntries)
{
cfgProperty_t	*p;
btint			w = 0;

	if(mp != NULL)
		w = mp->backend.vfsConfigPath(mp->vfs, path, maxEntries);
	p = cfgQueryWithName(NFS2_MODULE_NAME, 1, cfgFilePlist(gdata.dbFile));
	if(p != NULL && w < maxEntries)
		path[w++] = p;
	p = cfgQueryWithName(NFS2_MODULE_NAME, 1, cfgFilePlist(gdata.cfgFile));
	if(p != NULL && w < maxEntries)
		path[w++] = p;
	return w;
}

/* ------------------------------------------------------------------------- */

#define	PERM_NORMAL	MAIN_GUI_PERM_AUTHORIZED
#define	PERM_EXPERT	(MAIN_GUI_PERM_AUTHORIZED | MAIN_GUI_PERM_EXPERT)

static guiTypedesNumeric_t	numdesTimeout = {
/*	div		mpy		inc		min		max			unit	*/
	10,		1,		1,		1,		10000,		"seconds"
};

static guiTypedesNumeric_t	numdesRetries = {
/*	div		mpy		inc		min		max			unit	*/
	1,		1,		1,		-1,		1000,		""
};

static guiTypedesNumeric_t	numdesAttCache = {
/*	div		mpy		inc		min		max			unit	*/
	1,		1,		1,		0,		10,			"seconds"
};

static btu8	*lookupExtNames[] = {"Database", "Pseudo Inode", NULL};
btu8	*nfsLookupStrategyNames[] = {"database", "pseudoInode", NULL};
/* must match implementation of luNew() in nfslookup.c */
static guiTypedesEnum_t	enumdesLookup = {lookupExtNames, nfsLookupStrategyNames};

#define	ALLOW_NULL	0
#define	FLG_0		0

static guiEntry_t	nfs2BackendEntries[] = {
	{	MAIN_GUI_ENTRY_NUMBER,	"NFS Retries",
		&numdesRetries,			"nfs2ConfigRetries",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"nfsRetransmit",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"NFS Initial Timeout",
		&numdesTimeout,			"nfs2ConfigInitialTimeout",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"nfsTimeout",
	},
	{	MAIN_GUI_ENTRY_NUMBER,	"NFS Kernel Attribute Cache Time",
		&numdesAttCache,		"nfs2ConfigKernelCacheTime",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"nfsKernelCacheTime",
	},
	{	MAIN_GUI_ENTRY_ENUM,	"NFS File-Handle Lookup Strategy",
		&enumdesLookup,			"nfs2ConfigLookupStrategy",
		PERM_EXPERT,			FLG_0, ALLOW_NULL,	"lookupStrategy",
	},
};


/* ------------------------------------------------------------------------- */

static array_t	*nfs2BackendConfig(void)
{
array_t	*rval;
btint	i;

	rval = arrNew(NULL, ELEMENTS(nfs2BackendEntries));
	for(i=0;i<ELEMENTS(nfs2BackendEntries);i++)
		arrAddObject(rval, &nfs2BackendEntries[i]);
	return rval;
}

/* ------------------------------------------------------------------------- */
/* ------------------------- module implementation ------------------------- */
/* ------------------------------------------------------------------------- */

static struct{
	btu8	*name;
	void	*function;
} functionTable[] = {
	{FE_MNT_NEW,			nfsMpNew},
	{FE_MNT_RETAIN,			nfsMpRetain},
	{FE_MNT_RELEASE,		nfsMpRelease},
	{FE_MNT_SET_BACKEND,	nfsMpSetBackend},
	{FE_MNT_GET_BACKEND,	nfsMpGetBackend},
	{FE_MNT_SET_FAKEMODE,	nfsMpSetFakemode},
	{FE_MNT_MOUNT,			nfsMount},
	{FE_MNT_UMOUNT,			nfsUmount},
	{FE_MNT_IS_MOUNTED,		nfsMpIsMounted},
	{FE_MNT_IS_READONLY,	nfsMpIsReadonly},
	{FE_MNT_MOUNTPOINT,		nfsMpMountpoint},
	{FE_MNT_MOUNTED_BY,		nfsMpMountedBy},
	{FE_CFGENTRIES_FOR_BACKEND, nfs2BackendConfig},
};

static void	*getImp(btu8 *functionName)
{
btint	i;

	for(i=0;i<sizeof(functionTable)/sizeof(functionTable[0]);i++){
		if(strcmp(functionName, functionTable[i].name) == 0)
			return functionTable[i].function;
	}
	return NULL;
}

static collectionObjectI_t	nfsMpInterface = {
	nfsMpCompare,
	nfsMpHashVal,
	NULL,
	NULL
};

void	nfs2_init(void)
{
module_t	*mod;
btu32		t;
nfsCtx_t	*context = NULL;	/* only used for measuring sizes */

	t = time(NULL);
	incarnation = (btu16)t;	/* a somewhat unique ID of this daemon instance */
	mountCounter = (btu16)((t >> 8) ^ (t >> 16));	/* start counter at pseudo-random count */
	mountsById = htabNew(&nfsMpInterface, 16, 50);
	runningRequests = htabNew(&contextByXidInterface, 16, 40);
    if(nfsPool == NULL)
        nfsPool = poolNew(sizeof(nfsCtx_t), (btu8 *)&context->argument - (btu8 *)context);
    if(mntPool == NULL)
        mntPool = poolNew(sizeof(mntCtx_t), sizeof(mntCtx_t));
	if(mntSignal == NULL)
		mntSignal = procCondNew();
	if(createNfsSocket() < 0){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfs2: error creating NFS socket during startup: [%d] %s\n", errno, sysStrerror(errno));
	}
	mod = memCalloc(sizeof(*mod));
	mod->name = NFS2_MODULE_NAME;
	mod->moduleType = MOD_TYPE_FRONTEND;
	mod->implementationOf = getImp;
	mod->interfaceVersion = 0;
	mod->implementationVersion = MAINCONST_VERSION;
	modRegister(mod);
}

/* ------------------------------------------------------------------------- */
