/*
 * Name: nfslookup.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-11
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfslookup.c,v 1.5 1999/07/26 20:07:00 cs Exp $
 */

#include "nfslookup.h"
#include "lru.h"
#include "hashtable.h"
#include "backend.h"
#include "module.h"
#include "main.h"
#include "cfgparse.h"
#include "logging.h"
#include "notification.h"
#include "nfs2const.h"
#include "process.h"

extern btu8	*nfsLookupStrategyNames[];

static void	luFhRelease(nfsLookup_t *lu, luFileHandle_t *fh);

/* ------------------------------------------------------------------------- */

static void	checkOpenLimit(nfsLookup_t *lu)
{
btint			closeCount;
luFileHandle_t	*p, *next;

	if(lu->maxOpenFiles > 0){
		closeCount = lruObjectCount(lu->openFilesLru) - lu->maxOpenFiles + 1;
		for(p = lruOldestRelativeTo(lu->openFilesLru, NULL); closeCount > 0 && p != NULL; p = next){
			next = lruOldestRelativeTo(lu->openFilesLru, p);
			if(lu->backend.fileOpRunning(p->vfsFile))
				continue;	/* don't close files that have operations running */
			lu->backend.fileSetOpenMode(p->vfsFile, BE_FILE_CLOSED, BE_FILE_OPEN_RW, 0, NULL, NULL);	/* close in background */
			closeCount--;
		}
	}
}

/* ------------------------------------------------------------------------- */

static void	openModeChanged(void *obj, void *file)
{
nfsLookup_t 	*lu = obj;
luFileHandle_t	*fh, dummyFh;
btint			newOpenMode;
btbool			newIsOpen, oldIsOpen;

	dummyFh.vfsFile = file;
	fh = htabFindObjectEqualTo(lu->filesByVfsFile, &dummyFh);
	if(fh != NULL){
		newOpenMode = lu->backend.fileOpenMode(file);
		newIsOpen = newOpenMode != BE_FILE_CLOSED;
		oldIsOpen = fh->openMode != BE_FILE_CLOSED;
		fh->openMode = newOpenMode;
		if(newIsOpen && !oldIsOpen){
			checkOpenLimit(lu);
			lruAddObject(lu->openFilesLru, fh);
		}else if(oldIsOpen && !newIsOpen){
			lruRemoveObject(lu->openFilesLru, fh);
		}
	}
}

static void	fileWasInvalidated(void *obj, void *file)
{
nfsLookup_t 	*lu = obj;
luFileHandle_t	*fh, dummyFh;

	dummyFh.vfsFile = file;
	fh = htabFindObjectEqualTo(lu->filesByVfsFile, &dummyFh);
	if(fh != NULL){
		luFhRelease(lu, fh);	/* also remove our reference */
	}
}

/* ------------------------------------------------------------------------- */

static void	checkFileLimit(nfsLookup_t *lu)
{
luFileHandle_t	*f, *next;
btint			freeFiles;

    freeFiles = lruObjectCount(lu->filesLru) - lu->maxFiles + 1;
	f = lruOldestRelativeTo(lu->filesLru, NULL);
	while(freeFiles > 0 && f != NULL){
		next = lruOldestRelativeTo(lu->filesLru, f);
		if(f->openMode == BE_FILE_CLOSED && f->refCount == 1){	/* closed and no children */
		/* this frees the object, but can't free the parent because the parent
		 * must have at least two references:
		 */
			luFhRelease(lu, f);
			freeFiles--;
		}
		f = next;
	}
}

/* ------------------------------------------------------------------------- */

static void	luFhFree(nfsLookup_t *lu, luFileHandle_t *fh)
{
	if(fh != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "nfslookup: deleting cache for nfsfh = 0x%08x 0x%08x 0x%08x 0x%08x\n", ((btu32 *)fh->nfsfh)[0], ((btu32 *)fh->nfsfh)[1], ((btu32 *)fh->nfsfh)[2], ((btu32 *)fh->nfsfh)[3]);
		if(fh->openMode != BE_FILE_CLOSED)
			lruRemoveObject(lu->openFilesLru, fh);
		lruRemoveObject(lu->filesLru, fh);
		htabRemoveObjectEqualTo(lu->filesByFh, fh);
		htabRemoveObjectEqualTo(lu->filesByVfsFile, fh);
		if(fh->parent != NULL){
			luFhRelease(lu, fh->parent);
			fh->parent = NULL;
		}
		if(fh->vfsFile != NULL){
			lu->backend.fileRelease(fh->vfsFile);
			fh->vfsFile = NULL;
		}
		memFree(fh);
	}
}

static inline void	luFhRetain(luFileHandle_t *fh)
{
	fh->refCount++;
}

static void	luFhRelease(nfsLookup_t *lu, luFileHandle_t *fh)
{
	if(--fh->refCount == 0){
		luFhFree(lu, fh);
	}else if(fh->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */

void	luCacheFile(nfsLookup_t *lu, void *vfsFile)
{
luFileHandle_t	*volatile fh, *volatile parent = NULL, dummyFh;
void			*vfsParent;

	dummyFh.vfsFile = vfsFile;
	fh = htabFindObjectEqualTo(lu->filesByVfsFile, &dummyFh);
	if(fh != NULL){
		lruTouch(lu->filesLru, fh);
		return;		/* already cached */
	}
	vfsParent = lu->backend.fileParent(vfsFile);
	if(vfsParent != NULL && vfsParent != vfsFile){
		dummyFh.vfsFile = vfsParent;
		parent = htabFindObjectEqualTo(lu->filesByVfsFile, &dummyFh);
		if(parent == NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "*** nfslookup warning: can\'t cache file because parent not found\n");
			return;	/* can't cache */
		}
	}
EXC_TRY;
	fh = memCalloc(sizeof(*fh));
	fh->vfsFile = vfsFile;
	lu->backend.fileRetain(vfsFile);
	fh->refCount = 1;
	if(parent == NULL){
		luRootHandle(lu, fh->nfsfh, NULL, 0);
	}else{
		fh->parent = parent;
		luFhRetain(parent);
		luHandleOfFile(lu, parent->nfsfh, lu->backend.fileName(vfsFile), fh->nfsfh, NULL);
	}
	checkFileLimit(lu);
	if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
		logPrintf(LOGM_RESOURCE, "nfslookup: creating cache for nfsfh = 0x%08x 0x%08x 0x%08x 0x%08x\n", ((btu32 *)fh->nfsfh)[0], ((btu32 *)fh->nfsfh)[1], ((btu32 *)fh->nfsfh)[2], ((btu32 *)fh->nfsfh)[3]);
	lruAddObject(lu->filesLru, fh);
	htabAddObject(lu->filesByFh, fh);
	htabAddObject(lu->filesByVfsFile, fh);
	fh->openMode = lu->backend.fileOpenMode(vfsFile);
	if(fh->openMode != BE_FILE_CLOSED){
		lruAddObject(lu->openFilesLru, fh);
		checkOpenLimit(lu);
	}
EXC_CATCH;
	/* ingore all exceptions, leave file uncached */
EXC_ENDTRY;
}

/* ------------------------------------------------------------------------- */

nfsLookup_t	*luNew(void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries)
{
nfsLookup_t	*rval;

	switch(cfgEnumValue(cfgQueryFromArray("lookupStrategy", numEntries, cfgPath), nfsLookupStrategyNames, 0)){
	default:	/* case values must match nfsLookupStrategyNames */
	case 0:
		rval = nfsfhLookupTableNew(vfs, backend, cfgPath, numEntries);
		break;
	case 1:
		rval = nfsfhLookupPseudoNew(vfs, backend, cfgPath, numEntries);
		break;
	}
	luCacheFile(rval, rval->rootFile);
	return rval;
}

/* ------------------------------------------------------------------------- */

static btint	fileFhCompare(void *obj, void *otherObj)
{
luFileHandle_t	*this = obj, *other = otherObj;

	return memcmp(this->nfsfh, other->nfsfh, 28); /* don't compare mountIds */
}

static btu32	fileFhHashVal(void *obj)
{
luFileHandle_t	*this = obj;
btu32			*fh;

	fh = (btu32 *)this->nfsfh;
	return fh[0] ^ fh[1];
}

static btint	fileVfsCompare(void *obj, void *otherObj)
{
luFileHandle_t	*this = obj, *other = otherObj;

	return (btu8 *)this->vfsFile - (btu8 *)other->vfsFile;
}

static btu32	fileVfsHashVal(void *obj)
{
luFileHandle_t	*this = obj;
btint			addr;

	addr = (btint)this->vfsFile;
	return addr ^ (addr >> 8);
}

static collectionObjectI_t	filesByFhInterface = {
	fileFhCompare,
	fileFhHashVal,
	NULL,
	NULL
};

static collectionObjectI_t	filesByVfsFileInterface = {
	fileVfsCompare,
	fileVfsHashVal,
	NULL,
	NULL
};

/* ------------------------------------------------------------------------- */

static btint	dummyFileOpenMode(void *obj)
{
	return BE_FILE_CLOSED;
}

static void	dummyFileSetOpenMode(void *obj, btint mode, btint modeMask, btbool autoclose, void *user, void *caller)
{
beFileRes_t	res;

	res.err.code = res.err.class = 0;
	PROC_IMMEDIATE_RET(caller, &res);
}

static btbool	dummyFileOpRunning(void *obj)
{
	return 0;
}

/* ------------------------------------------------------------------------- */

void	luInit(nfsLookup_t *lu, nfsLookupInterface_t *interface, void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries)
{
luFileHandle_t	*dummy = NULL;

	lu->i = interface;
	lu->backend.vfsRetain = backend->implementationOf(BE_VFS_RETAIN);
	lu->backend.vfsRelease = backend->implementationOf(BE_VFS_RELEASE);
	lu->backend.vfsRootFile = backend->implementationOf(BE_VFS_ROOT_FILE);
	lu->backend.vfsCaseSensitive = backend->implementationOf(BE_VFS_CASE_SENSITIVE);
	lu->backend.vfsOpenNotification = backend->implementationOf(BE_VFS_OPEN_NOTIFICATION);
	lu->backend.vfsInvalidateNotification = backend->implementationOf(BE_VFS_INVALIDATE_NOTIFICATION);
	
	lu->backend.fileRetain = backend->implementationOf(BE_FILE_RETAIN);
	lu->backend.fileRelease = backend->implementationOf(BE_FILE_RELEASE);
	lu->backend.fileLookup = backend->implementationOf(BE_FILE_LOOKUP);
	lu->backend.fileOpenMode = backend->implementationOf(BE_FILE_OPENMODE);
	lu->backend.fileSetOpenMode = backend->implementationOf(BE_FILE_SET_OPENMODE);
	lu->backend.fileOpRunning = backend->implementationOf(BE_FILE_OP_RUNNING);
	lu->backend.fileName = backend->implementationOf(BE_FILE_NAME);
	lu->backend.fileParent = backend->implementationOf(BE_FILE_PARENT);
	lu->backend.fileReaddir = backend->implementationOf(BE_FILE_READDIR);
	
	if(lu->backend.fileOpenMode == NULL)
		lu->backend.fileOpenMode = dummyFileOpenMode;
	if(lu->backend.fileSetOpenMode == NULL)
		lu->backend.fileSetOpenMode = dummyFileSetOpenMode;
	if(lu->backend.fileOpRunning == NULL)
		lu->backend.fileOpRunning = dummyFileOpRunning;

	lu->vfs = vfs;
	lu->backend.vfsRetain(lu->vfs);
	lu->rootFile = lu->backend.vfsRootFile(lu->vfs);
	lu->filesLru = lruNew(0);
	lu->openFilesLru = lruNew((btint)&dummy->openFilesLru);
	lu->filesByFh = htabNew(&filesByFhInterface, 128, 50);
	lu->filesByVfsFile = htabNew(&filesByVfsFileInterface, 128, 50);
	lu->isCaseSensitive = lu->backend.vfsCaseSensitive(lu->vfs);
	lu->maxFiles = cfgIntValue(cfgQueryFromArray("cacheFiles", numEntries, cfgPath), 100);
	lu->maxOpenFiles = cfgIntValue(cfgQueryFromArray("cacheOpenFiles", numEntries, cfgPath), 20);
	lu->openNotify = lu->backend.vfsOpenNotification(lu->vfs);
	ncRetain(lu->openNotify);
	ncAddObserver(lu->openNotify, openModeChanged, lu);
	lu->invalidateNotify = lu->backend.vfsInvalidateNotification(lu->vfs);
	ncRetain(lu->invalidateNotify);
	ncAddObserver(lu->invalidateNotify, fileWasInvalidated, lu);
	lu->refCount = 1;
}

/* ------------------------------------------------------------------------- */

void	luExit(nfsLookup_t *lu)
{
array_t	*allHandles;
btint	cnt, i;

	if(lu != NULL){
		if(lu->openNotify != NULL){
			ncRemoveObserver(lu->openNotify, openModeChanged, lu);
			ncRelease(lu->openNotify);
			lu->openNotify = NULL;
		}
		if(lu->invalidateNotify != NULL){
			ncRemoveObserver(lu->invalidateNotify, fileWasInvalidated, lu);
			ncRelease(lu->invalidateNotify);
			lu->invalidateNotify = NULL;
		}
		if(lu->filesByFh != NULL){
			allHandles = htabAllObjects(lu->filesByFh);
			cnt = arrObjectCount(allHandles);
			for(i=0;i<cnt;i++){
				luFhRelease(lu, arrObjectAtIndex(allHandles, i));
			}
			arrFree(allHandles);
			htabFree(lu->filesByFh);
			lu->filesByFh = NULL;
		}
		if(lu->filesByVfsFile != NULL){
			htabFree(lu->filesByVfsFile);
			lu->filesByVfsFile = NULL;
		}
		if(lu->openFilesLru != NULL){
			lruFree(lu->openFilesLru);
			lu->openFilesLru = NULL;
		}
		if(lu->filesLru != NULL){
			lruFree(lu->filesLru);
			lu->filesLru = NULL;
		}
		if(lu->rootFile != NULL){
			lu->backend.fileRelease(lu->rootFile);
			lu->rootFile = NULL;
		}
		if(lu->vfs != NULL){
			lu->backend.vfsRelease(lu->vfs);
			lu->vfs = NULL;
		}
	}
}

/* ------------------------------------------------------------------------- */

void		luRetain(nfsLookup_t *lu)
{
	lu->refCount++;
}

void		luRelease(nfsLookup_t *lu)
{
	if(--lu->refCount == 0){
		lu->i->free(lu);
	}else if(lu->refCount < 0){
		DEBUG_TRAP();
	}
}

/* ------------------------------------------------------------------------- */
