/*
 * Name: nfslookup.h
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-11
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfslookup.h,v 1.3 1999/07/05 16:31:30 cs Exp $
 */

#ifndef __nfslookup_H_INCLUDED__
#define __nfslookup_H_INCLUDED__

#include "basictypes.h"
#include "lru.h"

/* General Description:
This module implements an abstract superclass for lookup strategies.
Lookup strategy classes provide a mapping between NFS file handles and
file pathes. This is required because many backend implementations don't
have the concept of inode numbers or at least can't find files just by
inode number.
*/

/* ------------------------------------------------------------------------- */

struct hashtable;
struct module;
struct cfgProperty;

/* ------------------------------------------------------------------------- */

typedef struct nfsLookupInterface{
	void	(*rootHandle)(void *lu, void *handle, btinum *inum, btu32 mountId);
	void	(*handleOfFile)(void *lu, void *parentHandle, btuni *name, void *destHandle, btinum *destInum);
	void	(*lookupFile)(void *lu, void *parentHandle, void *byUser, btbool touchOpen, void *caller);	/* callback: vfsFile */
	void	(*rename)(void *lu, void *fromDirHandle, btuni *fromName, void *toDirHandle, btuni *toName);
	void	(*markStale)(void *lu, void *parentHandle, btuni *name);
	void	(*free)(void *lu);
}nfsLookupInterface_t;

/* ------------------------------------------------------------------------- */

typedef struct luFileHandle{
    lruObject_t			allFilesLru;	/* lru-pointers for all files */
	lruObject_t			openFilesLru;	/* lru-pointers for open files */
	struct luFileHandle	*parent;
	btint				refCount;
	void				*vfsFile;
	btint				openMode;
	btu8				nfsfh[32];
}luFileHandle_t;

/* ------------------------------------------------------------------------- */

typedef struct nfsLookup{
	nfsLookupInterface_t	*i;
	btint					refCount;
	struct{
		void	(*vfsRetain)(void *obj);
		void	(*vfsRelease)(void *obj);
		void	*(*vfsRootFile)(void *obj);
		btbool	(*vfsCaseSensitive)(void *vfs);
		struct notificationCenter	*(*vfsOpenNotification)(void *vfs);
		struct notificationCenter	*(*vfsInvalidateNotification)(void *vfs);
		void	(*fileRetain)(void *obj);
		void	(*fileRelease)(void *obj);
		void	*(*fileLookup)(void *parent, btuni *name, void *user, void *caller);
		btint	(*fileOpenMode)(void *obj);
		void	(*fileSetOpenMode)(void *obj, btint mode, btint modeMask, btbool autoclose, void *user, void *caller);
		btbool	(*fileOpRunning)(void *obj);
		btuni	*(*fileName)(void *obj);
		void	*(*fileParent)(void *obj);
		void	(*fileReaddir)(void *obj, void *user, btint startIndex, btbool (*resultCallback)(void *callerCtx, btuni *name, btbool isDir), void *caller);
	}						backend;
	void					*vfs;
	void					*rootFile;	/* pointer to vfsFile */
	lru_t					*filesLru;
	lru_t					*openFilesLru;
	struct hashtable		*filesByFh;
	struct hashtable		*filesByVfsFile;
	btbool					isCaseSensitive;
	struct notificationCenter	*openNotify;
	struct notificationCenter	*invalidateNotify;
	btint					maxFiles;
	btint					maxOpenFiles;
}nfsLookup_t;

/* ------------------------------------------------------------------------- */

typedef struct luRes{
	btError	err;
	void	*vfsFile;
}luRes_t;

/* ------------------------------------------------------------------------- */

/* subclass methods: */
nfsLookup_t	*nfsfhLookupTableNew(void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries);
nfsLookup_t	*nfsfhLookupPseudoNew(void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries);


/* superclass methods: */
nfsLookup_t	*luNew(void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries);
/* returns object of appropriate subclass depending on configuration */
void		luInit(nfsLookup_t *lu, nfsLookupInterface_t *interface, void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries);
void		luExit(nfsLookup_t *lu);
void		luRetain(nfsLookup_t *lu);
void		luRelease(nfsLookup_t *lu);
void		luCacheFile(nfsLookup_t *lu, void *vfsFile);


/* polymorphic methods: */
#define	luRootHandle(lu, handle, inum, mountId)									(((nfsLookup_t *)(lu))->i->rootHandle(lu, handle, inum, mountId))
#define	luHandleOfFile(lu, parentHandle, name, destHandle, destInum)	(((nfsLookup_t *)(lu))->i->handleOfFile(lu, parentHandle, name, destHandle, destInum))
#define	luLookupFile(lu, handle, userContext, touchOpen, caller)		(((nfsLookup_t *)(lu))->i->lookupFile(lu, handle, userContext, touchOpen, caller))
#define	luRename(lu, fromDirHandle, fromName, toDirHandle, toName)		(((nfsLookup_t *)(lu))->i->rename(lu, fromDirHandle, fromName, toDirHandle, toName))
#define	luMarkStale(lu, parentHandle, name)								(((nfsLookup_t *)(lu))->i->markStale(lu, parentHandle, name))
#define	luFree(lu)														(((nfsLookup_t *)(lu))->i->free(lu))

/* ------------------------------------------------------------------------- */

#endif /* __nfslookup_H_INCLUDED__ */
