/*
 * Name: nfslupseudo.c
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-05
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: nfslupseudo.c,v 1.3 1999/06/11 20:31:38 cs Exp $
 */

/* General Description:
This subclass of nfsLookup implements the pseudo-inode lookup strategy. This
strategy is a hack because it can't provide full Unix compatibility in the
behaviour on file renames and because it has a slight chance to fail, but
it has the advantage of very little memory consumption. It works as follows:
NFS file handles are hash values over the file's path. To simplify lookup by
NFS handle, the handle also contains hash values over all parent directories
of a given file. Going through the directory hierarchy recursively, it should
always be possible to find a file in reasonable time. Together with a lookup
cache, the performance should be OK.
*/

#include "stdheaders.h"
#include "nfslookup.h"
#include "hashtable.h"
#include "process.h"
#include "blockpool.h"
#include "unicode.h"
#include "logging.h"
#include "backend.h"
#include "errcodes.h"

/* ------------------------------------------------------------------------- */

typedef struct nfsFh{
	btinum	inum;
	btu8	depth;
	btu8	componentHash[32 - sizeof(btu8) - sizeof(btinum) - sizeof(btu32)];	/* 'path' of this file as hash values */
	btu32	mountId;
}nfsFh_t;


typedef struct findContext{
    PROC_VARSET;
	nfsLookup_t		*lu;
	void			*dir;	/* pointer to vfs-file */
	btint			depth;
    nfsFh_t			*nfsFh;
	void			*user;
    btuni			*fileName;
    array_t			*childNames;
    int				cnt, i;
	btu32			contextCrc;
}findContext_t;

static blockpool_t	*findPool = NULL;

static void	findCtxFree(void *block)
{
    poolReturnBlock(findPool, block);
}

extern nfsLookupInterface_t	nfsfhPseudoInterface;	/* forward declaration */

/* ------------------------------------------------------------------------- */

static void	nfsfhLookupPseudoFree(void *obj)
{
nfsLookup_t	*this = obj;

	if(this != NULL){
		luExit(this);
		memFree(this);
	}
}

nfsLookup_t	*nfsfhLookupPseudoNew(void *vfs, struct module *backend, struct cfgProperty **cfgPath, btint numEntries)
{
nfsLookup_t	*volatile this = NULL;

    if(findPool == NULL)
        findPool = poolNew(sizeof(findContext_t), sizeof(findContext_t));
EXC_TRY;
	this = memCalloc(sizeof(nfsLookup_t));
	luInit(this, &nfsfhPseudoInterface, vfs, backend, cfgPath, numEntries);
EXC_CATCH;
	nfsfhLookupPseudoFree(this);
	EXC_RETHROW();
EXC_ENDTRY;
    return this;
}

/* ------------------------------------------------------------------------- */

static void	psRootHandle(void *this, void *dest, btinum *inum, btu32 mountId)
{
nfsFh_t		*handle = dest;

	if(handle != NULL){
		memset(handle, 0, 32);
		handle->inum = -1;
		handle->depth = 255;
		handle->mountId = mountId;
	}
	if(inum != NULL){
		*inum = -1;
	}
}

/* ------------------------------------------------------------------------- */

static void	psHandleOfFile(void *obj, void *parentHandle, btuni *name, void *destHandle, btinum *destInum)
{
nfsLookup_t		*lu = obj;
nfsFh_t			*parent = parentHandle, *fh = destHandle;
btu32 			crc = -1;
luFileHandle_t	*lufh = NULL;
static btuni	sep[2] = {UNI_CHAR_BACKSLASH, 0};
static btuni	dot[2] = {'.', 0};
static btuni	dotdot[3] = {'.', '.', 0};

	if(memcmp(name, dot, sizeof(dot)) == 0){
		crc = parent->inum;
		if(fh != NULL)
			*fh = *parent;
	}else if(memcmp(name, dotdot, sizeof(dotdot)) == 0){
		lufh = (luFileHandle_t *)((btu8 *)parentHandle - (btint)&lufh->nfsfh);
		lufh = htabFindObjectEqualTo(lu->filesByFh, lufh);
		if(lufh == NULL){
			if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
				logPrintf(LOGM_ERROR, "nfslupseudo: handleOfFile: can\'t find parent\n");
		}else{
			if(fh != NULL)
				memcpy(fh, lufh->nfsfh, 32);
			crc = ((nfsFh_t *)lufh->nfsfh)->inum;
		}
	}else{
		crc = parent->inum;
		uniStrAddHash(&crc, sep, lu->isCaseSensitive);
		uniStrAddHash(&crc, name, lu->isCaseSensitive);
		if(fh != NULL){
			*fh = *parent;
			fh->inum = crc;
			fh->depth++;
			if(fh->depth < sizeof(fh->componentHash)){
				fh->componentHash[(btint)fh->depth] = uniStrHash(name, lu->isCaseSensitive);
			}
		}
	}
	if(destInum != NULL)
		*destInum = crc;
}

/* ------------------------------------------------------------------------- */

static void	psLookupHandleInDir(nfsLookup_t *lu, void *dir, btint depth, nfsFh_t *nfsFh, btu32 contextCrc, void *user, void *caller);

static btbool	findInumResult(void *obj, btuni *name, btbool isDir)
{
findContext_t	*this = obj;
static btuni	sep[2] = {UNI_CHAR_BACKSLASH, 0};
btu32			crc;

    this->i++;
	crc = this->contextCrc;
	uniStrAddHash(&crc, sep, this->lu->isCaseSensitive);
	uniStrAddHash(&crc, name, this->lu->isCaseSensitive);
    if(crc == this->nfsFh->inum){
        this->fileName = uniStrDup(name);
        this->i = -1;
        return 0;
    }
    return 1;
}

static btbool	findDirWithHashResult(void *obj, btuni *name, btbool isDir)
{
findContext_t	*this = obj;
btu8			compHash;

    this->i++;
    if(isDir){
		if(this->depth >= ELEMENTS(this->nfsFh->componentHash)){
        	arrAddObject(this->childNames, uniStrDup(name));
		}else{
			compHash = uniStrHash(name, this->lu->isCaseSensitive);
			if(compHash == this->nfsFh->componentHash[this->depth]){
        		arrAddObject(this->childNames, uniStrDup(name));
			}
		}
    }
    return 1;
}

static void	psLookupHandleInDirProc(void *obj, void *arg)
{
findContext_t 				*this = obj;
beFileRes_t					*rval = arg;
luRes_t						res, *lrval = arg;
btuni						*name;
btu32						crc;
static btuni				sep[2] = {UNI_CHAR_BACKSLASH, 0};
static collectionObjectI_t	stringCollectionInterface = {
    NULL,
    NULL,
    NULL,
    (void *)memFree
};

	PROC_BEGIN
    if(this->depth == this->nfsFh->depth){	/* look for a file, not a directory */
		this->i = 0;
		do{
			PROC_WAITFOR_X(1, this->lu->backend.fileReaddir(this->dir, this->user, this->i, findInumResult, this));
		}while(rval->err.class == 0 && !rval->eof && this->i >= 0);
		if(this->fileName == NULL){
			if(rval->err.class != 0){
				if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
					logPrintf(LOGM_RARE, "nfsfhpseudo: psLookupHandleInDir: readdir for file failed\n");
				res.err = rval->err;
				PROC_RETURN(&res);
			}else{
				if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
					logPrintf(LOGM_RARE, "nfsfhpseudo: psLookupHandleInDir: file not found\n");
				res.err.class = ERRCLS_CIFS_LOCAL;
				res.err.code = ERR_CIFS_LOCstale;
				PROC_RETURN(&res);
			}
		}
		PROC_WAITFOR_X(2, this->lu->backend.fileLookup(this->dir, this->fileName, this->user, this));
		if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
			logPrintf(LOGM_RARE, "nfsfhpseudo: psLookupHandleInDir: mapped handle to name: ->%s<-\n", convUniToAscii(this->fileName));
		memFree(this->fileName);
		if(rval->err.class == 0)
			luCacheFile(this->lu, rval->file);
		res.vfsFile = rval->file;
		res.err = rval->err;
		PROC_RETURN(&res);
    }	/* look for a directory: */
    this->i = 0;
    this->childNames = arrNew(&stringCollectionInterface, 8);
    do{
        PROC_WAITFOR_X(3, this->lu->backend.fileReaddir(this->dir, this->user, this->i, findDirWithHashResult, this));
    }while(rval->err.class == 0 && !rval->eof);
    if(rval->err.class != 0){
        arrFree(this->childNames);
		if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
			logPrintf(LOGM_RARE, "nfsfhpseudo: psLookupHandleInDir: readdir for dir failed\n");
        PROC_RETURN(rval);
    }
    this->cnt = arrObjectCount(this->childNames);
    for(this->i=0;this->i<this->cnt;this->i++){
        name = arrObjectAtIndex(this->childNames, this->i);
        PROC_WAITFOR_X(4, this->lu->backend.fileLookup(this->dir, name, this->user, this));
        if(rval->err.class != 0 || rval->file == NULL)
            continue;	/* ignore errors as long as we find the file in the end */
		luCacheFile(this->lu, rval->file);
		crc = this->contextCrc;
		uniStrAddHash(&crc, sep, this->lu->isCaseSensitive);
		uniStrAddHash(&crc, arrObjectAtIndex(this->childNames, this->i), this->lu->isCaseSensitive);
        PROC_WAITFOR_X(5, psLookupHandleInDir(this->lu, rval->file, this->depth+1, this->nfsFh, crc, this->user, this));
        if(lrval->err.class == 0){
            arrFree(this->childNames);
            PROC_RETURN(lrval);
        }
    }
    arrFree(this->childNames);
	if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
		logPrintf(LOGM_RARE, "nfsfhpseudo: psLookupHandleInDir: dir not found\n");
	res.err.class = ERRCLS_CIFS_LOCAL;
	res.err.code = ERR_CIFS_LOCstale;
	res.vfsFile = NULL;
    PROC_RETURN(&res);
    PROC_END
}

static void	psLookupHandleInDir(nfsLookup_t *lu, void *dir, btint depth, nfsFh_t *nfsFh, btu32 contextCrc, void *user, void *caller)
{
findContext_t	*ctx;
luRes_t			res;

#if 0
    if(nfsFh->depth == 255){}		/* this is the root handle */
    /* we can't come across the root handle here because it's in the cache */
#endif
	if(depth > 240){
		if(LOG_MODE_IS_ACTIVE(LOGM_ERROR))
			logPrintf(LOGM_ERROR, "nfsfhpseudo: psLookupHandleInDir: recursion overflow\n");
		res.err.class = ERRCLS_CIFS_LOCAL;
		res.err.code = ERR_CIFS_LOCstale;
		res.vfsFile = NULL;
		PROC_IMMEDIATE_RET(caller, &res);
	}
	PROC_INIT(ctx, psLookupHandleInDirProc, caller, findPool, findCtxFree);
	ctx->lu = lu;
    ctx->nfsFh = nfsFh;
	ctx->depth = depth;
    ctx->dir = dir;
	ctx->user = user;
	ctx->contextCrc = contextCrc;
    PROC_START(ctx, NULL);
}

static void	psFileWithHandle(void *obj, void *handle, void *user, btbool touchOpen, void *caller)
{
nfsLookup_t		*lu = obj;
luFileHandle_t	*lufh = NULL;
luRes_t			res;

	lufh = (luFileHandle_t *)((btu8 *)handle - (btint)&lufh->nfsfh);
	lufh = htabFindObjectEqualTo(lu->filesByFh, lufh);
	if(lufh != NULL){
		if(LOG_MODE_IS_ACTIVE(LOGM_DEBUG))
			logPrintf(LOGM_DEBUG, "nfslupseudo: inum %d cached\n", ((nfsFh_t *)handle)->inum);
		res.err.class = res.err.code = 0;
		res.vfsFile = lufh->vfsFile;
 		if(touchOpen && lufh->openFilesLru.prevnext != NULL){
			lruTouch(lu->openFilesLru, lufh);
		}
		lu->backend.fileRetain(res.vfsFile);
        PROC_IMMEDIATE_RET(caller, &res);
	}
	if(LOG_MODE_IS_ACTIVE(LOGM_RARE))
		logPrintf(LOGM_RARE, "nfslupseudo: inum %d not cached\n", ((nfsFh_t *)handle)->inum);
    psLookupHandleInDir(lu, lu->rootFile, 0, handle, -1, user, caller);
}

/* ------------------------------------------------------------------------- */

static void	psFileRename(void *this, void *fromDirFh, btuni *fromName, void *toDirFh, btuni *toName)
{
	/* we don't and can't do anything here... */
}

/* ------------------------------------------------------------------------- */

static void	psFileDestroy(void *this, void *parentHandle, btuni *name)
{
	/* we don't keep any state and thus don't need to maintain it */
}

/* ------------------------------------------------------------------------- */

nfsLookupInterface_t	nfsfhPseudoInterface = {
	psRootHandle,
	psHandleOfFile,
	psFileWithHandle,
	psFileRename,
	psFileDestroy,
	nfsfhLookupPseudoFree,
};

/* ------------------------------------------------------------------------- */
