/*
 * Name: sunrpc.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1998-12-02
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: sunrpc.c,v 1.1.1.1 1999/04/12 13:56:45 cs Exp $
 */

#include "stdheaders.h"
#include "exception.h"
#include "logging.h"
#include "memalloc.h"
#include "sunrpc.h"


EXC_DEFINE(srpcExcEncodeOverflow);
EXC_DEFINE(srpcExcDecodeOverflow);
EXC_DEFINE(srpcExcUnexpected);

/* ------------------------------------------------------------------------- */
/* ------------------ message object creation/destruction ------------------ */
/* ------------------------------------------------------------------------- */

#define	MAX_USAGE_COUNT		512

static fifo_t	*unusedMessages = NULL;

srpcMsg_t	*srpcMsgNew(void)
{
srpcMsg_t	*msg;

	if(unusedMessages == NULL)
		unusedMessages = fifoNew();
	if((msg = fifoGetObject(unusedMessages)) == NULL){
		msg = memAlloc(sizeof(srpcMsg_t));
		msg->usageCount = 0;
		if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
			logPrintf(LOGM_RESOURCE, "sunrpc: allocating new msg 0x%08x\n", (int)msg);
	}
	msg->size = msg->readIndex = 0;
	msg->usageCount++;
	return msg;
}

void		srpcMsgFree(srpcMsg_t *msg)
{
	if(msg != NULL){
		if(msg->usageCount > MAX_USAGE_COUNT){
			if(LOG_MODE_IS_ACTIVE(LOGM_RESOURCE))
				logPrintf(LOGM_RESOURCE, "sunrpc: freeing msg 0x%08x\n", (int)msg);
			memFree(msg);
		}else{
			fifoAddObject(unusedMessages, msg);
		}
	}
}

static inline void	srpcMsgEnsureSize(srpcMsg_t *msg, btsize space)
{
	if(msg->size + space > sizeof(msg->data)){
		EXC_THROW_MSG(srpcExcEncodeOverflow, "message size overflow");
	}
}

/* ------------------------------------------------------------------------- */
/* ----------- first define encoding and decoding of primitives ------------ */
/* ------------------------------------------------------------------------- */

static inline void	srpcEncodeInt(struct srpcMsg *msg, btu32 x)
{
	srpcMsgEnsureSize(msg, 4);
	*(btu32 *)(msg->data + msg->size) = htonl(x);
	msg->size += 4;
}

static inline void	srpcEncodeFixedBlock(struct srpcMsg *msg, void *block, btsize len)
{
btsize	pad;

	pad = (4 - (len % 4)) % 4;
	srpcMsgEnsureSize(msg, len + pad);
	memcpy(msg->data + msg->size, block, len);
	msg->size += len;
	if(pad > 0){
		memset(msg->data + msg->size, 0, pad);
		msg->size += pad;
	}
}

static inline void	srpcEncodeVarBlock(struct srpcMsg *msg, void *block, btsize len)
{
btsize	pad;

	pad = (4 - (len % 4)) % 4;
	srpcMsgEnsureSize(msg, 4 + len + pad);
	*(btu32 *)(msg->data + msg->size) = htonl(len);
	msg->size += 4;
	memcpy(msg->data + msg->size, block, len);
	msg->size += len;
	if(pad > 0){
		memset(msg->data + msg->size, 0, pad);
		msg->size += pad;
	}
}

static inline void	srpcEncodeString(struct srpcMsg *msg, btu8 *s)
{
	srpcEncodeVarBlock(msg, s, strlen(s));
}

/* ------------------------------------------------------------------------- */

static inline btu32	srpcDecodeInt(struct srpcMsg *msg)
{
btu32	x;

	if(msg->size - msg->readIndex < 4){
		EXC_THROW_MSG(srpcExcDecodeOverflow, "error decoding int32");
	}
	x = ntohl(*(btu32 *)(msg->data + msg->readIndex));
	msg->readIndex += 4;
	return x;
}

static inline void	*srpcDecodeFixedBlock(struct srpcMsg *msg, btsize len)
{
btsize	pad;
void	*rval;

	pad = (4 - (len % 4)) % 4;
	if(msg->size - msg->readIndex < len + pad){
		EXC_THROW_MSG(srpcExcDecodeOverflow, "error decoding fixed block");
	}
	rval = msg->data + msg->readIndex;
	msg->readIndex += len + pad;
	return rval;
}

static inline void	*srpcDecodeVarBlock(struct srpcMsg *msg, btsize *lenp)
{
btsize	pad, len;
void	*rval;

	if(msg->size - msg->readIndex < 4){
		EXC_THROW_MSG(srpcExcDecodeOverflow, "error decoding var block: no length prefix");
	}
	len = ntohl(*(btu32 *)(msg->data + msg->readIndex));
	msg->readIndex += 4;
	pad = (4 - (len % 4)) % 4;
	if(msg->size - msg->readIndex < len + pad){
		EXC_THROW_MSG(srpcExcDecodeOverflow, "error decoding var block");
	}
	rval = msg->data + msg->readIndex;
	msg->readIndex += len + pad;
	*lenp = len;
	return rval;
}

static inline btu8	*srpcDecodeString(struct srpcMsg *msg)
{
btsize	pad, len;
btu8	*rval;

	if(msg->size - msg->readIndex < 4){
		EXC_THROW_MSG(srpcExcDecodeOverflow, "error decoding string: no length prefix");
	}
	len = ntohl(*(btu32 *)(msg->data + msg->readIndex));
	msg->readIndex += 4;
	pad = (4 - (len % 4)) % 4;
	if(msg->size - msg->readIndex < len + pad){
		EXC_THROW_MSG(srpcExcDecodeOverflow, "error decoding var block");
	}
	rval = msg->data + msg->readIndex;
	msg->readIndex += len + pad;
	if(pad > 0){	/* simple case: space for null-termination in padding */
		rval[len] = 0;
	}else{			/* nasty case: have to copy to end of message */
		srpcMsgEnsureSize(msg, len + 1);
		memcpy(msg->data + msg->size, rval, len);
		rval = msg->data + msg->size;
		msg->size += len;
		msg->data[msg->size++] = 0;	/* null terminate string */
	}
	return rval;
}

/* ------------------------------------------------------------------------- */
/* -------------------------- RPC (un)marshalling -------------------------- */
/* ------------------------------------------------------------------------- */


void	srpcEncodeRpcHeader(struct srpcMsg *msg, btu32 xid, btu32 isDenied, btu32 status)
{
	srpcEncodeInt(msg, xid);
	srpcEncodeInt(msg, 1);	/* this is a reply */
	srpcEncodeInt(msg, isDenied);
	if(isDenied){
		srpcEncodeInt(msg, status);	/* reason for deny */
		if(status == 0){			/* RPC_MISMATCH */
			srpcEncodeInt(msg, 2);	/* we support only RPC version 2 */
			srpcEncodeInt(msg, 2);
		}else{						/* AUTH_ERROR */
			srpcEncodeInt(msg, 1);	/* AUTH_BADCRED: bad credentials */
		}
	}else{
		srpcEncodeInt(msg, 0);		/* AUTH_NULL */
		srpcEncodeVarBlock(msg, NULL, 0);	/* null length block */
		srpcEncodeInt(msg, status);
		if(status == 2){	/* if PROG_MISMATCH (version mismatch) */
			srpcEncodeInt(msg, 2);	/* we support only NFS version 2 */
			srpcEncodeInt(msg, 2);
		}
	}
	/* results will be marshalled after this header if status == SUCCESS */
}

void	srpcDecodeRpcHeader(struct srpcMsg *msg, srpcRequestHeader_t *hdr)
{
btu32	type;
btu8	*cred;	/* msg wrapper around credentials for unmarshalling */
btsize	credSize, storedSize, storedReadIndex;

	hdr->xid = srpcDecodeInt(msg);
	type = srpcDecodeInt(msg);
	if(type != 0){	/* it's a reply */
		EXC_THROW_MSG(srpcExcUnexpected, "reply from remote not expected");
	}else{			/* must be call: type == 0 */
		hdr->rpcvers = srpcDecodeInt(msg);
		hdr->prog = srpcDecodeInt(msg);
		hdr->vers = srpcDecodeInt(msg);
		hdr->proc = srpcDecodeInt(msg);
		hdr->credFlavor = srpcDecodeInt(msg);
		cred = srpcDecodeVarBlock(msg, &credSize);
		storedSize = msg->size;
		storedReadIndex = msg->readIndex;
		msg->readIndex = cred - msg->data;
		msg->size = msg->readIndex + credSize;
		srpcDecodeInt(msg);	/* skip stamp */
		srpcDecodeVarBlock(msg, &credSize);	/* parse string as varblock to avoid copy if no padding */
		hdr->credUid = srpcDecodeInt(msg);
		hdr->credGid = srpcDecodeInt(msg);
		/* don't decode array of gids, not needed */
		msg->readIndex = storedReadIndex;
		msg->size = storedSize;
		hdr->verfFlavor = srpcDecodeInt(msg);
		hdr->verfData = srpcDecodeVarBlock(msg, &hdr->verfLen);
	}
}

/* ------------------------------------------------------------------------- */
/* ------------------------- NFS2 (un)marshalling -------------------------- */
/* ------------------------------------------------------------------------- */

static inline void	srpcEncodeNfstime(struct srpcMsg *msg, nfstime *t)
{
	srpcEncodeInt(msg, t->seconds);
	srpcEncodeInt(msg, t->useconds);
}

static inline void	srpcDecodeNfstime(struct srpcMsg *msg, nfstime *t)
{
	t->seconds = srpcDecodeInt(msg);
	t->useconds = srpcDecodeInt(msg);
}

static void	srpcEncodeFattr(struct srpcMsg *msg, fattr *fa)
{
	srpcEncodeInt(msg, fa->type);
	srpcEncodeInt(msg, fa->mode);
	srpcEncodeInt(msg, fa->nlink);
	srpcEncodeInt(msg, fa->uid);
	srpcEncodeInt(msg, fa->gid);
	srpcEncodeInt(msg, fa->size);
	srpcEncodeInt(msg, fa->blocksize);
	srpcEncodeInt(msg, fa->rdev);
	srpcEncodeInt(msg, fa->blocks);
	srpcEncodeInt(msg, fa->fsid);
	srpcEncodeInt(msg, fa->fileid);
	srpcEncodeNfstime(msg, &fa->atime);
	srpcEncodeNfstime(msg, &fa->mtime);
	srpcEncodeNfstime(msg, &fa->ctime);
}


static void	srpcDecodeSattr(struct srpcMsg *msg, sattr *sa)
{
	sa->mode = srpcDecodeInt(msg);
	sa->uid = srpcDecodeInt(msg);
	sa->gid = srpcDecodeInt(msg);
	sa->size = srpcDecodeInt(msg);
	srpcDecodeNfstime(msg, &sa->atime);
	srpcDecodeNfstime(msg, &sa->mtime);
}

void	srpcEncodeAttrstat(struct srpcMsg *msg, attrstat *as)
{
	srpcEncodeInt(msg, as->status);
	if(as->status == NFS_OK){
		srpcEncodeFattr(msg, &as->attrstat_u.attributes);
	}
}

void	srpcDecodeSattrargs(struct srpcMsg *msg, sattrargs *sa)
{
	memcpy(sa->file.data, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
	srpcDecodeSattr(msg, &sa->attributes);
}

void	srpcDecodeDiropargs(struct srpcMsg *msg, diropargs *da)
{
	memcpy(da->dir.data, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
	da->name = srpcDecodeString(msg);
}

void	srpcEncodeDiropres(struct srpcMsg *msg, diropres *dr)
{
	srpcEncodeInt(msg, dr->status);
	if(dr->status == NFS_OK){
		srpcEncodeFixedBlock(msg, dr->diropres_u.diropres.file.data, NFS_FHSIZE);
		srpcEncodeFattr(msg, &dr->diropres_u.diropres.attributes);
	}
}

void	srpcEncodeReadlinkres(struct srpcMsg *msg, readlinkres *rr)
{
	srpcEncodeInt(msg, rr->status);
	if(rr->status == NFS_OK){
		srpcEncodeString(msg, rr->readlinkres_u.data);
	}
}

void	srpcDecodeReadargs(struct srpcMsg *msg, readargs *ra)
{
	memcpy(ra->file.data, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
	ra->offset = srpcDecodeInt(msg);
	ra->count = srpcDecodeInt(msg);
	ra->totalcount = srpcDecodeInt(msg);
}

void	srpcEncodeReadres(struct srpcMsg *msg, readres *rr)
{
	srpcEncodeInt(msg, rr->status);
	if(rr->status == NFS_OK){
		srpcEncodeFattr(msg, &rr->readres_u.reply.attributes);
		srpcEncodeVarBlock(msg, rr->readres_u.reply.data.data_val, rr->readres_u.reply.data.data_len);
	}
}

void	srpcDecodeWriteargs(struct srpcMsg *msg, writeargs *wa)
{
	memcpy(wa->file.data, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
	wa->beginoffset = srpcDecodeInt(msg);
	wa->offset = srpcDecodeInt(msg);
	wa->totalcount = srpcDecodeInt(msg);
	wa->data.data_val = srpcDecodeVarBlock(msg, &wa->data.data_len);
}

void	srpcDecodeCreateargs(struct srpcMsg *msg, createargs *ca)
{
	srpcDecodeDiropargs(msg, &ca->where);
	srpcDecodeSattr(msg, &ca->attributes);
}

void	srpcDecodeRenameargs(struct srpcMsg *msg, renameargs *ra)
{
	srpcDecodeDiropargs(msg, &ra->from);
	srpcDecodeDiropargs(msg, &ra->to);
}

void	srpcDecodeLinkargs(struct srpcMsg *msg, linkargs *la)
{
	memcpy(la->from.data, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
	srpcDecodeDiropargs(msg, &la->to);
}

void	srpcDecodeSymlinkargs(struct srpcMsg *msg, symlinkargs *sa)
{
	srpcDecodeDiropargs(msg, &sa->from);
	sa->to = srpcDecodeString(msg);
	srpcDecodeSattr(msg, &sa->attributes);
}

void	srpcDecodeReaddirargs(struct srpcMsg *msg, readdirargs *ra)
{
	memcpy(ra->dir.data, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
	memcpy(ra->cookie, srpcDecodeFixedBlock(msg, NFS_COOKIESIZE), NFS_COOKIESIZE);
	ra->count = srpcDecodeInt(msg);
}

static void	srpcEncodeEntry(struct srpcMsg *msg, entry *entry)
{
	if(entry == NULL){
		srpcEncodeInt(msg, 0);
	}else{
		srpcEncodeInt(msg, 1);
		srpcEncodeInt(msg, entry->fileid);
		srpcEncodeString(msg, entry->name);
		srpcEncodeFixedBlock(msg, entry->cookie, NFS_COOKIESIZE);
		srpcEncodeEntry(msg, entry->nextentry);
	}
}

void	srpcEncodeReaddirres(struct srpcMsg *msg, readdirres *rr)
{
	srpcEncodeInt(msg, rr->status);
	if(rr->status == NFS_OK){
		srpcEncodeEntry(msg, rr->readdirres_u.reply.entries);
		srpcEncodeInt(msg, rr->readdirres_u.reply.eof);
	}
}

void	srpcEncodeStatfsres(struct srpcMsg *msg, statfsres *sr)
{
	srpcEncodeInt(msg, sr->status);
	if(sr->status == NFS_OK){
		srpcEncodeInt(msg, sr->statfsres_u.reply.tsize);
		srpcEncodeInt(msg, sr->statfsres_u.reply.bsize);
		srpcEncodeInt(msg, sr->statfsres_u.reply.blocks);
		srpcEncodeInt(msg, sr->statfsres_u.reply.bfree);
		srpcEncodeInt(msg, sr->statfsres_u.reply.bavail);
	}
}

void	srpcEncodeNfsstat(struct srpcMsg *msg, nfsstat *s)
{
	srpcEncodeInt(msg, *s);
}

void	srpcDecodeNfs_fh(struct srpcMsg *msg, nfs_fh *fh)
{
	memcpy(fh, srpcDecodeFixedBlock(msg, NFS_FHSIZE), NFS_FHSIZE);
}

/* ------------------------------------------------------------------------- */
