/*
 * Name: backend.h
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-13
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: backend.h,v 1.2 1999/07/05 16:31:32 cs Exp $
 */

#ifndef __backend_H_INCLUDED__
#define __backend_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This header defines datatypes and function names for backend modules. The
functions described here must be implemented by every backend (except
otherwise noted). Backends are the source of the information that will be
mapped into the filesystem space, e.g. a remote CIFS file system.
*/


/* ------------------------------------------------------------------------- */
/* ------------------------------ Data Types ------------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct beFileTime{
	btu32	seconds;	/* seconds part of file time */
	btu32	nseconds;	/* nanoseconds part of file time */
}beFileTime_t;
/* This datatype is used to communicate file modification/creation/whatever
 * dates or other time-like information.
 */


typedef struct beVfsRes{
	btError			err;
	void			*user;				/* result of authentication */
	btu32			bytesPerAU;			/* size of allocation unit */
	btu32			totalAU;			/* total AUs on filesystem */
	btu32			availableAU;		/* AUs available to user */
	beFileTime_t	timeGranularity;	/* granularity of file times */
	btu32			features;			/* features of the filesystem */
	btint			maxFilenameLen;		/* max. length of path components */
}beVfsRes_t;
/* This datatype holds the result of a backend function that operates on
 * the entire filesystem (and not on one particular file). Not all fields
 * of this structure are always used, see the particular function for a
 * list of valid return fields.
 */

/* bitmask values for beVfsRes_t->features (mostly from NFS3): */
#define	BE_VFS_CAN_HARDLINK			1
#define	BE_VFS_CAN_SYMLINK			2
#define	BE_VFS_CAN_SETTIMES			4	/* can set file modification times */
#define	BE_VFS_IS_CASE_SENSITIVE	8
#define	BE_VFS_IS_CASE_PRESERVING	16


typedef struct beFileAttr{
	btint			type;		/* see the following defines */
	btu32			unixMode;	/* Unix permissions, octal interpretation */
	btint			numLinks;	/* set to 0 if not known */
	btint			uid;
	btint			gid;
	btu64			size;
	btu64			allocatedSize; 
	btint			devMajor;	/* only valid if type indicates device */
	btint			devMinor;	/* only valid if type indicates device */
	beFileTime_t	atime;		/* last access */
	beFileTime_t	mtime;		/* last modification */
	beFileTime_t	ctime;		/* last attribute change */
}beFileAttr_t;
/* This datatype holds attributes of a file in the filesystem. The attributes
 * are Unix-centric because all frontend modules will map the files into
 * a Unix filesystem. It's therefore the responsibility of the backend to map
 * the data source's attribute system to Unix attributes. This structure is
 * only used if attibutes are read, not for setting attributes.
 */

/* Possible values for beFileAttr_t->type (related to NFS2 file types): */
#define	BE_FILETYPE_REG		1
#define	BE_FILETYPE_DIR		2
#define	BE_FILETYPE_BLK		3
#define	BE_FILETYPE_CHR		4
#define	BE_FILETYPE_LNK		5
#define	BE_FILETYPE_SOCK	6
#define	BE_FILETYPE_BAD		7
#define	BE_FILETYPE_FIFO	8


typedef struct beFileSetAttr{
	btbool			setUnixMode;
	btbool			setUid;
	btbool			setGid;
	btbool			setSize;
	btbool			setAtime;
	btbool			setMtime;

	btu32			unixMode;
	btint			uid;		/* owner of file; local uid */
	btint			gid;		/* group of file; local gid */
	btu64			size;		/* if set: truncate file to this size */
	beFileTime_t	atime;		/* last access */
	beFileTime_t	mtime;		/* last modification */
}beFileSetAttr_t;
/* This datatype contains the information for setting file attributes. It's
 * also very Unix-centric for the same reasons as beFileAttr_t is. The
 * boolean fields at the beginning of the structure indicate whether the
 * related field should be used to set the file's attribute or whether it
 * should stay untouched.
 */


typedef struct beFileRes{
	btError			err;
	void			*file;
	beFileAttr_t	*attr;
	btbool			eof;		/* for readdir */
	btsize			dataLen;	/* for read, write must write all */
	btuni			*path;		/* for readlink */
}beFileRes_t;
/* This datatype describes the result of all file-specific backend functions.
 * Not all fields are valid after execution of every backend function. See
 * the notes for each function for more details.
 */


/* ------------------------------------------------------------------------- */
/* ------------------------ Backend Function Names ------------------------- */
/* ------------------------------------------------------------------------- */

/* --------------------- Filesystem specific functions --------------------- */

#define	BE_VFS_FOR_PATH		"vfsForPath"
/* Prototype: void *vfsForPath(btuni *path);
 * This function creates (if it did not already exist) or returns an existing
 * representation for a filesystem.
 * Parameters: 'path' specifies the filesystem object to be accessed. The
 *    interpretation and syntax of this string is backend dependent. However,
 *    it must *uniquely* describe the accessed object within the scope of the
 *    module.
 * Returns: an opaque pointer to an object representing the filesystem.
 * Referencing: The returned object is retained and must be released by the
 *    caller if it's not needed any more.
 */
#define	BE_VFS_RETAIN		"vfsRetain"
/* Prototype: void vfsRetain(void *obj);
 * This function increments the reference count of a filesystem object (created
 * or looked up with vfsForPath()).
 * Parameters: 'obj' must be a pointer obtained with vfsForPath().
 */
#define	BE_VFS_RELEASE		"vfsRelease"
/* Prototype: void vfsRelease(void *obj);
 * This function decrements the reference count of a filesystem object (created
 * or looked up with vfsForPath()). If the reference count drops to zero, the
 * object is freed and all associated resources are released. A backend module
 * may decide to delay freeing and to cache unused filesystem object for
 * future lookups.
 * Parameters: 'obj' must be a pointer obtained with vfsForPath().
 */
#define	BE_CONFIG_PATH		"vfsConfigPath"
/* Prototype: btint vfsConfigPath(void *obj, struct cfgProperty *path[], btint maxEntries);
 * This function retrieves a path of configuration sections where to look up
 * configuration entities for this filesystem representation. It's used by the
 * frontend to read backend-specific configuration options.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath(), 'path'
 *    is a pointer to an array of property entries that are searched in the
 *    order of increasing index, 'maxEntries' is the size of the array 'path'
 *    in number of entries.
 * Returns: The number of entries actually returned in 'path'.
 * Note: The pointers returned in 'path' must not be stored because they
 *    change if the config-file is re-parsed.
 */
#define	BE_VFS_CONNECT		"vfsConnect"
/* Prototype: void	vfsConnect(void *obj, void *user, void *caller);
 * This function connects the filesystem object to the underlying data source.
 * If such a connection is not required (e.g. if the data source is the local
 * filesystem), the function should just call the callback with no error.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath(), 'user'
 *    is a pointer to a authenticated user-object obtained with
 *    userAuthenticate() and 'caller' is the caller's context used for the
 *    calling back.
 * Callback: beVfsRes_t with the field 'err' set.
 */
#define	BE_VFS_DISCONNECT	"vfsDisconnect"
/* Prototype: void vfsDisconnect(void *obj, void *caller);
 * This function disconnects the filesystem object from the underlying data
 * source, if there is such a thing as a "connection" in the backend.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath(),
 *    'caller' is the caller's context used for the callback.
 * Callback: beVfsRes_t with the field 'err' set.
 */
#define	BE_VFS_ROOT_FILE	"vfsRootFile"
/* Prototype: void *vfsRootFile(void *obj);
 * This function obtains the representation of the root directory for a
 * given filesystem.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath().
 * Returns: a pointer to an object representing the root directory.
 * Referencing: The returned object is retained, it must be released by
 *    the caller with fileRelease() if it's not needed any more.
 */
#define	BE_VFS_STATFS		"vfsStatfs"
/* Prototype: void vfsStatfs(void *obj, void *user, void *caller);
 * This function retrieves the filesystem's properties.
 * The filesystem must be in the connected state.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath(),
 *    'user' is representation of an authenticated user obtained wtih
 *    userAuthenticate() and 'caller' is the caller's context.
 * Callback: beVfsRes_t with the fields 'err', 'bytesPerAU', 'totalAU',
 *    'availableAU', 'timeGranularity', 'features' and 'maxFilenameLen' set.
 */
#define	BE_VFS_PATH			"vfsPath"
/* Prototype: btuni *vfsPath(void *obj);
 * This function returns the string that was used to create the filesystem
 * object. It may save the module's user the work of storing this string.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath().
 * Returns: A pointer to an unicode string uniquely identifying the
 *    filesystem. The string is valid as long as the filesystem object
 *    is not freed. It must not be modified by the caller.
 */
#define	BE_VFS_CASE_SENSITIVE	"vfsCaseSensitive"
/* Prototype: btbool vfsCaseSensitive(void *obj);
 * This function returns a flag indicating whether filenames are casesensitive
 * in the backend's filesystem. The same information can be obtained with
 * vfsStatfs(), but this function does not require the filesystem to be
 * connected.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath().
 * Returns: 1 if filenames are casesensitive, 0 if they are not.
 */
#define	BE_VFS_OPEN_NOTIFICATION	"vfsOpenNotification"
/* Prototype: struct notificationCenter *vfsOpenNotification(void *obj);
 * This function returns a pointer to a notification center which fires each
 * time a file changes it's open mode. The notification's parameter is a
 * pointer to the opening file object.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath().
 * Returns: a pointer to a notification center internally stored by the
 *    filesystem object.
 * Referencing: The returned notification center is *not* retained. If the
 *    caller stores a pointer, it should retain the object.
 */
#define	BE_VFS_INVALIDATE_NOTIFICATION	"vfsInvalidateNotification"
/* Prototype: struct notificationCenter *vfsInvalidateNotification(void *obj);
 * This function returns a pointer to a notification center which fires each
 * time a file is invalidated (e.g. if it is removed). The notification's
 * parameter is a pointer to the file object that became invalid.
 * The user of the backend module is expected to destroy all references to the
 * file and call fileRelease() the appropriate number of times.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath().
 * Returns: a pointer to a notification center internally stored by the
 *    filesystem object.
 * Referencing: The returned notification center is *not* retained. If the
 *    caller stores a pointer, it should retain the object.
 */
#define	BE_USER_AUTHENTICATE	"userAuthenticate"
/* Prototype: void userAuth(void *obj, btint unixUid, void *caller);
 * This function establishes a mapping between a local Unix user (referenced
 * by Unix UID) and a user representation on the filesystem. This mapping may
 * involve asking the user for a password, going through a mapping database,
 * or, if the filesystem's data source is the local filesystem, just a one by
 * one correspondence.
 * Parameters: 'obj' is a filesystem object obtained with vfsForPath(),
 *    'unixUid' is the Unix UID of the user that wants to access data and
 *    'caller' is the caller's context.
 * Callback: beVfsRes_t with the fields 'err' and 'user' set.
 * Referencing: The returned user-representation object is *not* retained.
 * If the caller stores a pointer, it should retain the object.
 */
#define	BE_USER_RETAIN		"userRetain"
/* Prototype: void userRetain(void *obj);
 * This function increments the reference count of a user-representation
 * object.
 * Parameters: 'obj' is a pointer to a user-representation object obtained with
 *    userAuthenticate().
 */
#define	BE_USER_RELEASE		"userRelease"	/* void userRelease(void *obj) */
/* Prototype: void userRelease(void *obj);
 * This function decrements the reference count of a user-representation
 * object.
 * Parameters: 'obj' is a pointer to a user-representation object obtained with
 *    userAuthenticate().
 */


/* ------------------- File/Directory specific functions ------------------- */

/* Note: Not all of these functions must be implemented by a backend.
 * If a function is not implemented, the associated filesystem access will
 * fail. A minimum functional implementation should at least implement:
 * BE_FILE_LOOKUP, BE_FILE_RETAIN, BE_FILE_RELEASE, BE_FILE_NFSATTR,
 * BE_FILE_READDIR, BE_FILE_NAME, BE_FILE_PARENT
 * The implementation pointers for other functions may be NULL.
 */

#define	BE_FILE_LOOKUP		"fileLookup"
/* Prototype: void fileLookup(void *parent, btuni *name, void *user, void *caller);
 * This function creates or returns an already created file object. The file
 * is specified by passing the parent directory's file object representation
 * and the name of the file.
 * Parameters: 'parent' is the parent directory's file object, obtained either
 *    with vfsRootFile() or with fileLookup(), 'name' is the name of the file
 *    that should be represented (within the parent directory's context) and
 *    'caller' is the caller's context for callbacks.
 * Callback: beFileRes_t with fields 'err' and 'file' set.
 * Referencing: The file object returned in beFileRes_t->file is retained. The
 *    caller must release it with fileRelease() when it does not reference it
 *    any more.
 */
#define	BE_FILE_RETAIN		"fileRetain"
/* Prototype: void fileRetain(void *obj);
 * This function increments the reference count for a file object.
 * Parameters: 'obj' is a file object obtained with fileLookup() or
 *    vfsRootFile().
 */
#define	BE_FILE_RELEASE		"fileRelease"
/* Prototype: void fileRelease(void *obj);
 * This function decrements the reference count for a file object. If the
 * reference count reaches zero, the file representation object is freed.
 * Parameters: 'obj' is a file object obtained with fileLookup() or
 *    vfsRootFile().
 */
#define	BE_FILE_GETATTR		"fileGetAttributes"
/* Prototype: void fileGetAttributes(void *obj, void *user, void *caller);
 * This function retrieves the file's attributes.
 * Parameters: 'obj' is a file object obtained with fileLookup()/vfsRootFile(),
 *    'user' is a user representation obtained with userAuthenticate() and
 *    'caller' is the caller's context.
 * Callback: beFileRes_t with fields 'err' and 'attr' set.
 */
#define	BE_FILE_SETATTR		"fileSetAttributes"
/* Prototype: void fileSetAttributes(void *obj, void *user, struct beFileSetAttr *attr, void *caller);
 * This function sets a file's attributes.
 * Parameters: 'obj' is a file object obtained with fileLookup()/vfsRootFile(),
 *    'user' is a user representation obtained with userAuthenticate(),
 *    'attr' is the description of the new attributes the file should have.
 *    Note that the backend module must copy 'attr' if it does not call the
 *    callback immediately. 'caller' is the caller's context.
 * Callback: beFileRes_t with the field 'err' set.
 */
#define	BE_FILE_READDIR		"fileReaddir"
/* Prototype: void fileReaddir(void *obj, void *user, btint startIndex, btbool (*resultCallback)(void *callerCtx, btuni *name, btu32 continuationIndex, btbool isDir), void *caller);
 * This function retrieves a directory listing. For each file found in the
 * directory 'obj', it calls the function 'resultCallback' which is passed as
 * a parameter. The listing may start at any offset given in 'startIndex'. A
 * start index of 0 means the beginning of the list. Subsequent indices for
 * continuation must be taken from the callback parameter.
 * fileReaddir() may return any number of directory entries it likes. However,
 * the caller may abort the list at any point by means of the return value of
 * 'resultCallback'.
 * Parameters: 'obj' is a directory file object for which the content should be
 *    listed, 'user' is a user representation obtained with userAuthenticate(),
 *    'startIndex' is where in the directory the listing should begin (used
 *    to continue aborted listings), 'resultCallback' is a function pointer to
 *    a function that's called for each directory entry with the following
 *    parameters: The caller's context, the filename of the entry, a
 *    continuation index for continuation after this entry and a flag whether
 *    the file is a directory as parameter. If it can't be known whether the
 *    file is a directory or not, the 'isDir' flag should be set to 1.
 *    If 'resultCallback' returns 0, the listing is aborted.
 *    'caller' is the caller's context for callbacks.
 * Callback: beFileRes_t with fields 'err' and 'eof' set.
 * Note: A good backend implementation should try to get as much information
 *    as possible during the directory listing. It should then store the
 *    information obtained in file representation objects, which are cached
 *    and returned by fileLookup() if requested. It's therefore not necessary
 *    to pass file attributes to the callback function 'resultCallback' because
 *    they can be obtained with fileGetAttributes() without performance loss.
 *    This is particularly important for NFS3, which has a readdir function
 *    that returns file attributes.
 *    The values passed for 'continuationIndex' to 'resultCallback' must all
 *    be smaller integers than the directory's file size attribute. This is
 *    to work around a bug on various NFS client implementations.
 */
#define	BE_FILE_CLOSEDIR	"fileClosedir"
/* Prototype: void fileClosedir(void *obj, void *caller);
 * This function is called to inform the backend that the module user does not
 * intend to continue current directory listings in the directory 'obj'. It is
 * advisory only. The backend module does not rely on the frontend to call this
 * function because many frontends won't call it. Backends should have a cache
 * of open searches and close them after a timeout or when too many searches
 * are open.
 * Parameters: 'obj' is a file object obtained with fileLookup()/vfsRootFile(),
 *    'caller' is the caller's context.
 * Callback: beFileRes_t with field 'err' set.
 */
#define	BE_FILE_CREATE		"fileCreate"
/* Prototype: void fileCreate(void *parent, btuni *name, btbool isDir, void *user, struct beFileSetAttr *attr, void *caller);
 * This function creates a new file or directory in the directory represented
 * by 'parent'.
 * Parameters: 'parent' is a directory file object obtained with
 *    fileLookup()/vfsRootFile(), 'name' is the name of the object that should
 *    be created, 'isDir' is a boolean variable indicating whether the created
 *    object should be a directory, 'user' is a user representation obtained
 *    with userAuthenticate(), 'attr' are the attributes the newly created
 *    object should have and 'caller' is the caller's context.
 * Callback: beFileRes_t with fields 'err' and 'file' set.
 * Referencing: The representation of the newly created object returned in
 *    beFileRes_t->file is retained. It must be released by the caller when it
 *    is no longer referenced.
 */
#define	BE_FILE_REMOVE		"fileRemoveChild"
/* Prototype: void fileRemoveChild(void *parent, btuni *name, btbool isDir, void *user, void *caller);
 * This function is used to delete a file or an empty directory. It also
 * invalidates any representations obtained for the deleted object and fires
 * the invalidate-notification accordingly.
 * Parameters: 'parent' is the representation of the objects parent directory,
 *    'name' is the name of the object that should be deleted, 'isDir' is a
 *    boolean variable indicating whether a directory or a file should be
 *    deleted, 'user' is a user representation obtained with userAuthenticate()
 *    and 'caller' is the caller's context.
 * Callback: beFileRes_t with field 'err' set.
 */
#define	BE_FILE_RENAME		"fileRename"
/* Prototype: void fileRename(void *fromDir, btuni *fromName, void *toDir, btuni *toName, void *user, void *caller);
 * This function moves a file or directory in the hierarchy and/or renames it.
 * Parameters: 'fromDir' is the representation of the object's parent
 *    directory, 'fromName' is the name of the object that should be moved,
 *    'toDir' is the representation of the parent for the file after moving,
 *    'toName' is the new name the object should have, 'user' is a user
 *    representation obtained with userAuthenticate() and 'caller' is the
 *    caller's context.
 * Callback: beFileRes_t with field 'err' set.
 * Note: If there is already a file at the destination position, this file
 *    is deleted (if possible). During this deletion process, an invalidate-
 *    notification is fired for the deleted object.
 */
#define	BE_FILE_RW			"fileRW"
/* Prototype: void fileRW(void *obj, btint operation, btbool writeThrough, btu8 *data, btoffset offset, btsize len, void *user, void *caller);
 * This function reads or writes data from/to a file. If the file is not
 * already open in an appropriate mode, the file is opened automatically in
 * the least required mode (open for read if read, open for read/write if
 * write) and with the autoclose option set. It's up to the backend module how
 * autmatic closing is implemented; it should resemble a closed file as good
 * as possible. If the file changes open mode, the open-notification is fired.
 * Parameters: 'obj' is the representation of the file that's accessed,
 *    'operation' is the operation that should be performed on the file
 *    (may be any of BE_FILE_OP_READ or BE_FILE_OP_WRITE), 'writeThrough' is
 *    a boolean variable that indicates whether a synchronous write is required
 *    (e.g. for NFS2) or an asynchronous write (delayed write from cache) is
 *    OK (e.g. for cached write of NFS3). 'data' is a pointer to the data that
 *    should be written or to an area where the read data should be placed.
 *    'offset' is the offset in the file in bytes where the operation should
 *    start, 'len' is the length of the requested operation in bytes, 'user' is
 *    a user representation obtained with userAuthenticate() and 'caller' is
 *    the caller's context for callbacks.
 * Callback: beFileRes_t with fields 'err' and 'dataLen' set. 'dataLen' is the
 *    length actually read in a read operation. Write operations don't succeed
 *    if they write less than requested.
 */
/* Constants representing file access operations: */
#define	BE_FILE_OP_NONE		0
#define	BE_FILE_OP_READ		1
#define	BE_FILE_OP_WRITE	2

#define	BE_FILE_HARDLINK	"fileHardLink"
/* Prototype: fileHardLink(void *from, void *toDir, btuni *name, void *user, void *caller);
 * This function creates a new hard link object in the filesystem. Please note
 * that supporting hard links correctly probably requires additional interface
 * functions because the various links to a file must all have the same inode
 * number and there's currently no concept of inode numbers in the backend
 * interface.
 * Parameters: 'from' is the representation of an existing object that should
 *    get an additional link, 'toDir' is the representation of the parent
 *    directory where the new link should be created, 'name' is the new
 *    object's name, 'user' is a user representation obtained with
 *    userAuthenticate() and 'caller' is the caller's context.
 * Callback: beFileRes_t with fields 'err' and 'file' set. 'file' points to
 *    the newly created link.
 * Referencing: beFileRes_t->file is retained. It must be released by the
 *    caller when it does not need the reference any more.
 */
#define	BE_FILE_SYMLINK		"fileSymlink"
/* Prototype: void fileSymlink(void *inDir, btuni *name, btuni *linkPath, void *user, struct beFileSetAttr *attr, void *caller);
 * This function creates a new symbolic link object in the filesystem.
 * Parameters: 'inDir' is the representation of the parent directory where the
 *    new object will be created, 'name' is the name of the new symbolic link,
 *    'linkPath' is the path to which the symbolic link points, 'user' is
 *    a user representation obtained with userAuthenticate(), 'attr' are the
 *    attributes that should be given to the newly created object and 'caller'
 *    is the caller's context for callbacks.
 * Callback: beFileRes_t with fields 'err' and 'file' set.
 * Referencing: beFileRes_t->file is a retained object. It must be released
 *     by the caller.
 */
#define	BE_FILE_READLINK	"fileReadlink"
/* Prototype: void fileReadlink(void *obj, void *user, void *caller);
 * This function retrieves the path a symbolic link object points to.
 * Parameters: 'obj' is the representation of the symbolic link object, 'user'
 *    is the user representation obtained with userAuthenticate() and 'caller'
 *    is the caller's context for callbacks.
 * Callback: beFileRes_t with fields 'err' and 'path' set.
 */
#define	BE_FILE_OPENMODE	"fileOpenMode"
/* Prototype: btint fileOpenMode(void *obj);
 * This function retrieves the current open mode of a file.
 * Parameters: 'obj' is the representation of the file object.
 * Returns: The current open mode.
 */
/* Constants representing open modes: */
#define	BE_FILE_CLOSED		0
#define	BE_FILE_OPEN_READ	1
#define	BE_FILE_OPEN_WRITE	2
#define	BE_FILE_OPEN_RW		3	/* logical OR of READ and WRITE */
#define	BE_FILE_AUTOCLOSE	4	/* bitmask value, may be ORed to READ, WRITE, RW */

#define	BE_FILE_SET_OPENMODE "fileSetOpenMode"
/* Prototype: void fileSetOpenMode(void *obj, btint mode, btint modeMask, btbool doAutoclose, void *user, void *caller);
 * This function is used to open or close a file. If the autoclose attribute
 * is set for an open file, the backend may close the file whenever it wants.
 * The backend should also close these files after a timeout or when other
 * processes request an (exclusive) open.
 * Parameters: 'obj' is the representation of the file object to be opened or
 *    closed, 'mode' is the new open mode requested, 'modeMask' is a bitmask
 *    that defines which bits of the open-mode are relevant and which are not.
 *    Relevant bits are marked with a '1', irrelevant bits with '0'. Irrelevant
 *    bits will not be changed from their current state. 'doAutoclose' tells
 *    the backend that it is responsible for closing the file. 'user' is a user
 *    representation obtained with userAuthenticate(). If the file should be
 *    closed, a NULL pointer may be passed for 'user'. The NULL user may always
 *    close files. 'caller' is the caller's context.
 * Callback: beFileRes_t with field 'err' set.
 */
#define	BE_FILE_OP_RUNNING	"fileOpRunning"
/* Prototype:  btbool fileOpRunning(void *obj);
 * This function returns whether there are any operations running for the given
 * file object, such as read, write, open, close, directory read or whatever.
 * Parameters: 'obj' is the representation of the queried object.
 * Returns: 1 if there are operations running, 0 otherwise.
 */
#define	BE_FILE_NAME		"fileName"
/* Prototype: btuni *fileName(void *obj);
 * This function retrieves the name of a given file object in it's parent
 * directory's context. It's the same name that was used in fileLookup().
 * This function may save the module user the work of storing this string.
 * Parameters: 'obj' is the representation of the queried object.
 * Returns: The name of the object. The name of the root directory may be a
 *    NULL pointer, but it may also be any string. The returned string is valid
 *    until the file object is freed.
 */
#define	BE_FILE_PARENT		"fileParent"
/* Prototype: void *fileParent(void *obj);
 * This function returns the representation of a file's parent directory.
 * Parameters: 'obj' is the representation of the file for which the parent
 *    is queried.
 * Returns: A pointer to the non-retained parent object. If the caller stores
 *    this pointer, it should retain the returned object.
 */


/* Note: This interface may be extended compatibly. If a frontend and a backend
 * agree on a set of extensions, they may use it. Extensions might include
 * methods to retrieve extended file attributes such as access control lists.
 */

#endif /* __backend_H_INCLUDED__ */
