/*
 * Name: frontend.h
 * Project: Sharity
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1999-02-13
 * Tabsize: 4
 * Copyright: (c) 1999 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: frontend.h,v 1.1.1.1 1999/04/12 13:56:43 cs Exp $
 */

#ifndef __frontend_H_INCLUDED__
#define __frontend_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This header defines datatypes and function names for frontend modules. The
functions described here must be implemented by every frontend. The
responsibility of frontend modules is to map the data represented by a backend
module into the operating system's filesystem. There may be NFS2, NFS3, kernel
VFS based or even CIFS based frontends.
*/

/* ------------------------------------------------------------------------- */
/* ------------------------------ Data Types ------------------------------- */
/* ------------------------------------------------------------------------- */

typedef struct feRes{
	btError	err;
}feRes_t;
/* This datatype holds the result of any frontend module functions. */

/* ------------------------------------------------------------------------- */
/* ------------------------ Frontend Function Names ------------------------ */
/* ------------------------------------------------------------------------- */

#define	FE_MNT_NEW			"mntNew"
/* Prototype: void *mntNew(btu8 *mountpoint);
 * This function creates a new mountpoint representation object. The mount
 * path must be known at object creation time and should be normalized (all
 * symbolic links resolved, /../ and /./ removed, etc.).
 * Parameters: 'mountpoint' is the ASCII path in the local filesystem where
 *    this object will be mounted.
 * Returns: A pointer to a retained mountpoint representation. The caller is
 *    responsible for releaseing the object when it's no longer needed.
 */
#define	FE_MNT_RETAIN		"mntRetain"
/* Prototype: void mntRetain(void *obj);
 * This function increments the reference count of a mountpoint representation.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew().
 */
#define	FE_MNT_RELEASE		"mntRelease"
/* Prototype: void mntRelease(void *obj);
 * This function decrements the reference count of a mountpoint representation.
 * If the reference count reaches zero, the object is freed.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew().
 */
#define	FE_MNT_SET_BACKEND	"mntSetBackend"
/* Prototype: void mntSetBackend(void *obj, struct module *backendMod, void *backendObj);
 * This function connects the frontend to a backend module.
 * Parameters: 'obj' is a mountpoint representation obtained with mntNew(),
 *    'backendMod' is the backend module description, 'backendObj' is an
 *    instance of a backend filesystem representation obtained with
 *    vfsForPath().
 */
#define	FE_MNT_GET_BACKEND	"mntGetBackend"
/* Prototype: void mntGetBackend(void *obj, struct module **backendMod, void **backendObj);
 * This function retrieves the data set with mntSetBackend() for the module
 * user's convenience.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew(),
 *    'backendMod' is a pointer to a location where a pointer to the backend
 *    module description will be stored and 'backendObj' is a pointer to a
 *    location where a pointer to the backend filesystem representation will
 *    be stored.
 */
#define	FE_MNT_SET_FAKEMODE	"mntSetFakeMode"
/* Prototype: void mntSetFakeMode(void *obj, btbool fakeMode);
 * This function is used to turn the fake-mode feature of the mountpoint on or
 * off. In fake-mode, all requests are answered with a minimum-damage reply,
 * which is an error in most cases. This is useful if the backend MUST NOT be
 * queried for whatever reason, but the frontend will have to answer requests.
 * The typical application is to make the mountpoint unmountable if the
 * backend module does not respond or to allow mounting although the backend
 * is not fully initialized.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew(),
 *    if 'fakeMode' is 1, the mountpoint is set into fake mode.
 */
#define	FE_MNT_MOUNT		"mntMount"
/* Prototype: void mntMount(void *obj, btint uid, btbool readOnly, void *caller);
 * This function actually establishes the mount. Calling mntNew() does not
 * change anything in the filesystem until mntMount() is called. NO access
 * permission check is performed! The caller is responsible for allowing only
 * privileged users to mount filesystems.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew(),
 *    'uid' is the user who establshes the mount, 'readOnly' is 1 if the
 *    filesystem should be mounted with the "read-only" attribute set and
 *    'caller' is the caller's context for callbacks.
 * Callback: feRes_t with field 'err' set.
 */
#define	FE_MNT_UMOUNT		"mntUmount"
/* Prototype: void mntUmount(void *obj, btint uid, void *caller);
 * This function unmounts a mountpoint representation previously mounted with
 * mntMount().
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew()
 *    and previously mounted with mntMount(), 'uid' is the local Unix user ID
 *    of the user who attempts the unmount and 'caller' is the caller's
 *    context.
 * Callback: feRes_t with field 'err' set.
 */
#define	FE_MNT_IS_MOUNTED	"mntIsMounted"
/* Prototype: btbool mntIsMounted(void *obj);
 * This function queries a mountpoint representation for whether it's already
 * mounted or not. This is a convenience function for the module user.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew().
 * Returns: 1 if it is mounted, 0 if it is not.
 */
#define	FE_MNT_IS_READONLY	"mntIsReadonly"
/* Prototype: btbool mntIsReadonly(void *obj);
 * This function queries a mountpoint representation for whether it has been
 * mounted read-only. It's a convenience method for the module user.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew().
 * Returns: 1 if it is mounted read-only, 0 otherwise.
 */
#define	FE_MNT_MOUNTPOINT	"mntMountpoint"
/* Prototype: btu8 *mntMountpoint(void *obj);
 * This function retrieves the string passed for 'mountpoint' at object
 * creation time. It's a convenience method for the module user who may
 * save the effort to store this string.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew().
 * Returns: A pointer to the mountpoint string as it was passed to mntNew().
 *    The string is valid as long as the mountpoint representation is not
 *    freed.
 */
#define	FE_MNT_MOUNTED_BY	"mntMountedBy"
/* Prototype: btint mntMountedBy(void *obj);
 * This function retrieves the Unix user ID of the user who mounted the
 * mountpoint.
 * Parameters: 'obj' is the mountpoint representation obtained with mntNew().
 * Returns: The Unix user ID of the user who mounted the mountpoint.
 */
#define	FE_CFGENTRIES_FOR_BACKEND	"mntCfgentriesForBackend"
/* struct array *mntCfgentriesForBackend(void);
 * This function returns an array of guiEntry_t objects that describe the part
 * of the frontend configuration that should be configurable on a per-backend
 * basis as part of the backend configuration.
 * Returns: An array of guiEntry_t structures that must be freed by the caller.
 */


#endif /* __frontend_H_INCLUDED__ */
