/*
 * Name: ipcclient.c
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-12
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: ipcclient.c,v 1.3 1999/06/09 13:36:55 cs Exp $
 */


#include "stdheaders.h"
#include "logging.h"
#include "ipcclient.h"
#include "ipc.h"
#include "uidsecure.h"
#include "system.h"
#include "errcodes.h"
#include "maininterface.h"
#include "cifsinterface.h"
#include "util.h"
#include "hashtable.h"
#include "memalloc.h"

/* ------------------------------------------------------------------------- */

typedef struct ipcClientConnection{
	ipcMsg_t	*msg;	/* the message object used for rx and tx */
	btint		sid;	/* sequence ID counter for outgoing requests */
	btint		fd;		/* file descriptor of connection */
	btcallback	requestHandler;		/* callback called for incoming requests */
	void		*requestCtx;		/* context of requestHandler */
	btint		currentFnNumber;	/* current IPC function number */
}ipcClientConnection_t;

static ipcClientConnection_t	client;

/* the following variable can be changed by the module user: */
btbool	ipcclientDebug = 0;

/* ------------------------------------------------------------------------- */

static btint	ipccModuleCompare(void *m1, void *m2)
{
ipccModuleInfo_t	*this = m1, *other = m2;

	return utilStrCmp(this->name, other->name, 1);
}

static btu32	ipccModuleHashVal(void *obj)
{
ipccModuleInfo_t	*mi = obj;

	return utilStrHash(mi->name, 1);
}

static collectionObjectI_t	moduleInfoInterface = {
	ipccModuleCompare,
	ipccModuleHashVal,
	NULL,
	NULL,
};

static hashtable_t	*moduleInfos = NULL;

/* ------------------------------------------------------------------------- */

void	ipccSetRequestHandler(btcallback handler, void *ctx)
{
	client.requestHandler = handler;
	client.requestCtx = ctx;
}

/* ------------------------------------------------------------------------- */

typedef struct pversionRange{
	btint	fnBase;
	btint	fnCount;
	btint	protocolVersion;
}pversionRange_t;

static array_t	*protVersions = NULL;

void	ipccSetProtocolVersion(btint fnBase, btint fnCount, btint version)
{
pversionRange_t	*pv;

	if(protVersions == NULL)
		protVersions = arrNew(NULL, 8);
	pv = memCalloc(sizeof(*pv));
	pv->fnBase = fnBase;
	pv->fnCount = fnCount;
	pv->protocolVersion = version;
	arrAddObject(protVersions, pv);
}

btint	ipccProtocolVersion(btint fnNumber)
{
btint			i, cnt;
pversionRange_t	*pv;

	if(protVersions == NULL)
		return 0;
	cnt = arrObjectCount(protVersions);
	for(i=0;i<cnt;i++){
		pv = arrObjectAtIndex(protVersions, i);
		if(fnNumber >= pv->fnBase && fnNumber - pv->fnBase < pv->fnCount)
			return pv->protocolVersion;
	}
	return 0;
}

/* ------------------------------------------------------------------------- */

btint	ipccConnect(void)
{
struct sockaddr_un	sockAddr;
int					rval, errstore;

	client.msg = ipcMsgNew();
	client.sid = 0;
	client.fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if(client.fd < 0){
		logPrintf(LOGM_ERROR, "UNIX: Error creating socket: [%d] %s\n", errno, sysStrerror(errno));
		exit(1);
	}
	sockAddr.sun_family = AF_UNIX;
	strcpy(sockAddr.sun_path, MAIN_IPC_SOCKET_PATH);
	uidsecBecomeRoot();	/* connect must be done as root */
	rval = connect(client.fd, (struct sockaddr *)&sockAddr, sizeof(sockAddr));
	errstore = errno;
	uidsecBecomeUser();
	if(rval != 0){
		logPrintf(LOGM_ERROR, "UNIX: Error connecting to daemon: [%d] %s\n", errstore, sysStrerror(errstore));
		exit(1);
	}
	return client.fd;
}

/* ------------------------------------------------------------------------- */

void	ipccStartProc(ipcMsg_t *msg, int msgId)
{
	ipcMsgClear(msg);
	ipcAddInt(msg, 0);	/* space for total length */
	ipcAddInt(msg, ipccProtocolVersion(msgId));
	ipcAddInt(msg, client.sid++);
	ipcAddInt(msg, msgId);
	ipcAddInt(msg, uidsecUserId());
	ipcAddInt(msg, getgid());
	client.currentFnNumber = msgId;
}

/* ------------------------------------------------------------------------- */

static void	dumpMessage(btu8 *title, btu8 *data, btsize len, btint initialLen)
{
int		i;
btu8	buffer[256];
int		lineLen = initialLen;

	while(len > 0){
		strcpy(buffer, title);
		for(i=0;i<lineLen && len>0;i++){
			sprintf(buffer + strlen(buffer), " %02x", *data);
			data++;
			len--;
		}
		lineLen = 16;	/* further lines will be 16 long */
		fprintf(stderr, "%s\n", buffer);
	}
}

/* ------------------------------------------------------------------------- */

void	ipccWriteMsg(ipcMsg_t *msg)
{
btsize	size, written, len;

	size = msg->size;
	msg->size = 0;		/* hack to insert total size at the beginning */
	ipcAddInt(msg, size);
	if(ipcclientDebug){
		dumpMessage("<- Tx:", msg->buffer, size, 4);
	}
	for(written = 0; size > 0;){
		if((len = write(client.fd, msg->buffer + written, size)) <= 0){
			if(len < 0)
				logPrintf(LOGM_ERROR, "UNIX: Error writing to daemon pipe: [%d] %s\n", errno, sysStrerror(errno));
			else
				logPrintf(LOGM_ERROR, "LOC: Broken pipe to daemon on write\n");
			exit(1);
		}
		written += len;
		size -= len;
	}
}

/* ------------------------------------------------------------------------- */

static void	readAll(btint fd, btu8 *buffer, btsize size)
{
btsize	len, alreadyRead;

	for(alreadyRead = 0; size > 0;){
		if((len = read(fd, buffer + alreadyRead, size)) <= 0){
			if(client.requestHandler != NULL)		/* inform callback about broken pipe */
				client.requestHandler(client.requestCtx, NULL);
			if(len < 0)
				logPrintf(LOGM_ERROR, "UNIX: Error reading from daemon pipe: [%d] %s\n", errno, sysStrerror(errno));
			else
				logPrintf(LOGM_ERROR, "LOC: Broken daemon pipe on read\n");
			exit(1);
		}
		alreadyRead += len;
		size -= len;
	}
}

btint	ipccReadMsg(ipcMsg_t *msg)
{
btsize	size;
btu32	pvers;

	ipcMsgClear(msg);
	ipcMsgEnsureSize(msg, 4);
	readAll(client.fd, msg->buffer, 4);
	msg->size = 4;
	if(ipcclientDebug){
		dumpMessage("-> Rx:", msg->buffer, 4, 4);
	}
	size = ipcGetInt(msg) - 4;
	ipcMsgClear(msg);
	ipcMsgEnsureSize(msg, size);
	msg->size = size;
	readAll(client.fd, msg->buffer, size);
	if(ipcclientDebug){
		dumpMessage("-> Rx:", msg->buffer, msg->size, 16);
	}
	pvers = ipcGetInt(msg);	/* protocol version */
	return pvers;
}

/* ------------------------------------------------------------------------- */

static void	protMismatchErr(btint ourProtocol, btint daemonProtocol)
{
	logPrintf(LOGM_ERROR, "LOC: Protocol version mismatch with daemon: daemon has %d instead of %d\n"
	"You are running a user interface version incompatible with the daemon!\n", daemonProtocol, ourProtocol);
	exit(1);
}

void	ipccRpc(ipcMsg_t *msg)
{
ipccRequestData_t	data;
btint				sentVers, pvers, expectedVers;

	ipccWriteMsg(msg);
	sentVers = ipccProtocolVersion(client.currentFnNumber);
	for(;;){
		pvers = ipccReadMsg(msg);
		data.sid = ipcGetInt(msg);
		data.procId = ipcGetInt(msg);
		if(data.procId == -1)	/* this is the reply we are waiting for */
			break;
		data.msg = msg;
		data.protocolVersion = pvers;
		if(client.requestHandler != NULL){
			expectedVers = ipccProtocolVersion(data.procId);
			if(expectedVers != 0 && pvers != expectedVers)
				protMismatchErr(expectedVers, pvers);
			client.requestHandler(client.requestCtx, &data);
		}else{	/* create a dummy error reply */
			ipcMsgClear(msg);
			ipcAddInt(msg, 0);		/* space for total length */
			ipcAddInt(msg, pvers);	/* reply with same version as request */
			ipcAddInt(msg, data.sid);
			ipcAddInt(msg, -1);
			ipcAddInt(msg, uidsecUserId());
			ipcAddInt(msg, getgid());
			ipcAddInt(msg, ERRCLS_FRAMEWORK);
			ipcAddInt(msg, ERR_FRWipcparameters);
		}
		ipccWriteMsg(msg);
	}
	if(sentVers != pvers)
		protMismatchErr(sentVers, pvers);
	ipcGetInt(msg);	/* skip uid */
	ipcGetInt(msg);	/* skip gid */
}

/* ------------------------------------------------------------------------- */

ipcMsg_t	*ipccCurrentRxMessage(void)
{
	return client.msg;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

static void	unmarshalResult(ipcMsg_t *msg, ipccRes_t *res)
{
	res->err.class = ipcGetInt(msg);
	res->err.code = ipcGetInt(msg);
	res->errmsg = ipcGetAscii(msg);
}

void	ipccPrintErr(ipccRes_t *res)
{
	logPrintf(LOGM_ERROR, "%s\n", res->errmsg);
}

/* ------------------------------------------------------------------------- */

void	ipccListModules(ipccRes_t *res)
{
	ipccStartProc(client.msg, MAIN_IPCFN_LIST_MODULES);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

ipccModuleInfo_t	*ipccGetModuleInfo(ipccRes_t *res, btu8 *moduleName)
{
btint				cnt, i;
ipccModuleInfo_t	*mi;

	if(moduleInfos == NULL){
		moduleInfos = htabNew(&moduleInfoInterface, 75, 8);
		ipccListModules(res);
		if(res->err.class != 0){
			return NULL;
		}
		cnt = ipcGetInt(client.msg);
		for(i=0;i<cnt;i++){
			mi = memCalloc(sizeof(*mi));
			mi->type = ipcGetInt(client.msg);
			mi->name = utilStrDup(ipcGetAscii(client.msg));
			ipcGetInt(client.msg);	/* interface version */
			ipcGetInt(client.msg);	/* implementation version */
			mi->fnBase = ipcGetInt(client.msg);
			mi->fnCount = ipcGetInt(client.msg);
			htabAddObject(moduleInfos, mi);
		}
	}
	res->err.code = res->err.class = 0;
	res->errmsg = "no Error";
	mi = htabFindObjectEqualTo(moduleInfos, &moduleName);
	if(mi == NULL){
		res->err.class =ERRCLS_FRAMEWORK;
		res->err.code = ERR_FRWmodulenotfound;
		res->errmsg = errDescription(&res->err);
	}
	return mi;
}

/* ------------------------------------------------------------------------- */

void	ipccSlashChars(ipccRes_t *res, btu8 *slash, btu8 *backsl, btsize bufsize)
{
	ipccStartProc(client.msg, MAIN_IPCFN_SLASH_CHARS);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	if(res->err.class == 0){
		strncpy(slash, ipcGetAscii(client.msg), bufsize);
		strncpy(backsl, ipcGetAscii(client.msg), bufsize);
	}else{
		strcpy(slash, "/");
		strcpy(backsl, "\\");
	}
}

/* ------------------------------------------------------------------------- */

void	ipccMount(ipccRes_t *res, btu8 *feModule, btu8 *fePath, btu8 *beModule, btu8 *bePath, btbool readOnly, btbool fakeMode)
{
	ipccStartProc(client.msg, MAIN_IPCFN_MOUNT);
	ipcAddAscii(client.msg, feModule);
	ipcAddAscii(client.msg, fePath);
	ipcAddAscii(client.msg, beModule);
	ipcAddAscii(client.msg, bePath);
	ipcAddInt(client.msg, readOnly);
	ipcAddInt(client.msg, fakeMode);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccUmount(ipccRes_t *res, btu8 *mountpoint, btbool fakeMode)
{
	ipccStartProc(client.msg, MAIN_IPCFN_UMOUNT);
	ipcAddAscii(client.msg, mountpoint);
	ipcAddInt(client.msg, fakeMode);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccRegisterForCallback(ipccRes_t *res, btu8 *modName)
{
	ipccStartProc(client.msg, MAIN_IPCFN_REGISTERCB);
	ipcAddAscii(client.msg, modName);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccUnregisterForCallback(ipccRes_t *res, btu8 *modName)
{
	ipccStartProc(client.msg, MAIN_IPCFN_UNREGISTERCB);
	ipcAddAscii(client.msg, modName);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccWillTerminate(ipccRes_t *res, btint termReason, btbool *doTerm)
{
	ipccStartProc(client.msg, MAIN_IPCFN_WILL_TERMINATE);
	ipcAddInt(client.msg, termReason);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	*doTerm = 1;
	if(res->err.class == 0)
		*doTerm = ipcGetInt(client.msg);
}

/* ------------------------------------------------------------------------- */

void	ipccGetObjectDescription(ipccRes_t *res, btint objectId)
{
	ipccStartProc(client.msg, MAIN_IPCFN_GET_OBDES);
	ipcAddInt(client.msg, objectId);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccConfigObjects(ipccRes_t *res)
{
	ipccStartProc(client.msg, MAIN_IPCFN_CONFIG_OBJECTS);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccGetCfgValue(ipccRes_t *res, btu8 *cfgPath, btbool dbOnly, btu8 **value)
{
	*value = NULL;
	ipccStartProc(client.msg, MAIN_IPCFN_GET_CFGVAL);
	ipcAddAscii(client.msg, cfgPath);
	ipcAddInt(client.msg, dbOnly);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	if(res->err.class == 0)
		*value = ipcGetAscii(client.msg);
}

/* ------------------------------------------------------------------------- */

void	ipccSetCfgValue(ipccRes_t *res, btu8 *cfgPath, btu8 *value)
{
	ipccStartProc(client.msg, MAIN_IPCFN_SET_CFGVAL);
	ipcAddAscii(client.msg, cfgPath);
	ipcAddAscii(client.msg, value);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccDelCfgValue(ipccRes_t *res, btu8 *cfgPath)
{
	ipccStartProc(client.msg, MAIN_IPCFN_DEL_CFGVAL);
	ipcAddAscii(client.msg, cfgPath);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccStoremnt(ipccRes_t *res, btu8 *mountpoint, btbool deleteStored)
{
	ipccStartProc(client.msg, MAIN_IPCFN_STOREMNT);
	ipcAddAscii(client.msg, mountpoint);
	ipcAddInt(client.msg, deleteStored);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccLookupList(ipccRes_t *res, btu8 *listName, btint *listId)
{
	ipccStartProc(client.msg, MAIN_IPCFN_LOOKUP_LIST);
	ipcAddAscii(client.msg, listName);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	*listId = ipcGetInt(client.msg);
}

/* ------------------------------------------------------------------------- */

void	ipccHelpForKey(ipccRes_t *res, btu8 *helpItem, btu8 **helpText)
{
	ipccStartProc(client.msg, MAIN_IPCFN_HELP_FOR_KEY);
	ipcAddAscii(client.msg, helpItem);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	*helpText = ipcGetAscii(client.msg);
}

/* ------------------------------------------------------------------------- */

void	ipccCifsLogin(ipccRes_t *res, btu8 *serverName)
{
ipccModuleInfo_t	*mi;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_LOGIN + mi->fnBase);
	ipcAddAscii(client.msg, serverName);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccCifsLogout(ipccRes_t *res, btu8 *serverName)
{
ipccModuleInfo_t	*mi;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_LOGOUT + mi->fnBase);
	ipcAddAscii(client.msg, serverName);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccCifsGetLicense(ipccRes_t *res, btu8 *lstring, btu8 **licenseType, bti32 *maxClients, bti32 *maxServers, btbool *srvCountLocal, btu8 **expiry)
{
ipccModuleInfo_t	*mi;
btsize				len;
btu8				*block;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_GET_LICENSE + mi->fnBase);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	if(res->err.class == 0){
		len = ipcGetBlock(client.msg, &block);
		if(len == 8){
			memcpy(lstring, block, 8);
			*licenseType = ipcGetAscii(client.msg);
			*maxClients = ipcGetInt(client.msg);
			*maxServers = ipcGetInt(client.msg);
			*srvCountLocal = ipcGetInt(client.msg);
			*expiry = ipcGetAscii(client.msg);
		}else{
			res->err.class = ERRCLS_FRAMEWORK;
			res->err.code = ERR_FRWipcparameters;
			res->errmsg = "error unmarshalling license data";
		}
	}
}

/* ------------------------------------------------------------------------- */

void	ipccCifsCurrentLicenseLoad(ipccRes_t *res, btint *servers, btint *clients)
{
ipccModuleInfo_t	*mi;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_LICENSE_LOAD + mi->fnBase);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
	if(res->err.class == 0){
		*clients = ipcGetInt(client.msg);
		*servers = ipcGetInt(client.msg);
	}else{
		*servers = *clients = 0;
	}
}

/* ------------------------------------------------------------------------- */

void	ipccCifsCurrentServers(ipccRes_t *res)
{
ipccModuleInfo_t	*mi;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_CURRENT_SERVERS + mi->fnBase);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccCifsSetLicense(ipccRes_t *res, btu8 ldata[12])
{
ipccModuleInfo_t	*mi;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_SET_LICENSE + mi->fnBase);
	ipcAddBlock(client.msg, ldata, 12);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */

void	ipccCifsStore(ipccRes_t *res, btu8 *serverName, btbool doDelete)
{
ipccModuleInfo_t	*mi;

	mi = ipccGetModuleInfo(res, "cifsFile");
	if(mi == NULL)
		return;
	ipccStartProc(client.msg, CIFS_IPCFN_STORE_LOGIN + mi->fnBase);
	ipcAddAscii(client.msg, serverName);
	ipcAddInt(client.msg, doDelete);
	ipccRpc(client.msg);
	unmarshalResult(client.msg, res);
}

/* ------------------------------------------------------------------------- */
