/*
 * Name: ipcclient.h
 * Project: CIFS Client
 * Author: Christian Starkjohann <cs@obdev.at>
 * Creation Date: 1997-12-12
 * Tabsize: 4
 * Copyright: (c) 1998 by Christian Starkjohann, all rights reserved.
 *     For details of the license see the file doc/License.txt.
 * This Revision: $Id: ipcclient.h,v 1.3 1999/06/09 13:36:55 cs Exp $
 */

#ifndef __ipcclient_H_INCLUDED__
#define __ipcclient_H_INCLUDED__

#include "basictypes.h"

/*
General Description:
This module contains RPC functions doing the marshalling and unmarshalling
called by the user interface commands. There's an RPC interface to the
daemon for user interface commands, if you didn't know....
*/

/* ------------------------------------------------------------------------- */

struct ipcMsg;

typedef struct ipccRequestData{
	btu32			sid;
	bti32			procId;
	struct ipcMsg	*msg;
	btint			protocolVersion;
}ipccRequestData_t;

typedef struct ipccRes{
	btError	err;
	btu8	*errmsg;
}ipccRes_t;

typedef struct ipccModuleInfo{
	btu8	*name;
	btint	type;
	btint	fnBase;
	btint	fnCount;
}ipccModuleInfo_t;
/* This structure is a container for what's known about a module */

/* ------------------------------------------------------------------------- */

extern btbool	ipcclientDebug;
/* If this variable is set to a non-zero value, the ipcclient module logs
 * all communication with the sharity daemon to standard error.
 */

void	ipccSetRequestHandler(btcallback handler, void *ctx);
/* The request handler is a function which is called if a client library
 * function receives a request from the daemon instead of an expected reply.
 * The handler is expected to process the request and send the reply back to
 * the daemon. This function sets the request handler function.
 * Parameters: 'handler' is a function pointer to the handler function and
 *    'ctx' is passed as the first parameter to the handler function.
 * Call parameters for handler: The first parameter is the context pointer
 * passed to this function and the second parameter is a pointer to a
 * ipccRequestData_t structure. If the daemon crahsed, the second parameter
 * is set to NULL to notify the client about the daemon death.
 */

void	ipccSetProtocolVersion(btint fnBase, btint fnCount, btint version);
/* This function tells the client library which protocol version is used for
 * which module in the inter-process communication to the Sharity daemon.
 * The module is specified in terms of an IPC function number range.
 * Parameters: 'fnBase' is the first IPC function number of the module,
 *    fnCount is the number of functions defined for the module and 'version'
 *    is the IPC protocol version implemented by the client. This number is
 *    used internally for version checks.
 */
btint	ipccProtocolVersion(btint fnNumber);
/* This function queries the protocol version (set with
 * ipccSetProtocolVersion()) for a given IPC function number.
 */

btint	ipccConnect(void);
void	ipccWriteMsg(struct ipcMsg *msg);
btint	ipccReadMsg(struct ipcMsg *msg);
/* This function reads a message from the pipe. It positions the read pointer
 * after the message length and protocol version entries.
 * Returns: The protocol version.
 */

/* creating/reading IPC messages by hand: */
struct ipcMsg	*ipccCurrentRxMessage(void);
void	ipccStartProc(struct ipcMsg *msg, int msgId);
void	ipccRpc(struct ipcMsg *msg);
/* This function executes the remote procedure call. The message in 'msg' is
 * sent to the daemon and the reply is received in the same message object.
 * If a request from the daemon is received instead of the reply, the function
 * passed in 'ipccSetRequestHandler()' (if any) is called. ipccRpc() must only
 * be called with messages initilized with 'ipccStartProc()' because
 * 'ipccStartProc()' passes information to 'ipccRpc()' in a static variable.
 */


/* some handy utility functions: */
void	ipccPrintErr(ipccRes_t *res);

/* ------------------------------------------------------------------------- */

ipccModuleInfo_t	*ipccGetModuleInfo(ipccRes_t *res, btu8 *moduleName);

void	ipccRegisterForCallback(ipccRes_t *res, btu8 *modName);
void	ipccUnregisterForCallback(ipccRes_t *res, btu8 *modName);
void	ipccWillTerminate(ipccRes_t *res, btint termReason, btbool *doTerm);
void	ipccListModules(ipccRes_t *res);
void	ipccSlashChars(ipccRes_t *res, btu8 *slash, btu8 *backsl, btsize bufsize);
void	ipccMount(ipccRes_t *res, btu8 *feModule, btu8 *fePath, btu8 *beModule, btu8 *bePath, btbool readOnly, btbool fakeMode);
void	ipccUmount(ipccRes_t *res, btu8 *mountpoint, btbool fakeMode);
void	ipccGetObjectDescription(ipccRes_t *res, btint objectId);
void	ipccConfigObjects(ipccRes_t *res);
void	ipccGetCfgValue(ipccRes_t *res, btu8 *cfgPath, btbool dbOnly, btu8 **value);
void	ipccSetCfgValue(ipccRes_t *res, btu8 *cfgPath, btu8 *value);
void	ipccDelCfgValue(ipccRes_t *res, btu8 *cfgPath);

void	ipccStoremnt(ipccRes_t *res, btu8 *mountpoint, btbool deleteStored);
void	ipccLookupList(ipccRes_t *res, btu8 *listName, btint *listId);
void	ipccHelpForKey(ipccRes_t *res, btu8 *helpItem, btu8 **helpText);

void	ipccCifsLogin(ipccRes_t *res, btu8 *serverName);
void	ipccCifsLogout(ipccRes_t *res, btu8 *serverName);
void	ipccCifsGetLicense(ipccRes_t *res, btu8 *lstring, btu8 **licenseType, bti32 *maxClients, bti32 *maxServers, btbool *srvCountLocal, btu8 **expiry);
void	ipccCifsCurrentLicenseLoad(ipccRes_t *res, btint *servers, btint *clients);
void	ipccCifsCurrentServers(ipccRes_t *res);
void	ipccCifsSetLicense(ipccRes_t *res, btu8 ldata[12]);
void	ipccCifsStore(ipccRes_t *res, btu8 *serverName, btbool doDelete);


/* ------------------------------------------------------------------------- */

#endif	/* __ipcclient_H_INCLUDED__ */
